/*                                                        -*- linux-c -*-
 *   _  _ ____ __ _ ___ ____ ____ __ _ _ _ _ | 
 * .  \/  |--| | \|  |  |--< [__] | \| | _X_ | s e c u r e  s y s t e m s
 *
 * .vt|ar5k - PCI/CardBus 802.11a WirelessLAN driver for Atheros AR5k chipsets
 *
 * $Id: vt_ar5kioctl.c,v 1.10 2003/08/22 17:35:38 reyk Exp $
 *
 * Copyright (c) 2002, .vantronix | secure systems
 *                     and Reyk Floeter <reyk@vantronix.net> 
 * Copyright (C) 2003 por .DyC Electronica 
 * | sistemas  inalambricos Mauricio Lange dyc_ar5k Mauricio Lange 
 * <mlange@dycelectronica.com.ar>, 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vt_ar5kconfig.h>
#include <vt_ar5k.h>

void vt_ar5k_ioctl_monitor(struct wireless_device *wdev, int val)
{
	struct net_device *dev = (struct net_device*)wdev->dev;
	vt_local_t *local = (vt_local_t *)dev->priv;

	VENTER();

	if(val == WIRELESS_ENABLE_MONITOR) {
		VPRINT("enable monitor mode\n");
		local->monitor = 1;
		local->dev_type = dev->type; /* store type */
		
		if(wdev->mac.header != WIRELESS_PRISM_HDR)
			dev->type = ARPHRD_IEEE80211;
		else 
			dev->type = ARPHRD_IEEE80211_PRISM;
	
		/* reset card to monitor mode */
		if(local->nic_active)
			VT_WLAN_OUT32(INIT_MONITOR_RX_FILTER, AR5K_RX_FILTER);
	} else {
		VPRINT("disable monitor mode\n");
		local->monitor = 0;
		dev->type = local->dev_type; /* store type */
		
		/* reset card to normal mode */
		if(local->nic_active) 
			VT_WLAN_OUT32(local->rx_filter, AR5K_RX_FILTER);
	}

	VEXIT();
}

int vt_ar5k_set_mode(struct wireless_device *wdev, int mode)
{
	struct net_device *dev = (struct net_device*)wdev->dev;
	vt_local_t *local = (vt_local_t*)dev->priv;
	u32 pcu = 0, beacon = 0, led = 0, addr_low, addr_high;
	u8 *bssid = NULL;
        VENTER();
	switch(mode) {
	case IW_MODE_MONITOR:
		if(!wdev->mac.cap.monitor)
			return(-ENOTSUPP);

		wireless_monitor(wdev, WIRELESS_ENABLE_MONITOR);
		return(0);

	case IW_MODE_INFRA:
		if(local->monitor)
			wireless_monitor(wdev, WIRELESS_DISABLE_MONITOR);

		pcu = AR5K_STA_ID1_PSPOLLDIS |
			AR5K_STA_ID1_DESC_ANT |
		    	AR5K_STA_ID1_PWR_SAV; 
		beacon = INIT_BCON_CNTRL_REG;
		led = AR5K_PCICFG_CLKRUNEN | 
			AR5K_PCICFG_LED_PEND | 
			AR5K_PCICFG_BCTL;
		bssid = (u8*)&wdev->mac.bssid;
		break;

	case IW_MODE_MASTER:
		if(!wdev->mac.cap.master)
			return(-ENOTSUPP);

		if(local->monitor)
			wireless_monitor(wdev, WIRELESS_DISABLE_MONITOR); 

		pcu = AR5K_STA_ID1_STA_AP | 
			AR5K_STA_ID1_PSPOLLDIS | 
			AR5K_STA_ID1_DESC_ANT;
		beacon = INIT_BCON_CNTRL_REG | 
			AR5K_BCR_APMODE;
		led = AR5K_PCICFG_LED_ACT | AR5K_PCICFG_BCTL;
		bssid = (u8*)&wdev->mac.address;
		break;

	case IW_MODE_ADHOC:
		if(local->monitor)
			wireless_monitor(wdev, WIRELESS_DISABLE_MONITOR); 

		pcu = AR5K_STA_ID1_AD_HOC | 
			AR5K_STA_ID1_PSPOLLDIS | 
			AR5K_STA_ID1_DESC_ANT;
		beacon = INIT_BCON_CNTRL_REG |
			AR5K_BCR_STAMODE;
		led = AR5K_PCICFG_CLKRUNEN | 
			AR5K_PCICFG_LED_PEND | 
			AR5K_PCICFG_BCTL;
		bssid = (u8*)&wdev->mac.bssid;
		break;

	default:
		printk(__FILE__ ": not implemented\n");
		return(-EINVAL);
	}

	/* set address */
	memcpy(&addr_low, &(wdev->mac.address[0]), 4);
	memcpy(&addr_high, &(wdev->mac.address[4]), 2);
	addr_high = 0x0000ffff & cpu_to_le32(addr_high);
	VT_WLAN_OUT32(cpu_to_le32(addr_low), AR5K_STA_ID0);
	VT_WLAN_OUT32(addr_high, AR5K_STA_ID1);

	/* set options */
	VT_WLAN_OUT32(pcu | addr_high, AR5K_STA_ID1);
	VT_WLAN_OUT32(beacon, AR5K_BCR);
	VT_WLAN_OUT32(led, AR5K_PCICFG);

	/* set bssid */
	memcpy(&addr_low, &(bssid[0]), 4);
	memcpy(&addr_high, &(bssid[4]), 2);
	addr_high = 0x0000ffff & cpu_to_le32(addr_high);
	VT_WLAN_OUT32(cpu_to_le32(addr_low), AR5K_BSS_ID0);
	VT_WLAN_OUT32(addr_high, AR5K_BSS_ID1);

	/* set the noise floor */
/*	VT_WLAN_OUT32(VT_WLAN_IN32(PHY_AGC_CONTROL) | PHY_AGC_CONTROL_NF,
		      PHY_AGC_CONTROL);
	
*/      VENTER();
        r__SetAGC(wdev);
	return(0);
}

int vt_ar5k_set_channel(struct wireless_device *wdev, struct iw_freq *freq)
{
	struct net_device *dev = (struct net_device*)wdev->dev;
	vt_local_t *local = (vt_local_t*)dev->priv;
	u32 data, data_orig, channel_ieee;
	int i;

	VENTER();

	/* set by frequency not implemented, use channel numbers */
	if(freq->e != 0) {
		channel_ieee = ((freq->m/100000)-5000)/5;
	} else {
		channel_ieee = freq->m;
	}

	/*
	 * OBSOLTE CODE: to be replaced by regulation domain settings
	 */

	/* FCC and Japan ,-) */
	if(local->conf->turbo_enable) {
		switch(channel_ieee) {
		case 3:
		case 42:
		case 50:
		case 58:
			break;
		default:
			VDEBUG("unsupported channel in turbo mode\n");
			return(-EOPNOTSUPP);
		}
	} else {
		if(channel_ieee < wdev->phy.min_freq.m ||
		   channel_ieee > wdev->phy.max_freq.m) {
			VDEBUG("unsupported channel in 802.11a mode\n");
			return(-EOPNOTSUPP);
		}
	}

	VT_WLAN_OUT32(PHY_ACTIVE_DIS, PHY_ACTIVE);
	udelay(1000);

	/** set the channel */
	data_orig = vt_ar5k_bitswap((channel_ieee - 24)/2, 5);
	data = (data_orig << 1) | (1 << 6) | 0x1;
	VT_WLAN_OUT32(data, PHY_BASE+(0x27<<2));
	VT_WLAN_OUT32(0, PHY_BASE+(0x30<<2));

	/** wait, activate phy, wait **/
	udelay(1000);
	VT_WLAN_OUT32(PHY_ACTIVE_EN, PHY_ACTIVE);
	udelay(1000);

	/** calibrate it and poll the bit going to 0 for completion **/
	VT_WLAN_OUT32(VT_WLAN_IN32(PHY_AGC_CONTROL) | PHY_AGC_CONTROL_CAL, 
		      PHY_AGC_CONTROL);

	for (i = 0; i < 10000; i++) {
		if ((VT_WLAN_IN32(PHY_AGC_CONTROL) 
		     & PHY_AGC_CONTROL_CAL) == 0)
			break;
		udelay(50);
	}

	/** set the noise floor **/
/*	VT_WLAN_OUT32(VT_WLAN_IN32(PHY_AGC_CONTROL) | PHY_AGC_CONTROL_NF,
		      PHY_AGC_CONTROL);
*/      r__SetAGC(wdev);
	/** save the current channel **/
	local->conf->hw_channel = local->conf->channel = channel_ieee;

	wdev->phy.freq.e = 1;
	wdev->phy.freq.m = (5000 + (5 * channel_ieee)) * 100000;

	VEXIT();

	return(0);
}

int vt_ar5k_set_network(struct wireless_device *wdev, int val)
{
	struct net_device *dev = (struct net_device*)wdev->dev;
	vt_local_t *local = (vt_local_t*)dev->priv;

	switch(val) {
	case WIRELESS_DISABLE_NETWORK:
		local->rx_filter = AR5K_RX_FILTER_ALL;
		memset(local->wdev->mac.bssid, 0, WIRELESS_ADDR_SIZE);
		memset(local->wdev->mac.desired_bssid, 0, WIRELESS_ADDR_SIZE);
		wdev->mac.searching = 0;
		VT_WLAN_OUT32(local->rx_filter, AR5K_RX_FILTER); 
/* XXX		VT_WLAN_OUT32(VT_WLAN_IN32(AR5K_IMR) & ~(AR5K_IMR_BMISS), AR5K_IMR); */
		break;

	case WIRELESS_FIND_NETWORK:
		memcpy(wdev->mac.bssid, wireless_no_addr, WIRELESS_ADDR_SIZE);
		memcpy(wdev->mac.desired_bssid, wireless_no_addr, WIRELESS_ADDR_SIZE);
		local->rx_filter =  INIT_SCAN_RX_FILTER;
		VT_WLAN_OUT32(local->rx_filter, AR5K_RX_FILTER); /* set receive filter */
		wdev->mac.searching = 1;
		break;

	case WIRELESS_FOUND_NETWORK:
		local->rx_filter = INIT_RX_FILTER; 
		VT_WLAN_OUT32(local->rx_filter, AR5K_RX_FILTER); /* set receive filter */
		wdev->mac.searching = 0;
		break;

	default:
		return(-EINVAL);
	}

	/* reset settings */
	if(wdev->set_mode)
		wdev->set_mode(wdev, wdev->mac.mode);

	return(0);
}

int vt_ar5k_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd) 
{
	vt_local_t *local = (vt_local_t *) dev->priv;
	struct wireless_device *wdev = (struct wireless_device*)local->wdev;
	struct iwreq *wrq = (struct iwreq *) ifr;
	unsigned long flags;
	int val, ret = 0;

	spin_lock_irqsave(&local->lock, flags);

	switch (cmd) {
	WIRELESS_IOCTLS: /* defined in linux/wdev.h */
		ret = wireless_ioctl(wdev, wrq, cmd);
		break;

	case SIOCGIWPRIV:
		if (wrq->u.data.pointer) {
			struct iw_priv_args privtab[] = {
				{ SIOCDEVPRIVATE + 0x0,
				  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
				  0, "led_mode" },
				{ SIOCDEVPRIVATE + 0x1,
				  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
				  0, "turbo_enable" },
				{ SIOCDEVPRIVATE + 0x2,
				  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
				  0, "swi" },
				{ SIOCDEVPRIVATE + 0x3,
				  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
				  0, "prism" },
				{ SIOCDEVPRIVATE + 0x4,
				  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
				  0, "reset" },
				{ SIOCDEVPRIVATE + 0x5,
				  IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1,
				  0, "dump" },
			};
			
			ret = verify_area(VERIFY_WRITE, wrq->u.data.pointer, sizeof(privtab));
			if (ret)
				break;
			
			wrq->u.data.length = sizeof(privtab) / sizeof(privtab[0]);
			if (copy_to_user(wrq->u.data.pointer, privtab, sizeof(privtab)))
				ret = -EFAULT;
		}
		break;
		
	case SIOCDEVPRIVATE + 0x0: /* led_mode */
		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}
		
		val = *((int*)wrq->u.name);
		
		/* store LED mode */
		local->led_mode = val ? 1 : -1;
		
		if (val > 0) {
			VT_WLAN_OUT32(AR5K_PCICFG_LED_ACT, AR5K_PCICFG); 
		} else {
			VT_WLAN_OUT32(AR5K_PCICFG_LED_PEND, AR5K_PCICFG); 
		}
		break;

	case SIOCDEVPRIVATE + 0x1: /* turbo_enable */
		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}

		val = *((int*)wrq->u.name);

		/* store turbo mode */
		local->conf->turbo_enable = val ? 1 : 0;
  
		if (val) {
			VT_WLAN_OUT32(PHY_FC_TURBO_MODE, PHY_FRAME_CONTROL);
		} else {
			VT_WLAN_OUT32(VT_WLAN_IN32(PHY_FRAME_CONTROL) & ~(PHY_FC_TURBO_MODE), PHY_FRAME_CONTROL);
		}

		wdev->phy.freq.e = 0;
		wdev->phy.freq.m = val ? turbo_channel : channel;
		ret = vt_ar5k_set_channel(wdev, &wdev->phy.freq);
		break;

	case SIOCDEVPRIVATE + 0x2: /* fire software interrupt - for testing */
		VT_WLAN_OUT32((VT_WLAN_IN32(AR5K_CR) | AR5K_CR_SWI), AR5K_CR); 
		break;

	case SIOCDEVPRIVATE + 0x3: /* enable prism header support */
		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}

		val = *((int*)wrq->u.name);

		/* use prism header? */
		wdev->mac.header = val ? WIRELESS_PRISM_HDR : WIRELESS_STD_HDR;

		/* reset monitor to new mode */
		if(local->monitor) { 
			wireless_monitor(wdev, WIRELESS_DISABLE_MONITOR);
			wireless_monitor(wdev, WIRELESS_ENABLE_MONITOR);
		}				

		break;

	case SIOCDEVPRIVATE + 0x4: /* reset - DANGEROUS!!! */
		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}

		val = *((int*)wrq->u.name);

		VDEBUG("WARNING: BE CAREFUL!\n");

		if((ret = vt_ar5k_nic_reset(local, val)) != 0)
			break;

		udelay(2300);

		break;

	case SIOCDEVPRIVATE + 0x5: /* dump like mad wifi dumps */
		if(!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}

		if(!(*((int*)wrq->u.name))) {
			ret = -EINVAL;
			break;
		}

		if(local->dump) 
			ret = local->dump(local);
		else 
			VDEBUG("dump not supported!\n");
		
		break;

	default:
#if VT_DEBUG > 1
		VPRINT("ioctl 0x%04X not supported!\n", cmd);
#endif
		ret = -EOPNOTSUPP;
		break;
	}

	spin_unlock_irqrestore(&local->lock, flags);

	return(ret);
}
