/*                                                        -*- linux-c -*-
 *   _  _ ____ __ _ ___ ____ ____ __ _ _ _ _ | 
 * .  \/  |--| | \|  |  |--< [__] | \| | _X_ | s e c u r e  s y s t e m s
 *
 * .vt|ar5k - PCI/CardBus 802.11a WirelessLAN driver for Atheros AR5k chipsets
 *
 * $Id: vt_ar5kmod.c,v 1.18 2003/12/02 17:26:28 root Exp root $
 *
 * Copyright (c) 2002, .vantronix | secure systems
 *                     and Reyk Floeter <reyk@vantronix.net> 
 * Copyright (C) 2003 por .DyC Electronica 
 * | sistemas  inalambricos Mauricio Lange dyc_ar5k Mauricio Lange 
 * <mlange@dycelectronica.com.ar>, 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vt_ar5kconfig.h>
#include <vt_ar5k.h>
#include <vt_ar5kinitregs.h>
/* #include <stdio.h> */
#include <linux/ieee80211.h>
#define DESCRIPTION "PCI/CardBus 802.11a WirelessLAN driver for Atheros AR5k chipsets" 
#define AUTHOR "Reyk Floeter <reyk@vantronix.net>, (C) 2002 by .vantronix | secure systems and Reyk Floeter Mauricio Lange <mlange@dycelectronica.com.ar>, (C) 2003 por .DyC Electronica | sistemas  inalambricos Mauricio Lange dyc_ar5k"
#define MODNAME "AR5k"

char *version = DESCRIPTION "\n" AUTHOR;
char *dev_info = "vt_ar5k";
pci_link_t *dev_list = NULL; 
int vt_ar5k_cards_found = -1;

/* atheros default- alike */
char vt_ar5k_mac_addr[6] = {0x00, 0x03, 0x7f, 0x00, 0x00, 0x23}; 
char vt_ar5k_bssid[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

/** generic driver options **/
struct pci_device_id pci_id_tbl[]  __devinitdata = {
	{PCIVENDOR_ATHEROS, CB_DEVICE_AR5000, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 
	 (unsigned long)"Atheros 802.11a CardBus Wireless LAN AR5000"},
	{PCIVENDOR_ATHEROS, CB_DEVICE_AR5001A, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 
	 (unsigned long)"Atheros 802.11a CardBus Wireless LAN AR5001A"},
	{PCIVENDOR_ATHEROS, CB_DEVICE_AR5001X, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 
	 (unsigned long)"Atheros 802.11a/b/g CardBus Wireless LAN AR5001X"},
	{PCIVENDOR_ATHEROS, CB_DEVICE_AR5001XPLUS, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 
	 (unsigned long)"Atheros 802.11a/b/g CardBus Wireless LAN AR5001X+"},
	{PCIVENDOR_ATHEROS, PCIDEVICE_AR5000, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 
	 (unsigned long)"Atheros 802.11a PCI Wireless LAN AR5000"},
	{PCIVENDOR_ATHEROS, PCIDEVICE_AR5001A, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 
	 (unsigned long)"Atheros 802.11a PCI Wireless LAN AR5001A"}, 
	{PCIVENDOR_ATHEROS, PCIDEVICE_AR5001X, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 
	 (unsigned long)"Atheros 802.11a/b/g PCI Wireless LAN AR5001X"}, 
	{PCIVENDOR_ATHEROS, PCIDEVICE_AR5001XPLUS, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 
	 (unsigned long)"Atheros 802.11a/b/g PCI Wireless LAN AR5001X+"}, 
	{0, 0, 0, 0, 0, 0, 0}
};

struct pci_driver vt_ar5k_pci_drv_id = {
	.name     = "ar5k", 
	.id_table = pci_id_tbl,
	.probe    = vt_ar5k_probe,
	.remove   = vt_ar5k_remove,
#ifdef CONFIG_PM
	.suspend  = vt_ar5k_suspend,
	.resume   = vt_ar5k_resume,
#endif /* CONFIG_PM */
};

/** module options **/
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_AUTHOR(AUTHOR);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
MODULE_DEVICE_TABLE(pci, pci_id_tbl);

int mtu = 1600 - sizeof(wlan_frame_hdr_t) - 4;
MODULE_PARM(mtu, "i");
MODULE_PARM_DESC(mtu, "maximum transfer unit");

char essid[IW_ESSID_MAX_SIZE + 1] = "test";
MODULE_PARM(essid, "c" __MODULE_STRING(IW_ESSID_MAX_SIZE));
MODULE_PARM_DESC(essid, "set ESSID (default: test)");

char reg_domain[AR5K_REG_SIZE];
MODULE_PARM(reg_domain, "c" __MODULE_STRING(AR5K_REG_SIZE));
MODULE_PARM_DESC(reg_domain, "please specify your regulation domain (no default)");

int channel = 36;
MODULE_PARM(channel, "i");
MODULE_PARM_DESC(channel, "select default channel (64)");

int turbo_channel = 58;
MODULE_PARM(turbo_channel, "i");
MODULE_PARM_DESC(turbo_channel, "select default turbo mode channel (58)");

int turbo_enable = 0;
MODULE_PARM(turbo_enable, "i");
MODULE_PARM_DESC(turbo_enable, "enable turbo mode (> 0)");

int led_mode = 0;
MODULE_PARM(led_mode, "i");
MODULE_PARM_DESC(led_mode, "LED mode (> 0 = activity status, < 1 = pending status, default 0)");

int opmode = 2;
MODULE_PARM(opmode, "i");
MODULE_PARM_DESC(opmode, "initial operation mode (1:ad-hoc, 2:infra [default], 3:master, 6:monitor)");

int vt_ar5k_probe(struct pci_dev *pdev, const struct pci_device_id *id) 
{
	unsigned long phymem;
	unsigned long mem = 0;
	pci_link_t *link = NULL;
	vt_local_t *local = NULL;
	struct net_device *dev = NULL;
	struct wireless_device *wdev = NULL;
	sta_config_t *conf = NULL;
	u32 memlen;
        INCHK
	VENTER();
	
	/** enable PCI device **/

	if (pci_enable_device(pdev))
		return(-EIO);

	if(!(phymem = pci_resource_start(pdev, 0)))
		goto err_out;

	/* enable bus master */
	pci_set_master(pdev);

	if(pci_request_regions(pdev, MODNAME)) {
		VINIT("cannot reserve PCI memory region\n");
		goto err_out;
	}

	memlen = pci_resource_len(pdev, 0);

	if((mem = 
	    (unsigned long) ioremap(phymem, memlen)) == 0) {
		VINIT("cannot remap PCI memory region\n");
		goto err_out_free_res;
	}

	/* enable DMA */
	if (pci_set_dma_mask(pdev, 0xffffffff)) {
                VINIT("no suitable DMA available.\n");
                goto fail;
	}

	/** init device data **/

	link = kmalloc(sizeof(pci_link_t), GFP_KERNEL);  
	local = kmalloc(sizeof(vt_local_t), GFP_KERNEL);
	dev = kmalloc(sizeof(struct net_device), GFP_KERNEL);
	wdev = kmalloc(sizeof(struct wireless_device), GFP_KERNEL);
	conf = kmalloc(sizeof(sta_config_t), GFP_KERNEL);

	if ((local != NULL) 
	    && (dev != NULL) 
	    && (wdev != NULL) 
	    && (link != NULL) 
	    && (conf != NULL)) {
		memset(link, 0, sizeof(pci_link_t));
		memset(local, 0, sizeof(vt_local_t));
		memset(dev, 0, sizeof(struct net_device));
		memset(wdev, 0, sizeof(struct wireless_device));
		memset(conf, 0, sizeof(sta_config_t));
		memset(&local->stats, 0, sizeof(struct net_device_stats));
	} else {
		VINIT("out of memory\n");
		goto fail;
	}

	vt_ar5k_cards_found++;
	
	link->priv = local;
	link->next = dev_list;
	dev_list = link;

	/* setup generic net device */
	dev->priv = local;
	dev->mem_start = mem;
	dev->mem_end = mem + memlen;
	dev->irq = local->irq = pdev->irq;
	dev->mtu = mtu;
	dev->priv = local;
	dev->open = vt_ar5k_open;
	dev->stop = vt_ar5k_stop;
	dev->do_ioctl = vt_ar5k_ioctl;
	dev->get_stats = vt_ar5k_get_stats;
	dev->hard_start_xmit = vt_ar5k_start_xmit;
	dev->get_wireless_stats = vt_ar5k_wireless_stats;
	dev->hard_header_len=6;
	memcpy(dev->name, "wlan%d", 7);
	pci_set_drvdata(pdev, dev);

	/* setup wireless device (see linux/wdev.h) */
	wdev->dev = dev;
	wdev->mac.mode = opmode;
	wdev->mac.proto = WIRELESS_PROTO_IEEE80211;
	wdev->mac.cap.master = 0; /* accesspoint mode is not implemented yet */
	wdev->mac.cap.monitor = 1; 
	wdev->set_channel = vt_ar5k_set_channel;
	wdev->set_mode = vt_ar5k_set_mode;
	wdev->set_network = vt_ar5k_set_network;
	wdev->set_monitor = vt_ar5k_ioctl_monitor;
	wdev->transmit = vt_ar5k_tx;

	VINIT("card %d: mem_start:0x%lx, len: %ld, irq:%d\n", 
	      vt_ar5k_cards_found,
	      dev->mem_start, dev->mem_end - dev->mem_start, dev->irq);
	
	switch(pdev->device & 0x00ff) {
	case ATHEROS_CHIPSET_AR5000:
		VINIT("card %d: AR5000 chipset\n", vt_ar5k_cards_found);
		
		wdev->phy.cap.ofdm = 1;
		wdev->phy.type = WIRELESS_PHY_MODE_OFDM;
		
		/* set frequencies
		 * AR5000 officially supports 5150-5350MHz 
		 */
		wdev->phy.min_freq.e = 0;
		wdev->phy.min_freq.m = 30;
		wdev->phy.max_freq.e = 0;
		wdev->phy.max_freq.m = 70;

		/* driver specific */
		local->eeprom_read = vt_ar5210_eeprom_read;
		local->chipset_init = vt_ar5210_init;
		local->dump = vt_ar5210_dump;
		local->reset_flags = 
			AR5K_RC_PCU | AR5K_RC_DMA | AR5K_RC_MAC | AR5K_RC_PHY;
		break;

	case ATHEROS_CHIPSET_AR5001A:
		VINIT("card %d: AR5001A chipset (not fully supported)\n", 
		      vt_ar5k_cards_found);
    
		wdev->phy.cap.ofdm = 1; 
		wdev->phy.type = WIRELESS_PHY_MODE_OFDM;

		/* set frequencies
		 * AR5001 officially supports 5150-5850MHz
		 */
		wdev->phy.min_freq.e = 0;
		wdev->phy.min_freq.m = 30;
		wdev->phy.max_freq.e = 0;
		wdev->phy.max_freq.m = 170;

		/* driver specific */
		local->eeprom_read = vt_ar5211_eeprom_read;
		local->chipset_init = vt_ar5211_init;
		local->reset_flags = AR5K_RC_PCU | AR5K_RC_DMA;
		break;

	case ATHEROS_CHIPSET_AR5001X:
		VINIT("card %d: AR5001X chipset (not fully supported)\n", 
		      vt_ar5k_cards_found);
    
		wdev->phy.cap.ofdm = 1; 
		wdev->phy.cap.dsss = 1; /* supports 802.11b */
		wdev->phy.type = WIRELESS_PHY_MODE_OFDM;

		/* set frequencies
		 * AR5001 officially supports 5150-5850MHz
		 * (and 2412-2484MHz) 
		 */
		wdev->phy.min_freq.e = 0;
		wdev->phy.min_freq.m = 30;
		wdev->phy.max_freq.e = 0;
		wdev->phy.max_freq.m = 170;

		/* driver specific */
		local->eeprom_read = vt_ar5211_eeprom_read;
		local->chipset_init = vt_ar5211_init;
		local->reset_flags = AR5K_RC_PCU | AR5K_RC_DMA;
		break;

	case ATHEROS_CHIPSET_AR5001XPLUS:
		VINIT("card %d: AR5001X+ chipset (not fully supported)\n", 
		      vt_ar5k_cards_found);
    
		wdev->phy.cap.ofdm = 1; 
		wdev->phy.cap.dsss = 1; /* supports 802.11b */
		wdev->phy.type = WIRELESS_PHY_MODE_OFDM;

		/* set frequencies
		 * AR5001 officially supports 5150-5850MHz
		 * (and 2412-2484MHz) 
		 */
		wdev->phy.min_freq.e = 0;
		wdev->phy.min_freq.m = 30;
		wdev->phy.max_freq.e = 0;
		wdev->phy.max_freq.m = 170;

		/* driver specific */
		local->eeprom_read = vt_ar5212_eeprom_read;
		local->chipset_init = vt_ar5212_init;
		local->reset_flags = AR5K_RC_PCU | AR5K_RC_DMA;
		break;

	default:
		VINIT("card %d: UNKNOWN CHIPSET (0x%04x)\n", 
		      vt_ar5k_cards_found,
		      local->pdev->device);
		goto fail;
	}

	/* setup local info structure */
	local->dev = dev;
	local->pdev = pdev;
	local->wdev = wdev;
	local->link = link;
	local->conf = conf;
	local->io_port = phymem;

	spin_lock_init(&local->lock);
	
	ether_setup(dev);
	netif_stop_queue(dev);
	SET_MODULE_OWNER(dev);
	
	if (vt_ar5k_config(local)) {
		VPRINT("configuration failed\n");
		goto fail;
	}
	
	/* calls register_netdev */
	if(wireless_register(wdev) != 0) {
		VPRINT("wireless_register() failed!\n");
		goto fail;
	}

	if(vt_ar5k_nic_init(local)) {
		VPRINT("nic initialization failed!\n");
		wireless_unregister(wdev);
		goto fail;
	}

	if(vt_ar5k_config_registers(local)) {
		VPRINT("register configuration failed!\n");
		wireless_unregister(wdev);
		goto fail;
	}

	vt_ar5k_init_proc(local);
	
	local->nic_ready = 1;

	VPRINT("card initialized successfully (status 0x%08lx)\n", 
	       (unsigned long)VT_WLAN_IN32(AR5K_CFG));
	
	VT_WLAN_OUT32(PHY_ACTIVE_DIS, PHY_ACTIVE);
	udelay(1000);

	VEXIT();
	
	return(0);

 fail:
	if (mem != 0) iounmap((void *) mem);
	if (link != NULL) kfree(link);
	if (local != NULL) kfree(local);
	if (dev != NULL) kfree(dev);
	if (conf != NULL) kfree(conf);

 err_out_free_res:
	release_mem_region(phymem, pci_resource_len(pdev, 0));
	/* pci_release_regions(pdev); */

 err_out:	
	pci_disable_device(pdev);
	
	VPRINT("card initialization failed!\n");

	return(-ENODEV);
}

void __devexit vt_ar5k_remove(struct pci_dev *pdev) 
{
	struct net_device *dev = (struct net_device*) pci_get_drvdata(pdev);
	vt_local_t *local = (vt_local_t*) dev->priv;
	pci_link_t *link = (pci_link_t*)local->link;
	
	VENTER();
		
	if(local->nic_active)
		vt_ar5k_stop(dev);

	/* disable rx */
	VT_WLAN_OUT32((VT_WLAN_IN32(AR5K_CR) | AR5K_CR_RXD | ~AR5K_CR_RXE),
		      AR5K_CR); /* disable receive bit */
	VT_WLAN_OUT32(AR5K_DIAG_RX_DIS, AR5K_DIAG_SW); 
	VT_WLAN_OUT32(AR5K_RX_FILTER_ALL, AR5K_RX_FILTER);
	vt_ar5k_free_rx_queue(local);

	/* disable tx */
	VT_WLAN_OUT32(AR5K_DIAG_TX_DIS, AR5K_DIAG_SW); 
	vt_ar5k_free_tx_queue(local);

	VPRINT("resetting device\n");
	vt_ar5k_nic_exit(local);

	VPRINT("unregistering net device\n");
	vt_ar5k_remove_proc(local);

	/* unregister net device */
	wireless_unregister(local->wdev);

	iounmap((void*)(dev->mem_start));
	kfree(link);
	kfree(local->conf);
	kfree(local);
	kfree(dev);
	
	pci_release_regions(pdev);
	pci_disable_device(pdev);
	pci_set_drvdata(pdev,NULL);
	
	vt_ar5k_cards_found--;

	VEXIT();
}

int vt_ar5k_nic_init(vt_local_t *local) 
{
	struct net_device *dev = (struct net_device*) local->dev;
	int i;
        int ret;
	INCHK
	VENTER();
	local->int_count = 0;

#ifdef USE_GPIO
	/* enable gpio[2] and set it to value 1 */
	if(vt_ar5k_gpio_out(local, 2) || 		
	   vt_ar5k_gpio_set(local, 2, 1))
		return(1);
#endif	

	/* 1st reset: pci interface */
	if((ret = vt_ar5k_nic_reset(local, AR5K_RC_PCI)) != 0)
		return(ret);

	udelay(2300);

	/** step 1: reset and wakeup procedure... **/
	if (vt_ar5k_nic_resume(local)) {
		VPRINT("nic wakeup failed\n");
		return(1);
	}	

	/* get MAC revision */
	local->mac_rev = (VT_WLAN_IN16(AR5K_SREV) & AR5K_SREV_ID_M);
	
	/* enable turbo mode? this is an ar5k enhancement to 802.11a */
	if(local->conf->turbo_enable) {
		VT_WLAN_OUT32(PHY_FC_TURBO_MODE, PHY_FRAME_CONTROL);
	} else {
		VT_WLAN_OUT32(0, PHY_FRAME_CONTROL);
	}

	/* 2nd reset: chipset */
	if((ret = vt_ar5k_nic_reset(local, local->reset_flags)) != 0)
		return(ret);

	udelay(2300);
	
	/* wakeup again */
	if (vt_ar5k_nic_resume(local))
		return(1);
	
	/* 3rd reset: warm reset */
	if((ret = vt_ar5k_nic_reset(local, 0)) != 0)
		return(ret);

	/* some gpio stuff... (???) */
#ifdef USE_GPIO
	if(vt_ar5k_gpio_out(local, 2) ||
	   vt_ar5k_gpio_set(local, 2, 1)) 
		return(1);
	udelay(5000); /* sleep for 5ms */
	VT_WLAN_OUT32(0x14, (PHY_BASE+(0x34<<2))); /* clear a2's reg5 */
	vt_ar5k_gpio_set(local, 2, 0); /* gpio[2] = 0 */
	vt_ar5k_gpio_set(local, 2, 1); /* gpio[2] = 1 (burnin) */
	vt_ar5k_gpio_set(local, 2, 0); /* gpio[2] = 0 */
#endif	

	/* get PHY revision */
	local->phy_rev = VT_WLAN_IN16(PHY_CHIP_ID);

	/* messy business reading */
	VT_WLAN_OUT32(0x00001c16, PHY_BUSY_REG1);
	for (i=0; i<4; i++) {
		VT_WLAN_OUT32(0x00010000, PHY_BUSY_REG0);
	}
	
	local->a2_rev = 
		(u16)(vt_ar5k_bitswap(((VT_WLAN_IN16(PHY_BASE + (256<<2)) >> 28) 
				       & 0xf), 4) + 1);
	
	/* gpio.... AGAIN */
	if ((local->gpio0_bit = vt_ar5k_gpio_get(local, 0)) == 0) {
		/* already closed, force no rf */
		vt_ar5k_gpio_int(local, 1);
	} else {
		vt_ar5k_gpio_int(local, 0);
	}

	VPRINT("starting card in %s mode\n", 
	       turbo_enable ? "turbo" : "standard");

	printk("%s (%s): MAC revision 0x%04x\n"
	       "%s (%s): PHY revision 0x%04x\n",
	       dev_info, dev->name, local->mac_rev,
	       dev_info, dev->name, local->phy_rev);

	VEXIT();
	
	return(0);
}

int vt_ar5k_nic_exit(vt_local_t *local) 
{
	VENTER();
	
	if (vt_ar5k_nic_suspend(local))
		return(1);

	VEXIT();
	
	return(0);
}

int vt_ar5k_nic_suspend(vt_local_t *local)
{
	struct net_device *dev = (struct net_device*) local->dev;
	unsigned long flags;
	u32 i;
	INCHK
	VENTER();

	spin_lock_irqsave(&local->lock, flags);

	local->nic_ready = 0;

	VT_WLAN_OUT32(AR5K_SCR_SLE_FSLEEP, AR5K_SCR);
	
	for (i = 0; i < POWER_UP_TIME / 200; i++) {
		if ((VT_WLAN_IN16(AR5K_PCICFG) & AR5K_PCICFG_SPWR_DN) == 0) {
			spin_unlock_irqrestore(&local->lock, flags);
			return(0);
		}
		udelay (200);
		VT_WLAN_OUT32(AR5K_SCR_SLE_FSLEEP, AR5K_SCR);
	}
	
	spin_unlock_irqrestore(&local->lock, flags);

	VEXIT();
	
	return(1);
}

int vt_ar5k_nic_resume(vt_local_t *local)
{
	struct net_device *dev = (struct net_device *)local->dev;
	unsigned long flags;
	u32 i;

	VENTER();
	
	spin_lock_irqsave(&local->lock, flags);

	local->nic_ready = 1;

	VT_WLAN_OUT32(AR5K_SCR_SLE_FWAKE, AR5K_SCR);
	
	for (i = 0; i < 2000 / 200; i++) {
		if ((VT_WLAN_IN16(AR5K_PCICFG) & AR5K_PCICFG_SPWR_DN) == 0) {
			spin_unlock_irqrestore(&local->lock, flags);
			return(0);		
		}
		udelay (200);
		VT_WLAN_OUT32(AR5K_SCR_SLE_FWAKE, AR5K_SCR);
	}
	
	spin_unlock_irqrestore(&local->lock, flags);

	VEXIT();
	
	return(1);
}

#ifdef CONFIG_PM
int vt_ar5k_resume(struct pci_dev *pdev)
{
	struct net_device *dev = pci_get_drvdata (pdev);
	vt_local_t *local = (vt_local_t *)dev->priv;

	VENTER();

#if 1	
	if (netif_running(dev)) {
		VPRINT("resume failed - %s is running\n", dev->name);
		return(0);
	}

	netif_device_attach(dev);
#endif

	if(vt_ar5k_nic_resume(local)) {
		VPRINT("resume failed\n");
		return(1);
	}
	
	VEXIT();
	
	return(0);
}

int vt_ar5k_suspend(struct pci_dev *pdev, u32 state)
{
	struct net_device *dev = pci_get_drvdata (pdev);
	vt_local_t *local = (vt_local_t *)dev->priv;

#if 1
	if (netif_running(dev)) {
		VPRINT("suspend failed - %s is running\n", dev->name);
		return(0);
	}

	netif_device_detach(dev);
#endif

	if(vt_ar5k_nic_suspend(local)) {
		VPRINT("suspend failed\n");
		return(1);
	}

	return(0);
}
#endif /* CONFIG_PM */

int vt_ar5k_nic_reset(vt_local_t *local, u32 val) 
{
	struct net_device *dev = (struct net_device*) local->dev;
	u32 mask = val ? val : ~0, read;
	int i, result = 1;
        INCHK
	VENTER();

	/* read added before writing, 
	   'cause kernel hanged when reinserting module */
	read = VT_WLAN_IN32(AR5K_RC);
	VT_WLAN_OUT32(val, AR5K_RC);
	
	/* need to wait at least 128 clocks when reseting PCI before read */
	udelay(15);
	
	val  &= AR5K_RC_PCU | AR5K_RC_DMA | AR5K_RC_MAC | AR5K_RC_PHY;
	mask &= AR5K_RC_PCU | AR5K_RC_DMA | AR5K_RC_MAC | AR5K_RC_PHY;
	
	for (i = 0; i < 10000; i++) {
		read = VT_WLAN_IN16(AR5K_RC);
		if ((read & mask) != val) 
			udelay(50);
		else {
			result = 0;
			break;
		}
	}

	if (result)	
		VPRINT("nic reset failed %x != %x failed\n", val,
		       (VT_WLAN_IN16(AR5K_RC) & mask));
	
	VEXIT();
	return(result);
}

#ifdef USE_SOFTIRQ
static void vt_ar5k_rx_tasklet(unsigned long data)
{
	vt_local_t *local = (vt_local_t*)data;
	INCHK;VENTER();
	vt_ar5k_rx(local);
}

static void vt_ar5k_tx_interrupt_tasklet(unsigned long data)
{
	vt_local_t *local = (vt_local_t*)data;
	INCHK;VENTER();
	vt_ar5k_tx_interrupt(local);
}

static void vt_ar5k_tx_ok_tasklet(unsigned long data)
{
	vt_local_t *local = (vt_local_t*)data;
	struct wireless_device *wdev = (struct wireless_device*)local->wdev;
	INCHK;VENTER();
	spin_lock(&local->lock);	
	if(wdev->transmit_ok)
		wdev->transmit_ok(wdev);
	spin_unlock(&local->lock);
}
#endif

int vt_ar5k_config(vt_local_t *local) 
{
        int ret = 0;
        INCHK
	VENTER();

	/** local data configuration **/
	local->led_mode = led_mode;
	local->rx_filter = INIT_RX_FILTER;
	local->monitor = 0;
	local->rx_drop = 0; /* drop invalid frames */
	local->tx_active = 0;
	local->change_count = 0;

	/** station data configuration **/
	local->conf->turbo_enable = turbo_enable;
	local->conf->qos_enabled = 0; /* TODO */
	local->conf->cwmin = 0; /* TODO */
	local->conf->aifs = 0; /* TODO */

	/** setup software interrupts **/
#ifdef USE_SOFTIRQ
	VT_TASKLET_INIT(&local->rx_tasklet,
			vt_ar5k_rx_tasklet, (unsigned long) local);
	VT_TASKLET_INIT(&local->tx_interrupt_tasklet,
			vt_ar5k_tx_interrupt_tasklet, (unsigned long) local);
	VT_TASKLET_INIT(&local->tx_ok_tasklet,
			vt_ar5k_tx_ok_tasklet, (unsigned long) local);
#endif

	VEXIT();

	return(ret);
}

int vt_ar5k_config_registers(vt_local_t *local) 
{	
	struct net_device *dev = (struct net_device*)local->dev;
	struct wireless_device *wdev = (struct wireless_device*)local->wdev;
	int ret;
	u8 mac[WIRELESS_ADDR_SIZE];
	u32 data, channel_ieee;
        INCHK
	VENTER();

	/** read mac address from eeprom or use default address on failure **/
	if((ret = vt_ar5k_eeprom_read_mac(local, (u8*)&mac)) == 0)
		memcpy(vt_ar5k_mac_addr, mac, WIRELESS_ADDR_SIZE);

	memcpy(local->conf->mac_perm_addr, vt_ar5k_mac_addr, WIRELESS_ADDR_SIZE);
	memcpy(wdev->mac.address, vt_ar5k_mac_addr, WIRELESS_ADDR_SIZE);
	memcpy(wdev->mac.desired_bssid, vt_ar5k_bssid, WIRELESS_ADDR_SIZE);
	memcpy(wdev->mac.essid, essid, WIRELESS_SSID_SIZE);
	memcpy(&dev->dev_addr, wdev->mac.address, WIRELESS_ADDR_SIZE);

	VPRINT("hardware address " MACSTR " (%s)\n", MAC2STR(dev->dev_addr),
	       ret ? "default" : "from EEPROM");

	ret = 0;

	/** set mac address, mode, essid and  bssid **/
	wdev->set_mode(wdev, opmode);
	
	/** reset tx/rx descriptor pointers **/
	VT_WLAN_OUT32(0, AR5K_TXDP0);
	VT_WLAN_OUT32(0, AR5K_TXDP1);
	VT_WLAN_OUT32(0, AR5K_RXDP);

	/** reset interrupt register **/
	(void)VT_WLAN_IN32(AR5K_ISR); /* reset */
	VT_WLAN_OUT32(0, AR5K_IMR);
	VT_WLAN_OUT32(AR5K_IER_DISABLE, AR5K_IER);
	local->nic_ready = 0;

	/** set dma size - default is 128byte **/
	VT_WLAN_OUT32(AR5K_DMASIZE_128B, AR5K_TXCFG);
 	VT_WLAN_OUT32(AR5K_DMASIZE_128B, AR5K_RXCFG); 
	VT_WLAN_OUT32(0, AR5K_CFG); /* big endian */
	
	/** set timeouts **/
	VT_WLAN_OUT32(INIT_TOPS, AR5K_TOPS);
	VT_WLAN_OUT32(INIT_RXNOPKT, AR5K_RXNPTO);
	VT_WLAN_OUT32(INIT_RPGTO, AR5K_RPGTO);
	VT_WLAN_OUT32(INIT_TXNOPKT, AR5K_TXNPTO);

	/** disable forced sleep **/
	VT_WLAN_OUT32(0, AR5K_SFR);

	/** reset mac configuration and status register **/
	VT_WLAN_OUT32(cpu_to_le32(INIT_CONFIG_STATUS), AR5K_CFG);

	/** reset mib counters **/
	VT_WLAN_OUT32(0, AR5K_MIBC);
	
	/** set 802.11e tunable cwmin **/
	if (!local->conf->qos_enabled) {
		local->conf->cwmin = INIT_CWMIN;
	}

	data = 1;
	while (data < local->conf->cwmin)
		data = (data << 1) | 1;

	/** set retry limits **/
	data = (data << AR5K_RETRY_LMT_CW_S)
		| (INIT_SLG_RETRY << AR5K_RETRY_LMT_SLG_S)
		| (INIT_SSH_RETRY << AR5K_RETRY_LMT_SSH_S);
	
	if (local->conf->swretry_enabled) {
		data |= ((local->conf->limit_tx_retries / 2) << AR5K_RETRY_LMT_LG_S)
			| (local->conf->limit_tx_retries / 2);
	} else {
		data |= (INIT_LG_RETRY << AR5K_RETRY_LMT_LG_S) | INIT_SH_RETRY;
	}
	VT_WLAN_OUT32(data, AR5K_RETRY_LMT);

	/* set initial filters */
	VT_WLAN_OUT32(AR5K_RX_FILTER_ALL, AR5K_RX_FILTER);
	VT_WLAN_OUT32(0, AR5K_MCAST_FIL0);
	VT_WLAN_OUT32(0x00001140, AR5K_MCAST_FIL1);
	VT_WLAN_OUT32(0, AR5K_TX_MASK0);
	VT_WLAN_OUT32(0, AR5K_TX_MASK1);
	VT_WLAN_OUT32(1, AR5K_CLR_TMASK);

	/** set transmit fifo threshold **/
	VT_WLAN_OUT32(INIT_TX_FIFO_THRESHOLD, AR5K_TRIG_LEV);

	/** reset pcu control register **/
	VT_WLAN_OUT32(0, AR5K_DIAG_SW);

	(void)VT_WLAN_IN32(AR5K_BSR); /* reset */
	VT_WLAN_OUT32(INIT_RSSI_THR, AR5K_RSSI_THR);
	VT_WLAN_OUT32(INIT_NEXT_BEACON_TIME, AR5K_TIMER0);
	VT_WLAN_OUT32(INIT_TSF_CLOCK_LOW, AR5K_TSF_L32);
	VT_WLAN_OUT32(INIT_DMA_BEACON_ALERT, AR5K_TIMER1);
	VT_WLAN_OUT32(INIT_SW_BCON_ALERT, AR5K_TIMER2);
	VT_WLAN_OUT32(0, AR5K_CFP_DUR);
	VT_WLAN_OUT32(INIT_CFP_INTERVAL, AR5K_CFP_PERIOD);
	VT_WLAN_OUT32(1, AR5K_TIMER3);
	VT_WLAN_OUT32(INIT_BEACON_CONTROL, AR5K_BEACON); /* enable beacons */

	if (!local->conf->qos_enabled) 
		local->conf->aifs = 2;

	/** set turbo/base mode parameters **/
	if(local->conf->turbo_enable) {
		/** turbo mode **/
		VT_WLAN_OUT32(INIT_SLOT_TIME_TURBO, AR5K_SLOT_TIME);
		VT_WLAN_OUT32(INIT_ACK_CTS_TIMEOUT_TURBO, AR5K_TIME_OUT);
		VT_WLAN_OUT32(INIT_TRANSMIT_LATENCY_TURBO, AR5K_USEC);
		data = ((INIT_SIFS_TURBO + local->conf->aifs * INIT_SLOT_TIME_TURBO) << AR5K_IFS0_DIFS_S)
			| INIT_SIFS_TURBO;
		VT_WLAN_OUT32(data, AR5K_IFS0);
		VT_WLAN_OUT32(INIT_PROTO_TIME_CNTRL_TURBO, AR5K_IFS1);
		VT_WLAN_OUT32((VT_WLAN_IN32(PHY_BASE+(17<<2)) & ~0x7F) |
			      0x38, PHY_BASE+(17<<2));
		VT_WLAN_OUT32(PHY_FC_SERVICE_ERR | PHY_FC_TX_UNDER_ERR |
			      PHY_FC_ILLLEN_ERR | PHY_FC_ILLRATE_ERR |
			      PHY_FC_PARITY_ERR | PHY_FC_TIMING_ERR |
			      0x2020 | PHY_FC_TURBO_MODE | PHY_FC_TURBO_SHORT,
			      PHY_FRAME_CONTROL);
	} else {
		/** base mode **/
		VT_WLAN_OUT32(INIT_SLOT_TIME, AR5K_SLOT_TIME);
		VT_WLAN_OUT32(INIT_ACK_CTS_TIMEOUT, AR5K_TIME_OUT);
		VT_WLAN_OUT32(INIT_TRANSMIT_LATENCY, AR5K_USEC);
		data = ((INIT_SIFS + local->conf->aifs * INIT_SLOT_TIME) << AR5K_IFS0_DIFS_S)
			| INIT_SIFS;
		VT_WLAN_OUT32(data, AR5K_IFS0);
		VT_WLAN_OUT32(INIT_PROTO_TIME_CNTRL, AR5K_IFS1);
		VT_WLAN_OUT32((VT_WLAN_IN32(PHY_BASE+(17<<2)) & ~0x7F) |
			      0x1C, PHY_BASE+(17<<2));
		VT_WLAN_OUT32(PHY_FC_SERVICE_ERR | PHY_FC_TX_UNDER_ERR | 
			      PHY_FC_ILLLEN_ERR | PHY_FC_ILLRATE_ERR |
			      PHY_FC_PARITY_ERR | PHY_FC_TIMING_ERR | 0x1020,
			      PHY_FRAME_CONTROL);
	}

	if(local->chipset_init)
		ret = local->chipset_init(local);

	channel_ieee = turbo_enable ? turbo_channel : channel;

	/* TODO: set IEEE channel numbers by reg domain */

	/* TODO: set power and antenna values if they were not set by the INI */

	wdev->phy.freq.e = 0;
	wdev->phy.freq.m = channel_ieee;
	vt_ar5k_set_channel(wdev, &wdev->phy.freq);

	/* XXX trouble? */
	vt_ar5k_set_network(wdev, WIRELESS_DISABLE_NETWORK);

	VEXIT();

	return(ret);
}

u32 vt_ar5k_bitswap(u32 val, int bit_count) 
{
	u32 retval;
	u32 bit;
	int i;
	INCHK
	VENTER();

	retval = 0;
	
	for (i = 0; i < bit_count; i++) {
		bit = (val >> i) & 1;
		retval = (retval << 1) | bit;
	}
	
	VEXIT();

	return retval;
}

int vt_ar5k_eeprom_read_mac(vt_local_t *local, u8 *omac)
{
	struct net_device *dev = (struct net_device*) local->dev;
	u32 total = 0, offset;
	u16 data = 0;
	int octet = 0, ret = 1;
	u8 mac[WIRELESS_ADDR_SIZE];
        INCHK
	VENTER();

	/** called from an early stage... be careful **/
	if(dev == NULL) {
		VPRINT("NULL dev pointer.\n");
		return(ret);
	}

	if(omac == NULL) {
		VPRINT("NULL mac pointer.\n");
		return(ret);
	}

	if(local->eeprom_read == NULL) {
		VPRINT("NULL eeprom read funtion.\n");
		return(ret);
	}

	if (VT_WLAN_IN32(AR5K_CFG) & AR5K_CFG_EEBS) {
		VPRINT("eeprom is busy.\n");
		memset(&mac, 0, WIRELESS_ADDR_SIZE);
		return(ret);
	}

	/* hack to let it work with ar5210 + ar5211 */
	if ((ret = local->eeprom_read(local, 0x20, &data)) != 0) 
		return(ret);

	for(offset = 0x1f; offset >= 0x1d; offset--) {
		if ((ret = local->eeprom_read(local, offset, &data)) != 0)
			return(ret);

		total += data;
		mac[octet + 1] = data & 0xff;
		mac[octet] = data >> 8;
		octet += 2;
	}

	memcpy(omac, mac, WIRELESS_ADDR_SIZE);

	if ((!total) || total == (3*0xffff)) {
		VPRINT("total==%d\n", total);
		return(ret); 
	}
	
	ret = 0;

	VEXIT();

	return(ret);
}

void vt_ar5k_free_rx_queue(vt_local_t *local)
{
	struct net_device *dev = (struct net_device*)local->dev;
	u32 tmp_cr;
        INCHK
	VENTER();
	
	/* disable RX descriptors */
	tmp_cr = VT_WLAN_IN32(AR5K_CR) & (~AR5K_CR_RXE);
	VT_WLAN_OUT32(tmp_cr, AR5K_CR);

	/* write null pointer to RXDP register (just to make full clearing) */
	VT_WLAN_OUT32(0, AR5K_RXDP);
	
	if(local->rx_desc_size)
		pci_free_consistent(local->pdev, local->rx_desc_size, local->rx_desc, local->rx_desc_dma);
	if(local->rx_buf_size)
		pci_free_consistent(local->pdev, local->rx_buf_size, local->rx_buf, local->rx_buf_dma);

	VEXIT();
}

int vt_ar5k_setup_rx_queue(vt_local_t *local)
{
	struct net_device *dev = (struct net_device*)local->dev;
	ar5k_desc_t *ptr = NULL;
	ar5k_buf_t *buf;
	dma_addr_t desc_addr, buf_addr;
	int i = 0;
        INCHK
	VENTER();

	local->rx_count = MAX_RX_DESC_BUF;

	/* rx buffer */
	local->rx_buf_size = local->rx_count * sizeof(ar5k_buf_t);
	local->rx_buf = pci_alloc_consistent(local->pdev, 
					     local->rx_buf_size, 
					     &local->rx_buf_dma);
	
	
	if(!local->rx_buf_dma) {
		local->rx_buf_size = 0;
		VDEBUG ("local->rx_buf_dma == NULL\n");
		return -ENOMEM;
	}
	printk("memoria dedicada al buffer dma: %u bytes",local->rx_buf_size);
	memset(local->rx_buf, 0, local->rx_buf_size);
	
	/* rx descriptor list */
	local->rx_desc_size = ((local->rx_count * sizeof(ar5k_desc_t)) 
			       + AR5K_DESC_ALIGN - 1) & ~(AR5K_DESC_ALIGN - 1);

	local->rx_desc = pci_alloc_consistent(local->pdev, 
					      local->rx_desc_size, 
					      &local->rx_desc_dma);
	
	if(!local->rx_desc_dma) {
		pci_free_consistent(local->pdev, local->rx_buf_size, local->rx_buf, local->rx_buf_dma);
		local->rx_buf_size = 0;
		local->rx_desc_size = 0;
		VDEBUG ("local->rx_desc_dma == NULL\n");
		return(-ENOMEM);
	}

	memset(local->rx_desc, 0, local->rx_desc_size);

	/* init buffer list */
	ptr = local->rx_first = local->rx_desc;
	buf = local->rx_buf;
	desc_addr = local->rx_desc_dma;
	buf_addr = local->rx_buf_dma;

	for(i = 0; i < local->rx_count; i++) {
		ptr->s.rx_compl_stat.done = 0;
		ptr->u.rx.buf_len = sizeof(ar5k_buf_t);
		ptr->u.rx.inter_req = 1;

		/* set receive buffer pointer */
		ptr->data = buf;
		ptr->buffer_ptr = buf_addr;

		desc_addr = desc_addr + sizeof(ar5k_desc_t);
		buf_addr = buf_addr + sizeof(ar5k_buf_t);

		if(i < (local->rx_count - 1)) {
			ptr->next = ptr + 1;
			ptr->link_ptr = desc_addr;
		} else {
			ptr->next = local->rx_first;
			ptr->link_ptr = local->rx_desc_dma; /* rx ring */
		}
		buf++;
		ptr++;
	}

	VT_WLAN_OUT32((u32)local->rx_desc_dma, AR5K_RXDP); /* set rx descriptor */
	VDEBUG ("set dma rx descriptor 0x%08x\n", (u32)local->rx_desc_dma);

	VEXIT();

	return(0);
}

irqreturn_t vt_ar5k_interrupt(int irq, void *dev_instance, struct pt_regs *regs) 
{
	struct net_device *dev = (struct net_device*)dev_instance;
	vt_local_t *local = (vt_local_t *)dev->priv;

	spin_lock(&local->lock);

	if(VT_WLAN_IN16(AR5K_INTPEND) & AR5K_INTPEND_TRUE) {
		u32 int_reg;
		
		VENTER();

		VT_WLAN_OUT32(AR5K_IER_DISABLE, AR5K_IER); /* disable interrupt */
		int_reg = VT_WLAN_IN16(AR5K_ISR);
		VDEBUG("real irq: 0x%08x\n", int_reg);

		local->int_count++;
	
		if(int_reg == AR5K_IMR_X) {
			VDEBUG("[%ld] interrupt... card removed (mask 0x%08x)\n", 
			       local->int_count, local->int_mask);
			local->nic_ready = 0;
			vt_ar5k_stop(dev);
			spin_unlock(&local->lock);
			RETURN_IRQ_HANDLED;
		}
	
		if(int_reg & AR5K_IMR_DPERR) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_DPERR\n", 
			       local->int_count);
		}
	
		if(int_reg & AR5K_IMR_SSERR) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_SSERR\n", 
			       local->int_count);
		}
	
		if(int_reg & AR5K_IMR_MCABT) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_MCABT\n", 
			       local->int_count);
		}
		if(int_reg & AR5K_IMR_RXNOPKT) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_RXNOPKT\n", 
			       local->int_count);
		}
	
		if(int_reg & AR5K_IMR_TXERR ||
		   int_reg & AR5K_IMR_TXOK) {
			if(int_reg & AR5K_IMR_TXERR) {
				VDEBUG("[%ld] interrupt... AR5K_IMR_TXERR\n", 
				       local->int_count);
				local->stats.tx_errors++;
			} else { 
				VDEBUG("[%ld] interrupt... AR5K_IMR_TXOK\n", 
				       local->int_count);
				local->stats.tx_packets++;
			}
#ifdef USE_SOFTIRQ
			VT_TASKLET_SCHEDULE(&local->tx_interrupt_tasklet);
#else
			vt_ar5k_tx_interrupt(local);
#endif
		}
	
		if(int_reg & AR5K_IMR_RXORN) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_RXORN\n", 
			       local->int_count);

			local->stats.rx_over_errors++;
		}
	
		if(int_reg & AR5K_IMR_RXEOL) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_RXEOL\n", 
			       local->int_count);
	
			local->stats.rx_packets++;
			local->stats.rx_dropped++;
		}
	
		if(int_reg & AR5K_IMR_RXERR) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_RXERR\n", 
			       local->int_count);
	
			local->stats.rx_errors++;
		}
	
		if(int_reg & AR5K_IMR_RXOK) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_RXOK\n", 
			       local->int_count);
#ifdef USE_SOFTIRQ
			VT_TASKLET_SCHEDULE(&local->rx_tasklet);
#else
			vt_ar5k_rx(local);
#endif
		}
	
		if(int_reg & AR5K_IMR_SWI) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_SWI\n", 
			       local->int_count);
		}
	
#ifdef USE_GPIO
		if(int_reg & AR5K_IMR_GPIO) {
			VDEBUG("[%ld] interrupt... AR5K_IMR_GPIO\n", 
			       local->int_count);
	
			vt_ar5k_gpio_get(local, 0) ? 
				vt_ar5k_gpio_int(local, 0) :
				vt_ar5k_gpio_int(local, 1);
		}
#endif
	
		VT_WLAN_OUT32(local->int_mask, AR5K_IMR); /* reset interrupt mask */
		VT_WLAN_OUT32(AR5K_IER_ENABLE, AR5K_IER); /* enable interrupt */

		VEXIT();
	} else {
		/* interrupt does not belong to us */
		spin_unlock(&local->lock);
		RETURN_IRQ_NONE; 
	}

	spin_unlock(&local->lock);

	RETURN_IRQ_HANDLED;
}

int vt_ar5k_frame_rx(vt_local_t *local, ar5k_desc_t *desc, int len)
{
	int ret = 0;
	struct sk_buff *skb = NULL;
	struct wireless_device *wdev = (struct wireless_device*)local->wdev;
	u8 *data,*aux=NULL;
	int i,j;
        unsigned long flags;
	
        INCHK;VENTER();
	data = desc->data->buffer;
//        printk("entrando a decap...\n");
	if(wdev->decap && 
	   ((skb = wdev->decap(wdev, data, &len)) != NULL)
	   ) {
		/* receive packet */
	i=skb->len;
//	printk("despues de decap: %d bytes\n",i);
//	printk("proto: %x, datalen %d\n",skb->protocol,skb->data_len);
/*        LINEA();
	aux=kmalloc(i,GFP_KERNEL);
	memcpy(aux,skb->data,i);*/
	j=(skb->len);
//	for(i=0;i<j;i++)
//	  printk("%02x ",aux[i]);
#if USE_SOFTIRQ
//		printk("netif:%d\n",
		netif_rx_ni(skb);
#else
		printk("iria netif_rx(skb);");
#endif

		/* update byte count */
		local->stats.rx_bytes += skb->len;//ret;
//		printk("paso!!!!\n");
	} else {
//		printk("drop!!!!\n");
		local->stats.rx_dropped++; 
	}
/*        LINEA();	
	if(aux!=NULL) 
	  kfree(aux);*/

	return(0);
}

#ifdef VT_DEBUG
#if VT_DEBUG > 1
static void inline vt_ar5k_dump(char *info, u8 *data, u32 datalen) 
{
	int i;

	VENTER();

	printk(KERN_DEBUG "%s data [%d]: 0x", info, datalen);
	for(i = 0; i < datalen; i++) {
		printk("%02x", data[i]);
	}
	printk("\n");

	VEXIT();
}
#endif

#if VT_DEBUG < 3
static void vt_ar5k_dump_rx_frame(vt_local_t *local, int id, 
				  ar5k_desc_t *desc)
{
	u32 datalen1;
	u8 i;
	struct net_device *dev = (struct net_device*) local->dev;
	wlan_frame_hdr_t *hdr;

	if(desc == NULL)
		return;

	VPRINT("rx (int_count=%lu,id=%u,done=%u,len=%u,rate=%u,sig=%u,antenna=%u,more=%u,timestamp=%u)\n", 
	       local->int_count,
	       id, 
	       desc->s.rx_compl_stat.done,
	       desc->s.rx_compl_stat.data_len,
	       desc->s.rx_compl_stat.rcv_rate,
	       desc->s.rx_compl_stat.rcv_sig_strength,
	       desc->s.rx_compl_stat.rcv_antenna,
	       desc->s.rx_compl_stat.more,
	       desc->s.rx_compl_stat.rcv_timestamp);
	
	hdr = (wlan_frame_hdr_t*)desc->data->buffer;

	if(hdr == NULL || 
	   desc->s.rx_compl_stat.data_len < sizeof(wlan_frame_hdr_t))
		{
		if(hdr==NULL) printk("el buffer es NULL!");
		else printk("tam header %d, len datos %d",sizeof(wlan_frame_hdr_t),
			desc->s.rx_compl_stat.data_len);
		return;
		}

//	wireless_ieee80211_dump_hdr("frame hdr", (u8*)hdr, desc->s.rx_compl_stat.data_len);

//#if VT_DEBUG > 1
/*
        printk("MAC header:");
	printk("%02x%02x|",(u8*)desc->data->buffer[0],(u8*)desc->data->buffer[1]); //FC
	printk("%02x%02x|",(u8*)desc->data->buffer[2],(u8*)desc->data->buffer[3]); //Dur-ID
        for(i=0;i<6;i++)
	  printk("%02x",(u8*)desc->data->buffer[i+4]);  //AD1
	printk("|");	  
        for(i=0;i<6;i++)
	  printk("%02x",(u8*)desc->data->buffer[i+10]);  //AD2
	printk("|");	  
        for(i=0;i<6;i++)
	  printk("%02x",(u8*)desc->data->buffer[i+16]);  //AD3 
	printk("|");
	printk("%02x%02x|",(u8*)desc->data->buffer[22],(u8*)desc->data->buffer[23]); //SEQ
        for(i=0;i<6;i++)
	  printk("%02x",(u8*)desc->data->buffer[i+24]);  //AD4 
        printk("|\n");

	if(desc->s.rx_compl_stat.data_len>30)
	{
	datalen1=desc->s.rx_compl_stat.data_len-30; 
	printk("frame body: [%d]",datalen1);
	for(i = 0; i < datalen1; i++) {
		printk("%02x ", (u8*)desc->data->buffer[i+30]);  //lo que queda, incluyendo FCS
	}
	printk("\n");
	}	     
*/		     

//#endif

}
#endif
#endif


int chequea_ack(ar5k_desc_t *desc)
{
 u8 *fcsb=(u8*)desc->data->buffer;
 u32 fcs;
 u16 i;
 wlan_frame_hdr_t *hdr=(wlan_frame_hdr_t *)desc->data->buffer; 
 if(hdr->control.type==1 && hdr->control.subtype==0xD)
 {
 i=((fcsb[10]<<8)&0xff00)|fcsb[11];
 fcs=(i<<16)&0xffff0000;
 i=((fcsb[12]<<8)&0xff00)|fcsb[13];
 fcs=fcs|i;
// for(i=0;i<14;i++)
//  printk("%02x.",fcsb[i]);   
   return 1;
   }
 else return 0;
 
}
void muestra_todo(u8 *buf,int len)
{
int i;
//printk("raw data:\n");
 for(i=0;i<len;i++)
 {
//  printk("%02x",buf[i]);
 } 
//printk("\n");
}
void vt_ar5k_rx(vt_local_t *local)
{
	struct net_device *dev = (struct net_device*) local->dev;
	ar5k_desc_t *desc;
	unsigned long flags;
	int i = 0;
        INCHK
	VENTER();
        
	/* disable rx */
	VT_WLAN_OUT32((VT_WLAN_IN32(AR5K_DIAG_SW) | AR5K_DIAG_RX_DIS), AR5K_DIAG_SW);
//	printk("posible recepcion...\n");
	desc = local->rx_desc;
	for(i = 0; i < local->rx_count; i++) {
		int len = desc->s.rx_compl_stat.data_len;

#ifdef VT_DEBUG		
#if VT_DEBUG == 2
//		if(len >= sizeof(wlan_frame_hdr_t))
//			vt_ar5k_dump_rx_frame(local, i, desc);
#endif
#endif

		/* fragmentation is not yet supported */
		if(desc->s.rx_compl_stat.more) {
			printk("fragmentation not supported!\n");
			local->rx_drop = 1;
			memset(&desc->s, 0, sizeof(ar5k_rx_status_t));
			desc++;
			continue;
		}

		/* drop last packet of a fragmented frame */
		if(local->rx_drop) {
//			printk("dropping packet!\n");
			local->stats.rx_dropped++; 
			local->rx_drop = 0;
			memset(&desc->s, 0, sizeof(ar5k_rx_status_t));
			desc++;
			continue;
		}

		if(desc->s.rx_compl_stat.done) {
//			printk("rx done, len %d\n",len);
			if ((len <= RX_BUF_LEN) &&
			  (len > sizeof(wlan_frame_hdr_t))) {
				   					   
				if(desc->s.rx_compl_stat.frame_rcv_ok) {
#ifdef VT_DEBUG
#if VT_DEBUG < 3
					//vt_ar5k_dump_rx_frame(local, i, desc);
//					muestra_todo(desc->data->buffer,len);
#endif
#endif
					vt_ar5k_frame_rx(local, desc, len);

					local->stats.rx_packets++;

					
				} else {
					local->stats.rx_dropped++; /* frame not ok */
//					printk("frame not ok!");
				}
			}
//			if(chequea_ack(desc))
//			  printk("Recibi ACK!\n");
			memset(&desc->s, 0, sizeof(ar5k_rx_status_t));
			desc->u.rx.buf_len = sizeof(ar5k_buf_t);
			desc->u.rx.inter_req = 1;
		}
		desc++;
	}
	
	/* re-enable rx */
	VT_WLAN_OUT32((VT_WLAN_IN32(AR5K_DIAG_SW) & (~AR5K_DIAG_RX_DIS)), AR5K_DIAG_SW);
        
	VEXIT();
}

void vt_ar5k_tx_interrupt(vt_local_t *local)
{
	struct net_device *dev = (struct net_device*)local->dev;
	unsigned long flags;
	ar5k_desc_t *desc;
        INCHK;VENTER();
	spin_lock_irqsave(&local->lock, flags);

	/* disable tx (XXX) */
	VT_WLAN_OUT32((VT_WLAN_IN32(AR5K_DIAG_SW) | AR5K_DIAG_TX_DIS), AR5K_DIAG_SW);

	/* disable transmit bit */
	VT_WLAN_OUT32((VT_WLAN_IN32(AR5K_CR) | AR5K_CR_TXD0), AR5K_CR); 
	
	desc = local->tx_desc;

	if(desc) {
		VDEBUG("[%ld] tx (done=%u,seq=%u,acksig=%u,ok=%u,eretries=%u,underrun=%u,filtered=%u,srcount=%u,lrcount=%u,timestamp=%u)\n", 
		       local->int_count,
		       desc->s.tx_compl_stat.done,
		       desc->s.tx_compl_stat.seq_num,
		       desc->s.tx_compl_stat.ack_sig_strength,
		       desc->s.tx_compl_stat.frm_xmit_ok,
		       desc->s.tx_compl_stat.excessive_retries,
		       desc->s.tx_compl_stat.fifo_underrun,
		       desc->s.tx_compl_stat.filtered,
		       desc->s.tx_compl_stat.sort_retry_count,
		       desc->s.tx_compl_stat.long_retry_count,
		       desc->s.tx_compl_stat.send_timestamp);
	} else {
		VPRINT("tx desc == NULL!\n");
	}
	
	/* reset tx status */
	memset(&desc->s, 0, sizeof(ar5k_tx_status_t));
	desc->u.tx.inter_req = 0;
	
	/* reset tx buffer
	 * XXX there is a known problem with some network drivers
	 *     that you can find data fragments from older packets
	 *     in received frames... 
	 */
	memset(local->tx_buf, 0, local->tx_buf_size);

	local->tx_active = 0;

	/* re-enable tx */
	VT_WLAN_OUT32((VT_WLAN_IN32(AR5K_DIAG_SW) & (~AR5K_DIAG_TX_DIS)), AR5K_DIAG_SW);
	
#ifdef USE_SOFTIRQ
	VT_TASKLET_SCHEDULE(&local->tx_ok_tasklet);
#else
	if(local->wdev->transmit_ok)
		local->wdev->transmit_ok(local->wdev);
#endif
	spin_unlock_irqrestore(&local->lock, flags);
}

void vt_ar5k_free_tx_queue(vt_local_t *local)
{
        INCHK
	VENTER();

	if(local->tx_desc_size)
		pci_free_consistent(local->pdev, local->tx_desc_size, local->tx_desc, local->tx_desc_dma);
	if(local->tx_buf_size)
		pci_free_consistent(local->pdev, local->tx_buf_size, local->tx_buf, local->tx_buf_dma);

	VEXIT();
}

int vt_ar5k_setup_tx_queue(vt_local_t *local)
{
        INCHK
	VENTER();

	/* tx0 buffer */
	local->tx_buf_size = mtu + sizeof(wlan_frame_hdr_t) + 4;
	local->tx_buf = pci_alloc_consistent(local->pdev, 
					     local->tx_buf_size, 
					     &local->tx_buf_dma);

	if(!local->tx_buf_dma) {
		local->tx_buf_size = 0;
		return(-ENOMEM);
	}

	memset(local->tx_buf, 0, local->tx_buf_size);

	/* tx descriptor */
	local->tx_desc_size = sizeof(ar5k_desc_t);
	local->tx_desc = pci_alloc_consistent(local->pdev, 
					      local->tx_desc_size, 
					      &local->tx_desc_dma);

	if(!local->tx_desc_dma) {
		pci_free_consistent(local->pdev, local->tx_buf_size, local->tx_buf, local->tx_buf_dma);
		local->tx_buf_size = 0;
		local->tx_desc_size = 0;
		return(-ENOMEM);
	}

	memset(local->tx_desc, 0, local->tx_desc_size);

	local->tx_desc->link_ptr = 0; /* local->tx_desc_dma; */
	local->tx_desc->buffer_ptr = local->tx_buf_dma;
	local->tx_desc->data = local->tx_buf;

	VEXIT();

	return(0);
}

#ifdef VT_DEBUG
static void vt_ar5k_dump_tx_frame(vt_local_t *local, ar5k_desc_t *desc)
{
	struct net_device *dev = (struct net_device*) local->dev;
	wlan_frame_hdr_t *hdr;

	if(desc == NULL)
		return;

	VPRINT("tx (int_count=%lu,frm_len=%u,hdr_len=%u,xmit_rate=%u,ant=%u,clr=%u,type=%u,intr=%u,buf_len=%u,more=%u,buf=%lu)\n",
	       local->int_count,
	       desc->u.tx.frame_len,
	       desc->u.tx.hdr_len,
	       desc->u.tx.xmit_rate,
	       desc->u.tx.ant_mode_xmit,
	       desc->u.tx.clear_dest_mask,
	       desc->u.tx.frm_type,
	       desc->u.tx.inter_req,
	       desc->u.tx.buf_len,
	       desc->u.tx.more,
	       (unsigned long)desc->buffer_ptr);
	
	hdr = (wlan_frame_hdr_t*)desc->data->buffer;

	if(hdr == NULL || 
	   desc->u.tx.frame_len < sizeof(wlan_frame_hdr_t))
		return;

//	wireless_ieee80211_dump_hdr("tx frame", (u8*)hdr, desc->u.tx.frame_len);

#if VT_DEBUG > 1
	vt_ar5k_dump("tx frame", (u8*)desc->data->buffer, 
		     desc->u.tx.frame_len);

#endif

}
#endif

int vt_ar5k_tx(struct net_device *dev, struct sk_buff *skb, int encap)
{
	vt_local_t *local = (vt_local_t *)dev->priv;
	struct wireless_device *wdev = (struct wireless_device*)local->wdev;
	unsigned long flags;
	ar5k_desc_t *desc = NULL;
	u8 *buf;
	u8 *datatemp;
	u8 lengi;
	
	int ret = 0;
        INCHK
	VENTER();
//	printk("a transmitir\n");
	spin_lock_irqsave(&local->lock, flags);

	if(skb->len < ETH_HLEN || skb->len > mtu) {
		ret = -EINVAL;
		goto out;
	}

	if(local->tx_active) {
		ret = -EAGAIN;
		goto out;
	}

	/* clear tx buffer */
	memset(local->tx_buf, 0, local->tx_buf_size);

	desc = local->tx_desc;
        /*aca hay algo que debo definir, y es que el frame header tiene
	longitudes distintas dependiendo de si el frame es tipo data, management o control
	header len: type data: 30
		    type management:  24
		    type control: rts=20,cts=14,ack=14,pspoll=20,cfend=20,cfe+ack=20			 
	*/
	/* init atheros descriptor */
	if(encap==WIRELESS_ENCAP)
	{
	desc->u.tx.frame_len =  (24//((sizeof(wlan_frame_hdr_t)) 
				+ 6 //ethersnap
				+ skb->len
				- 12 //borrando los mac al principio
				+4 //FCS ??
				+ AR5K_DESC_ALIGN - 1) & ~(AR5K_DESC_ALIGN - 1);
				//el tamao del frame es siempre multiplo de 4
	desc->u.tx.hdr_len =  (24//(sizeof(wlan_frame_hdr_t) 
			      + AR5K_DESC_ALIGN - 1) & ~(AR5K_DESC_ALIGN - 1);
			      //el tamao del header es siempre multiplo de 4
	}
	else
	{
	desc->u.tx.frame_len =  ((sizeof(wlan_frame_hdr_t)) 
				+ skb->len
				+ AR5K_DESC_ALIGN - 1) & ~(AR5K_DESC_ALIGN - 1);
				//el tamao del frame es siempre multiplo de 4
	desc->u.tx.hdr_len =  (sizeof(wlan_frame_hdr_t) 
			      + AR5K_DESC_ALIGN - 1) & ~(AR5K_DESC_ALIGN - 1);
			      //el tamao del header es siempre multiplo de 4
	}
	desc->u.tx.xmit_rate = AR5K_XMIT_RATE_OFDM_24MBPS; /* 24Mbps base rate */
	desc->u.tx.frm_type = AR5K_FRAME_NORMAL;
	desc->u.tx.inter_req = 1;
	desc->u.tx.buf_len = desc->u.tx.frame_len - 4; /* ??? buf_len > frame_len doesn't work! */
	//ademas, debe ser que framelen-4 es justamente la long de datos menos FCS
	desc->u.tx.more = 0;
	
	lengi=skb->len-12;
        /*
	if( (datatemp=kmalloc(lengi,GFP_KERNEL))!=NULL)
	{
//        printk("datos crudos a transmitir:\n");	
	memcpy(datatemp,skb->data+12,skb->len-12);
//	for(i=0;i<lengi;i++)
//	  printk("%02x ",datatemp[i]);
//	printk("\n");
	kfree(datatemp);
	} */
	
	if(encap == WIRELESS_ENCAP) {
		if(wdev->encap)
			ret = wdev->encap(wdev, skb, 
					  (u8*)local->tx_buf,
					  local->tx_buf_size);
		if(!ret) 
			goto out;
		else
			ret = 0;
	} else {
		/* copy payload */
		buf = (u8*)local->tx_buf;
		memcpy(buf, skb->data, skb->len);
	}
//	crc32((u8*)local->tx_buf,desc->u.tx.buf_len);
#ifdef VT_DEBUG
	vt_ar5k_dump_tx_frame(local, desc);
#endif
	lengi=desc->u.tx.frame_len;
/*	if( (datatemp=kmalloc(lengi,GFP_KERNEL))!=NULL)	{
//	printk("lo que realmente se transmite: (%d bytes)\n",lengi);
	memcpy(datatemp,(u8*)local->tx_buf,lengi);
//	for(i=0;i<lengi;i++)
//	  printk("%02x ",datatemp[i]);
//	printk("\n");
	kfree(datatemp);
	}               */
       
	/* set tx descriptor */
	VT_WLAN_OUT32((u32)local->tx_desc_dma, AR5K_TXDP0);

	/* enable transmit bit */
	VT_WLAN_OUT32((VT_WLAN_IN32(AR5K_CR) | AR5K_CR_TXE0), AR5K_CR);

	local->tx_active = 1;
//        printk("en teoria, tx ok!\n");
 out:
	spin_unlock_irqrestore(&local->lock, flags);

/*       	if(skb)
        {
//            LINEA();
	     dev_kfree_skb(skb);
             }
*/
	VEXIT();

	return(ret);
}

int vt_ar5k_start_xmit(struct sk_buff *skb, struct net_device *dev) 
{
	vt_local_t *local = (vt_local_t *)dev->priv;
//	printk("imprimira?\n");
        INCHK;VENTER();
	if(local->wdev->mac.associated)
		return(vt_ar5k_tx(dev, skb, WIRELESS_ENCAP));
	else
		return(0);
}

int vt_ar5k_nic_open(struct net_device *dev, u32 rx_filter) 
{
	int ret = 0;
	vt_local_t *local = (vt_local_t *)dev->priv;
	u32 tmp_cr;
        INCHK
	VENTER();

	if (request_irq(dev->irq, vt_ar5k_interrupt, SA_SHIRQ, dev->name,
			dev)) {
		VPRINT("request_irq failed\n");
		return(-EINVAL);
	}

	udelay(1000);
	VT_WLAN_OUT32(PHY_ACTIVE_EN, PHY_ACTIVE);
	udelay(1000);
	
	if(!local->change_count++) {		
		/* init rx ring */
		if((ret = vt_ar5k_setup_rx_queue(local)) != 0)
			return(ret);
		
		/* setup tx ring */
		if((ret = vt_ar5k_setup_tx_queue(local)) != 0)
			return(ret);

		/* get and store interrupt mask */
		VT_WLAN_OUT32(INIT_INTERRUPT_MASK, AR5K_IMR);
		local->int_mask = VT_WLAN_IN16(AR5K_IMR);
		
		/* enable rx */
		/*	tmp_cr = VT_WLAN_IN32(AR5K_CR) | ~AR5K_CR_RXD | AR5K_CR_RXE; */
		tmp_cr = AR5K_CR_RXE;
		VT_WLAN_OUT32(tmp_cr, AR5K_CR); /* enable receive bit */
		VT_WLAN_OUT32(VT_WLAN_IN32(AR5K_DIAG_SW) & ~(AR5K_DIAG_RX_DIS), AR5K_DIAG_SW); 
		VT_WLAN_OUT32(rx_filter, AR5K_RX_FILTER);
		
		/* enable tx */
		VT_WLAN_OUT32(VT_WLAN_IN32(AR5K_DIAG_SW) & ~(AR5K_DIAG_TX_DIS), AR5K_DIAG_SW); 
	}

	ret = vt_ar5k_set_network(local->wdev, WIRELESS_FIND_NETWORK);

	/* enable interrupts */
	VT_WLAN_OUT32(AR5K_IER_ENABLE, AR5K_IER);

	netif_device_attach(dev);
	netif_start_queue(dev);

	local->nic_active = 1;

	VEXIT();

	return(ret);
}

void vt_ar5k_nic_stop(struct net_device *dev)
{
	vt_local_t *local = (vt_local_t *)dev->priv;
        INCHK
	VENTER();

	if(netif_running(dev)) {
		netif_stop_queue(dev);
		netif_device_detach(dev);
	}

	/* disable interrupts */
	VT_WLAN_OUT32(AR5K_IER_DISABLE, AR5K_IER);

	/* disable phy */
	VT_WLAN_OUT32(PHY_ACTIVE_DIS, PHY_ACTIVE);
	udelay(1000);

	free_irq(dev->irq, dev);

	local->nic_active = 0;

	vt_ar5k_set_network(local->wdev, WIRELESS_DISABLE_NETWORK);

	VEXIT();
}

int vt_ar5k_open(struct net_device *dev) 
{
	vt_local_t *local = (vt_local_t *)dev->priv;
	unsigned long flags;
	int ret = 0;
	u32 rx_filter;
        INCHK
	VENTER();

	/* check if card is ready or has been removed */
	if(!local->nic_ready || local == NULL)
		return(-ENODEV);

	VDEBUG("net_device opened...\n");

	spin_lock_irqsave(&local->lock, flags);

	/* monitor mode? */
	rx_filter = local->monitor ? AR5K_RX_PROM : local->rx_filter;
	ret = vt_ar5k_nic_open(dev, rx_filter);

	spin_unlock_irqrestore(&local->lock, flags);

	/* MOD_INC_USE_COUNT; obsolete... (replace it?) */

	VEXIT();

	return(ret);
}

int vt_ar5k_stop(struct net_device *dev) 
{
	vt_local_t *local = (vt_local_t *)dev->priv;
	unsigned long flags;
        INCHK
	VENTER();

	/* check if card is ready or has been removed */
	if(local == NULL)
		return(-ENODEV);

	VDEBUG("net_device stopped...\n");

	spin_lock_irqsave(&local->lock, flags);

	if(local->nic_ready)
		vt_ar5k_nic_stop(dev);
	else {
		if(netif_running(dev)) {
			netif_stop_queue(dev);
			netif_device_detach(dev);
		}
	}

	spin_unlock_irqrestore(&local->lock, flags);

	/* MOD_DEC_USE_COUNT; obsolete... (replace it?) */

	VEXIT();

	return(0);
}

struct net_device_stats *vt_ar5k_get_stats(struct net_device *dev) 
{
	vt_local_t *local = (vt_local_t*)dev->priv;
	VENTER();

	if(local == NULL) {
		VPRINT("priv data should not be NULL!\n");
		return(NULL);
	}

	VEXIT();
	return &local->stats;
} 

/** MODULE INIT **/
int __init vt_ar5k_init(void) 
{
	int ret = 0;

	VENTER();
	
	VINIT("%s\n", version);

#ifndef WIRELESS_MODULE	
	if((ret = init_wireless()) != 0) {
		printk("cannot initialize wireless layer!\n");
		return(ret);
	}
#endif

	if(vt_ar5k_reg_init() != 0)
		return(-EINVAL);
	
	if (pci_register_driver(&vt_ar5k_pci_drv_id) <= 0) {
		VINIT("no devices found, driver not installed.\n");
		pci_unregister_driver(&vt_ar5k_pci_drv_id);
		return -ENODEV;
	}
	
	VEXIT();
	
	return(ret);
}

/** MODULE EXIT **/
void __exit vt_ar5k_exit(void) 
{
	VENTER();
	
	VINIT("unregistering pci device\n");
	pci_unregister_driver(&vt_ar5k_pci_drv_id);

#ifndef WIRELESS_MODULE	
	exit_wireless();
#endif

	VINIT("driver unloaded\n");
	
	VEXIT();
}

module_init(vt_ar5k_init)
module_exit(vt_ar5k_exit)
