/*                                                        -*- linux-c -*-
 *   _  _ ____ __ _ ___ ____ ____ __ _ _ _ _ | 
 * .  \/  |--| | \|  |  |--< [__] | \| | _X_ | s e c u r e  s y s t e m s
 *
 * wdev.h - common wireless stack implementation for linux
 *
 * $Id: wdev.c,v 1.20 2003/12/02 17:24:39 root Exp root $
 *
 * Copyright (c) 2003, .vantronix | secure systems
 *                     and Reyk Floeter <reyk@vantronix.net> 
 * Copyright (C) 2003 por .DyC Electronica 
 * | sistemas  inalambricos Mauricio Lange dyc_ar5k Mauricio Lange 
 * <mlange@dycelectronica.com.ar>, 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/wireless.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/skbuff.h>
#include <asm/uaccess.h>

#include <linux/wdev.h>
#include <linux/ieee80211.h>
#include <vt_ar5k.h>
#include <vt_types.h>


#define DEBUG

#ifndef WIRELESS_EXT
#error "you have to enable the common wireless extensions"
#endif

#define DESCRIPTION "Common wireless stack implementation for Linux"
#define AUTHOR "Reyk Floeter <reyk@vantronix.net>, (C) 2003 by .vantronix | secure systems and Reyk Floeter Mauricio Lange <mlange@dycelectronica.com.ar>, (C) 2003 por .DyC Electronica | sistemas  inalambricos Mauricio Lange dyc_ar5k"
#define MODNAME "wdev"

/* if compiled as a standalone module */
#ifdef WIRELESS_MODULE
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_AUTHOR(AUTHOR);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif

static char *version = DESCRIPTION "\n" AUTHOR;
struct proc_dir_entry *proc_wireless;
u8 wireless_no_addr[WIRELESS_ADDR_SIZE];

#ifndef MAC2STR
#define MAC2STR(a) (a)[0], (a)[1], (a)[2], (a)[3], (a)[4], (a)[5]
#define MACSTR "%02X:%02X:%02X:%02X:%02X:%02X"
#endif

/* 
 * common functions to handle a wireless device 
 */

int veces_entrado=0;
int cuenta=0;


static int wireless_proc_phy(char *page, char **start, off_t off,
			     int count, int *eof, void *data) 
{
	char *p = page;
	struct wireless_device *wdev = (struct wireless_device*)data;

	if (off != 0) {
		*eof = 1;
		return 0;
	}
	p += sprintf(p, "Physical layer:\n");
	p += sprintf(p, "Type:      %d (FHSS:%d,DSSS:%d,OFDM:%d,MISC:%d)\n",
		     wdev->phy.type, 
		     WIRELESS_PHY_MODE_FHSS,
		     WIRELESS_PHY_MODE_DSSS,
		     WIRELESS_PHY_MODE_OFDM,
		     WIRELESS_PHY_MODE_MISC);
	p += sprintf(p, "Flags:     FHSS:%u,DSSS:%u,OFDM:%u,MISC:%u\n",
		     wdev->phy.cap.fhss,
		     wdev->phy.cap.dsss,
		     wdev->phy.cap.ofdm,
		     wdev->phy.cap.misc);
	p += sprintf(p, "Channel:   %d\n",
		     wdev->phy.freq.m);
		     
	return(p - page);
}

static int wireless_proc_mac(char *page, char **start, off_t off,
			     int count, int *eof, void *data) 
{
	char *p = page;
	struct wireless_device *wdev = (struct wireless_device*)data;

	if (off != 0) {
		*eof = 1;
		return 0;
	}
        p += sprintf(p, "BSSID ACTUAL " MACSTR,MAC2STR(wdev->mac.bssid));
	p += sprintf(p, "MAC layer:\n");
	p += sprintf(p, "Protocol:  %d (IEEE 802.11:%d,WLSEC:%d,CUSTOM:%d)\n",
		     wdev->mac.proto,
		     WIRELESS_PROTO_IEEE80211,
		     WIRELESS_PROTO_WLSEC,
		     WIRELESS_PROTO_CUSTOM);
	p += sprintf(p, "Encoding:  %d (NONE:%d,WEP:%d,WLSEC ESP:%d,CUSTOM:%d)\n",
		     wdev->mac.encoding,
		     WIRELESS_ENCODING_NONE,
		     WIRELESS_ENCODING_WEP,
		     WIRELESS_ENCODING_WLSEC_ESP,
		     WIRELESS_ENCODING_CUSTOM);
	p += sprintf(p, "Searching: %d\n",
		     wdev->mac.searching);

	return(p - page);
}

int wireless_register(struct wireless_device *wdev)
{
	struct iw_statistics *iw_stats;
	int ret = 0;

	/* need the network device */
	if(wdev->dev == NULL || 
	   wdev->set_channel == NULL || 
	   wdev->set_mode == NULL ||
	   wdev->set_network == NULL)
		return(-ENODEV);

	/* TODO: implement dynamic protocol registration */
	switch(wdev->mac.proto) {
	case WIRELESS_PROTO_IEEE80211:
		wdev->init = wireless_ieee80211_init;
		wdev->exit = wireless_ieee80211_exit;
		wdev->encap = wireless_ieee80211_encap;
		wdev->decap = wireless_ieee80211_decap;
		wdev->transmit_ok = wireless_ieee80211_transmit_ok;
		break;
	case WIRELESS_PROTO_CUSTOM:
		/* handlers provided by driver */
		break;
	default:
		printk(__FILE__ ": protocol not implemented\n");
		return(-ENOENT);
	}

	switch(wdev->mac.encoding) {
	case WIRELESS_ENCODING_WEP:
		wdev->encoding_init = wireless_wep_init;
		wdev->encoding_exit = wireless_wep_exit;
		wdev->encode = wireless_wep_encode;
		wdev->decode = wireless_wep_decode;
		break;
	case WIRELESS_ENCODING_NONE:
	case WIRELESS_ENCODING_CUSTOM:
		/* custom protocol / no host based encoding support */
		break;
	default:
		printk(__FILE__ ": encoding protocol not implemented\n");
		return(-ENOENT);
	}

	/* register netdevice */
	if(register_netdev(wdev->dev) != 0) {
		printk(__FILE__ ": register_netdev() failed!\n");
		return(-ENODEV);
	}

	/* initialize proc structure */
	if(proc_wireless == NULL) {
		ret = -ENOENT;
		goto err_dev;
	}

	wdev->proc = proc_mkdir(wdev->dev->name, proc_wireless);

	create_proc_read_entry("phy", 0, wdev->proc,
			       wireless_proc_phy, wdev);
	create_proc_read_entry("mac", 0, wdev->proc,
			       wireless_proc_mac, wdev);
	
	/* initialize protocol */
	if(wdev->init)
		if((ret = wdev->init(wdev)) != 0)
			goto err_fail;

	if(wdev->encoding_init)
		if((ret = wdev->encoding_init(wdev)) != 0)
			goto err_fail;

	if((iw_stats = 
	    kmalloc(sizeof(struct iw_statistics), GFP_KERNEL)) == NULL) {
		ret = -ENOMEM;
		goto err_fail;
	}

	memset(iw_stats, 0, sizeof(struct iw_statistics));
	wdev->iw_stats = iw_stats;

	printk(__FILE__ ": registered new wireless device '%s'!\n", 
	       wdev->dev->name);

	goto err_ok;

 err_fail:
	remove_proc_entry(wdev->dev->name, proc_wireless);		

 err_dev:
	unregister_netdev(wdev->dev);
	
 err_ok:
	return(ret);
}

void wireless_monitor(struct wireless_device *wdev, int val)
{
	if(wdev->mac.cap.monitor &&
	   wdev->set_monitor)
		wdev->set_monitor(wdev, val);
}

void wireless_unregister(struct wireless_device *wdev)
{
	/* remove proc entries */
	remove_proc_entry("phy", wdev->proc);
	remove_proc_entry("mac", wdev->proc);
	remove_proc_entry(wdev->dev->name, proc_wireless);

	/* unregister net device */
	if(wdev->dev)
		unregister_netdev(wdev->dev);

	/* free protocol */
	if(wdev->exit)
		wdev->exit(wdev); 

	/* free statistics */
	if(wdev->iw_stats)
		kfree(wdev->iw_stats);

	return;
}

/* 
 * IEEE 802.11 protocol handlers
 */

int wireless_ieee80211_init(struct wireless_device *wdev)
{
	wireless_ieee80211_t *priv;
        INCHK;VENTER();
	/* allocate private data */
	if((priv = kmalloc(sizeof(wireless_ieee80211_t), GFP_KERNEL)) == NULL)
		return(-ENOMEM);

	memset(priv, 0, sizeof(wireless_ieee80211_t));
	priv->seq++;
	wdev->mac.priv = priv;
	wdev->mac.cap.nwid = 0;
	wdev->mac.cap.essid = 1;
	wdev->mac.cap.bssid = 1;
 	priv->status = WIRELESS_IEEE80211_OFFLINE;
	/* disabled per default (should be enabled by the driver if supported):
	 * wdev->mac.cap.master 
	 * wdev->mac.cap.monitor
	 */

	return(0);
}

void wireless_ieee80211_exit(struct wireless_device *wdev)
{
	if(wdev->mac.priv == NULL) {
		printk(__FILE__ ": protocol not initialized\n");
		return;
	}

	/* free private data */
	kfree(wdev->mac.priv);

	return;
}

int wireless_ieee80211_encap(struct wireless_device *wdev, 
			     struct sk_buff *skb, 
			     void *buf, int buflen)
{
	wireless_ieee80211_t *priv = (wireless_ieee80211_t*)wdev->mac.priv;
	wlan_frame_hdr_t *frame = NULL;
	u8 *data;
	u8 ethersnap[6]={0xaa,0xaa,3,0,0,0};
/*  esto estaba para probar transmision, antes de que arreglar
    asignaciones de direccion
	const u8 dire[6]={0,0x9,0x5b,0x0c,0x85,0x33};
	const u8 bssid1[6]={0x72,0x2d,0x26,0x0c,0x75,0xf6};
	const u8 sa[6]={0,0x30,0xab,0x1a,0xb6,0xbc};
*/
        INCHK;VENTER();
	if(wdev->mac.priv == NULL) {
		printk(__FILE__ ": protocol not initialized\n");
		return(-EINVAL);
	}

	if(buflen < skb->len + sizeof(wlan_frame_hdr_t)) {
		printk(__FILE__ ": invalid buffer (to small)");
		return(-EINVAL);
	}

	/* XXX TODO */

	/* init 802.11 header */
	frame = (wlan_frame_hdr_t*)buf;
	frame->control.type = ieee_802_11_frame_type_Data;
	frame->control.subtype = ieee_802_11_frame_subtype_Data;
	frame->seq.num = ++(priv->seq);
	frame->control.to_ds=0;
	frame->control.from_ds=0;
	frame->control.duration=32000;
	/* ... */



	/* FIX FIX FIX */
	memcpy(&frame->address1, wdev->mac.dest_address, WIRELESS_ADDR_SIZE);
	memcpy(&frame->address2, wdev->mac.address, WIRELESS_ADDR_SIZE);
	memcpy(&frame->address3, wdev->mac.bssid, WIRELESS_ADDR_SIZE);

	/* copy payload */
	data = (u8*)buf + 24; //sizeof(wlan_frame_hdr_t);
	/* el tamao del header para data frames es siempre 30*/
	memcpy(data, ethersnap, 6);	
	memcpy(data+6, skb->data+12, skb->len-12); 

	return(skb->len + sizeof(wlan_frame_hdr_t));
}

struct sk_buff *wireless_ieee80211_monitor(struct wireless_device *wdev, 
					   void *buf, int *rlen)
{
	struct net_device *dev = (struct net_device*)wdev->dev;
	struct sk_buff *skb = NULL;
	int len;
	u8 *packet, *data;
	u32 csum; /* crc32 */
        INCHK;VENTER();
	if(wdev->mac.priv == NULL || buf == NULL || rlen == NULL) {
		*rlen = -EFAULT;
		return(NULL);
	}

	/* substract checksum */
	len = *rlen;
	len -= 4; 

	data = buf;

	if(wdev->mac.header == WIRELESS_PRISM_HDR) {
		int plen = len + sizeof(linux_wlan_ng_prism_hdr_t) - 4;
		linux_wlan_ng_prism_hdr_t phdr;

		skb = dev_alloc_skb(plen + 2); 
		skb_reserve(skb, 2);

		if(skb == NULL)
			goto drop;

		/* partially ripped from linux hostap driver ,)
		 * -> http://hostap.epitest.fi */
		memset(&phdr, 0, sizeof(linux_wlan_ng_prism_hdr_t));
#define LWNG_DID_BASE (4 | (1 << 6)) /* section 4, group 1 */
		phdr.msgcode = LWNG_DID_BASE;
		phdr.msglen = sizeof(linux_wlan_ng_prism_hdr_t);
		memcpy(phdr.devname, dev->name, sizeof(phdr.devname));
#define LWNG_SETVAL(f,i,s,l,d) \
phdr.f.did = LWNG_DID_BASE | (i << 12); phdr.f.status = s; phdr.f.len = l; \
phdr.f.data = d
		LWNG_SETVAL(hosttime, 1, 0, 4, jiffies);
		LWNG_SETVAL(mactime, 2, 0, 4, jiffies); /* le32_to_cpu(desc->s.rx_compl_stat.rcv_timestamp)); */
		LWNG_SETVAL(channel, 3, 0, 4, wdev->phy.freq.m);
		LWNG_SETVAL(rssi, 4, 1 /* no value */, 4, 0);
		LWNG_SETVAL(sq, 5, 1 /* no value */, 4, 0);
		LWNG_SETVAL(signal, 6, 0, 4, 0); /* le32_to_cpu(desc->s.rx_compl_stat.rcv_sig_strength)); */
		LWNG_SETVAL(noise, 7, 1 /* no value */, 4, 0);
		LWNG_SETVAL(rate, 8, 0, 4, 0); /* le32_to_cpu(desc->s.rx_compl_stat.rcv_rate)); */
		LWNG_SETVAL(istx, 9, 0, 4, 0);
		LWNG_SETVAL(frmlen, 10, 0, 4, len);
#undef LWNG_SETVAL
#undef LWNG_DID_BASE

		/* send socket packetfer */
		packet = skb_put(skb, plen);
		memset(packet, 0, plen);
		memcpy(packet, &phdr, sizeof(linux_wlan_ng_prism_hdr_t)); /* prepend prism header */
		memcpy(packet + sizeof(linux_wlan_ng_prism_hdr_t), data, len);
		memcpy(&csum, packet + plen, 4); /* crc32 */
	} else {
		skb = dev_alloc_skb(len + 2); 
		skb_reserve(skb, 2); 

		if(skb == NULL)
			goto drop;

		/* send socket buffer */
		packet = skb_put(skb, len);
		memset(packet, 0, len);
		memcpy(packet, data, len);
		memcpy(&csum, packet + len, 4); /* crc32 */
	}

	skb->dev = dev;
	skb->mac.raw = packet; /* address 1 */				
	skb->pkt_type = PACKET_HOST;
	skb->protocol = __constant_htons(ETH_P_802_2);
	skb->csum = csum;

	*rlen = skb->len;
	return(skb);

 drop:
//	printk("tire paquete\n");
	*rlen = -EINVAL;
	return(NULL);
}

int wireless_ieee80211_parse_elements(struct wireless_device *wdev,
				      u8 *buf, int len,
				      wlan_elems_t *elems)
{
	size_t left = len;
	u8 *pos = buf;
	int unknown = 0;
	INCHK;VENTER();
	
	memset(elems, 0, sizeof(*elems));

	while (left >= 2) {
		u8 id, elen;

		id = *pos++;
		elen = *pos++;
		left -= 2;

		if (elen > left) {
#ifdef DEBUG
			printk(__FILE__ ": IEEE 802.11 elems parse failed "
			       "(id=%d elen=%d left=%d len=%d)\n", id, elen, left, len);
#endif
			return(1);
		}

		switch (id) {
		case IEEE80211_EID_SSID:
			elems->ssid = pos;
			elems->ssid_len = elen;
			break;
		case IEEE80211_EID_SUPP_RATES:
			elems->supp_rates = pos;
			elems->supp_rates_len = elen;
			break;
		case IEEE80211_EID_FH_PARAMS:
			elems->fh_params = pos;
			elems->fh_params_len = elen;
			break;
		case IEEE80211_EID_DS_PARAMS:
			elems->ds_params = pos;
			elems->ds_params_len = elen;
			break;
		case IEEE80211_EID_CF_PARAMS:
			elems->cf_params = pos;
			elems->cf_params_len = elen;
			break;
		case IEEE80211_EID_TIM:
			elems->tim = pos;
			elems->tim_len = elen;
			break;
		case IEEE80211_EID_IBSS_PARAMS:
			elems->ibss_params = pos;
			elems->ibss_params_len = elen;
			break;
		case IEEE80211_EID_CHALLENGE:
			elems->challenge = pos;
			elems->challenge_len = elen;
			break;
		default:
#ifdef DEBUG
			printk(__FILE__ ": IEEE 802.11 elems parse ignored "
			       "unknown element (id=%d elen=%d)\n",
			       id, elen);
#endif
			unknown++;
			break;
		}

		left -= elen;
		pos += elen;
	}

	if(left)
		return(1);

	return(unknown ? 1 : 0);
}

/* to be removed by a generic ...send_mgmt... */
static int wireless_ieee80211_send_authentication(struct wireless_device *wdev,
						  u8 *bssid, u8 *mac)
{
	struct net_device *dev = wdev->dev;
	wireless_ieee80211_t *priv = (wireless_ieee80211_t*)wdev->mac.priv;
	wlan_frame_hdr_t *frame;
	int len = sizeof(wlan_frame_hdr_t) + sizeof(frame->p.mgmt.u.auth);
	u8 buf[len], *packet;
	struct sk_buff *skb = NULL;

	/** TODO: HAS TO BE FIXED **/

	/* init 802.11 header */
	INCHK;VENTER();
	memset(&buf, 0, sizeof(buf));
	frame = (wlan_frame_hdr_t*)&buf;
	frame->control.type = ieee_802_11_frame_type_Management;
	frame->control.subtype = ieee_802_11_frame_subtype_Authentication;
	frame->seq.num = priv->seq++;
	/* ... */
	
	memcpy(&frame->address1, bssid, WIRELESS_ADDR_SIZE);
	memcpy(&frame->address2, mac, WIRELESS_ADDR_SIZE);
	memcpy(&frame->address3, bssid, WIRELESS_ADDR_SIZE);

	/* init management frame */
	frame->p.mgmt.u.auth.auth_alg = 0;
	frame->p.mgmt.u.auth.auth_transaction = 1;
	frame->p.mgmt.u.auth.status_code = 0;

	/* allocate socket buffer */
	skb = dev_alloc_skb(len + 2); 
	skb_reserve(skb, 2); 
	
	if(skb == NULL)
		return(-EINVAL);
	
	packet = skb_put(skb, len);
	memset(packet, 0, len);
	memcpy(skb->data, buf, len);

	skb->dev = dev;
	skb->mac.raw = packet; /* address 1 */
	wdev->mac.associated=1;    
	return(wdev->transmit(dev, skb, WIRELESS_RAW));
}

/* to be removed by a generic ...send_mgmt... */
#if 1
static int wireless_ieee80211_send_association(struct wireless_device *wdev,
					       u8 *bssid, u8 *mac)
{
	struct net_device *dev = wdev->dev;
	wireless_ieee80211_t *priv = (wireless_ieee80211_t*)wdev->mac.priv;
	wlan_frame_hdr_t *frame;
	int len = sizeof(wlan_frame_hdr_t) + sizeof(frame->p.mgmt.u.auth);
	u8 buf[len], *packet;
	struct sk_buff *skb = NULL;

	/** TODO: HAS TO BE FIXED **/

	/* init 802.11 header */
	INCHK;VENTER();
	memset(&buf, 0, sizeof(buf));
	frame = (wlan_frame_hdr_t*)&buf;
	frame->control.type = ieee_802_11_frame_type_Management;
	frame->control.subtype = ieee_802_11_frame_subtype_Association_Req;
	frame->seq.num = priv->seq++;
	/* ... */
	
	memcpy(&frame->address1, bssid, WIRELESS_ADDR_SIZE);
	memcpy(&frame->address2, mac, WIRELESS_ADDR_SIZE);
	memcpy(&frame->address3, bssid, WIRELESS_ADDR_SIZE);

	/* init management frame */
	frame->p.mgmt.u.assoc_req.capab_info = 0;
	frame->p.mgmt.u.assoc_req.listen_interval = 0;

	/* allocate socket buffer */
	skb = dev_alloc_skb(len + 2); 
	skb_reserve(skb, 2); 
	
	if(skb == NULL)
		return(-EINVAL);
	
	packet = skb_put(skb, len);
	memset(packet, 0, len);
	memcpy(skb->data, buf, len);

	skb->dev = dev;
	skb->mac.raw = packet; /* address 1 */
	wdev->mac.associated=1;    
	return(wdev->transmit(dev, skb, WIRELESS_RAW));
}
#endif

int wireless_ieee80211_transmit_ok(struct wireless_device *wdev)
{
	/* nothing to do */
	INCHK;VENTER();
	return(0);
}

struct sk_buff *wireless_ieee80211_rx_mgmt(struct wireless_device *wdev,
					void *buf, int *len)
{
	wireless_ieee80211_t *priv = (wireless_ieee80211_t*)wdev->mac.priv;
	wlan_frame_hdr_t *hdr = (wlan_frame_hdr_t*)buf;
	wlan_frame_mgmt_t *mgmt;
	wlan_elems_t elems;
	wlan_reason_strings_t reasons[] = 
		IEEE_802_11_LINK_STATUS_FAILURE_REASON_STRINGS
	struct sk_buff *skb = NULL;
	u8 *frm;
        INCHK;VENTER();
	mgmt = (wlan_frame_mgmt_t*)&hdr->p.mgmt;

	switch(wdev->mac.mode) {
	case IW_MODE_INFRA:
		printk(__FILE__ ": management frame 0x%02x\n", hdr->control.subtype);		
		switch(hdr->control.subtype) {		
		case ieee_802_11_frame_subtype_Beacon:
			*len -=	sizeof(wlan_frame_hdr_t) + WIRELESS_IEEE80211_FCS; 
			frm = (u8*)buf + sizeof(wlan_frame_hdr_t); 
			if(wireless_ieee80211_parse_elements(wdev, 
							     frm,
							     *len,
							     &elems) == 1) {
				*len = -EINVAL;
				break;
			}

			/* find access point... */
			if(wdev->mac.searching) {
				char current_essid[IW_ESSID_MAX_SIZE + 1];

				if(elems.ssid_len > IW_ESSID_MAX_SIZE) {
					*len = -EINVAL;
					break;
				}
				
				memset(&current_essid, 0, sizeof(current_essid));
				memcpy(&current_essid, elems.ssid, elems.ssid_len);

#ifdef DEBUG
				printk(__FILE__ ": desired ssid '%s' <-> beacon ssid '%s'\n", 
				       current_essid, wdev->mac.essid);
				       
#endif

				if(strcmp(current_essid, 
					  wdev->mac.essid) == 0) {
					/* found ;) */
					memcpy(wdev->mac.bssid, hdr->address2, WIRELESS_ADDR_SIZE);
					wdev->mac.network_timeout = jiffies;

					if(wdev->set_network) {
						wdev->set_network(wdev, WIRELESS_FOUND_NETWORK);
					}

					if(wireless_ieee80211_send_authentication(wdev, 
										  (u8*)&wdev->mac.bssid,
										  (u8*)&wdev->mac.address) == 0)
						priv->status = WIRELESS_IEEE80211_AUTHENTICATING;
				}
			}
			break;
		case ieee_802_11_frame_subtype_Authentication:
		     printk(__FILE__": recibi frame Authentication\n");
		     break;
		case ieee_802_11_frame_subtype_Association_Resp:
		     printk(__FILE__": recibi frame Association_Resp\n");		
		     break;
		case ieee_802_11_frame_subtype_Reassociation_Resp:
		     printk(__FILE__": recibi frame Reassociation_Resp\n");		
		     break;
		case ieee_802_11_frame_subtype_Probe_Req:
		     printk(__FILE__":recibi frame Probe Req\n");
		     break;
		case ieee_802_11_frame_subtype_Probe_Resp:
		     printk(__FILE__": recibi frame Probe_Resp\n");		
		     break;
		case ieee_802_11_frame_subtype_Disassociation:
			printk(__FILE__ ": received management frame 0x%02x\n", 
			       hdr->control.subtype);
			*len = -ENOENT;
			break;
		case ieee_802_11_frame_subtype_Deauthentication:
			printk(__FILE__ ": received deauthentication frame "
			       "(reason 0x%02x '%s')\n", 
			       mgmt->u.deauth.reason_code,
			       reasons[mgmt->u.deauth.reason_code].description);
			*len = -ENOENT;
			break;
		default:
			printk("received unknown management frame 0x%02x\n", hdr->control.subtype);
			break;
		}
		break;
	case IW_MODE_ADHOC: 
		printk(__FILE__ ": management frame 0x%02x\n", hdr->control.subtype);		
		switch(hdr->control.subtype) {		
		case ieee_802_11_frame_subtype_Beacon:
			*len -=	sizeof(wlan_frame_hdr_t) + WIRELESS_IEEE80211_FCS; 
			frm = (u8*)buf + sizeof(wlan_frame_hdr_t); 
			if(wireless_ieee80211_parse_elements(wdev, 
							     frm,
							     *len,
							     &elems) == 1) {
				*len = -EINVAL;
				break;
			}

			/* find access point... */
			if(wdev->mac.searching) {
				char current_essid[IW_ESSID_MAX_SIZE + 1];

				if(elems.ssid_len > IW_ESSID_MAX_SIZE) {
					*len = -EINVAL;
					break;
				}
				
				memset(&current_essid, 0, sizeof(current_essid));
				memcpy(&current_essid, elems.ssid, elems.ssid_len);

#ifdef DEBUG
				printk(__FILE__ ": desired ssid '%s' <-> beacon ssid '%s'\n", 
				       current_essid, wdev->mac.essid);
				       
#endif

				if(strcmp(current_essid, 
					  wdev->mac.essid) == 0) {
					/* found ;) */
/*					memcpy(wdev->mac.bssid, hdr->address2, WIRELESS_ADDR_SIZE);
en realidad, bssid es el numero generado con el string essid, entonces, seria asi: */
					memcpy(wdev->mac.bssid, hdr->address3, WIRELESS_ADDR_SIZE);
/* y guardo la direccion del destino en dest_address */					
					memcpy(wdev->mac.dest_address, hdr->address2, WIRELESS_ADDR_SIZE);
					wdev->mac.network_timeout = jiffies;

					if(wdev->set_network) {
						wdev->set_network(wdev, WIRELESS_FOUND_NETWORK);
					}

					if(wireless_ieee80211_send_authentication(wdev, 
										  (u8*)&wdev->mac.bssid,
										  (u8*)&wdev->mac.address) == 0)
						priv->status = WIRELESS_IEEE80211_AUTHENTICATING;
				}
			}
			break;
		case ieee_802_11_frame_subtype_Authentication:
		     printk(__FILE__": recibi frame Authentication\n");
		     break;
		case ieee_802_11_frame_subtype_Association_Resp:
		     printk(__FILE__": Association_Resp no sirve aca\n");		
		     break;
		case ieee_802_11_frame_subtype_Reassociation_Resp:
		     printk(__FILE__": Reassociation_Resp no sirve aca\n");		
		     break;
		case ieee_802_11_frame_subtype_Probe_Req:
		     printk(__FILE__":recibi frame Probe Req\n");
		     break;     
		case ieee_802_11_frame_subtype_Probe_Resp:
		     printk(__FILE__": recibi frame Probe_Resp\n");		
		     break;
		case ieee_802_11_frame_subtype_Disassociation:
			printk(__FILE__ ": received management frame 0x%02x\n", 
			       hdr->control.subtype);
			*len = -ENOENT;
			break;
		case ieee_802_11_frame_subtype_Deauthentication:
			printk(__FILE__ ": received deauthentication frame "
			       "(reason 0x%02x '%s')\n", 
			       mgmt->u.deauth.reason_code,
			       reasons[mgmt->u.deauth.reason_code].description);
			*len = -ENOENT;
			break;
		default:
			printk("received unknown management frame 0x%02x\n", hdr->control.subtype);
			break;
		}

		/* TODO */
		break;
	case IW_MODE_MASTER:
		if(!wdev->mac.cap.master)
			break;
		/* TODO...  */
	default:
		*len = -ENOENT;
		break;
	}	

	return(skb);
}

struct sk_buff *wireless_ieee80211_rx_data(struct wireless_device *wdev,
					   void *buf, int *len)
{
	struct net_device *dev = (struct net_device*)wdev->dev;
	struct sk_buff *skb = NULL;
	u8 *data=NULL,*aux;
	int plen;
	int i;
	int auxlen;
	wlan_frame_hdr_t *hdr;
	u32 csum;
	auxlen=*len;
	auxlen=auxlen-4;
        INCHK;VENTER();
	hdr = (wlan_frame_hdr_t*)buf;
	aux=buf;
//	printk("buffer copiado.\n");
	plen=*len;
	plen=plen-24-6; //longitud neta de datos es len-24 o len-30
//        LINEA();
	skb=dev_alloc_skb(plen);
	if(skb==NULL) 
	    goto drop; 
	skb_reserve(skb,2);
//	printk("reservado skb\n");
	data=skb_put(skb,plen-4);   //defino datos en skbuff
	if(data==NULL) 
	    goto drop;	    
//	printk("puesto puntero a data\n");
	skb->dev=dev;
	memset(data,0,plen); //limpio memoria
	if(wdev->mac.mode==IW_MODE_ADHOC)
   	    memcpy(data,buf+30,plen-4); //asumo que la longitud len es del total del buffer
	else
	    memcpy(data,buf+36,plen-4);	    
	//ademas, los datos empiezan +30bytes
//	printk("payload copiado en data\n");
	memcpy(&csum,buf+(auxlen),4);
//	printk("checksm copiado\n");
        i=data[0]<<8|data[1];
	skb->protocol=__constant_htons(i);
        skb_pull(skb,2);

	skb->mac.raw=skb->data;	
	skb->pkt_type=PACKET_HOST;

	skb->csum=csum;	
	skb->ip_summed=CHECKSUM_NONE;
	return(skb);
drop:
//	printk("fallo alloc rxdata\n");
/*	if(skb)
	    dev_kfree_skb_irq(skb);
*/	    
	*len = -ENOENT;
	return(NULL);
}

struct sk_buff *wireless_ieee80211_decap(struct wireless_device *wdev,
					 void *buf, int *len)
{
	wlan_frame_hdr_t *hdr;
	struct sk_buff *skb = NULL;

	/* monitor mode (if supported by hardware) */
//	printk("1..");
        INCHK;VENTER();
	if(wdev->mac.cap.monitor && wdev->mac.mode == IW_MODE_MONITOR) 
		return(wireless_ieee80211_monitor(wdev, buf, len));

	/* small frame */
//	printk("2..");
	if(*len < sizeof(wlan_frame_hdr_t)) {
		*len = -EINVAL;
		return(NULL);
	}

	hdr = (wlan_frame_hdr_t*)buf;
//        printk("3..y frame tipo: %d\n",hdr->control.type);
	switch(hdr->control.type) {
	case ieee_802_11_frame_type_Management:
		skb = wireless_ieee80211_rx_mgmt(wdev, buf, len);
//		printk("4..");
		break;
	case ieee_802_11_frame_type_Control:
	//printk("5..");
	case ieee_802_11_frame_type_Data:
	//printk("6..");
		skb = wireless_ieee80211_rx_data(wdev, buf, len);
		break;
	case ieee_802_11_frame_type_Reserved:
	//printk("7..");
	default:
	//printk("8..");
		*len = -ENOENT;
	}
	//printk("9..");
	return(skb);
}

void wireless_ieee80211_dump_hdr(char *info, u8 *buf, int len)
{
	wlan_frame_hdr_t *hdr = (wlan_frame_hdr_t*)buf;

	if(len < sizeof(wlan_frame_hdr_t)) {
		printk(__FILE__ ": invalid IEEE 802.11 header (to small)");
		return;
	}

#ifndef DEBUG
	if(hdr->control.subtype == ieee_802_11_frame_subtype_Beacon)
		return;
#endif


        printk("%1X|%1X|%2X|%1X|%1X|%1X|%1X|%1X|%1X|%1X|%1X\n",
	hdr->control.version,hdr->control.type,hdr->control.subtype,hdr->control.to_ds,
	hdr->control.from_ds,hdr->control.more_frag,hdr->control.retry,hdr->control.pwr_mgt,
	hdr->control.more_data,hdr->control.wep,hdr->control.order);
	printk("addr1=" MACSTR ",addr2=" MACSTR ",addr3=" MACSTR "\n",
	MAC2STR(hdr->address1),MAC2STR(hdr->address2),MAC2STR(hdr->address3));
/*	printk(__FILE__ ": IEEE 802.11 header (%s):\n"
	       "   order=%u,wep=%u,retry=%u,from_ds=%u,to_ds=%u,\n"
	       "   subtype=%u,type=%u,version=%u,seq=%u,\n"
	       "   addr1=" MACSTR ",addr2=" MACSTR ",addr3=" MACSTR "\n",
	       info,
	       hdr->control.order,
	       hdr->control.wep,
	       hdr->control.retry,
	       hdr->control.from_ds,
	       hdr->control.to_ds,
	       hdr->control.subtype,
	       hdr->control.type,
	       hdr->control.version,
	       hdr->seq.num,
	       MAC2STR(hdr->address1),
	       MAC2STR(hdr->address2),
	       MAC2STR(hdr->address3));*/
}

/*
 * (host based) encoding (TODO)
 */

int wireless_wep_init(struct wireless_device *wdev)
{
	wireless_encoding_wep_t *priv;

	/* allocate private data */
	if((priv = kmalloc(sizeof(wireless_encoding_wep_t), GFP_KERNEL)) == NULL)
		return(-ENOMEM);

	memset(priv, 0, sizeof(wireless_encoding_wep_t));
	wdev->mac.encoding_priv = priv;

	return(0);
}

void wireless_wep_exit(struct wireless_device *wdev)
{
	if(wdev->mac.encoding_priv == NULL) {
		printk(__FILE__ ": encoding protocol not initialized\n");
		return;
	}

	/* free private data */
	kfree(wdev->mac.encoding_priv);
}

int wireless_wep_encode(struct wireless_device *wdev, void *buf, int len)
{
	if(wdev->mac.priv == NULL) {
		printk(__FILE__ ": protocol not initialized\n");
		return(-EINVAL);
	}

	return(-ENOENT);
}

int wireless_wep_decode(struct wireless_device *wdev, void *buf, int len)
{
	if(wdev->mac.priv == NULL) {
		printk(__FILE__ ": protocol not initialized\n");
		return(-EINVAL);
	}

	return(-ENOENT);
}

int wireless_ioctl(struct wireless_device *wdev, 
		   struct iwreq *wrq, int cmd) {
	int ret = 0;
	char name[IFNAMSIZ];
      VENTER();
        INCHK
	switch (cmd) {
	case SIOCGIWNAME:
		sprintf(name, IFNAMSIZ-1, "%s-%s", 
			 wireless_proto(wdev->mac.proto), 
			 wireless_type(wdev->phy.type));
		name[IFNAMSIZ] = '\0';
		memcpy(wrq->u.name, name, sizeof(wrq->u.name));
		break;

	case SIOCSIWFREQ:
		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}

		/* driver callback */
		if(wdev->set_channel)
			ret = wdev->set_channel(wdev, &wrq->u.freq);
		break;

	case SIOCGIWFREQ:
		memcpy(&wrq->u.freq, &wdev->phy.freq, sizeof(struct iw_freq));
		break;

	case SIOCSIWMODE:
		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}
		
		/* driver callback */
		if(wdev->set_mode && 
		   (ret = wdev->set_mode(wdev, wrq->u.mode) != 0))
			break;

		wdev->mac.mode = wrq->u.mode;
		break;
			
	case SIOCSIWAP:
		if(!wdev->mac.cap.bssid) {
			ret = -ENOTSUPP;
			break;
		}

		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}

		memcpy(&wdev->mac.desired_bssid, &wrq->u.ap_addr.sa_data, 
		       WIRELESS_ADDR_SIZE);
		break;

	case SIOCGIWAP:
		if(!wdev->mac.cap.bssid) {
			ret = -ENOTSUPP;
			break;
		}

		memcpy(&wrq->u.ap_addr.sa_data, &wdev->mac.bssid, 
		       ETH_ALEN);
		break;

	case SIOCSIWESSID:
		if(!wdev->mac.cap.essid) {
			ret = -ENOTSUPP;
			break;
		}

		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}

		if (!wrq->u.essid.pointer)
			ret = -EINVAL;
		else if (wrq->u.essid.length > IW_ESSID_MAX_SIZE)
			ret = -E2BIG;
		else {
			if (&wrq->u.essid.flags == 0) {
				/* any */
				memset(wdev->mac.essid, 0, IW_ESSID_MAX_SIZE);
			} else {
				u8 ssid[IW_ESSID_MAX_SIZE];
				if(copy_from_user(&ssid, wrq->u.essid.pointer,
						  wrq->u.essid.length)) {
					ret = -EFAULT;
					break;
				}

				if(memcmp(&wdev->mac.essid, &ssid, wrq->u.essid.length) == 0)
					break;

				memset(&wdev->mac.essid, 0, IW_ESSID_MAX_SIZE);
				memcpy(&wdev->mac.essid, &ssid, wrq->u.essid.length);
				
				if(wdev->set_network) {
					ret = wdev->set_network(wdev, WIRELESS_FIND_NETWORK);
					if(ret == 0)
						wdev->mac.searching = 1;
				}
			}
		}
		break;

	case SIOCGIWESSID:
		if(!wdev->mac.cap.essid) {
			ret = -ENOTSUPP;
			break;
		}
		
		wrq->u.essid.length = strlen(wdev->mac.essid);

		if(wrq->u.essid.length > 0)
			wrq->u.essid.flags = 1; 

		if(copy_to_user(wrq->u.essid.pointer, 
				wdev->mac.essid,
				wrq->u.essid.length))
			ret = -EFAULT;
		break;

	case SIOCGIWNWID:
		if(!wdev->mac.cap.nwid) {
			ret = -ENOTSUPP;
			break;
		}
			
		wrq->u.nwid.value = wdev->mac.desired_nwid;

		if(wrq->u.nwid.value > 0) {
			wrq->u.nwid.fixed = 1;
			wrq->u.nwid.fixed = 0;
		}
				
		break;

	case SIOCSIWNWID:
		if(!wdev->mac.cap.nwid) {
			ret = -ENOTSUPP;
			break;
		}

		if (!capable(CAP_NET_ADMIN)) {
			ret = -EPERM;
			break;
		}
			
		if(wdev->mac.desired_nwid == wrq->u.nwid.value)
			break;

		wdev->mac.desired_nwid = wrq->u.nwid.value;
		break;

	case SIOCGIWMODE:
		wrq->u.mode = wdev->mac.mode;
		break;
	}		
      VEXIT();
	return(ret);
}

/*
 * create or re-use existing proc base directory
 */

static struct proc_dir_entry* init_procbase(const char* name, struct proc_dir_entry* parent)
{
	struct proc_dir_entry* base;
	int len = strlen(name);

	for(base = parent->subdir; base; base = base->next)
		if ((base->namelen == len) && (!memcmp(base->name, name, len)))
			break;
	if (base == NULL) {
		base = proc_mkdir(name, parent);
#ifdef WIRELESS_MODULE
		if (base)
			base->owner = THIS_MODULE;
#endif
	}
	return base;
}

/*
 * remove proc base directory iff empty
 */

static void remove_procbase(struct proc_dir_entry* entry)
{
	if (entry) {
		if (entry->nlink <= 2
#ifdef WIRELESS_MODULE
		    && entry->owner == THIS_MODULE
#endif
		)
			remove_proc_entry(entry->name, entry->parent);
	}
}

/*
 * global wireless data
 */

int init_wireless(void)
{
	if(proc_net == NULL)
		return(-ENOENT);

	printk(__FILE__ ": %s\n", version);

	proc_wireless = init_procbase("wdev", proc_net);
	memset(&wireless_no_addr, 0xff, WIRELESS_ADDR_SIZE);

	return(0);
}

void exit_wireless(void)
{
	remove_procbase(proc_wireless);	
	proc_wireless = NULL;
}

#ifdef WIRELESS_MODULE
module_init(init_wireless)
module_exit(exit_wireless)
#endif
