/*                                                        -*- linux-c -*-
 *   _  _ ____ __ _ ___ ____ ____ __ _ _ _ _ | 
 * .  \/  |--| | \|  |  |--< [__] | \| | _X_ | s e c u r e  s y s t e m s
 *
 * ieee80211.h - common IEEE 802.11 protocol definitions
 *
 * NOTE: this file extends linux/802_11.h
 *
 * $Id: ieee80211.h,v 1.4 2003/08/22 17:35:38 reyk Exp $
 *
 * Copyright (c) 2003, .vantronix | secure systems
 *                     and Reyk Floeter <reyk@vantronix.net> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef _EXT_IEEE_80211_H
#define _EXT_IEEE_80211_H

#include <linux/802_11.h>
#include <linux/wdev.h>

/* list of eleemnt ids */
#define IEEE80211_EID_SSID 0
#define IEEE80211_EID_SUPP_RATES 1 
#define IEEE80211_EID_FH_PARAMS 2	
#define	IEEE80211_EID_DS_PARAMS 3	/* ds parameter set */
#define	IEEE80211_EID_CF_PARAMS 4	/* cf parameter set */
#define IEEE80211_EID_TIM 5 
#define IEEE80211_EID_IBSS_PARAMS 6 
#define IEEE80211_EID_CHALLENGE 16 
/* bits 17-31 reserved for challenge txt extension */

/* max size for element values */
#define	IEEE80211_MAC_ADDR_SIZE WIRELESS_ADDR_SIZE
#define IEEE80211_EID_SSID_SIZE WIRELESS_SSID_SIZE
#define IEEE80211_EID_INFO_SIZE 255 
#define IEEE80211_EID_RATE_SIZE 8 
#define IEEE80211_EID_BITMAP_SIZE 251 
#define IEEE80211_EID_CHLG_TEXT_SIZE 253 
#define IEEE80211_EID_CF_PARAM_SIZE 6 
#define IEEE80211_EID_IBSS_PARAM_SIZE 2 

/* note:
 * bit fields are ordered for little endian.
 * please use the corresponding swap() function for big endian. 
 */

typedef struct wlan_cap_info {
	u16 ess:1,
	    ibss:1, 
	    cf_pollable:1,
	    cf_poll_req:1,
	    privacy:1,
	    reserved:11;
} __attribute__ ((__packed__)) wlan_cap_info_t;

typedef struct wlan_ssid {
	u8 element_id;
	u8 length;
	u8 ssid[IEEE80211_EID_SSID_SIZE];
} __attribute__ ((__packed__)) wlan_ssid_t;

typedef struct wlan_rate_set {
	u8 element_id;
	u8 length; 
	u8 rates[IEEE80211_EID_RATE_SIZE]; 
} __attribute__ ((__packed__)) wlan_rate_set_t;

typedef struct wlan_elems {
	u8 *ssid;
	u8 ssid_len;
	u8 *supp_rates;
	u8 supp_rates_len;
	u8 *fh_params;
	u8 fh_params_len;
	u8 *ds_params;
	u8 ds_params_len;
	u8 *cf_params;
	u8 cf_params_len;
	u8 *tim;
	u8 tim_len;
	u8 *ibss_params;
	u8 ibss_params_len;
	u8 *challenge;
	u8 challenge_len;
} __attribute__ ((__packed__)) wlan_elems_t;

/** 802.11 header **/
typedef struct wlan_frame_control {
        u16 version:2,
	    type:2,
	    subtype:4,
	    to_ds:1,
	    from_ds:1,
	    more_frag:1,
	    retry:1,
	    pwr_mgt:1,
	    more_data:1,
	    wep:1,
	    order:1;
        u16 duration:15,
	    clear_bit:1;
} __attribute__ ((__packed__)) wlan_frame_control_t;

typedef struct wlan_sequence {
	u16 frag:4,
	    num:12;
} __attribute__ ((__packed__)) wlan_seq_t;

typedef struct wlan_frame_mgmt {
	union {
		struct {
			/* followed by SSID and supported rates */
		} __attribute__ ((__packed__)) probe_req;
		struct {
			u8 timestamp[8];
			u16 beacon_int;
			u16 capab_info;
			/* followed by some of SSID, Supported rates,
			 * FH Params, DS Params, CF Params, IBSS Params */
		} __attribute__ ((__packed__)) probe_resp;
		struct {
			u16 auth_alg;
			u16 auth_transaction;
			u16 status_code;
			/* possibly followed by Challenge text */
		} __attribute__ ((__packed__)) auth;
		struct {
			u16 reason_code;
		} __attribute__ ((__packed__)) deauth;
		struct {
			u16 capab_info;
			u16 listen_interval;
			/* followed by SSID and Supported rates */
		} __attribute__ ((__packed__)) assoc_req;
		struct {
			u16 capab_info;
			u16 status_code;
			u16 aid;
			/* followed by Supported rates */
		} __attribute__ ((__packed__)) assoc_resp, reassoc_resp;
		struct {
			u16 capab_info;
			u16 listen_interval;
			u8 current_ap[6];
			/* followed by SSID and Supported rates */
		} __attribute__ ((__packed__)) reassoc_req;
		struct {
			u16 reason_code;
		} __attribute__ ((__packed__)) disassoc;
		struct {
			u8 timestamp[8];
			u16 beacon_int;
			u16 capab_info;
			/* followed by some of SSID, Supported rates,
			 * FH Params, DS Params, CF Params, IBSS Params, TIM */
		} __attribute__ ((__packed__)) beacon;
	} u;
} __attribute__ ((__packed__)) wlan_frame_mgmt_t;

typedef struct wlan_frame_header {
	wlan_frame_control_t control;
	u8 address1[IEEE80211_MAC_ADDR_SIZE];
	u8 address2[IEEE80211_MAC_ADDR_SIZE];
	u8 address3[IEEE80211_MAC_ADDR_SIZE];
	wlan_seq_t seq;
	union {
		wlan_frame_mgmt_t mgmt;
		u8 data[0];
	} p;
} __attribute__ ((__packed__)) wlan_frame_hdr_t;

/* use reasons defined in linux/802_11.h:
 * wlan_reason_strings_t reasons[] = 
 *	IEEE_802_11_LINK_STATUS_FAILURE_REASON_STRINGS;
 */
typedef struct wlan_reason_strings {
	u16 value;
	u8 reserved;
	char *description;
} __attribute__ ((__packed__)) wlan_reason_strings_t;

#ifdef __KERNEL__
extern int wireless_ieee80211_parse_elements(struct wireless_device *wdev,
					    u8 *buf, int len,
					    wlan_elems_t *elems);
extern struct sk_buff *wireless_ieee80211_rx_mgmt(struct wireless_device *wdev,
						  void *buf, int *len);
extern struct sk_buff *wireless_ieee80211_rx_data(struct wireless_device *wdev,
						  void *buf, int *len);
#endif /* __KERNEL__ */
#endif /* _EXT_IEEE_802_11_H */
