/*                                                        -*- linux-c -*-
 *   _  _ ____ __ _ ___ ____ ____ __ _ _ _ _ | 
 * .  \/  |--| | \|  |  |--< [__] | \| | _X_ | s e c u r e  s y s t e m s
 *
 * .vt|ar5k - PCI/CardBus 802.11a WirelessLAN driver for Atheros AR5k chipsets
 *
 * $Id: vt_ar5210.c,v 1.4 2003/08/07 19:07:34 reyk Exp $
 *
 * Copyright (c) 2002, .vantronix | secure systems
 *                     and Reyk Floeter <reyk@vantronix.net> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vt_ar5kconfig.h>
#include <vt_ar5k.h>

int vt_ar5210_eeprom_read(vt_local_t *local, u32 offset, u16 *data)
{
	struct net_device *dev = (struct net_device*) local->dev;
	int timeout = 10000;   
	u32 status;
	u32 pcicfg;

	VENTER();

	/** enable eeprom access */
	pcicfg = VT_WLAN_IN32(AR5K_PCICFG);
	VT_WLAN_OUT32(pcicfg | AR5K_PCICFG_EEAE, AR5K_PCICFG);
	
	/* prime read pump */
	(void)VT_WLAN_IN32(AR5210_EEPROM_BASE + (4 * offset));
	
	while(timeout > 0) {
		udelay(1);
		status = VT_WLAN_IN32(AR5210_EEPROM_STATUS);
		if(status & AR5210_EEPROM_STAT_RDDONE) {
			if(status & AR5210_EEPROM_STAT_RDERR) {
				VPRINT("eeprom read access failed!\n");
				return(1);
			}
			status = VT_WLAN_IN32(AR5210_EEPROM_RDATA);
			*data = (u16)(status & 0xffff);
			return(0);
		}
		timeout--;
	}
	
	VPRINT("eeprom read timeout!\n");

	VEXIT();

	return(1);
}

int vt_ar5210_init(vt_local_t *local)
{
	struct net_device *dev = (struct net_device*)local->dev;
	const wlan_register_val_t *init_regs;
	u32 i;

	VENTER();
	
	init_regs = vt_ar5k_register_init;
	local->init_regs.count = 0;
	
	while (init_regs[local->init_regs.count].offset != 0xffffffff) {
		if (local->init_regs.count > MAX_REG_COUNT) {
			VPRINT("init register overflow!\n");
			return(1);
		}
		local->init_regs.values[local->init_regs.count].offset =
			init_regs[local->init_regs.count].offset;
		local->init_regs.values[local->init_regs.count].value =
			init_regs[local->init_regs.count].value;
		local->init_regs.count++;
	}
	
	/** write initial phyreg values **/
	for (i = 0; i < local->init_regs.count; i++) 	{
		VT_WLAN_OUT32(local->init_regs.values[i].value,
			      local->init_regs.values[i].offset);
	}

	VEXIT();

	return(0);
}

int vt_ar5210_dump(vt_local_t *local)
{
	struct net_device *dev = (struct net_device*)local->dev;
	unsigned long flags;

	spin_lock_irqsave(&local->lock, flags);

	/* dump as the mad wifi driver dumps */
	VPRINT("register state:\n");
	printk("TXDP0    %08x  TXDP1    %08x  CR       %08x  RXDP     %08x\n"
	       "CFG      %08x  ISR      %08x  IMR      %08x  IER      %08x\n"
	       "BCR      %08x  TXCFG    %08x  RXCFG    %08x  MIBC     %08x\n"
	       "TOPS     %08x  RXNOFR   %08x  TXNOFR   %08x  RPGTO    %08x\n"
	       "RFCNT    %08x  MISC     %08x  RC       %08x  SCR      %08x\n"
	       "INTPEND  %08x  SFR      %08x  PCICFG   %08x  GPIOCR   %08x\n"
	       "SREV     %08x  STA_ID0  %08x  STA_ID1  %08x  BSS_ID0  %08x\n"
	       "BSS_ID1  %08x  SLOT_TIM %08x  TIME_OUT %08x  RSSI_THR %08x\n",
	       VT_WLAN_IN32(AR5K_TXDP0), VT_WLAN_IN32(AR5K_TXDP1), VT_WLAN_IN32(AR5K_CR), VT_WLAN_IN32(AR5K_RXDP),
	       VT_WLAN_IN32(AR5K_CFG), VT_WLAN_IN32(AR5K_ISR), VT_WLAN_IN32(AR5K_IMR), VT_WLAN_IN32(AR5K_IER),
	       VT_WLAN_IN32(AR5K_BCR), VT_WLAN_IN32(AR5K_TXCFG), VT_WLAN_IN32(AR5K_RXCFG), VT_WLAN_IN32(AR5K_MIBC),
	       VT_WLAN_IN32(AR5K_TOPS), VT_WLAN_IN32(AR5K_RXNPTO), VT_WLAN_IN32(AR5K_TXNPTO), VT_WLAN_IN32(AR5K_RPGTO),
	       VT_WLAN_IN32(AR5K_RPCNT), VT_WLAN_IN32(AR5K_MACMISC), VT_WLAN_IN32(AR5K_RC), VT_WLAN_IN32(AR5K_SCR),
	       VT_WLAN_IN32(AR5K_INTPEND), VT_WLAN_IN32(AR5K_SFR), VT_WLAN_IN32(AR5K_PCICFG), VT_WLAN_IN32(AR5K_GPIOCR),
	       VT_WLAN_IN32(AR5K_SREV), VT_WLAN_IN32(AR5K_STA_ID0), VT_WLAN_IN32(AR5K_STA_ID1), VT_WLAN_IN32(AR5K_BSS_ID0),
	       VT_WLAN_IN32(AR5K_BSS_ID1), VT_WLAN_IN32(AR5K_SLOT_TIME), VT_WLAN_IN32(AR5K_TIME_OUT), VT_WLAN_IN32(AR5K_RSSI_THR));

	printk("RETRY_LM %08x  USEC     %08x  BEACON   %08x  CFP_PER  %08x\n"
	       "TIMER0   %08x  TIMER1   %08x  TIMER2   %08x  TIMER3   %08x\n"
	       "IFS0     %08x  IFS1     %08x  CFP_DUR  %08x  RX_FILTR %08x\n"
	       "MCAST_0  %08x  MCAST_1  %08x  TXMASK0  %08x  TXMASK1  %08x\n"
	       "TRIGLEV  %08x  DIAG_SW  %08x  TSF_L32  %08x  TSF_U32  %08x\n"
	       "LAST_TST %08x  RETRYCNT %08x  BACKOFF  %08x  NAV      %08x\n"
	       "RTS_OK   %08x  RTS_FAIL %08x  ACK_FAIL %08x  FCS_FAIL %08x\n"
	       "BEAC_CNT %08x\n",
	       VT_WLAN_IN32(AR5K_RETRY_LMT), VT_WLAN_IN32(AR5K_USEC), VT_WLAN_IN32(AR5K_BEACON), VT_WLAN_IN32(AR5K_CFP_PERIOD),
	       VT_WLAN_IN32(AR5K_TIMER0), VT_WLAN_IN32(AR5K_TIMER1), VT_WLAN_IN32(AR5K_TIMER2), VT_WLAN_IN32(AR5K_TIMER3),
	       VT_WLAN_IN32(AR5K_IFS0), VT_WLAN_IN32(AR5K_IFS1), VT_WLAN_IN32(AR5K_CFP_DUR), VT_WLAN_IN32(AR5K_RX_FILTER),
	       VT_WLAN_IN32(AR5K_MCAST_FIL0), VT_WLAN_IN32(AR5K_MCAST_FIL1), VT_WLAN_IN32(AR5K_TX_MASK0), VT_WLAN_IN32(AR5K_TX_MASK1),
	       VT_WLAN_IN32(AR5K_TRIG_LEV), VT_WLAN_IN32(AR5K_DIAG_SW), VT_WLAN_IN32(AR5K_TSF_L32), VT_WLAN_IN32(AR5K_TSF_U32),
	       VT_WLAN_IN32(AR5K_LAST_TSTP), VT_WLAN_IN32(AR5K_RETRY_CNT), VT_WLAN_IN32(AR5K_BACKOFF), VT_WLAN_IN32(AR5K_NAV),
	       VT_WLAN_IN32(AR5K_RTS_OK), VT_WLAN_IN32(AR5K_RTS_FAIL), VT_WLAN_IN32(AR5K_ACK_FAIL), VT_WLAN_IN32(AR5K_FCS_FAIL),
	       VT_WLAN_IN32(AR5K_BEACON_CNT));

	spin_unlock_irqrestore(&local->lock, flags);

	return(0);
}
