/*                                                        -*- linux-c -*-
 *   _  _ ____ __ _ ___ ____ ____ __ _ _ _ _ | 
 * .  \/  |--| | \|  |  |--< [__] | \| | _X_ | s e c u r e  s y s t e m s
 *
 * .vt|ar5k - PCI/CardBus 802.11a WirelessLAN driver for Atheros AR5k chipsets
 *
 * $Id: vt_ar5kreg.c,v 1.3 2003/05/22 12:48:38 reyk Exp $
 *
 * Copyright (c) 2002, .vantronix | secure systems
 *                     and Reyk Floeter <reyk@vantronix.net> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vt_ar5kconfig.h>
#include <vt_ar5k.h>

/* frequency = 5000 + 5 * channel MHz */

/* channel numbers and restrictions by regulation domain: 
 *
 * FCC:
 * U-NII lower band:  36,40,44,48
 * U-NII middle band: 52,56,60,64
 * U-NII upper band:  149,153,157,161 
 *
 * Germany:
 * with TPC and DFS:
 * 1. range 5150 - 5350 MHz: 30,34,38,42,46,50,54,58,62,66,70
 *                           up to 20mW EIRP, indoor use only
 * 2. range 5470 - 5725 MHz: 94,98,102,106,110,114,118,122,126,130,134,138,142
 *                           up to 1W EIRP, indoor and outdoor use
 * without TPC and DFS:
 * special range 5150 - 5250 MHz: 30,34,38,42,46
 *                                60mW EIRP with TPC, 30mW EIRP without TPC,
 *                                indoor use only
 *
 * Japan:
 * Indoor use only?
 */

int vt_ar5k_reg_init(void) {
	if(strlen(reg_domain) < 1) {
		VINIT("please specify your regulation domain!\n");
		VINIT("reg_domain=(fcc|etsi|de|japan)\n");
		return(-EINVAL);
	} else {
		if(strcmp(reg_domain, "fcc") == 0) {
			/* XXX TODO: settings for FCC */
		} else if (strcmp(reg_domain, "etsi") == 0) {
			/* XXX TODO: settings for ETSI */
		} else if (strcmp(reg_domain, "de") == 0) {
			/* XXX TODO: settings for Germany */
		} else if (strcmp(reg_domain, "japan") == 0) {
			/* XXX TODO: settings for Japan */
		} else if (strcmp(reg_domain, "NONE") == 0) {
			/* XXX only for development - to be removed in the final driver! */
		} else {
			VINIT("regulation domain '%s' is currently not supported!\n", reg_domain);
			return(-EINVAL);
		}

		VINIT("regulation domain '%s'\n", reg_domain);
	}

	return(0);
}
