/*
 * Compaq TC1000 Touchscreen Device
 * Version 0.1
 * Copyright (c) Frank Bruno <fbruno@attbi.com>
 *
 * Can be freely distributed and used under the terms of the GNU GPL.
 * 
 * This module file is a hack. It is the DSDT init routine for the 
 * COM1 interface of the Via vt82c686b chipset. This should really go
 * in the serial.c file somewhere, or ideally the ACPI interpretter
 * Should handle all of this.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/kmod.h>
#include <linux/pci.h>
#include <linux/fs.h>       /* The character device 
                             * definitions are here */

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,44))
#include <linux/wrapper.h>  /* A wrapper which does 
                             * next to nothing at
                             * at present, but may 
                             * help for compatibility
                             * with future versions 
                             * of Linux */
#endif

const unsigned short confindex=0x3F0,confdata=0x3F1;
const unsigned char spcidx=0xEE;
const unsigned char cr2=0xE2;
const unsigned char cr3=0xE3;
const unsigned char cr6=0xE6;
const unsigned char cr7=0xE7;
const unsigned char cr8=0xE8;
const unsigned char cre=0xEE;

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0))
spinlock_t driver_lock = SPIN_LOCK_UNLOCKED;
#endif

static int fpi2002_init(void)
{
  struct pci_dev *pcidev = NULL;
  unsigned char confval,val;
  unsigned char local0, local1;
  /* See if we have a via chip */
  pcidev = pci_find_device (PCI_VENDOR_ID_VIA,PCI_DEVICE_ID_VIA_82C686,NULL);
  if (pcidev) {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0))
    spin_lock_irq(&driver_lock);
#else
    cli();
#endif
    /* start config ACPI ENCG */
    pci_read_config_byte(pcidev,0x85,&confval);
    confval |= 0x2;
    pci_write_config_byte (pcidev, 0x85,confval);
    /* Load CR2 IO address */
    outb(cr2,confindex); /* set index */
    local0 = (unsigned char) inb(confdata);
    local0 = local0 & 0xfb;
    outb(cr2,confindex);
    outb(local0,confdata);
    /*    local0 = (0xf8 << 2) & 0xfe; // IO low
	  local1 = (0x03 << 6);
	  local0 = local0 | local1; */
    outb(cr7,confindex); /* set index */
    outb(0xfe,confdata); /* 0x3f8 >> 2  (IO port) */
    /* Load IRQ */
    pci_read_config_byte(pcidev,0x52,&confval);
    local0 = (confval & 0xf0) | 0x04;
    pci_write_config_byte (pcidev, 0x52,local0);
    outb(cr2,confindex);
    outb(local0,confdata);
    local0 = local0 | 0x04;
    outb(cr2,confindex);
    local0 = (unsigned char) inb(confdata);
    local0 = local0 | 0x04;
    outb(cr2,confindex); /* set index */
    outb(local0,confdata);
    /* EXCG */
    pci_read_config_byte(pcidev,0x85,&confval);
    confval &= 0xfd;
    pci_write_config_byte (pcidev, 0x85,confval);

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0))
    spin_unlock_irq(&driver_lock);
#else
    sti();
#endif
    printk (KERN_INFO "Enabled FPI2002 Interface\n");
  }
  else {
    printk (KERN_INFO "Couldn't locate VIA chipset\n");
    return -1;
  }
  return 0;
}

static void fpi2002_exit(void)
{
  struct pci_dev *pcidev = NULL;
  unsigned char confval,val;
  unsigned char local0, local1;
  pcidev = pci_find_device (PCI_VENDOR_ID_VIA,PCI_DEVICE_ID_VIA_82C686,NULL);
  if (pcidev) {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0))
    spin_lock_irq(&driver_lock);
#else
    cli();
#endif
    /* start config */
    pci_read_config_byte(pcidev,0x85,&confval);
    confval |= 0x2;
    pci_write_config_byte (pcidev, 0x85,confval);

    outb(cr2,confindex); /* set index */
    local0 = (unsigned char) inb(confdata);
    outb(cr2,confindex); /* set index */
    outb(local0,confdata); 
    outb(cr7,confindex); /* set index */
    outb(0x0,confdata); 
    pci_read_config_byte(pcidev,0x52,&confval);
    local0 = (confval & 0xf0);
    pci_write_config_byte (pcidev, 0x52,local0);
    
    /* EXCG */
    pci_read_config_byte(pcidev,0x85,&confval);
    confval &= 0xfd;
    pci_write_config_byte (pcidev, 0x85,confval);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0))
    spin_unlock_irq(&driver_lock);
#else
    sti();
#endif
    printk (KERN_INFO "FPI2002 disabled\n");
  }
}

module_init(fpi2002_init);
module_exit(fpi2002_exit);
MODULE_DESCRIPTION("FPI2002 port enabler");
MODULE_AUTHOR("Frank Bruno <fbruno@attbi.com>");

