/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_cfgparm.h 1.26 2004/07/22 18:34:49EDT sf_support Exp  $
 */

#ifndef _H_LPFC_CFGPARAM
#define _H_LPFC_CFGPARAM

#define LPFC_MAX_LUN                    256	/* max nunber of LUNs
						   supported */
#define LPFC_MAX_TARGET                 256	/* max nunber of targets
						   supported */
#define LPFC_MAX_DISC_THREADS           64	/* max outstanding discovery els
						   requests */
#define LPFC_MAX_NS_RETRY               3	/* Try to get to the NameServer
						   3 times and then give up. */
#define LPFC_MAX_HBA_Q_DEPTH            10240	/* max cmds allowed per hba */
#define LPFC_DFT_HBA_Q_DEPTH            2048	/* max cmds per hba */
#define LPFC_LC_HBA_Q_DEPTH             1024	/* max cmds per low cost hba */
#define LPFC_LP101_HBA_Q_DEPTH          128	/* max cmds per low cost hba */

/* LPFC specific parameters start at LPFC_CORE_NUM_OF_CFG_PARAM */
#define LPFC_CFG_LOG_VERBOSE             0	/* log-verbose */
#define LPFC_CFG_DFT_LUN_Q_DEPTH         1	/* lun_queue_depth */
#define LPFC_CFG_LINKDOWN_TMO            2	/* linkdown-tmo */
#define LPFC_CFG_NODEV_TMO               3	/* nodev-tmo */
#define LPFC_CFG_DFT_HBA_Q_DEPTH         4	/* dft_hba_q_depth */
#define LPFC_CFG_AUTOMAP                 5	/* automap */
#define LPFC_CFG_FCP_CLASS               6	/* fcp-class */
#define LPFC_CFG_USE_ADISC               7	/* use-adisc */
#define LPFC_CFG_ACK0                    8	/* ack0 */
#define LPFC_CFG_TOPOLOGY                9	/* topology */
#define LPFC_CFG_SCAN_DOWN               10	/* scan-down */
#define LPFC_CFG_LINK_SPEED              11	/* link-speed */
#define LPFC_CFG_CR_DELAY                12	/* cr-delay */
#define LPFC_CFG_CR_COUNT                13	/* cr-count */
#define LPFC_CFG_FDMI_ON                 14	/* fdmi-on-count */
#define LPFC_CFG_BINDMETHOD              15	/* fcp-bind-method */
#define LPFC_CFG_DISC_THREADS            16	/* discovery-threads */

/* Note: The following define LPFC_TOTAL_NUM_OF_CFG_PARAM represents the total
	 number of user configuration params. This define is used to specify the
	 number of entries in the array lpfc_icfgparam[].
 */
#define LPFC_TOTAL_NUM_OF_CFG_PARAM      17

typedef enum tagenumparamidx {
	log_verbose, lun_queue_depth, linkdown_tmo, 
	nodev_tmo,  hba_queue_depth, automap, fcp_class, use_adisc, ack0,
	topology, scan_down, link_speed, cr_delay, cr_count, fdmi_on,
	fcp_bind_method, discovery_threads,
} enumparamidx;

/* the icfgparam structure - internal use only */
typedef struct ICFGPARAM {
	char *a_string;
	uint32_t a_low;
	uint32_t a_hi;
	uint32_t a_default;
} icfgparam;

#ifdef LPFC_DEF_ICFG
icfgparam lpfc_icfgparam[LPFC_TOTAL_NUM_OF_CFG_PARAM] = {
	/* general driver parameters */
	{"log_verbose", 0, 0xffff, 0},
	{"lun_queue_depth", 1, 128, 30},
	{"linkdown_tmo", 0, 255, 30}, 
	{"nodev_tmo", 0, 255, 30},
	{"hba_queue_depth", 0, 10240, 0},
	/* Start of product specific (lpfc) config params */
	{"automap", 0, 1, 1},
	{"fcp_class", 2, 3, 3},
	{"use_adisc", 0, 1, 0},
	/* Fibre Channel specific parameters */
	{"ack0", 0, 1, 0},
	{"topology", 0, 6, 0},
	{"scan_down", 0, 1, 1}, 
	{"link_speed", 0, 2, 0}, 
	{"cr_delay", 0, 63, 0},
	{"cr_count", 1, 255, 1}, 
	{"fdmi_on", 0, 2, 0}, 
	{"fcp_bind_method", 1, 4, 2},
	{"discovery_threads", 1, LPFC_MAX_DISC_THREADS, 1},
};
#endif				/* LPFC_DEF_ICFG */

#endif				/* _H_LPFC_CFGPARAM */
