/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_defaults.c 1.2 2004/07/08 14:48:58EDT sf_support Exp  $
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/blkdev.h>
#include <linux/ctype.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/pci.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <linux/timer.h>
#include <linux/utsname.h>
#include <asm/byteorder.h>

#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>

#include "lpfc_sli.h"
#include "lpfc_disc.h"
#include "lpfc_scsi.h"
#include "lpfc_cfgparm.h"
#include "lpfc.h"

/*
 - Setup FCP persistent bindings.
 * These variables are defined to provide backward compatibility with
 * previous versions of the driver. Although a significant aid, they do
 * not guarantee nor fully provide persistent device naming and
 * addressing. The Linux 2.6 kernel community recommends the use of
 * functions such as device mapper (http://sources.redhat.com/dm/) or
 * devlabel (http://linux.dell.com/devlabel/devlabel.html) to provide
 * persistent device naming.

# fcp-bind-WWPN binds a specific WorldWide PortName to a target id,
# fcp-bind-WWNN binds a specific WorldWide NodeName to a target id,
# fcp-bind-DID binds a specific DID to a target id.
# Each array must end with a NULL pointer.
# Binding method must match with the bind method of that HBA, else the
# binding will be ignored.
# scan-down should NOT be set to 2 when one of these binding methods
# is used.
# WWNN, WWPN and DID are hexadecimal values.
# WWNN must be 16 digit BCD with leading 0s.
# WWPN must be 16 digit BCD with leading 0s.
# DID must be 6 digit BCD with leading 0s.
# The SCSI ID to bind to consists of two parts, the lpfc interface
# to bind to, and the target number for that interface.
# Thus lpfc0t2 specifies target 2 on interface lpfc0.
#
# Here are some examples:
#                                   WWNN             SCSI ID
# char *lpfc_fcp_bind_WWNN[MAX_FC_BINDINGS]={"22000020370b8275:lpfc0t1",
#                                            "22000020370b8998:lpfc0t2",
#                                            0};
#
#                                   WWPN             SCSI ID
# char *lpfc_fcp_bind_WWPN[MAX_FC_BINDINGS]={"22000020370b8275:lpfc0t1",
#                                            "22000020370b8998:lpfc0t2",
#                                            0};
#
#                                   DID   SCSI ID
# char *lpfc_fcp_bind_DID[MAX_FC_BINDINGS]={"0000dc:lpfc0t1",
#                                           "0000e0:lpfc0t2",
#                                           0};
#
*/
char *lpfc_fcp_bind_WWNN[MAX_FC_BINDINGS]   =  {0};
char *lpfc_fcp_bind_WWPN[MAX_FC_BINDINGS]   =  {0};
char *lpfc_fcp_bind_DID[MAX_FC_BINDINGS]    =  {0};
