/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_fcp.c 1.308 2004/07/22 20:34:06EDT sf_support Exp  $
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/blkdev.h>
#include <linux/ctype.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/pci.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <linux/timer.h>
#include <linux/utsname.h>


#include <asm/byteorder.h>

#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>

#include "lpfc_sli.h"
#include "lpfc_disc.h"
#include "lpfc_scsi.h"
/* Configuration parameters defined */
#define LPFC_DEF_ICFG
#include "lpfc_cfgparm.h"
#include "lpfc.h"
#include "lpfc_fcp.h"
#include "lpfc_hw.h"
#include "lpfc_logmsg.h"
#include "lpfc_mem.h"
#include "lpfc_version.h"
#include "lpfc_compat.h"
#include "lpfc_crtn.h"

static char *lpfc_drvr_name = LPFC_DRIVER_NAME;

#define FC_EXTEND_TRANS_A 1

/* Binding Definitions: Max string size  */
#define FC_MAX_DID_STRING       6
#define FC_MAX_WW_NN_PN_STRING 16

static struct list_head lpfc_hba_list = LIST_HEAD_INIT(lpfc_hba_list);

uint16_t lpfc_num_hba;	/* count of hba configed */

static struct scsi_transport_template *lpfc_transport_template = 0;




static const char *
lpfc_info(struct Scsi_Host *host)
{
	struct lpfc_hba    *phba = (struct lpfc_hba *) host->hostdata[0];
	int len;
	static char  lpfcinfobuf[128];
	
	memset(lpfcinfobuf,0,128);
	if (phba && phba->pcidev){
	        lpfc_get_hba_model_desc(phba, 0, lpfcinfobuf);
		len = strlen(lpfcinfobuf);	
		snprintf(lpfcinfobuf + len,
			128-len,
	       		" on PCI bus %02x device %02x irq %d",
			phba->pcidev->bus->number,
		 	phba->pcidev->devfn,
			phba->pcidev->irq);
	}
	return lpfcinfobuf;
}


static ssize_t
lpfc_drvr_version_show(struct class_device *cdev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, LPFC_MODULE_DESC "\n");
}

static ssize_t
lpfc_info_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	return snprintf(buf, PAGE_SIZE, "%s\n",lpfc_info(host));
}

static ssize_t
lpfc_serialnum_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	return snprintf(buf, PAGE_SIZE, "%s\n",phba->SerialNumber);
}

static ssize_t
lpfc_fwrev_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	char fwrev[32];
	lpfc_decode_firmware_rev(phba, fwrev, 1);
	return snprintf(buf, PAGE_SIZE, "%s\n",fwrev);
}

static void
lpfc_jedec_to_ascii(int incr, char hdw[])
{
	int i, j;
	for (i = 0; i < 8; i++) {
		j = (incr & 0xf);
		if (j <= 9)
			hdw[7 - i] = 0x30 +  j;
		 else
			hdw[7 - i] = 0x61 + j - 10;
		incr = (incr >> 4);
	}
	hdw[8] = 0;
	return;
}

static ssize_t
lpfc_hdw_show(struct class_device *cdev, char *buf)
{
	char hdw[9];
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	lpfc_vpd_t *vp = &phba->vpd;
	lpfc_jedec_to_ascii(vp->rev.biuRev, hdw);
	return snprintf(buf, PAGE_SIZE, "%s\n", hdw);	
}
static ssize_t
lpfc_events_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	int i = 0, len = 0, get = phba->hba_event_put;
	struct lpfc_hba_event *rec;
	
	if (get == phba->hba_event_get)
		 return snprintf(buf, PAGE_SIZE, "None\n");
	
	for (i = 0; i < MAX_HBAEVT; i++) {
		if (get == 0)
			get = MAX_HBAEVT;
		get--;
		rec = &phba->hbaevt[get];
		switch (rec->fc_eventcode) {
			case 0:
				len += snprintf(buf+len, PAGE_SIZE-len, "---------");
				break;
			case HBA_EVENT_RSCN:
				len += snprintf(buf+len, PAGE_SIZE-len, "RSCN     ");
				break;
			case HBA_EVENT_LINK_UP:
				len += snprintf(buf+len, PAGE_SIZE-len,
						 "LINK UP  ");
				break;
			case HBA_EVENT_LINK_DOWN:
				len += snprintf(buf+len, PAGE_SIZE-len,
							"LINK DOWN");
				break;
			default:
				len += snprintf(buf+len, PAGE_SIZE-len, "?????????");
				break;

		}
		len += snprintf(buf+len, PAGE_SIZE-len, " %d,%d,%d,%d\n",
				 rec->fc_evdata1, rec->fc_evdata2,
				 rec->fc_evdata3, rec->fc_evdata4);
	}
	return len;
}
static ssize_t
lpfc_paramidx_show(struct class_device *cdev, char *buf, int paramidx)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	int val = 0;
	if (phba){
		val = phba->config[paramidx];
		return snprintf(buf, PAGE_SIZE, "%d\n",
				phba->config[paramidx]);
	}
	return 0;
}

static ssize_t
lpfc_paramidx_store(struct class_device *cdev, const char *buf, int paramidx)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	int val=0;
	extern icfgparam lpfc_icfgparam[];
	if (sscanf(buf, "%d", &val) != 1)
		return 0;
	if (phba){
		if (val >= lpfc_icfgparam[paramidx].a_low && val <= lpfc_icfgparam[paramidx].a_hi) {
			phba->config[paramidx] = val;
			return strlen(buf);
		}
	}
	return 0;
}

#define lpfc_param_show(attr)	\
static ssize_t \
lpfc_##attr##_show(struct class_device *cdev, char *buf) \
{ \
	enumparamidx i = attr; \
	return lpfc_paramidx_show(cdev, buf, i); \
}

#define lpfc_param_store(attr) \
static ssize_t \
lpfc_##attr##_store(struct class_device *cdev, const char *buf, size_t count) \
{ \
	enumparamidx i = attr; \
	return lpfc_paramidx_store(cdev, buf, i); \
}

#define LPFC_ATTR_R(name, val, desc) \
static int lpfc_##name = val;\
module_param(lpfc_##name, int, 0);\
MODULE_PARM_DESC(lpfc_##name, desc);\
lpfc_param_show(name)\
static CLASS_DEVICE_ATTR(lpfc_##name, S_IRUGO , lpfc_##name##_show, NULL)

#define LPFC_ATTR_RW(name, val, desc) \
static int lpfc_##name = val;\
module_param(lpfc_##name, int, 0);\
MODULE_PARM_DESC(lpfc_##name, desc);\
lpfc_param_show(name)\
lpfc_param_store(name)\
static CLASS_DEVICE_ATTR(lpfc_##name, S_IRUGO | S_IWUSR,\
			 lpfc_##name##_show, lpfc_##name##_store)

static CLASS_DEVICE_ATTR(info, S_IRUGO, lpfc_info_show, NULL);
static CLASS_DEVICE_ATTR(serialnum, S_IRUGO, lpfc_serialnum_show, NULL);
static CLASS_DEVICE_ATTR(fwrev, S_IRUGO, lpfc_fwrev_show, NULL);
static CLASS_DEVICE_ATTR(hdw, S_IRUGO, lpfc_hdw_show, NULL);
static CLASS_DEVICE_ATTR(events, S_IRUGO, lpfc_events_show, NULL);
static CLASS_DEVICE_ATTR(lpfc_drvr_version, S_IRUGO, lpfc_drvr_version_show,
			 NULL);

/*
# lpfc_log_verbose: Only turn this flag on if you are willing to risk being
# deluged with LOTS of information.
# You can set a bit mask to record specific types of verbose messages:
#
# LOG_ELS                       0x1        ELS events
# LOG_DISCOVERY                 0x2        Link discovery events
# LOG_MBOX                      0x4        Mailbox events
# LOG_INIT                      0x8        Initialization events
# LOG_LINK_EVENT                0x10       Link events
# LOG_IP                        0x20       IP traffic history
# LOG_FCP                       0x40       FCP traffic history
# LOG_NODE                      0x80       Node table events
# LOG_MISC                      0x400      Miscellaneous events
# LOG_SLI                       0x800      SLI events
# LOG_CHK_COND                  0x1000     FCP Check condition flag
# LOG_LIBDFC                    0x2000     LIBDFC events
# LOG_ALL_MSG                   0xffff     LOG all messages
*/
LPFC_ATTR_RW(log_verbose, 0x0, "Verbose logging bit-mask");

/*
# lun_queue_depth:  This parameter is used to limit the number of outstanding
# commands per FCP LUN. Value range is [1,128]. Default value is 30.
*/
LPFC_ATTR_R(lun_queue_depth, 30,
	    "Max number of FCP commands we can queue to a specific LUN");

/*
# Some disk devices have a "select ID" or "select Target" capability.
# From a protocol standpoint "select ID" usually means select the
# Fibre channel "ALPA".  In the FC-AL Profile there is an "informative
# annex" which contains a table that maps a "select ID" (a number
# between 0 and 7F) to an ALPA.  By default, for compatibility with
# older drivers, the lpfc driver scans this table from low ALPA to high
# ALPA.
#
# Turning on the scan-down variable (on  = 1, off = 0) will
# cause the lpfc driver to use an inverted table, effectively
# scanning ALPAs from high to low. Value range is [0,1]. Default value is 1.
#
# (Note: This "select ID" functionality is a LOOP ONLY characteristic
# and will not work across a fabric. Also this parameter will take
# effect only in the case when ALPA map is not available.)
*/
LPFC_ATTR_RW(scan_down, 1,
	     "Start scanning for devices from highest ALPA to lowest");

/*
# lpfc_linkdown_tmo: Determine how long the driver will wait to begin
# linkdown processing when a cable has been pulled or the link has
# otherwise become inaccessible.  Linkdown processing includes failing back
# cmds to the target driver that have been waiting around for the link
# to come back up.  There's a tradeoff here:  small values of the timer
# cause the link to appear to "bounce", while large values of the
# timer can delay redundent HBA recovery in a fault tolerant environment.
# Units are in seconds. A value of 0 means never failback cmds until the
# link comes up. Value range is [0,255]. Default value is 30.
*/
LPFC_ATTR_RW(linkdown_tmo, 30,
	     "Seconds driver will wait before deciding link is really down");

/*
# lpfc_nodev_tmo: If set, it will hold all I/O errors on devices that disappear
# until the timer expires. Value range is [0,255]. Default value is 30.
*/
LPFC_ATTR_RW(nodev_tmo, 30,
	     "Seconds driver will hold I/O waiting for a device to come back");

/*
# lpfc_topology:  link topology for init link
#            0x0  = attempt loop mode then point-to-point
#            0x02 = attempt point-to-point mode only
#            0x04 = attempt loop mode only
#            0x06 = attempt point-to-point mode then loop
# Set point-to-point mode if you want to run as an N_Port.
# Set loop mode if you want to run as an NL_Port. Value range is [0,0x6].
# Default value is 0.
*/
LPFC_ATTR_R(topology, 0, "Select Fibre Channel topology");

/*
# lpfc_link_speed: Link speed selection for initializing the Fibre Channel
# connection.
#       0 = auto select (default)
#       1 = 1 Gigabaud
#       2 = 2 Gigabaud
# Value range is [0,2]. Default value is 0.
*/
LPFC_ATTR_R(link_speed, 0, "Select link speed");

/*
# lpfc_fcp_class:  Determines FC class to use for the FCP protocol.
# Value range is [2,3]. Default value is 3.
*/
LPFC_ATTR_R(fcp_class, 3,
	     "Select Fibre Channel class of service for FCP sequences");

/*
# lpfc_use_adisc: Use ADISC for FCP rediscovery instead of PLOGI. Value range
# is [0,1]. Default value is 0.
*/
LPFC_ATTR_RW(use_adisc, 0,
	     "Use ADISC on rediscovery to authenticate FCP devices");

/*
# lpfc_ack0: Use ACK0, instead of ACK1 for class 2 acknowledgement. Value
# range is [0,1]. Default value is 0.
*/
LPFC_ATTR_R(ack0, 0, "Enable ACK0 support");

/*
# If automap is set, SCSI IDs for all FCP nodes without
# persistent bindings will be automatically generated.
# If new FCP devices are added to the network when the system is down,
# there is no guarantee that these SCSI IDs will remain the same
# when the system is booted again.
# The bind method of the port is used as the binding method of
# automap devices to preserve SCSI IDs between link down and link up.
# If automap is 0, only devices with persistent bindings will be
# recognized by the system. User can change the automap property
# of port instance X by changing the value of lpfcX_automap parameter.
# Value range is [0,1]. Default value is 1.
*/
LPFC_ATTR_R(automap, 1,
	    "Automatically bind FCP devices as they are discovered");

/*
# lpfc_fcp_bind_method: It specifies the method of binding to be used for each
# port. This  binding method is used for persistent binding and automaped
# binding. A value of 1 will force WWNN binding, value of 2 will force WWPN
# binding, value of 3 will force DID binding and value of 4 will force the
# driver to derive binding from ALPA. Any persistent binding whose type does
# not match with the bind method of the port will be ignored. Value range
# is [1,4]. Default value is 2.
*/
LPFC_ATTR_R(fcp_bind_method, 2, "Select the bind method to be used");

/*
# lpfc_cr_delay & lpfc_cr_count: Default values for I/O colaesing
# cr_delay (msec) or cr_count outstanding commands. cr_delay can take
# value [0,63]. cr_count can take value [0,255]. Default value of cr_delay
# is 0. Default value of cr_count is 0. The cr_count feature is disabled if
# cr_delay is set to 0.
*/
static int lpfc_cr_delay = 0;
module_param(lpfc_cr_delay, int , 0);
MODULE_PARM_DESC(lpfc_cr_delay, "A count of milliseconds after which an"
		"interrupt response is generated");

static int lpfc_cr_count = 1;
module_param(lpfc_cr_count, int, 0);
MODULE_PARM_DESC(lpfc_cr_count, "A count of I/O completions after which an"
		"interrupt response is generated");

/*
# lpfc_fdmi_on: controls FDMI support.
#       0 = no FDMI support
#       1 = support FDMI without attribute of hostname
#       2 = support FDMI with attribute of hostname
# Value range [0,2]. Default value is 0.
*/
LPFC_ATTR_RW(fdmi_on, 0, "Enable FDMI support");

/*
# Specifies the maximum number of ELS cmds we can have outstanding (for
# discovery). Value range is [1,64]. Default value = 1.
*/
static int lpfc_discovery_threads = 1;
module_param(lpfc_discovery_threads, int, 0);
MODULE_PARM_DESC(lpfc_discovery_threads, "Maximum number of ELS commands"
		 "during discovery");

int count;			/* thrown away */
extern char *lpfc_fcp_bind_WWNN[MAX_FC_BINDINGS];
extern char *lpfc_fcp_bind_WWPN[MAX_FC_BINDINGS];
extern char *lpfc_fcp_bind_DID[MAX_FC_BINDINGS];
module_param_array(lpfc_fcp_bind_WWPN, charp, count, 0);
module_param_array(lpfc_fcp_bind_WWNN, charp, count, 0);
module_param_array(lpfc_fcp_bind_DID, charp, count, 0);

static void
lpfc_get_port_id(struct scsi_device *sdev)
{
	struct lpfc_target *target = sdev->hostdata;
	if (sdev->host->transportt && target->pnode)
		fc_port_id(sdev) = target->pnode->nlp_DID;
}
static void
lpfc_get_node_name(struct scsi_device *sdev)
{
	struct lpfc_target *target = sdev->hostdata;
	union {
		uint64_t bits64val;
		struct lpfc_name bytes8val;
		} name;
	if (sdev->host->transportt && target->pnode){
		memcpy(&name.bytes8val,
		       &target->pnode->nlp_nodename,
		       sizeof(struct lpfc_name));
		fc_node_name(sdev) = name.bits64val;
	}
}
static void
lpfc_get_port_name(struct scsi_device *sdev)
{
	struct lpfc_target *target = sdev->hostdata;
	union {
		uint64_t bits64val;
		struct lpfc_name bytes8val;
		} name;
	if (sdev->host->transportt && target->pnode){
		memcpy(&name.bytes8val,
		        &target->pnode->nlp_portname,
	        	sizeof(struct lpfc_name));
		fc_port_name(sdev) = name.bits64val;
	}
}
static struct fc_function_template lpfc_transport_functions = {
	.get_port_id  = lpfc_get_port_id,
	.show_port_id = 1,
	.get_node_name = lpfc_get_node_name,
	.show_node_name = 1,
	.get_port_name = lpfc_get_port_name,
	.show_port_name = 1,
};
/*
 * Retrieve lpfc_hba * matching instance (board no)
 * If found return lpfc_hba *
 * If not found return NULL
 */
struct lpfc_hba *
lpfc_get_phba_by_inst(int inst)
{
	struct lpfc_hba * phba;
	int    found = 0;

	phba = NULL;
	list_for_each_entry(phba, &lpfc_hba_list, hba_list) {
		if (phba->brd_no == inst) {
			found = 1;
			break;
		}
	}
	
	if (!found)
		return(NULL);
	else
		return(phba);

}	

static int
lpfc_proc_info(struct Scsi_Host *host,
	       char *buf, char **start, off_t offset, int count, int rw)
{

	struct lpfc_hba *phba = NULL;
	struct pci_dev *pdev;
	char fwrev[32];
	lpfc_vpd_t *vp;
	struct lpfc_nodelist *ndlp;
	char hdw[9];
	int len = 0, found = 0;
	
	/* Sufficient bytes to hold a port or node name. */
	uint8_t name[sizeof (struct lpfc_name)];

	/* If rw = 0, then read info
	 * If rw = 1, then write info (NYI)
	 */
	if (rw)
		return -EINVAL;


	list_for_each_entry(phba, &lpfc_hba_list, hba_list) {
		if (phba == (struct lpfc_hba *)host->hostdata[0]) {
			found = 1;
			break;
		}
	}

	if (!found) {
		return sprintf(buf, "Cannot find adapter for requested host "
			       "number.\n");
	}

	vp = &phba->vpd;
	pdev = phba->pcidev;

	len += snprintf(buf, PAGE_SIZE, LPFC_MODULE_DESC "\n");

	len += snprintf(buf + len, PAGE_SIZE-len, "%s\n",
			lpfc_info(phba->host));

	len += snprintf(buf + len, PAGE_SIZE-len, "SerialNum: %s\n",
			phba->SerialNumber);

	lpfc_decode_firmware_rev(phba, fwrev, 1);
	len += snprintf(buf + len, PAGE_SIZE-len, "Firmware Version: %s\n",
			fwrev);

	len += snprintf(buf + len, PAGE_SIZE-len, "Hdw: ");
	/* Convert JEDEC ID to ascii for hardware version */
	lpfc_jedec_to_ascii(vp->rev.biuRev, hdw);
	len += snprintf(buf + len, PAGE_SIZE-len, hdw);

	len += snprintf(buf + len, PAGE_SIZE-len, "\nVendorId: 0x%x\n",
		       ((((uint32_t) pdev->device) << 16) |
			(uint32_t) (pdev->vendor)));

	/* A Fibre Channel node or port name is 8 octets long and delimited by
	 * colons.
	 */
	len += snprintf(buf + len, PAGE_SIZE-len, "Portname: ");
	memcpy (&name[0], &phba->fc_portname, sizeof (struct lpfc_name));
	len += snprintf(buf + len, PAGE_SIZE-len,
			"%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
			name[0], name[1], name[2], name[3], name[4], name[5],
			name[6], name[7]);

	len += snprintf(buf + len, PAGE_SIZE-len, "   Nodename: ");
	memcpy (&name[0], &phba->fc_nodename, sizeof (struct lpfc_name));
	len += snprintf(buf + len, PAGE_SIZE-len,
			"%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
			name[0], name[1], name[2], name[3], name[4], name[5],
			name[6], name[7]);

	switch (phba->hba_state) {
	case LPFC_INIT_START:
	case LPFC_INIT_MBX_CMDS:
	case LPFC_LINK_DOWN:
		len += snprintf(buf + len, PAGE_SIZE-len, "\n\nLink Down\n");
		break;
	case LPFC_LINK_UP:
	case LPFC_LOCAL_CFG_LINK:
		len += snprintf(buf + len, PAGE_SIZE-len, "\n\nLink Up\n");
		break;
	case LPFC_FLOGI:
	case LPFC_FABRIC_CFG_LINK:
	case LPFC_NS_REG:
	case LPFC_NS_QRY:
	case LPFC_BUILD_DISC_LIST:
	case LPFC_DISC_AUTH:
	case LPFC_CLEAR_LA:
		len += snprintf(buf + len, PAGE_SIZE-len,
				"\n\nLink Up - Discovery\n");
		break;
	case LPFC_HBA_READY:
		len += snprintf(buf + len, PAGE_SIZE-len,
				"\n\nLink Up - Ready:\n");
		len += snprintf(buf + len, PAGE_SIZE-len, "   PortID 0x%x\n",
				phba->fc_myDID);
		if (phba->fc_topology == TOPOLOGY_LOOP) {
			if (phba->fc_flag & FC_PUBLIC_LOOP)
				len += snprintf(buf + len, PAGE_SIZE-len,
						"   Public Loop\n");
			else
				len += snprintf(buf + len, PAGE_SIZE-len,
						"   Private Loop\n");
		} else {
			if (phba->fc_flag & FC_FABRIC)
				len += snprintf(buf + len, PAGE_SIZE-len,
						"   Fabric\n");
			else
				len += snprintf(buf + len, PAGE_SIZE-len,
						"   Point-2-Point\n");
		}
		if (phba->fc_linkspeed == LA_2GHZ_LINK)
			len += snprintf(buf + len, PAGE_SIZE-len,
					"   Current speed 2G\n");
		else
			len += snprintf(buf + len, PAGE_SIZE-len,
					"   Current speed 1G\n\n");

		/* Loop through the list of mapped nodes and dump the known node
		   information. */
		list_for_each_entry(ndlp, &phba->fc_nlpmap_list, nlp_listp) {
			if (ndlp->nlp_state == NLP_STE_MAPPED_NODE){
				len += snprintf(buf + len, PAGE_SIZE -len,
						"lpfc%dt%02x DID %06x WWPN ",
						ndlp->nlp_Target->pHba->brd_no,
						ndlp->nlp_sid, ndlp->nlp_DID);

				/* A Fibre Channel node or port name is 8 octets
				 * long and delimited by colons.
				 */
				memcpy (&name[0], &ndlp->nlp_portname,
					sizeof (struct lpfc_name));
				len += snprintf(buf + len, PAGE_SIZE-len,
						"%02x:%02x:%02x:%02x:%02x:%02x:"
						"%02x:%02x",
						name[0], name[1], name[2],
						name[3], name[4], name[5],
						name[6], name[7]);

				len += snprintf(buf + len, PAGE_SIZE-len,
						" WWNN ");
				memcpy (&name[0], &ndlp->nlp_nodename,
					sizeof (struct lpfc_name));
				len += snprintf(buf + len, PAGE_SIZE-len,
						"%02x:%02x:%02x:%02x:%02x:%02x:"
						"%02x:%02x\n",
						name[0], name[1], name[2],
						name[3], name[4], name[5],
						name[6], name[7]);

			}
			if (PAGE_SIZE - len < 90)
				break;
		}

		if (&ndlp->nlp_listp != &phba->fc_nlpmap_list)
			len += snprintf(buf+len, PAGE_SIZE-len, "...\n");

	}
	return (len);
}

static int
lpfc_slave_alloc(struct scsi_device *scsi_devs)
{
	struct lpfc_hba *phba;
	struct lpfc_target *target;

	/* Store the lun pointer in the scsi_device hostdata pointer provided
	 * the driver has already discovered the target/lun id.  Undiscovered
	 * target/lun information is handled dynamically via the driver's 
	 * discovery state machine.
	 */
	phba = (struct lpfc_hba *) scsi_devs->host->hostdata[0];
	target = lpfc_find_target(phba, scsi_devs->id, scsi_devs->lun, 1);
	if (target)
		scsi_devs->hostdata = target;
		
	/* Always return success from this routine since the driver handles
	 * target/lun discovery and resource allocation dynamically.  Failing
	 * this call causes the midlayer to dump an error message per target 
	 * based on conflicting information from the driver.  Although the 
	 * driver could allocate a target instance here for each scsi device 
	 * passed by the midlayer, a TUR or SCSI_INQUIRY issued to
	 * queuecommand would just fail and result in lpfc_slave_destroy freeing
	 * the memory.  Returning success avoids this situation.
	 */
	return 0;
}

static int
lpfc_slave_configure(struct scsi_device *sdev)
{
	struct lpfc_hba *phba = (struct lpfc_hba *)sdev->host->hostdata[0];

	if (sdev->tagged_supported) {
		sdev->current_tag = 0;
		sdev->queue_depth = phba->config[LPFC_CFG_DFT_LUN_Q_DEPTH];
	} else
		sdev->queue_depth = 16;
	return 0;
}

static void
lpfc_slave_destroy(struct scsi_device *sdev)
{
        /* 
	 * Since the driver allocated no memory in lpfc_slave_alloc,
	 * no resources have to be released here.  Just set the hostdata
	 * pointer to 0 provided the driver previously set it to a 
	 * nonzero value.
	 */
	sdev->hostdata = 0;
}

static int
lpfc_biosparam(struct scsi_device *sdev, struct block_device *bdev,
	       sector_t capacity, int ip[])
{
	int size = capacity;

	ip[0] = 64;
	ip[1] = 32;
	ip[2] = size >> 11;
	if (ip[2] > 1024) {
		ip[0] = 255;
		ip[1] = 63;
		ip[2] = size / (ip[0] * ip[1]);
#ifndef FC_EXTEND_TRANS_A
		if (ip[2] > 1023)
			ip[2] = 1023;
#endif
	}
	return (0);
}
static struct class_device_attribute *lpfc_host_attrs[] = {
	&class_device_attr_info,
	&class_device_attr_serialnum,
	&class_device_attr_fwrev,
	&class_device_attr_hdw,
	&class_device_attr_events,
	&class_device_attr_lpfc_drvr_version,
	&class_device_attr_lpfc_log_verbose,
	&class_device_attr_lpfc_lun_queue_depth,
	&class_device_attr_lpfc_linkdown_tmo,
	&class_device_attr_lpfc_nodev_tmo,
	&class_device_attr_lpfc_automap,
	&class_device_attr_lpfc_fcp_class,
	&class_device_attr_lpfc_use_adisc,
	&class_device_attr_lpfc_ack0,	
	&class_device_attr_lpfc_topology,	
	&class_device_attr_lpfc_scan_down,
	&class_device_attr_lpfc_link_speed,
	&class_device_attr_lpfc_fdmi_on,
	&class_device_attr_lpfc_fcp_bind_method,
	NULL,
};
static struct scsi_host_template driver_template = {
	.module			= THIS_MODULE,
	.name			= LPFC_DRIVER_NAME,
	.info			= lpfc_info,
	.queuecommand		= lpfc_queuecommand,
	.eh_abort_handler	= lpfc_abort_handler,
	.eh_device_reset_handler= lpfc_reset_lun_handler,
	.eh_bus_reset_handler	= lpfc_reset_bus_handler,
	.slave_alloc		= lpfc_slave_alloc,
	.slave_configure	= lpfc_slave_configure,
	.slave_destroy		= lpfc_slave_destroy,
	.bios_param		= lpfc_biosparam,
	.proc_info		= lpfc_proc_info,
	.proc_name		= LPFC_DRIVER_NAME,
	.this_id		= -1,
	.sg_tablesize		= SG_ALL,
	.cmd_per_lun		= 30,
	.shost_attrs		= lpfc_host_attrs,
	.use_clustering		= ENABLE_CLUSTERING,
};

void
lpfc_setup_slim_access(struct lpfc_hba *phba)
{
	phba->MBslimaddr = phba->slim_memmap_p;
	phba->HAregaddr = (uint32_t *) (phba->ctrl_regs_memmap_p) +
		HA_REG_OFFSET;
	phba->HCregaddr = (uint32_t *) (phba->ctrl_regs_memmap_p) +
		HC_REG_OFFSET;
	phba->CAregaddr = (uint32_t *) (phba->ctrl_regs_memmap_p) +
		CA_REG_OFFSET;
	phba->HSregaddr = (uint32_t *) (phba->ctrl_regs_memmap_p) +
		HS_REG_OFFSET;
	return;
}


uint32_t
lpfc_intr_prep(struct lpfc_hba * phba)
{
	uint32_t ha_copy;

	/* Ignore all interrupts during initialization. */
	if (phba->hba_state < LPFC_LINK_DOWN)
		return (0);

	/* Read host attention register to determine interrupt source */
	ha_copy = readl(phba->HAregaddr);

	/* Clear Attention Sources, except ERATT (to preserve status) & LATT
	 *    (ha_copy & ~(HA_ERATT | HA_LATT));
	 */
	writel((ha_copy & ~(HA_LATT | HA_ERATT)), phba->HAregaddr);
	return (ha_copy);
}				/* lpfc_intr_prep */


static int
lpfc_sli_setup(struct lpfc_hba * phba)
{
	int i, totiocb = 0;
	struct lpfc_sli *psli = &phba->sli;
	LPFC_RING_INIT_t *pring;

	psli->sliinit.num_rings = MAX_CONFIGURED_RINGS;
	psli->fcp_ring = LPFC_FCP_RING;
	psli->next_ring = LPFC_FCP_NEXT_RING;
	psli->ip_ring = LPFC_IP_RING;

	for (i = 0; i < psli->sliinit.num_rings; i++) {
		pring = &psli->sliinit.ringinit[i];
		switch (i) {
		case LPFC_FCP_RING:	/* ring 0 - FCP */
			/* numCiocb and numRiocb are used in config_port */
			pring->numCiocb = SLI2_IOCB_CMD_R0_ENTRIES;
			pring->numRiocb = SLI2_IOCB_RSP_R0_ENTRIES;
			pring->numCiocb += SLI2_IOCB_CMD_R1XTRA_ENTRIES;
			pring->numRiocb += SLI2_IOCB_RSP_R1XTRA_ENTRIES;
			pring->numCiocb += SLI2_IOCB_CMD_R3XTRA_ENTRIES;
			pring->numRiocb += SLI2_IOCB_RSP_R3XTRA_ENTRIES;
			pring->iotag_ctr = 0;
			pring->iotag_max =
			    (phba->config[LPFC_CFG_DFT_HBA_Q_DEPTH] * 2);
			pring->fast_iotag = pring->iotag_max;
			pring->num_mask = 0;
			break;
		case LPFC_IP_RING:	/* ring 1 - IP */
			/* numCiocb and numRiocb are used in config_port */
			pring->numCiocb = SLI2_IOCB_CMD_R1_ENTRIES;
			pring->numRiocb = SLI2_IOCB_RSP_R1_ENTRIES;
			pring->num_mask = 0;
			break;
		case LPFC_ELS_RING:	/* ring 2 - ELS / CT */
			/* numCiocb and numRiocb are used in config_port */
			pring->numCiocb = SLI2_IOCB_CMD_R2_ENTRIES;
			pring->numRiocb = SLI2_IOCB_RSP_R2_ENTRIES;
			pring->fast_iotag = 0;
			pring->iotag_ctr = 0;
			pring->iotag_max = 4096;
			pring->num_mask = 4;
			pring->prt[0].profile = 0;	/* Mask 0 */
			pring->prt[0].rctl = FC_ELS_REQ;
			pring->prt[0].type = FC_ELS_DATA;
			pring->prt[0].lpfc_sli_rcv_unsol_event =
			    lpfc_els_unsol_event;
			pring->prt[1].profile = 0;	/* Mask 1 */
			pring->prt[1].rctl = FC_ELS_RSP;
			pring->prt[1].type = FC_ELS_DATA;
			pring->prt[1].lpfc_sli_rcv_unsol_event =
			    lpfc_els_unsol_event;
			pring->prt[2].profile = 0;	/* Mask 2 */
			/* NameServer Inquiry */
			pring->prt[2].rctl = FC_UNSOL_CTL;
			/* NameServer */
			pring->prt[2].type = FC_COMMON_TRANSPORT_ULP;
			pring->prt[2].lpfc_sli_rcv_unsol_event =
			    lpfc_ct_unsol_event;
			pring->prt[3].profile = 0;	/* Mask 3 */
			/* NameServer response */
			pring->prt[3].rctl = FC_SOL_CTL;
			/* NameServer */
			pring->prt[3].type = FC_COMMON_TRANSPORT_ULP;
			pring->prt[3].lpfc_sli_rcv_unsol_event =
			    lpfc_ct_unsol_event;
			break;
		}
		totiocb += (pring->numCiocb + pring->numRiocb);
	}
	if (totiocb > MAX_SLI2_IOCB) {
		/* Too many cmd / rsp ring entries in SLI2 SLIM */
		lpfc_printf_log(phba,
			        KERN_ERR,
			        LOG_INIT,
			        "%d:0462 Too many cmd / rsp ring entries in SLI2 SLIM Data: x%x x%x\n",
			        phba->brd_no,
			        totiocb, MAX_SLI2_IOCB);
	}

#ifdef USE_HGP_HOST_SLIM
	psli->sliinit.sli_flag = LPFC_HGP_HOSTSLIM;
#else
	psli->sliinit.sli_flag = 0;
#endif

	return (0);
}

irqreturn_t
lpfc_intr_handler(int irq, void *dev_id, struct pt_regs * regs)
{
	struct lpfc_hba *phba;
	int intr_status;

	/*
	 * Get the driver's phba structure from the dev_id and
	 * assume the HBA is not interrupting.
	 */
	phba = (struct lpfc_hba *) dev_id;

	if (phba) {
		/* Call SLI to handle the interrupt event. */
		intr_status = lpfc_sli_intr(phba);
		if (intr_status == 0)
			return IRQ_HANDLED;
	}

	return IRQ_NONE;

} /* lpfc_intr_handler */

static int
lpfc_get_bind_type(struct lpfc_hba * phba)
{
	int bind_type = phba->config[LPFC_CFG_BINDMETHOD];
	int ret = LPFC_BIND_WW_NN_PN;

	switch (bind_type) {
	case 1:
		phba->fcp_mapping = FCP_SEED_WWNN;
		break;

	case 2:
		phba->fcp_mapping = FCP_SEED_WWPN;
		break;

	case 3:
		phba->fcp_mapping = FCP_SEED_DID;
		ret = LPFC_BIND_DID;
		break;

	case 4:
		phba->fcp_mapping = FCP_SEED_DID;
		ret = LPFC_BIND_DID;
		break;
	}

	return (ret);
}

static int 
lpfc_get_uint8_t(char *p_in, uint8_t * p_out)
{
        uint32_t tmp;

        if(1 != sscanf(p_in,"%x", &tmp))
                return(0);
        *p_out = tmp;
        return(1);
}

static int 
lpfc_extract_int(char *pchar_in, int * pint_out)
{
int int_tmp;
char c;
			/* Get 1st digit */
	c = *pchar_in++;
	if (!isdigit(c))
		return (0);
	int_tmp = c - 0x30;

			/* Get 2nd lpfc digit */
	c = *pchar_in;
	if (!isdigit(c))
		goto convert_it;
	pchar_in++;
	int_tmp = (int_tmp * 10) + c - 0x30;

			/* Get 3rd lpfc digit */
	c = *pchar_in;
	if (!isdigit(c))
		goto convert_it;
	pchar_in++;
	int_tmp = (int_tmp * 10) + c - 0x30;

	if (int_tmp < 0)
		return (0);
convert_it:
	*pint_out = int_tmp;
	
  return(1);
}

static int 
lpfc_extract_name(struct lpfc_hba * phba,
		int entry, 
		char *pcharin, 
		struct lpfc_name * pnameout,
		int * pinstance, 
		int * ptarget, 
		int type)
{
char dummy[40];
char * pcharI = pcharin;
int in_len = 8;
int in_size, count;
int rstatus = LPFC_SYNTAX_ERR_ASC_CONVERT;
char c;
uint8_t IEEE[4];        /* FC IEEE address */

	memset(pnameout, '\0', sizeof(struct lpfc_name));
	if (type == LPFC_BIND_DID) 
		in_len = 4;
	
	in_size = 2*in_len;
	pcharI = strchr(pcharin, ':');
	if(!pcharI) {
		rstatus = LPFC_SYNTAX_ERR_EXP_COLON;
		goto syntax_error;
	}
	memcpy(dummy, pcharin, in_size);
	pcharI = &dummy[in_size];
	
	for(count = in_len - 1; count >= 0; count--) {
		c = *(pcharI); 
		*(pcharI) = '\0';
		pcharI -=2;
		if (1 != lpfc_get_uint8_t(pcharI, &IEEE[0])){
			goto input_not_ascii;
		}
		if(count == 0){
			pnameout->IEEEextMsn = IEEE[0];
			pnameout->nameType = IEEE[0] >> 4;
		}
		else
		if(count == 1)
			pnameout->IEEEextLsb = IEEE[0];
		else
		if(count > 1)
			pnameout->IEEE[count - 2] = IEEE[0];
		*(pcharI + 2) = c; 
	}
	pcharI = strchr(pcharin, ':');
	pcharI ++;

	pcharI = strstr(pcharI, "lpfc");
	if(!pcharI) { 
		rstatus = LPFC_SYNTAX_ERR_EXP_LPFC;
		goto syntax_error;
	}
	pcharI +=4;

	/* Parse lpfc number */
	if(lpfc_extract_int(pcharI, pinstance) != 1) {
		rstatus = LPFC_SYNTAX_ERR_INV_LPFC_NUM;
		goto syntax_error;
	}
	if ((phba = lpfc_get_phba_by_inst(*pinstance)) == NULL) {
		/* Skip this entry */
		return (LPFC_SYNTAX_OK_BUT_NOT_THIS_BRD);
	}
	pcharI+=2;

	/* Parse target number */
	if(lpfc_extract_int(pcharI, ptarget) != 1) {
		rstatus = LPFC_SYNTAX_ERR_INV_TARGET_NUM;
		goto syntax_error;
	}
	return (LPFC_SYNTAX_OK);	/* Success */

syntax_error:
	/* For syntax error code definitions see LPFC_SYNTAX_ERR_ defines. 
	   WWNN binding entry <num>: Syntax error code <code> */
	switch(phba->fcp_mapping) {
		case FCP_SEED_WWPN:
			lpfc_printf_log(phba,
					KERN_ERR,
					LOG_INIT,
					"%d:0430 WWPN binding entry %d: Syntax error code %d\n",
					phba->brd_no,
					entry, rstatus);
		break;
		case FCP_SEED_WWNN:
			lpfc_printf_log(phba,
					KERN_ERR,
					LOG_INIT,
					"%d:0431 WWNN binding entry %d: Syntax error code %d\n",
					phba->brd_no,
					entry, rstatus);
		break;
		case FCP_SEED_DID:
			lpfc_printf_log(phba,
					KERN_ERR,
					LOG_INIT,
					"%d:0432 DID binding entry %d: Syntax error code %d\n",
					phba->brd_no,
					entry, rstatus);
		break;
	}

	return(0);

input_not_ascii:
	/* Convert ASC to hex. Input char seq not ASC hex. */
	lpfc_printf_log(phba,
	       KERN_ERR,
	       LOG_MISC,
	       "%d:1213 Convert ASC to hex. Input char seq not ASC hex.\n",
	       phba->brd_no);
	return (LPFC_SYNTAX_ERR_ASC_CONVERT);
}

/******************************************************************************
* Function name : lpfc_parse_binding_entry
*
* Description   : Parse binding entry for WWNN & WWPN
*
* ASCII Input string example: 2000123456789abc:lpfc1t0
*
* Return        :  0              = Success
*                  Greater than 0 = Binding entry syntax error. SEE defs
*                                   LPFC_SYNTAX_ERR_XXXXXX.
******************************************************************************/
static int
lpfc_parse_binding_entry(struct lpfc_hba * phba, char **arrayp, u_int type)
{
	int instance = 0, target = 0;
	int entry;
	char *pchar;
	int ret;
	struct lpfc_name lpfc_n;
	struct lpfc_bindlist *blp;

	for (entry=0; entry < MAX_FC_BINDINGS; entry++)	{
		memset(&lpfc_n, 0, sizeof(struct lpfc_name));
		pchar = arrayp[entry];
		if(pchar) {
			ret = lpfc_extract_name(phba, entry, pchar, &lpfc_n, &instance, &target, type);
			if(ret == LPFC_SYNTAX_OK) { // Success
				/* Loop through all BINDLIST entries and 
				   find the next available entry. */
				if ((blp = mempool_alloc(phba->bind_mem_pool, GFP_ATOMIC)) == 0) 
					return(-1);
				
				memset(blp, 0, sizeof (struct lpfc_bindlist));
				blp->nlp_bind_type = phba->fcp_mapping;
				blp->nlp_sid = (target & 0xff);
				switch (phba->fcp_mapping) {
				case FCP_SEED_WWPN:
					memcpy(&blp->nlp_portname, &lpfc_n, sizeof (struct lpfc_name));
				break;
				case FCP_SEED_WWNN:
					memcpy(&blp->nlp_nodename, &lpfc_n, sizeof (struct lpfc_name));
				break;
				case FCP_SEED_DID:
					memcpy(&blp->nlp_DID, &lpfc_n, sizeof (struct lpfc_name));
				break;
				}
		                lpfc_nlp_bind(phba, blp);
			}
		}
	}
	return (0);
}				
static int
lpfc_bind_setup(struct lpfc_hba * phba)
{
	char **arrayp = 0;
	u_int cnt = 0;
	int type;

	/* Check if there are any WWNN / scsid bindings  */
	type = lpfc_get_bind_type(phba);

	switch (phba->fcp_mapping) {
	case FCP_SEED_WWNN:
		arrayp = lpfc_fcp_bind_WWNN;
		cnt = 0;
		while (arrayp[cnt] != 0)
			cnt++;
		if (cnt && (*arrayp != 0)) {
			if(lpfc_parse_binding_entry(phba, arrayp, type)) 
				/* WWPN binding entry: node table full */
				lpfc_printf_log(phba,
						KERN_ERR,
						LOG_INIT,
						"%d:0434 WWNN binding entry: node table full\n",
						phba->brd_no);
		}
		break;

	case FCP_SEED_WWPN:
		arrayp = lpfc_fcp_bind_WWPN;
		cnt = 0;
		while (arrayp[cnt] != 0)
			cnt++;
		if (cnt && (*arrayp != 0)) {
			if(lpfc_parse_binding_entry(phba, arrayp, type)) 
				/* WWPN binding entry: node table full */
				lpfc_printf_log(phba,
						KERN_ERR,
						LOG_INIT,
						"%d:0433 WWPN binding entry: node table full\n",
						phba->brd_no);
			
		}
		break;

	case FCP_SEED_DID:
		if (phba->config[LPFC_CFG_BINDMETHOD] != 4) {
			arrayp = lpfc_fcp_bind_DID;
			cnt = 0;
			while (arrayp[cnt] != 0)
				cnt++;
			if (cnt && (*arrayp != 0)) {
				if(lpfc_parse_binding_entry(phba, arrayp, type)) 
					/* WWPN binding entry: node table full */
					lpfc_printf_log(phba,
							KERN_ERR,
							LOG_INIT,
							"%d:0435 DID binding entry: node table full\n",
							phba->brd_no);
			}
		}
		break;
	}

	if (cnt && (*arrayp != 0) &&
	    (phba->config[LPFC_CFG_BINDMETHOD] == 4)) 
		/* Using ALPA map with Persistent binding - ignoring ALPA map */
		lpfc_printf_log(phba,
			         KERN_ERR,
			         LOG_INIT,
			         "%d:0411 fcp_bind_method is 4 with Persistent binding - ignoring fcp_bind_method Data: x%x x%x\n",
			         phba->brd_no,
			        phba->config[LPFC_CFG_BINDMETHOD],
				phba->fcp_mapping);
	

	return (0);
}

static void
lpfc_wakeup_event(struct lpfc_hba * phba, fcEVTHDR_t * ep)
{
	ep->e_mode &= ~E_SLEEPING_MODE;
	switch (ep->e_mask) {
	case FC_REG_LINK_EVENT:
		wake_up_interruptible(&phba->linkevtwq);
		break;
	case FC_REG_RSCN_EVENT:
		wake_up_interruptible(&phba->rscnevtwq);
		break;
	case FC_REG_CT_EVENT:
		wake_up_interruptible(&phba->ctevtwq);
		break;
	}
	return;
}

static uint32_t
lpfc_get_cfgparam(int brd, int param)
{
	uint32_t value = (uint32_t)-1;
	
	switch (param) {
	case LPFC_CFG_LOG_VERBOSE:	/* log-verbose */
		value = lpfc_log_verbose;
		break;
	case LPFC_CFG_AUTOMAP:	/* automap */
		value = lpfc_automap;
		break;
	case LPFC_CFG_BINDMETHOD:	/* bind-method */
		value = lpfc_fcp_bind_method;
		break;
	case LPFC_CFG_CR_DELAY:	/* cr_delay */
		value = lpfc_cr_delay;
		break;
	case LPFC_CFG_CR_COUNT:	/* cr_count */
		value = lpfc_cr_count;
		break;
	case LPFC_CFG_DFT_LUN_Q_DEPTH:	/* lun_queue_depth */
		value = lpfc_lun_queue_depth;
		break;
	case LPFC_CFG_FCP_CLASS:	/* fcp-class */
		value = lpfc_fcp_class;
		break;
	case LPFC_CFG_USE_ADISC:	/* use-adisc */
		value = lpfc_use_adisc;
		break;
	case LPFC_CFG_ACK0:	/* ack0 */
		value = lpfc_ack0;
		break;
	case LPFC_CFG_TOPOLOGY:	/* topology */
		value = lpfc_topology;
		break;
	case LPFC_CFG_SCAN_DOWN:	/* scan-down */
		value = lpfc_scan_down;
		break;
	case LPFC_CFG_LINKDOWN_TMO:	/* linkdown-tmo */
		value = lpfc_linkdown_tmo;
		break;
	case LPFC_CFG_NODEV_TMO:	/* nodev-tmo */
		value = lpfc_nodev_tmo;
		break;
	case LPFC_CFG_LINK_SPEED:	/* link-speed */
		value = lpfc_link_speed;
		break;
	case LPFC_CFG_FDMI_ON:	/* fdmi-on */
		value = lpfc_fdmi_on;
		break;
	case LPFC_CFG_DISC_THREADS:	/* discovery-threads */
		value = lpfc_discovery_threads;
		break;
	default:
		break;
	}

	return (value);
}
/******************************************************************************
* Function name : lpfc_config_setup
*
* Description   : Called from attach to setup configuration parameters for
*                 adapter
*                 The goal of this routine is to fill in all the
*                 members of the CfgParam structure for all configuration
*                 parameters.
* Example:
* phba->config[LPFC_CFG_XXX] = (uint32_t)value;
* value might be a define, a global variable, phba->config[LPFC_CFG_XXX].a_default,
* or some other enviroment specific way of initializing config parameters.
******************************************************************************/

static int
lpfc_config_setup(struct lpfc_hba * phba)
{
	struct lpfc_sli *psli;
	int i;
	int brd;

	psli = &phba->sli;
	brd = phba->brd_no;

	/*
	 * Read the configuration parameters. Also set to default if
	 * parameter value is out of allowed range.
	 */
	for (i = 0; i < LPFC_TOTAL_NUM_OF_CFG_PARAM; i++) {
		phba->config[i] = lpfc_get_cfgparam(brd, i);

		if (i == LPFC_CFG_DFT_HBA_Q_DEPTH)
			continue;

		if ((phba->config[i] >= lpfc_icfgparam[i].a_low) &&
		    (phba->config[i] <= lpfc_icfgparam[i].a_hi)) {
			/* we continue if the range check is satisfied
			 * however LPFC_CFG_TOPOLOGY has holes and needs
			 * to be checked for correctness.
			 */
			if (i == LPFC_CFG_TOPOLOGY) {
				/* odd values 1,3,5 are out */
				if (!(phba->config[i] & 1))
					continue;
			} else
				continue;
		}

		/* The cr_count feature is disabled if cr_delay is set to 0.  So
		   do not bother user with messages about cr_count if cr_delay
		   is 0 */
		if (i == LPFC_CFG_CR_COUNT)
			if (phba->config[LPFC_CFG_CR_DELAY] == 0)
				continue;

		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_INIT,
				"%d:0413 Configuration parameter lpfc_%s out of range [%d,%d]. Using default value %d\n",
				phba->brd_no,
				lpfc_icfgparam[i].a_string,
				lpfc_icfgparam[i].a_low,
				lpfc_icfgparam[i].a_hi,
				lpfc_icfgparam[i].a_default);

		phba->config[i] = lpfc_icfgparam[i].a_default;

	}

	switch (phba->pcidev->device) {
	case PCI_DEVICE_ID_LP101:
		phba->config[LPFC_CFG_DFT_HBA_Q_DEPTH] =
			LPFC_LP101_HBA_Q_DEPTH;
		break;
	case PCI_DEVICE_ID_RFLY:
	case PCI_DEVICE_ID_PFLY:
	case PCI_DEVICE_ID_TFLY:
		phba->config[LPFC_CFG_DFT_HBA_Q_DEPTH] = LPFC_LC_HBA_Q_DEPTH;
		break;
	default:
		phba->config[LPFC_CFG_DFT_HBA_Q_DEPTH] = LPFC_DFT_HBA_Q_DEPTH;
	}

	if (phba->config[LPFC_CFG_DISC_THREADS]) {
		/*
		 * Set to FC_NLP_REQ if automap is set to 0 since order of
		 * discovery does not matter if everything is persistently
		 * bound.
		 */
		if (phba->config[LPFC_CFG_AUTOMAP] == 0) {
			phba->config[LPFC_CFG_DISC_THREADS] =
			    LPFC_MAX_DISC_THREADS;
		}
	}


	return (0);
}

int
lpfc_put_event(struct lpfc_hba * phba, uint32_t evcode, uint32_t evdata0,
	       uint32_t evdata1, uint32_t evdata2, uint32_t evdata3)
{
	fcEVT_t *ep;
	fcEVTHDR_t *ehp = phba->fc_evt_head;
	int found = 0;
	void *fstype = 0;
	struct lpfc_dmabuf *mp;
	struct lpfc_sli_ct_request *ctp;
	struct lpfc_hba_event *rec;
	uint32_t evtype;

	switch (evcode) {
		case HBA_EVENT_RSCN:
			evtype = FC_REG_RSCN_EVENT;
			break;
		case HBA_EVENT_LINK_DOWN:
		case HBA_EVENT_LINK_UP:
			evtype = FC_REG_LINK_EVENT;
			break;
		default:
			evtype = FC_REG_CT_EVENT;
	}

	if (evtype == FC_REG_RSCN_EVENT || evtype == FC_REG_LINK_EVENT) {
		rec = &phba->hbaevt[phba->hba_event_put];
		rec->fc_eventcode = evcode;
		rec->fc_evdata1 = evdata0;
		rec->fc_evdata2 = evdata1;
		rec->fc_evdata3 = evdata2;
		rec->fc_evdata4 = evdata3;

		phba->hba_event_put++;
		if (phba->hba_event_put >= MAX_HBAEVT)
			phba->hba_event_put = 0;

		if (phba->hba_event_put == phba->hba_event_get) {
			phba->hba_event_missed++;
			phba->hba_event_get++;
			if (phba->hba_event_get >= MAX_HBAEVT)
				phba->hba_event_get = 0;
		}
	}

	if (evtype == FC_REG_CT_EVENT) {
		mp = (struct lpfc_dmabuf *) evdata1;
		ctp = (struct lpfc_sli_ct_request *) mp->virt;
		fstype = (void *)(ulong) (ctp->FsType);
	}

	while (ehp && ((ehp->e_mask != evtype) || (ehp->e_type != fstype)))
		ehp = (fcEVTHDR_t *) ehp->e_next_header;

	if (!ehp)
		return (0);

	ep = ehp->e_head;

	while (ep && !(found)) {
		if (ep->evt_sleep) {
			switch (evtype) {
			case FC_REG_CT_EVENT:
				if ((ep->evt_type ==
				     (void *)(ulong) FC_FSTYPE_ALL)
				    || (ep->evt_type == fstype)) {
					found++;
					ep->evt_data0 = evdata0; /* tag */
					ep->evt_data1 = evdata1; /* buffer
								    ptr */
					ep->evt_data2 = evdata2; /* count */
					ep->evt_sleep = 0;
					if (ehp->e_mode & E_SLEEPING_MODE) {
						ehp->e_flag |=
						    E_GET_EVENT_ACTIVE;
						lpfc_wakeup_event(phba, ehp);
					}
					/* For FC_REG_CT_EVENT just give it to
					   first one found */
				}
				break;
			default:
				found++;
				ep->evt_data0 = evdata0;
				ep->evt_data1 = evdata1;
				ep->evt_data2 = evdata2;
				ep->evt_sleep = 0;
				if ((ehp->e_mode & E_SLEEPING_MODE)
				    && !(ehp->e_flag & E_GET_EVENT_ACTIVE)) {
					ehp->e_flag |= E_GET_EVENT_ACTIVE;
					lpfc_wakeup_event(phba, ehp);
				}
				/* For all other events, give it to every one
				   waiting */
				break;
			}
		}
		ep = ep->evt_next;
	}
	if (evtype == FC_REG_LINK_EVENT)
		phba->nport_event_cnt++;

	return (found);
}

void
lpfc_nodev(unsigned long l)
{
	return;
}

/*
 * This is only called to handle FC discovery events. Since this a rare
 * occurance, we allocate an LPFC_DISC_EVT_t structure here instead of
 * embedding it in the IOCB.
 */
int
lpfc_discq_post_event(struct lpfc_hba * phba, void *arg1, void *arg2,
		      uint32_t evt)
{
	LPFC_DISC_EVT_t  *evtp;
	unsigned long flags;

	/* All Mailbox completions and LPFC_ELS_RING rcv ring IOCB events
	 * will be queued to DPC for processing
	 */
	evtp = (LPFC_DISC_EVT_t *) kmalloc(sizeof(LPFC_DISC_EVT_t), GFP_ATOMIC);
	if (!evtp)
		return 0;

	evtp->evt_arg1  = arg1;
	evtp->evt_arg2  = arg2;
	evtp->evt       = evt;
	evtp->evt_listp.next = 0;
	evtp->evt_listp.prev = 0;

	/* Queue the event to the DPC to be processed later */
	spin_lock_irqsave(&phba->dpc_lock, flags);
	list_add_tail(&evtp->evt_listp, &phba->dpc_disc);
	spin_unlock_irqrestore(&phba->dpc_lock, flags);
	if (phba->dpc_wait)
		up(phba->dpc_wait);

	return 1;
}


static void
lpfc_evt_iocb_free(struct lpfc_hba * phba, struct lpfc_iocbq * saveq)
{
	struct lpfc_iocbq  *rspiocbp, *tmpiocbp;

	/* Free up iocb buffer chain for cmd just processed */
	list_for_each_entry_safe(rspiocbp, tmpiocbp,
		&saveq->list, list) {
		list_del(&rspiocbp->list);
		mempool_free( rspiocbp, phba->iocb_mem_pool);
	}
	mempool_free( saveq, phba->iocb_mem_pool);
}

int
lpfc_disc_done(struct lpfc_hba * phba)
{
	struct lpfc_sli *psli = &phba->sli;
	LPFC_DISC_EVT_t  *evtp, *next_evtp;
	LPFC_MBOXQ_t *pmb;
	struct lpfc_iocbq  *cmdiocbp, *saveq;
	LPFC_RING_MASK_t *func;
	LIST_HEAD(local_dpc_disc);
	unsigned long      flags;

	spin_lock_irqsave(&phba->dpc_lock, flags);
	list_splice_init(&phba->dpc_disc, &local_dpc_disc);
	spin_unlock_irqrestore(&phba->dpc_lock, flags);
	/* check discovery event list */
	list_for_each_entry_safe(evtp, next_evtp, &local_dpc_disc, evt_listp) {
		list_del(&evtp->evt_listp);

		spin_lock_irqsave(&(phba->drvrlock), flags);
		switch(evtp->evt) {
		case LPFC_EVT_MBOX:
			pmb = (LPFC_MBOXQ_t *)(evtp->evt_arg1);
			(pmb->mbox_cmpl) (phba, pmb);
			break;
		case LPFC_EVT_SOL_IOCB:
			cmdiocbp = (struct lpfc_iocbq *)(evtp->evt_arg1);
			saveq = (struct lpfc_iocbq *)(evtp->evt_arg2);
			(cmdiocbp->iocb_cmpl) (phba, cmdiocbp, saveq);
			lpfc_evt_iocb_free(phba, saveq);
			break;
		case LPFC_EVT_UNSOL_IOCB:
			func = (LPFC_RING_MASK_t *)(evtp->evt_arg1);
			saveq = (struct lpfc_iocbq *)(evtp->evt_arg2);
			(func->lpfc_sli_rcv_unsol_event) (phba,
	 		&psli->ring[LPFC_ELS_RING], saveq);
			lpfc_evt_iocb_free(phba, saveq);
			break;
		}
		spin_unlock_irqrestore(&(phba->drvrlock), flags);

		kfree(evtp);
	}
	return(0);
}

static int
lpfc_do_dpc(void *p)
{
	DECLARE_MUTEX_LOCKED(sem);
	struct lpfc_hba *phba = (struct lpfc_hba *)p;
	unsigned long flags;

	lock_kernel();

	daemonize("lpfc_dpc_%d", phba->brd_no);
	allow_signal(SIGHUP);

	phba->dpc_wait = &sem;
	set_user_nice(current, -20);

	unlock_kernel();

	complete(&phba->dpc_startup);

	while (1) {
		if (down_interruptible(&sem))
			break;

		if (signal_pending(current) )
			break;

		if (phba->dpc_kill)
			break;

		spin_lock_irqsave(&phba->dpc_lock, flags);
		if (!list_empty(&phba->dpc_disc)) {
			spin_unlock_irqrestore(&phba->dpc_lock, flags);
			lpfc_disc_done(phba);
		}
		else {
			spin_unlock_irqrestore(&phba->dpc_lock, flags);
		}
	}
	/*
	* Zero out semaphore we were waiting on.
	*/
	phba->dpc_wait = NULL;

	complete_and_exit(&phba->dpc_exiting, 0);

	return(0);
}

static int __devinit
lpfc_pci_probe_one(struct pci_dev *pdev, const struct pci_device_id *pid)
{
	struct Scsi_Host *host;
	struct lpfc_hba  *phba;
	struct lpfc_sli  *psli;
	unsigned long iflag;
	uint32_t timeout;
	unsigned long bar0map_len, bar2map_len;
	int error = -ENODEV, i, retval;

	if (pci_enable_device(pdev))
		goto out;
	if (pci_request_regions(pdev, LPFC_DRIVER_NAME))
		goto out_disable_device;

	/*
	 * Allocate space for adapter info structure
	 */
	phba = kmalloc(sizeof(*phba), GFP_KERNEL);
	if (!phba)
		goto out_release_regions;

	memset(phba, 0, sizeof (struct lpfc_hba));
	INIT_LIST_HEAD(&phba->freebufList);

	/* Initialize timers used by driver */
	init_timer(&phba->fc_estabtmo);
	phba->fc_estabtmo.function = lpfc_establish_link_tmo;
	init_timer(&phba->fc_disctmo);
	phba->fc_disctmo.function = lpfc_disc_timeout;
	init_timer(&phba->fc_linkdown);
	phba->fc_linkdown.function = lpfc_linkdown_timeout;
	init_timer(&phba->fc_fdmitmo);
	phba->fc_fdmitmo.function = lpfc_fdmi_tmo;
	init_timer(&phba->els_tmofunc);
	phba->els_tmofunc.function = lpfc_els_timeout_handler;
	init_timer(&phba->scsi_tmofunc);
	phba->scsi_tmofunc.function = lpfc_scsi_timeout_handler;
	psli = &phba->sli;
	init_timer(&psli->mbox_tmo);
	psli->mbox_tmo.function = lpfc_mbox_timeout;

	/* Set everything to the defaults */
	for (i = 0; i < LPFC_TOTAL_NUM_OF_CFG_PARAM; i++)
		phba->config[i] = lpfc_icfgparam[i].a_default;
	
	/* Assign an unused board number */
	i = 0;
	while (1) {
		if (lpfc_get_phba_by_inst(i) == NULL ) {
			phba->brd_no = i;
			break;
		}
		i++;
	}

	/* Add adapter structure to list */
	list_add_tail(&phba->hba_list, &lpfc_hba_list);

	/* Initialize all internally managed lists. */
	INIT_LIST_HEAD(&phba->fc_nlpmap_list);
	INIT_LIST_HEAD(&phba->fc_nlpunmap_list);
	INIT_LIST_HEAD(&phba->fc_plogi_list);
	INIT_LIST_HEAD(&phba->fc_adisc_list);
	INIT_LIST_HEAD(&phba->fc_nlpbind_list);

	/* Initialize plxhba - LINUX specific */
	phba->pcidev = pdev;
	init_waitqueue_head(&phba->linkevtwq);
	init_waitqueue_head(&phba->rscnevtwq);
	init_waitqueue_head(&phba->ctevtwq);

	pci_set_master(pdev);
	retval = pci_set_mwi(pdev);
	if (retval)
		dev_printk(KERN_WARNING, &pdev->dev,
			   "Warning: pci_set_mwi returned %d\n", retval);
	
	/* Configure DMA attributes. */
	if (dma_set_mask(&phba->pcidev->dev, 0xffffffffffffffffULL) &&
		dma_set_mask(&phba->pcidev->dev, 0xffffffffULL))
			goto out_list_del;

	/*
	 * Get the physical address of Bar0 and Bar2 and the number of bytes
	 * required by each mapping.
	 */
	phba->pci_bar0_map = pci_resource_start(phba->pcidev, 0);
	bar0map_len        = pci_resource_len(phba->pcidev, 0);
	
	phba->pci_bar2_map = pci_resource_start(phba->pcidev, 2);
	bar2map_len        = pci_resource_len(phba->pcidev, 2);

	/* Map HBA SLIM and Control Registers to a kernel virtual address. */
	phba->slim_memmap_p      = ioremap(phba->pci_bar0_map, bar0map_len);
	phba->ctrl_regs_memmap_p = ioremap(phba->pci_bar2_map, bar2map_len);

	/* Setup SLI2 interface */
	if (!phba->slim2p.virt) {
		/*
		 * Allocate memory for SLI-2 structures
		 */
		phba->slim2p.virt = dma_alloc_coherent(&phba->pcidev->dev,
						       sizeof (SLI2_SLIM_t),
						       &(phba->slim2p.phys),
						       GFP_ATOMIC);
		
		if (!phba->slim2p.virt)
			goto out_iounmap;

		/* The SLIM2 size is stored in the next field */
		phba->slim_size = sizeof (SLI2_SLIM_t);
		memset(phba->slim2p.virt, 0, sizeof (SLI2_SLIM_t));
	}

	lpfc_num_hba++;
	lpfc_config_setup(phba);	/* Setup configuration parameters */

       	/* Register this board */
	host = scsi_host_alloc(&driver_template, sizeof (unsigned long));

	if (!host) {
		printk (KERN_WARNING "%s%d: scsi_host_alloc failed during "
			"attach\n", lpfc_drvr_name, phba->brd_no);
		error = -ENOMEM;
		goto out_cleanup_timers;
	}

	phba->host = host;
	host->can_queue = phba->config[LPFC_CFG_DFT_HBA_Q_DEPTH] - 10;
	host->max_id = LPFC_MAX_TARGET;

	host->unique_id = lpfc_num_hba;

	host->max_lun = LPFC_MAX_LUN; 

	/* Adapter ID - tell midlayer not to reserve an ID for us */
	host->this_id = -1;


	/* Initialize all per HBA locks */
	spin_lock_init(&phba->drvrlock);
	spin_lock_init(&phba->hiprilock);

	lpfc_sli_setup(phba);	/* Setup SLI Layer to run over lpfc HBAs */
	lpfc_sli_queue_setup(phba);	/* Initialize the SLI Layer */

	if (!lpfc_mem_alloc(phba))
		goto out_dec_nhbas;

	lpfc_bind_setup(phba);	/* Setup binding configuration parameters */

	/* Initialize HBA structure */
	phba->fc_edtov = FF_DEF_EDTOV;
	phba->fc_ratov = FF_DEF_RATOV;
	phba->fc_altov = FF_DEF_ALTOV;
	phba->fc_arbtov = FF_DEF_ARBTOV;

	/* Set the FARP and XRI timeout values now since they depend on
	   fc_ratov. */
	phba->fc_ipfarp_timeout = (3 * phba->fc_ratov);
	phba->fc_ipxri_timeout = (3 * phba->fc_ratov);

	spin_lock_init(&phba->dpc_lock);
	INIT_LIST_HEAD(&phba->dpc_disc);
	init_completion(&phba->dpc_startup);
	init_completion(&phba->dpc_exiting);

	/*
	* Startup the kernel thread for this host adapter
	*/
	phba->dpc_kill = 0;
	phba->dpc_pid = kernel_thread(lpfc_do_dpc, phba, 0);
	if (phba->dpc_pid < 0) {
		error = phba->dpc_pid;
		goto out_free_mem;
	}
	wait_for_completion(&phba->dpc_startup);



	/* Call SLI to initialize the HBA. */
	error = lpfc_sli_hba_setup(phba);
	if (error)
		goto out_hba_down;



	/*
	 * Setup the scsi timeout handler with a
	 * timeout value = greater of (2*RATOV, 5).
	 */
	timeout = (phba->fc_ratov << 1) > 5 ? (phba->fc_ratov << 1) : 5;
	phba->scsi_tmofunc.expires = jiffies + HZ * timeout;
	phba->scsi_tmofunc.data = (unsigned long)phba;
	add_timer(&phba->scsi_tmofunc);


	/*
	 * Starting with 2.4.0 kernel, Linux can support commands longer
	 * than 12 bytes. However, scsi_register() always sets it to 12.
	 * For it to be useful to the midlayer, we have to set it here.
	 */
	host->max_cmd_len = 16;

	/*
	 * Queue depths per lun
	 */
	host->transportt = lpfc_transport_template;
	host->hostdata[0] = (unsigned long)phba;
	pci_set_drvdata(pdev, host);
	error = scsi_add_host(host, &pdev->dev);
	if (error)
		goto out_put_host;
	scsi_scan_host(host);
	return 0;
out_put_host:
	scsi_host_put(host);
out_hba_down:
	lpfc_sli_hba_down(phba);

out_cleanup_timers:
	/* Stop any timers that were started during this attach. */
	spin_lock_irqsave(&(phba->drvrlock), iflag);
	lpfc_stop_timer(phba);
	spin_unlock_irqrestore(&(phba->drvrlock), iflag);

	/* Kill the kernel thread for this host */
	if (phba->dpc_pid >= 0) {
		phba->dpc_kill = 1;
		wmb();
		kill_proc(phba->dpc_pid, SIGHUP, 1);
		wait_for_completion(&phba->dpc_exiting);
	}

	free_irq(phba->pcidev->irq, phba);
out_free_mem:
	lpfc_mem_free(phba);
out_dec_nhbas:
	lpfc_num_hba--;
	if (phba->slim2p.virt) {
		dma_free_coherent(&pdev->dev,
				  phba->slim_size,
				  phba->slim2p.virt,
				  phba->slim2p.phys);
	}
out_iounmap:
	iounmap(phba->ctrl_regs_memmap_p);
	iounmap(phba->slim_memmap_p);
out_list_del:
	list_del_init(&phba->hba_list);
	kfree(phba);
out_release_regions:
	pci_release_regions(pdev);
out_disable_device:
	pci_disable_device(pdev);
out:
	return error;
}

int
lpfc_stop_timer(struct lpfc_hba * phba)
{
	struct lpfc_sli *psli = &phba->sli;

	/* Instead of a timer, this has been converted to a
	 * deferred procedding list.
	 */
	while (!list_empty(&phba->freebufList)) {
		struct lpfc_dmabuf *mp;

		mp = (struct lpfc_dmabuf *)(phba->freebufList.next);
		if (mp) {
			lpfc_mbuf_free(phba, mp->virt, mp->phys);
			list_del(&mp->list);
			kfree((void *)mp);
		}
	}

	if (phba->fc_estabtmo.expires) {
		del_timer_sync(&phba->fc_estabtmo);
		phba->fc_estabtmo.expires = 0;
	}
	if (phba->fc_disctmo.expires) {
		del_timer_sync(&phba->fc_disctmo);
		phba->fc_disctmo.expires = 0;
	}
	if (phba->fc_linkdown.expires) {
		del_timer_sync(&phba->fc_linkdown);
		phba->fc_linkdown.expires = 0;
	}
	if (phba->fc_fdmitmo.expires) {
		del_timer_sync(&phba->fc_fdmitmo);
		phba->fc_fdmitmo.expires = 0;
	}
	if (phba->els_tmofunc.expires) {
		del_timer_sync(&phba->els_tmofunc);
		phba->els_tmofunc.expires = 0;
	}
	if (phba->scsi_tmofunc.expires) {
		del_timer_sync(&phba->scsi_tmofunc);
		phba->scsi_tmofunc.expires = 0;
	}
	psli = &phba->sli;
	if (psli->mbox_tmo.expires) {
		del_timer_sync(&psli->mbox_tmo);
		psli->mbox_tmo.expires = 0;
	}
	return(1);
}
static void __devexit
lpfc_pci_remove_one(struct pci_dev *pdev)
{
	struct Scsi_Host *host = pci_get_drvdata(pdev);
	struct lpfc_hba  *phba = (struct lpfc_hba *)host->hostdata[0];
	unsigned long iflag;

	scsi_remove_host(phba->host);
	list_del(&phba->hba_list);

	/*
	 *  detach the board
	 */

	spin_lock_irqsave(&(phba->drvrlock), iflag);
	lpfc_stop_timer(phba);
	spin_unlock_irqrestore(&(phba->drvrlock), iflag);

	/* Kill the kernel thread for this host */
	if (phba->dpc_pid >= 0) {
		phba->dpc_kill = 1;
		wmb();
		kill_proc(phba->dpc_pid, SIGHUP, 1);
		wait_for_completion(&phba->dpc_exiting);
	}

	/*
	 * Bring down the SLI Layer. This step disable all interrupts,
	 * clears the rings, discards all mailbox commands, and resets
	 * the HBA.
	 */
	lpfc_sli_hba_down(phba);	

	/* Release the irq reservation */
	free_irq(phba->pcidev->irq, phba);

	lpfc_cleanup(phba, 0);
	lpfc_scsi_free(phba);
	lpfc_mem_free(phba);

	/* Free resources associated with SLI2 interface */
	if (phba->slim2p.virt) {
		dma_free_coherent(&pdev->dev,
				  phba->slim_size,
				  phba->slim2p.virt,
				  phba->slim2p.phys);
	}

	/* unmap adapter SLIM and Control Registers */
	iounmap(phba->ctrl_regs_memmap_p);
	iounmap(phba->slim_memmap_p);
	

	pci_release_regions(phba->pcidev);
	pci_disable_device(phba->pcidev);

	kfree(phba);

	scsi_host_put(phba->host);
	pci_set_drvdata(pdev, 0);
	lpfc_num_hba--;
}





static struct pci_device_id lpfc_id_table[] = {
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_VIPER,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_THOR,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_PEGASUS,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_CENTAUR,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_DRAGONFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_SUPERFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_RFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_PFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_TFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{ 0 }
};
MODULE_DEVICE_TABLE(pci, lpfc_id_table);


static struct pci_driver lpfc_driver = {
	.name		= LPFC_DRIVER_NAME,
	.id_table	= lpfc_id_table,
	.probe		= lpfc_pci_probe_one,
	.remove		= __devexit_p(lpfc_pci_remove_one),
};

static int __init
lpfc_init(void)
{
	int rc;

	printk(LPFC_MODULE_DESC "\n");


	lpfc_transport_template =
		fc_attach_transport(&lpfc_transport_functions);
	if (!lpfc_transport_template)
		return -ENODEV;
	rc = pci_module_init(&lpfc_driver);
	return rc;

}

static void __exit
lpfc_exit(void)
{
	pci_unregister_driver(&lpfc_driver);
	fc_release_transport(lpfc_transport_template);
}
module_init(lpfc_init);
module_exit(lpfc_exit);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(LPFC_MODULE_DESC);
MODULE_AUTHOR("Emulex Corporation - tech.support@emulex.com");
MODULE_VERSION("0:" LPFC_DRIVER_VERSION);
