/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_hbadisc.c 1.126 2004/07/22 13:23:47EDT sf_support Exp  $
 */

#include <linux/version.h>
#include <linux/blkdev.h>
#include <linux/dma-mapping.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>

#include "lpfc_sli.h"
#include "lpfc_disc.h"
#include "lpfc_scsi.h"
#include "lpfc_cfgparm.h"
#include "lpfc.h"
#include "lpfc_crtn.h"
#include "lpfc_fcp.h"
#include "lpfc_hw.h"
#include "lpfc_logmsg.h"
#include "lpfc_mem.h"

static int lpfc_matchdid(struct lpfc_hba *, struct lpfc_nodelist *, uint32_t);
static void lpfc_free_tx(struct lpfc_hba *, struct lpfc_nodelist *);
static int lpfc_linkup(struct lpfc_hba *);
static void lpfc_mbx_cmpl_config_link(struct lpfc_hba *, LPFC_MBOXQ_t *);
static void lpfc_mbx_cmpl_read_sparam(struct lpfc_hba *, LPFC_MBOXQ_t *);
static struct lpfc_nodelist *lpfc_findnode_scsiid(struct lpfc_hba *, uint32_t);


/* AlpaArray for assignment of scsid for scan-down and bind_method */
uint8_t lpfcAlpaArray[] = {
	0xEF, 0xE8, 0xE4, 0xE2, 0xE1, 0xE0, 0xDC, 0xDA, 0xD9, 0xD6,
	0xD5, 0xD4, 0xD3, 0xD2, 0xD1, 0xCE, 0xCD, 0xCC, 0xCB, 0xCA,
	0xC9, 0xC7, 0xC6, 0xC5, 0xC3, 0xBC, 0xBA, 0xB9, 0xB6, 0xB5,
	0xB4, 0xB3, 0xB2, 0xB1, 0xAE, 0xAD, 0xAC, 0xAB, 0xAA, 0xA9,
	0xA7, 0xA6, 0xA5, 0xA3, 0x9F, 0x9E, 0x9D, 0x9B, 0x98, 0x97,
	0x90, 0x8F, 0x88, 0x84, 0x82, 0x81, 0x80, 0x7C, 0x7A, 0x79,
	0x76, 0x75, 0x74, 0x73, 0x72, 0x71, 0x6E, 0x6D, 0x6C, 0x6B,
	0x6A, 0x69, 0x67, 0x66, 0x65, 0x63, 0x5C, 0x5A, 0x59, 0x56,
	0x55, 0x54, 0x53, 0x52, 0x51, 0x4E, 0x4D, 0x4C, 0x4B, 0x4A,
	0x49, 0x47, 0x46, 0x45, 0x43, 0x3C, 0x3A, 0x39, 0x36, 0x35,
	0x34, 0x33, 0x32, 0x31, 0x2E, 0x2D, 0x2C, 0x2B, 0x2A, 0x29,
	0x27, 0x26, 0x25, 0x23, 0x1F, 0x1E, 0x1D, 0x1B, 0x18, 0x17,
	0x10, 0x0F, 0x08, 0x04, 0x02, 0x01
};

/* lpfc_nport_recov is only called by lpfc_linkdown below */
static void
lpfc_nport_recov(struct lpfc_hba *phba, struct lpfc_target *targetp)
{
	unsigned long tmo;


	if (phba->config[LPFC_CFG_NODEV_TMO]){
		targetp->targetFlags |= FC_NPR_ACTIVE;
		if (targetp->tmofunc.expires) {
			del_timer_sync(&targetp->tmofunc);
			targetp->tmofunc.expires = 0;
		}

		if (phba->config[LPFC_CFG_NODEV_TMO] >
		    phba->config[LPFC_CFG_LINKDOWN_TMO]) {
			tmo = phba->config[LPFC_CFG_NODEV_TMO];
		} else {
			tmo = phba->config[LPFC_CFG_LINKDOWN_TMO];
		}

		targetp->tmofunc.expires = jiffies + HZ * tmo;
		targetp->tmofunc.data = (unsigned long)targetp;
		add_timer(&targetp->tmofunc);
	} else {
		/* flush the target */
		lpfc_sli_abort_iocb_tgt(phba,
					&phba->sli.ring[phba->sli.fcp_ring],
					targetp->scsi_id);
	}
}

int
lpfc_linkdown(struct lpfc_hba * phba)
{
	struct lpfc_sli       *psli;
	struct lpfc_nodelist  *ndlp, *next_ndlp;
	struct list_head *listp;
	struct list_head *node_list[4];
	LPFC_MBOXQ_t     *mb;
	struct lpfc_target *targetp;
	int               rc, i;
	unsigned long     flags;

	psli = &phba->sli;
	phba->hba_state = LPFC_LINK_DOWN;
	phba->fc_flag |= FC_LNK_DOWN;


	lpfc_put_event(phba, HBA_EVENT_LINK_DOWN, phba->fc_myDID, 0, 0, 0);

	/* Clean up any firmware default rpi's */
	if ((mb = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
		lpfc_unreg_did(phba, 0xffffffff, mb);
		if (lpfc_sli_issue_mbox(phba, mb, (MBX_NOWAIT | MBX_STOP_IOCB))
		    == MBX_NOT_FINISHED) {
			mempool_free( mb, phba->mbox_mem_pool);
		}
	}

	/* Cleanup any outstanding RSCN activity */
	lpfc_els_flush_rscn(phba);

	/* Cleanup any outstanding ELS commands */
	lpfc_els_flush_cmd(phba);

	/*
	 * If this function was called by the lpfc_do_dpc, don't recurse into
	 * the routine again.  If not, just process any outstanding
	 * discovery events.
	 */
	spin_lock_irqsave(&phba->dpc_lock, flags);
	if (!list_empty(&phba->dpc_disc)) {
		spin_unlock_irqrestore(&phba->dpc_lock, flags);
		lpfc_disc_done(phba);
	}
	else {
		spin_unlock_irqrestore(&phba->dpc_lock, flags);
	}

	/* Handle linkdown timer logic.   */

	if (!(phba->fc_flag & FC_LD_TIMER)) {
		/* Should we start the link down watchdog timer */
		if (phba->config[LPFC_CFG_LINKDOWN_TMO] == 0){
			phba->fc_flag |= (FC_LD_TIMER | FC_LD_TIMEOUT);
			phba->hba_flag |= FC_LFR_ACTIVE;
		} else {
			phba->fc_flag |= FC_LD_TIMER;
			phba->hba_flag |= FC_LFR_ACTIVE;
			if (phba->fc_linkdown.expires) {
				del_timer_sync(&phba->fc_linkdown);
				phba->fc_linkdown.expires = 0;
			}
				phba->fc_linkdown.expires = jiffies + HZ *
					phba->config[LPFC_CFG_LINKDOWN_TMO];
				phba->fc_linkdown.data = (unsigned long)phba;
				add_timer(&phba->fc_linkdown);
		}
	}

	/* Issue a LINK DOWN event to all nodes */
	node_list[0] = &phba->fc_plogi_list;
	node_list[1] = &phba->fc_adisc_list;
	node_list[2] = &phba->fc_nlpunmap_list;
	node_list[3] = &phba->fc_nlpmap_list;
	for (i = 0; i < 4; i++) {
		listp = node_list[i];
		if (list_empty(listp))
			continue;

		list_for_each_entry_safe(ndlp, next_ndlp, listp, nlp_listp) {
			/* Fabric nodes are not handled thru state machine for
			   link down */
			if (!(ndlp->nlp_type & NLP_FABRIC)) {
				lpfc_set_failmask(phba, ndlp,
						  LPFC_DEV_LINK_DOWN,
						  LPFC_SET_BITMASK);
			}

			rc = lpfc_disc_state_machine(phba, ndlp, 0,
						     NLP_EVT_DEVICE_UNK);
			if ((rc != NLP_STE_FREED_NODE)
			    && (phba->config[LPFC_CFG_USE_ADISC] == 0)) {
				/* If we are not using ADISC when the link comes
				 * back up, we might as well free all the nodes
				 * right now.
				 */
				targetp = ndlp->nlp_Target;
				/* If we were a FCP target, go into NPort
				 * Recovery mode to give it a chance to come
				 * back.
				 */
				if (targetp) {
					lpfc_nport_recov(phba, targetp);
				}

				lpfc_findnode_did(phba,
						  (NLP_SEARCH_ALL |
						   NLP_SEARCH_DEQUE),
						  ndlp->nlp_DID);
				lpfc_nlp_remove(phba, ndlp);
			}
		}
	}

	/* Setup myDID for link up if we are in pt2pt mode */
	if (phba->fc_flag & FC_PT2PT) {
		phba->fc_myDID = 0;
		if ((mb = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			lpfc_config_link(phba, mb);
			if (lpfc_sli_issue_mbox
			    (phba, mb, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mb, phba->mbox_mem_pool);
			}
		}
		phba->fc_flag &= ~(FC_PT2PT | FC_PT2PT_PLOGI);
	}
	phba->fc_flag &= ~FC_LBIT;

	/* Turn off discovery timer if its running */
	lpfc_can_disctmo(phba);


	/* Must process IOCBs on all rings to handle ABORTed I/Os */
	return (0);
}

static int
lpfc_linkup(struct lpfc_hba * phba)
{
	struct lpfc_nodelist *ndlp, *next_ndlp;
	struct list_head *listp;
	struct list_head *node_list[4];
	int i;

	phba->hba_state = LPFC_LINK_UP;
	phba->hba_flag |= FC_NDISC_ACTIVE;
	phba->fc_flag &= ~(FC_LNK_DOWN | FC_PT2PT | FC_PT2PT_PLOGI |
			   FC_RSCN_MODE | FC_NLP_MORE | FC_DELAY_DISC |
			   FC_RSCN_DISC_TMR | FC_RSCN_DISCOVERY | FC_LD_TIMER |
			   FC_LD_TIMEOUT);
	phba->fc_ns_retry = 0;


	lpfc_put_event(phba, HBA_EVENT_LINK_UP, phba->fc_myDID,
			  phba->fc_topology, 0, phba->fc_linkspeed);

	if (phba->fc_linkdown.expires) {
		del_timer_sync(&phba->fc_linkdown);
		phba->fc_linkdown.expires = 0;
	}

	/*
	 * Clean up old Fabric, NameServer and other NLP_FABRIC logins.
	 */
	list_for_each_entry_safe(ndlp, next_ndlp, &phba->fc_nlpunmap_list,
				nlp_listp) {
		if (ndlp->nlp_type & NLP_FABRIC) {
			ndlp->nlp_flag &= ~(NLP_UNMAPPED_LIST |
					    NLP_TGT_NO_SCSIID);
			phba->fc_unmap_cnt--;
			list_del(&ndlp->nlp_listp);
			lpfc_nlp_remove(phba, ndlp);
		}
	}

	/* Mark all nodes for LINK UP */
	node_list[0] = &phba->fc_plogi_list;
	node_list[1] = &phba->fc_adisc_list;
	node_list[2] = &phba->fc_nlpunmap_list;
	node_list[3] = &phba->fc_nlpmap_list;
	for (i = 0; i < 4; i++) {
		listp = node_list[i];
		if (list_empty(listp))
			continue;

		list_for_each_entry(ndlp, listp, nlp_listp) {
			lpfc_set_failmask(phba, ndlp, LPFC_DEV_LINK_DOWN,
					  LPFC_CLR_BITMASK);
		}
	}

	return (0);
}

/*
 * This routine handles processing a READ_LA mailbox
 * command upon completion. It is setup in the LPFC_MBOXQ
 * as the completion routine when the command is
 * handed off to the SLI layer.
 */
void
lpfc_mbx_cmpl_read_la(struct lpfc_hba * phba, LPFC_MBOXQ_t * pmb)
{
	struct lpfc_dmabuf *mp;
	struct lpfc_sli *psli;
	READ_LA_VAR *la;
	LPFC_MBOXQ_t *mbox;
	MAILBOX_t *mb;
	uint32_t control;
	int i;

	psli = &phba->sli;
	mb = &pmb->mb;
	/* Check for error */
	if (mb->mbxStatus) {
		/* READ_LA mbox error <mbxStatus> state <hba_state> */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_LINK_EVENT,
				"%d:1307 READ_LA mbox error x%x state x%x\n",
				phba->brd_no,
				mb->mbxStatus, phba->hba_state);

		lpfc_linkdown(phba);
		phba->hba_state = LPFC_HBA_ERROR;

		/* turn on Link Attention interrupts */
		psli->sliinit.sli_flag |= LPFC_PROCESS_LA;
		control = readl(phba->HCregaddr);
		control |= HC_LAINT_ENA;
		writel(control, phba->HCregaddr);
		readl(phba->HCregaddr); /* flush */
		return;
	}
	la = (READ_LA_VAR *) & pmb->mb.un.varReadLA;

	mp = (struct lpfc_dmabuf *) (pmb->context1);

	/* The mailbox was populated by the HBA.  Flush it to main store for the
	 * driver.  Note that all context buffers are from the driver's
	 * dma pool and have length LPFC_BPL_SIZE.
	 */
	pci_dma_sync_single_for_cpu(phba->pcidev, mp->phys, LPFC_BPL_SIZE,
				    PCI_DMA_FROMDEVICE);

	/* Get Loop Map information */
	if (mp) {
		memcpy(&phba->alpa_map[0], mp->virt, 128);
	} else {
		memset(&phba->alpa_map[0], 0, 128);
	}

	if (la->pb)
		phba->fc_flag |= FC_BYPASSED_MODE;
	else
		phba->fc_flag &= ~FC_BYPASSED_MODE;

	if (((phba->fc_eventTag + 1) < la->eventTag) ||
	    (phba->fc_eventTag == la->eventTag)) {
		phba->fc_stat.LinkMultiEvent++;
		if (la->attType == AT_LINK_UP) {
			if (phba->fc_eventTag != 0) {

				lpfc_linkdown(phba);
			}
		}
	}

	phba->fc_eventTag = la->eventTag;

	if (la->attType == AT_LINK_UP) {
		phba->fc_stat.LinkUp++;
		/* Link Up Event <eventTag> received */
		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_LINK_EVENT,
				"%d:1303 Link Up Event x%x received Data: x%x x%x x%x x%x\n",
				phba->brd_no,
				la->eventTag, phba->fc_eventTag,
				la->granted_AL_PA, la->UlnkSpeed,
				phba->alpa_map[0]);

		if (la->UlnkSpeed == LA_2GHZ_LINK)
			phba->fc_linkspeed = LA_2GHZ_LINK;
		else
			phba->fc_linkspeed = 0;

		if ((phba->fc_topology = la->topology) == TOPOLOGY_LOOP) {

			if (la->il) {
				phba->fc_flag |= FC_LBIT;
			}

			phba->fc_myDID = la->granted_AL_PA;

			i = la->un.lilpBde64.tus.f.bdeSize;
			if (i == 0) {
				phba->alpa_map[0] = 0;
			} else {
				if (phba->config[LPFC_CFG_LOG_VERBOSE]
				    & LOG_LINK_EVENT) {
					int numalpa, j, k;
					union {
						uint8_t pamap[16];
						struct {
							uint32_t wd1;
							uint32_t wd2;
							uint32_t wd3;
							uint32_t wd4;
						} pa;
					} un;

					numalpa = phba->alpa_map[0];
					j = 0;
					while (j < numalpa) {
						memset(un.pamap, 0, 16);
						for (k = 1; j < numalpa; k++) {
							un.pamap[k - 1] =
							    phba->alpa_map[j +
									   1];
							j++;
							if (k == 16)
								break;
						}
						/* Link Up Event ALPA map */
						lpfc_printf_log(phba,
							KERN_WARNING,
							LOG_LINK_EVENT,
							"%d:1304 Link Up Event ALPA map Data: x%x x%x x%x x%x\n",
							phba->brd_no,	
							un.pa.wd1, un.pa.wd2,
							un.pa.wd3, un.pa.wd4);
					}
				}
			}
		} else {
			phba->fc_myDID = phba->fc_pref_DID;
			phba->fc_flag |= FC_LBIT;
		}

		lpfc_linkup(phba);
		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			lpfc_read_sparam(phba, mbox);
			mbox->mbox_cmpl = lpfc_mbx_cmpl_read_sparam;
			lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB));
		}

		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			phba->hba_state = LPFC_LOCAL_CFG_LINK;
			lpfc_config_link(phba, mbox);
			mbox->mbox_cmpl = lpfc_mbx_cmpl_config_link;
			lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB));
		}
	} else {
		phba->fc_stat.LinkDown++;
		/* Link Down Event <eventTag> received */
		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_LINK_EVENT,
				"%d:1305 Link Down Event x%x received Data: x%x x%x x%x\n",
				phba->brd_no,
				la->eventTag, phba->fc_eventTag,
				phba->hba_state, phba->fc_flag);

		lpfc_linkdown(phba);

		/* turn on Link Attention interrupts - no CLEAR_LA needed */
		psli->sliinit.sli_flag |= LPFC_PROCESS_LA;
		control = readl(phba->HCregaddr);
		control |= HC_LAINT_ENA;
		writel(control, phba->HCregaddr);
		readl(phba->HCregaddr); /* flush */
	}

	pmb->context1 = 0;
	lpfc_mbuf_free(phba, mp->virt, mp->phys);
	kfree(mp);
	mempool_free( pmb, phba->mbox_mem_pool);
	return;
}

static void
lpfc_mbx_cmpl_config_link(struct lpfc_hba * phba, LPFC_MBOXQ_t * pmb)
{
	struct lpfc_sli *psli;
	MAILBOX_t *mb;

	psli = &phba->sli;
	mb = &pmb->mb;
	/* Check for error */
	if (mb->mbxStatus) {
		/* CONFIG_LINK mbox error <mbxStatus> state <hba_state> */
		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_MBOX,
				"%d:0306 CONFIG_LINK mbxStatus error x%x HBA state x%x\n",
				phba->brd_no,
				mb->mbxStatus, phba->hba_state);

		lpfc_linkdown(phba);
		phba->hba_state = LPFC_HBA_ERROR;
		goto out;
	}

	if (phba->hba_state == LPFC_LOCAL_CFG_LINK) {
		if (phba->fc_topology == TOPOLOGY_LOOP) {
			/* If we are public loop and L bit was set */
			if ((phba->fc_flag & FC_PUBLIC_LOOP) &&
			    !(phba->fc_flag & FC_LBIT)) {
				/* Need to wait for FAN - use discovery timer
				 * for timeout.  hba_state is identically
				 * LPFC_LOCAL_CFG_LINK while waiting for FAN
				 */
				lpfc_set_disctmo(phba);
				mempool_free( pmb, phba->mbox_mem_pool);
				return;
			}
		}

		/* Start discovery by sending a FLOGI hba_state is identically
		 * LPFC_FLOGI while waiting for FLOGI cmpl
		 */
		phba->hba_state = LPFC_FLOGI;
		lpfc_set_disctmo(phba);
		lpfc_initial_flogi(phba);
		mempool_free( pmb, phba->mbox_mem_pool);
		return;
	}

out:
	/* CONFIG_LINK bad hba state <hba_state> */
	lpfc_printf_log(phba,
			KERN_ERR,
			LOG_DISCOVERY,
			"%d:0200 CONFIG_LINK bad hba state x%x\n",
			phba->brd_no, phba->hba_state);

	if (phba->hba_state != LPFC_CLEAR_LA) {
		lpfc_clear_la(phba, pmb);
		pmb->mbox_cmpl = lpfc_mbx_cmpl_clear_la;
		if (lpfc_sli_issue_mbox(phba, pmb, (MBX_NOWAIT | MBX_STOP_IOCB))
		    == MBX_NOT_FINISHED) {
			mempool_free( pmb, phba->mbox_mem_pool);
			lpfc_disc_flush_list(phba);
			psli->ring[(psli->ip_ring)].flag &=
				~LPFC_STOP_IOCB_EVENT;
			psli->ring[(psli->fcp_ring)].flag &=
				~LPFC_STOP_IOCB_EVENT;
			psli->ring[(psli->next_ring)].flag &=
				~LPFC_STOP_IOCB_EVENT;
			phba->hba_state = LPFC_HBA_READY;
		}
	} else {
		mempool_free( pmb, phba->mbox_mem_pool);
	}
	return;
}

static void
lpfc_mbx_cmpl_read_sparam(struct lpfc_hba * phba, LPFC_MBOXQ_t * pmb)
{
	struct lpfc_sli *psli;
	MAILBOX_t *mb;
	struct lpfc_dmabuf *mp;

	psli = &phba->sli;
	mb = &pmb->mb;

	/* Check for error */
	if (mb->mbxStatus) {
		/* READ_SPARAM mbox error <mbxStatus> state <hba_state> */
		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_MBOX,
				"%d:0319 READ_SPARAM mbxStatus error x%x hba state x%x>\n",
				phba->brd_no,
				mb->mbxStatus, phba->hba_state);

		lpfc_linkdown(phba);
		phba->hba_state = LPFC_HBA_ERROR;
		goto out;
	}

	mp = (struct lpfc_dmabuf *) pmb->context1;

	/* The mailbox was populated by the HBA.  Flush it to main store for the
	 * driver.  Note that all context buffers are from the driver's
	 * dma pool and have length LPFC_BPL_SIZE.
	 */
	pci_dma_sync_single_for_cpu(phba->pcidev, mp->phys, LPFC_BPL_SIZE,
			PCI_DMA_FROMDEVICE);

	memcpy((uint8_t *) & phba->fc_sparam, (uint8_t *) mp->virt,
	       sizeof (struct serv_parm));
	memcpy((uint8_t *) & phba->fc_nodename,
	       (uint8_t *) & phba->fc_sparam.nodeName,
	       sizeof (struct lpfc_name));
	memcpy((uint8_t *) & phba->fc_portname,
	       (uint8_t *) & phba->fc_sparam.portName,
	       sizeof (struct lpfc_name));
	memcpy(phba->phys_addr, phba->fc_portname.IEEE, 6);
	lpfc_mbuf_free(phba, mp->virt, mp->phys);
	kfree(mp);
	mempool_free( pmb, phba->mbox_mem_pool);
	return;

out:
	if (phba->hba_state != LPFC_CLEAR_LA) {
		lpfc_clear_la(phba, pmb);
		pmb->mbox_cmpl = lpfc_mbx_cmpl_clear_la;
		if (lpfc_sli_issue_mbox(phba, pmb, (MBX_NOWAIT | MBX_STOP_IOCB))
		    == MBX_NOT_FINISHED) {
			mempool_free( pmb, phba->mbox_mem_pool);
			lpfc_disc_flush_list(phba);
			psli->ring[(psli->ip_ring)].flag &=
			    ~LPFC_STOP_IOCB_EVENT;
			psli->ring[(psli->fcp_ring)].flag &=
			    ~LPFC_STOP_IOCB_EVENT;
			psli->ring[(psli->next_ring)].flag &=
			    ~LPFC_STOP_IOCB_EVENT;
			phba->hba_state = LPFC_HBA_READY;
		}
	} else {
		mempool_free( pmb, phba->mbox_mem_pool);
	}
	return;
}

/*
 * This routine handles processing a CLEAR_LA mailbox
 * command upon completion. It is setup in the LPFC_MBOXQ
 * as the completion routine when the command is
 * handed off to the SLI layer.
 */
void
lpfc_mbx_cmpl_clear_la(struct lpfc_hba * phba, LPFC_MBOXQ_t * pmb)
{
	struct lpfc_sli *psli;
	struct lpfc_nodelist *ndlp;
	MAILBOX_t *mb;
	uint32_t control;

	psli = &phba->sli;
	mb = &pmb->mb;
	/* Since we don't do discovery right now, turn these off here */
	psli->ring[psli->ip_ring].flag &= ~LPFC_STOP_IOCB_EVENT;
	psli->ring[psli->fcp_ring].flag &= ~LPFC_STOP_IOCB_EVENT;
	psli->ring[psli->next_ring].flag &= ~LPFC_STOP_IOCB_EVENT;
	/* Check for error */
	if ((mb->mbxStatus) && (mb->mbxStatus != 0x1601)) {
		/* CLEAR_LA mbox error <mbxStatus> state <hba_state> */
		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_MBOX,
				"%d:0320 CLEAR_LA mbxStatus error x%x hba state x%x\n",
				phba->brd_no,
				mb->mbxStatus, phba->hba_state);

		phba->hba_state = LPFC_HBA_ERROR;
		goto out;
	}

	phba->num_disc_nodes = 0;
	/* go thru PLOGI list and issue ELS PLOGIs */
	if (phba->fc_plogi_cnt) {
		list_for_each_entry(ndlp, &phba->fc_plogi_list, nlp_listp) {
			if (ndlp->nlp_state == NLP_STE_UNUSED_NODE) {
				ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
				lpfc_issue_els_plogi(phba, ndlp, 0);
				ndlp->nlp_flag |= NLP_DISC_NODE;
				phba->num_disc_nodes++;
				if (phba->num_disc_nodes >=
				    phba->config[LPFC_CFG_DISC_THREADS]) {
					if (phba->fc_plogi_cnt >
					    phba->num_disc_nodes)
						phba->fc_flag |= FC_NLP_MORE;
					break;
				}
			}
		}
	}
	if (phba->num_disc_nodes == 0) {
		phba->hba_flag &= ~FC_NDISC_ACTIVE;
	}
	phba->hba_state = LPFC_HBA_READY;

out:
	/* Device Discovery completes */
	lpfc_printf_log(phba,
			 KERN_INFO,
			 LOG_DISCOVERY,
			 "%d:0225 Device Discovery completes\n",
			 phba->brd_no);

	phba->hba_flag &= ~FC_LFR_ACTIVE;

	mempool_free( pmb, phba->mbox_mem_pool);
	if (phba->fc_flag & FC_ESTABLISH_LINK) {
		phba->fc_flag &= ~FC_ESTABLISH_LINK;
	}
	if (phba->fc_estabtmo.expires) {
		del_timer_sync(&phba->fc_estabtmo);
		phba->fc_estabtmo.expires = 0;
	}
	lpfc_can_disctmo(phba);

	/* turn on Link Attention interrupts */
	psli->sliinit.sli_flag |= LPFC_PROCESS_LA;
	control = readl(phba->HCregaddr);
	control |= HC_LAINT_ENA;
	writel(control, phba->HCregaddr);
	readl(phba->HCregaddr); /* flush */

	return;
}

/*
 * This routine handles processing a REG_LOGIN mailbox
 * command upon completion. It is setup in the LPFC_MBOXQ
 * as the completion routine when the command is
 * handed off to the SLI layer.
 */
void
lpfc_mbx_cmpl_reg_login(struct lpfc_hba * phba, LPFC_MBOXQ_t * pmb)
{
	struct lpfc_sli *psli;
	MAILBOX_t *mb;
	struct lpfc_dmabuf *mp;
	struct lpfc_nodelist *ndlp;

	psli = &phba->sli;
	mb = &pmb->mb;

	ndlp = (struct lpfc_nodelist *) pmb->context2;
	mp = (struct lpfc_dmabuf *) (pmb->context1);

	/* The mailbox was populated by the HBA.  Flush it to main store for the
	 * driver.  Note that all context buffers are from the driver's
	 * dma pool and have length LPFC_BPL_SIZE.
	 */
	pci_dma_sync_single_for_cpu(phba->pcidev, mp->phys, LPFC_BPL_SIZE,
			PCI_DMA_FROMDEVICE);

	pmb->context1 = 0;

	/* Good status, call state machine */
	lpfc_disc_state_machine(phba, ndlp, pmb, NLP_EVT_CMPL_REG_LOGIN);
	lpfc_mbuf_free(phba, mp->virt, mp->phys);
	kfree(mp);
	mempool_free( pmb, phba->mbox_mem_pool);

	return;
}

/*
 * This routine handles processing a Fabric REG_LOGIN mailbox
 * command upon completion. It is setup in the LPFC_MBOXQ
 * as the completion routine when the command is
 * handed off to the SLI layer.
 */
void
lpfc_mbx_cmpl_fabric_reg_login(struct lpfc_hba * phba, LPFC_MBOXQ_t * pmb)
{
	struct lpfc_sli *psli;
	MAILBOX_t *mb;
	struct lpfc_dmabuf *mp;
	struct lpfc_nodelist *ndlp;
	struct lpfc_nodelist *ndlp_fdmi;


	psli = &phba->sli;
	mb = &pmb->mb;

	ndlp = (struct lpfc_nodelist *) pmb->context2;
	mp = (struct lpfc_dmabuf *) (pmb->context1);

	/* The mailbox was populated by the HBA.  Flush it to main store for the
	 * driver.  Note that all context buffers are from the driver's
	 * dma pool and have length LPFC_BPL_SIZE.
	 */
	pci_dma_sync_single_for_cpu(phba->pcidev, mp->phys, LPFC_BPL_SIZE,
			PCI_DMA_FROMDEVICE);

	pmb->context1 = 0;

	if (ndlp->nlp_rpi != 0)
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);
	ndlp->nlp_rpi = mb->un.varWords[0];
	lpfc_addnode_rpi(phba, ndlp, ndlp->nlp_rpi);
	ndlp->nlp_type |= NLP_FABRIC;
	lpfc_nlp_unmapped(phba, ndlp);
	ndlp->nlp_state = NLP_STE_PRLI_COMPL;

	if (phba->hba_state == LPFC_FABRIC_CFG_LINK) {
		/* This NPort has been assigned an NPort_ID by the fabric as a
		 * result of the completed fabric login.  Issue a State Change
		 * Registration (SCR) ELS request to the fabric controller
		 * (SCR_DID) so that this NPort gets RSCN events from the
		 * fabric.
		 */
		lpfc_issue_els_scr(phba, SCR_DID, 0);

		/* Allocate a new node instance.  If the pool is empty, just
		 * start the discovery process and skip the Nameserver login
		 * process.  This is attempted again later on.  Otherwise, issue
		 * a Port Login (PLOGI) to the NameServer
		 */
		if ((ndlp = mempool_alloc(phba->nlp_mem_pool, GFP_ATOMIC))
		    == 0) {
			lpfc_disc_start(phba);
		} else {
			lpfc_nlp_init(phba, ndlp, NameServer_DID);
			ndlp->nlp_type |= NLP_FABRIC;
			ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
			lpfc_issue_els_plogi(phba, ndlp, 0);
			if (phba->config[LPFC_CFG_FDMI_ON]) {
				if ((ndlp_fdmi = mempool_alloc(
						       phba->nlp_mem_pool,
						       GFP_ATOMIC))) {
					lpfc_nlp_init(phba, ndlp_fdmi, 
						FDMI_DID);
					ndlp_fdmi->nlp_type |= NLP_FABRIC;
					ndlp_fdmi->nlp_state =
					    NLP_STE_PLOGI_ISSUE;
					lpfc_issue_els_plogi(phba, ndlp_fdmi,
							     0);
				}
			}
		}
	}

	lpfc_mbuf_free(phba, mp->virt, mp->phys);
	kfree(mp);
	mempool_free( pmb, phba->mbox_mem_pool);

	return;
}

/*
 * This routine handles processing a NameServer REG_LOGIN mailbox
 * command upon completion. It is setup in the LPFC_MBOXQ
 * as the completion routine when the command is
 * handed off to the SLI layer.
 */
void
lpfc_mbx_cmpl_ns_reg_login(struct lpfc_hba * phba, LPFC_MBOXQ_t * pmb)
{
	struct lpfc_sli *psli;
	MAILBOX_t *mb;
	struct lpfc_dmabuf *mp;
	struct lpfc_nodelist *ndlp;

	psli = &phba->sli;
	mb = &pmb->mb;

	ndlp = (struct lpfc_nodelist *) pmb->context2;
	mp = (struct lpfc_dmabuf *) (pmb->context1);

	/* The mailbox was populated by the HBA.  Flush it to main store for the
	 * driver.  Note that all context buffers are from the driver's
	 * dma pool and have length LPFC_BPL_SIZE.
	 */
	pci_dma_sync_single_for_cpu(phba->pcidev, mp->phys, LPFC_BPL_SIZE,
			PCI_DMA_FROMDEVICE);

	pmb->context1 = 0;

	if (ndlp->nlp_rpi != 0)
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);
	ndlp->nlp_rpi = mb->un.varWords[0];
	lpfc_addnode_rpi(phba, ndlp, ndlp->nlp_rpi);
	ndlp->nlp_type |= NLP_FABRIC;
	lpfc_nlp_unmapped(phba, ndlp);
	ndlp->nlp_state = NLP_STE_PRLI_COMPL;

	if (phba->hba_state < LPFC_HBA_READY) {
		/* Link up discovery requires Fabrib registration. */
		lpfc_ns_cmd(phba, ndlp, SLI_CTNS_RNN_ID);
		lpfc_ns_cmd(phba, ndlp, SLI_CTNS_RSNN_NN);
		lpfc_ns_cmd(phba, ndlp, SLI_CTNS_RFT_ID);
	}

	phba->fc_ns_retry = 0;
	/* Good status, issue CT Request to NameServer */
	if (lpfc_ns_cmd(phba, ndlp, SLI_CTNS_GID_FT)) {
		/* Cannot issue NameServer Query, so finish up discovery */
		lpfc_disc_start(phba);
	}

	lpfc_mbuf_free(phba, mp->virt, mp->phys);
	kfree(mp);
	mempool_free( pmb, phba->mbox_mem_pool);

	return;
}

/* Put blp on the bind list */
int
lpfc_nlp_bind(struct lpfc_hba * phba, struct lpfc_bindlist * blp)
{
	struct lpfc_target *targetp;

	/* Put it at the end of the bind list */
	list_add_tail(&blp->nlp_listp, &phba->fc_nlpbind_list);
	phba->fc_bind_cnt++;
	targetp = phba->device_queue_hash[blp->nlp_sid];
	if (targetp) {
		targetp->pnode = 0;
	}

	/* Add scsiid <sid> to BIND list */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_NODE,
			"%d:0903 Add scsiid %d to BIND list Data: x%x x%x x%x x%p\n",
			phba->brd_no,
			blp->nlp_sid, phba->fc_bind_cnt, blp->nlp_DID,
			blp->nlp_bind_type, blp);

	return (0);
}

/* Put blp on the plogi list */
int
lpfc_nlp_plogi(struct lpfc_hba * phba, struct lpfc_nodelist * nlp)
{
	struct lpfc_bindlist *blp;
	struct lpfc_sli *psli;

	psli = &phba->sli;
	blp = 0;

	/* Check to see if this node exists on any other list */
	if (nlp->nlp_flag & NLP_LIST_MASK) {
		if (nlp->nlp_flag & NLP_MAPPED_LIST) {
			nlp->nlp_flag &= ~NLP_MAPPED_LIST;
			phba->fc_map_cnt--;
			list_del(&nlp->nlp_listp);
			phba->nport_event_cnt++;

			/* Must call before binding is removed */
			lpfc_set_failmask(phba, nlp, LPFC_DEV_DISCONNECTED,
					  LPFC_SET_BITMASK);

			blp = nlp->nlp_listp_bind;
			if (blp) {
				blp->nlp_Target = nlp->nlp_Target;
			}
			/* flush the target */
			lpfc_sli_abort_iocb_tgt(phba,
						&phba->sli.ring[phba->sli.
								fcp_ring],
						nlp->nlp_Target->scsi_id);

			nlp->nlp_listp_bind = 0;
			nlp->nlp_sid = 0;
			nlp->nlp_Target = 0;
			nlp->nlp_flag &= ~NLP_SEED_MASK;

		} else if (nlp->nlp_flag & NLP_UNMAPPED_LIST) {
			nlp->nlp_flag &=
			    ~(NLP_UNMAPPED_LIST | NLP_TGT_NO_SCSIID);
			phba->fc_unmap_cnt--;
			list_del(&nlp->nlp_listp);
			phba->nport_event_cnt++;

		} else if (nlp->nlp_flag & NLP_PLOGI_LIST) {
			return (0);	/* Already on plogi list */
		} else if (nlp->nlp_flag & NLP_ADISC_LIST) {
			nlp->nlp_flag &= ~NLP_ADISC_LIST;
			phba->fc_adisc_cnt--;
			list_del(&nlp->nlp_listp);
		}
	}

	/* Put it at the end of the plogi list */
	list_add_tail(&nlp->nlp_listp, &phba->fc_plogi_list);
	phba->fc_plogi_cnt++;
	nlp->nlp_flag |= NLP_PLOGI_LIST;

	/* Add NPort <did> to PLOGI list */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_NODE,
			"%d:0904 Add NPort x%x to PLOGI list Data: x%x x%p\n",
			phba->brd_no,
			nlp->nlp_DID, phba->fc_plogi_cnt, blp);

	if (blp) {
		lpfc_nlp_bind(phba, blp);
	}
	return (0);
}

/* Put nlp on the adisc list */
int
lpfc_nlp_adisc(struct lpfc_hba * phba, struct lpfc_nodelist * nlp)
{
	struct lpfc_bindlist *blp;
	struct lpfc_sli *psli;
	struct lpfc_target *targetp;
	unsigned long tmo;

	blp = 0;
	psli = &phba->sli;
	targetp = nlp->nlp_Target;

	if ((targetp) && (targetp->tmofunc.expires == 0)) {
		targetp->targetFlags |= FC_NPR_ACTIVE;
		
		if (phba->config[LPFC_CFG_NODEV_TMO] >
		    phba->config[LPFC_CFG_LINKDOWN_TMO]) {
			tmo = phba->config[LPFC_CFG_NODEV_TMO];
		} else {
			tmo = phba->config[LPFC_CFG_LINKDOWN_TMO];
		}

		targetp->tmofunc.expires = jiffies + HZ * tmo;
		targetp->tmofunc.data = (unsigned long)targetp;
		add_timer(&targetp->tmofunc);
	}

	/* Check to see if this node exist on any other list */
	if (nlp->nlp_flag & NLP_LIST_MASK) {
		if (nlp->nlp_flag & NLP_MAPPED_LIST) {
			nlp->nlp_flag &= ~NLP_MAPPED_LIST;
			phba->fc_map_cnt--;
			list_del(&nlp->nlp_listp);
			phba->nport_event_cnt++;

			/* Must call before binding is removed */
			lpfc_set_failmask(phba, nlp, LPFC_DEV_DISAPPEARED,
					  LPFC_SET_BITMASK);

			blp = nlp->nlp_listp_bind;
			if (blp) {
				blp->nlp_Target = nlp->nlp_Target;
			}
			if ((nlp->nlp_Target) &&
			    (!(phba->fc_flag & FC_RSCN_MODE))) {
				/* flush the target */
				lpfc_sli_abort_iocb_tgt(phba,
							&phba->sli.ring[phba->
								sli.fcp_ring],
							nlp->nlp_Target->
							scsi_id);
			}

			nlp->nlp_listp_bind = 0;
			nlp->nlp_Target = 0;

			/* Keep sid since LPFC_DEV_DISAPPEARED
			 * is a non-fatal error
			 */
			nlp->nlp_flag &= ~NLP_SEED_MASK;

		} else if (nlp->nlp_flag & NLP_UNMAPPED_LIST) {
			nlp->nlp_flag &=
			    ~(NLP_UNMAPPED_LIST | NLP_TGT_NO_SCSIID);
			phba->fc_unmap_cnt--;
			list_del(&nlp->nlp_listp);
			phba->nport_event_cnt++;


		} else if (nlp->nlp_flag & NLP_PLOGI_LIST) {
			nlp->nlp_flag &= ~NLP_PLOGI_LIST;
			phba->fc_plogi_cnt--;
			list_del(&nlp->nlp_listp);
		} else if (nlp->nlp_flag & NLP_ADISC_LIST) {
			return (0);	/* Already on adisc list */
		}
	}

	/* Put it at the end of the adisc list */
	list_add_tail(&nlp->nlp_listp, &phba->fc_adisc_list);
	phba->fc_adisc_cnt++;
	nlp->nlp_flag |= NLP_ADISC_LIST;

	/* Add NPort <did> to ADISC list */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_NODE,
			"%d:0905 Add NPort x%x to ADISC list Data: x%x x%p\n",
			phba->brd_no,
			nlp->nlp_DID, phba->fc_adisc_cnt, blp);

	if (blp) {
		lpfc_nlp_bind(phba, blp);
	}

	return (0);
}

/*
 * Put nlp on the unmapped list
 * NOTE: - update nlp_type to NLP_FC_NODE
 */
int
lpfc_nlp_unmapped(struct lpfc_hba * phba, struct lpfc_nodelist * nlp)
{
	struct lpfc_bindlist *blp;

	blp = 0;

	/* Check to see if this node exists on any other list */
	if (nlp->nlp_flag & NLP_LIST_MASK) {
		if (nlp->nlp_flag & NLP_MAPPED_LIST) {
			nlp->nlp_flag &= ~NLP_MAPPED_LIST;
			phba->fc_map_cnt--;
			list_del(&nlp->nlp_listp);

			/* Must call before binding is removed */
			lpfc_set_failmask(phba, nlp, LPFC_DEV_DISAPPEARED,
					  LPFC_SET_BITMASK);

			blp = nlp->nlp_listp_bind;
			if (blp) {
				blp->nlp_Target = nlp->nlp_Target;
			}
			/* flush the target */
			lpfc_sli_abort_iocb_tgt(phba,
						&phba->sli.ring[phba->sli.
								fcp_ring],
						nlp->nlp_Target->scsi_id);
			nlp->nlp_listp_bind = 0;
			nlp->nlp_Target = 0;
			/* Keep sid since LPFC_DEV_DISAPPEARED
			 * is a non-fatal error
			 */
			nlp->nlp_flag &= ~NLP_SEED_MASK;
		} else if (nlp->nlp_flag & NLP_UNMAPPED_LIST) {
			return (0);	/* Already on unmapped list */
		} else if (nlp->nlp_flag & NLP_PLOGI_LIST) {
			nlp->nlp_flag &= ~NLP_PLOGI_LIST;
			phba->fc_plogi_cnt--;
			list_del(&nlp->nlp_listp);
		} else if (nlp->nlp_flag & NLP_ADISC_LIST) {
			nlp->nlp_flag &= ~NLP_ADISC_LIST;
			phba->fc_adisc_cnt--;
			list_del(&nlp->nlp_listp);
		}
	}

	/* Put it at the end of the unmapped list */
	list_add_tail(&nlp->nlp_listp, &phba->fc_nlpunmap_list);
	phba->nport_event_cnt++;
	phba->fc_unmap_cnt++;
	nlp->nlp_type |= NLP_FC_NODE;
	nlp->nlp_flag |= NLP_UNMAPPED_LIST;

	/* Add NPort <did> to UNMAP list */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_NODE,
			"%d:0906 Add NPort x%x to UNMAP list Data: x%x x%p\n",
			phba->brd_no,
			nlp->nlp_DID, phba->fc_unmap_cnt, blp);

	if (blp) {
		lpfc_nlp_bind(phba, blp);
	}
	return (0);
}

/*
 * Put nlp on the mapped list
 * NOTE: - update nlp_type to NLP_FCP_TARGET
 *       - attach binding entry to context2
 */
int
lpfc_nlp_mapped(struct lpfc_hba * phba, struct lpfc_nodelist * nlp,
		struct lpfc_bindlist * blp)
{
	struct lpfc_target *targetp;

	/* Check to see if this node exists on any other list */
	if (nlp->nlp_flag & NLP_LIST_MASK) {
		if (nlp->nlp_flag & NLP_MAPPED_LIST) {
			return (0);	/* Already on mapped list */
		} else if (nlp->nlp_flag & NLP_UNMAPPED_LIST) {
			nlp->nlp_flag &=
			    ~(NLP_UNMAPPED_LIST | NLP_TGT_NO_SCSIID);
			phba->fc_unmap_cnt--;
			list_del(&nlp->nlp_listp);
		} else if (nlp->nlp_flag & NLP_PLOGI_LIST) {
			nlp->nlp_flag &= ~NLP_PLOGI_LIST;
			phba->fc_plogi_cnt--;
			list_del(&nlp->nlp_listp);
		} else if (nlp->nlp_flag & NLP_ADISC_LIST) {
			nlp->nlp_flag &= ~NLP_ADISC_LIST;
			phba->fc_adisc_cnt--;
			list_del(&nlp->nlp_listp);
		}
	}

	/* Put it at the end of the mapped list */
	list_add_tail(&nlp->nlp_listp, &phba->fc_nlpmap_list);
	phba->nport_event_cnt++;
	phba->fc_map_cnt++;
	nlp->nlp_flag |= NLP_MAPPED_LIST;
	nlp->nlp_type |= NLP_FCP_TARGET;
	nlp->nlp_sid = blp->nlp_sid;
	nlp->nlp_Target = blp->nlp_Target;
	nlp->nlp_listp_bind = blp;
	targetp = phba->device_queue_hash[nlp->nlp_sid];

	/* Add NPort <did> to MAPPED list scsiid <sid> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_NODE,
			"%d:0907 Add NPort x%x to MAP list scsiid %d Data: x%x x%p\n",
			phba->brd_no,
			nlp->nlp_DID, nlp->nlp_sid, phba->fc_map_cnt,
			blp);

	if (targetp) {
		nlp->nlp_Target = targetp;
		targetp->pnode = nlp;
	}

	return (0);
}

/*
 * Start / ReStart rescue timer for Discovery / RSCN handling
 */
void
lpfc_set_disctmo(struct lpfc_hba * phba)
{
	uint32_t tmo;

	tmo = ((phba->fc_ratov * 2) + 1);

	/* Turn off discovery timer if its running */
	if (phba->fc_disctmo.expires) {
		del_timer_sync(&phba->fc_disctmo);
	}
	phba->fc_disctmo.expires = jiffies + HZ * tmo;
	phba->fc_disctmo.data = (unsigned long) phba;
	add_timer(&phba->fc_disctmo);

	/* Start Discovery Timer state <hba_state> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_DISCOVERY,
			"%d:0247 Start Discovery Timer state x%x Data: x%x x%lx x%x x%x\n",
			phba->brd_no,
			phba->hba_state, tmo, (unsigned long)&phba->fc_disctmo,
			phba->fc_plogi_cnt, phba->fc_adisc_cnt);

	return;
}

/*
 * Cancel rescue timer for Discovery / RSCN handling
 */
int
lpfc_can_disctmo(struct lpfc_hba * phba)
{
	int rc;

	rc = 0;

	/* Turn off discovery timer if its running */
	if (phba->fc_disctmo.expires) {
		del_timer_sync(&phba->fc_disctmo);
		phba->fc_disctmo.expires = 0;
		rc = 1;
	}

	/* Cancel Discovery Timer state <hba_state> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_DISCOVERY,
			"%d:0248 Cancel Discovery Timer state x%x Data: x%x x%x x%x x%x\n",
			phba->brd_no,
			phba->hba_state, phba->fc_flag, rc, phba->fc_plogi_cnt,
			phba->fc_adisc_cnt);

	return (rc);
}

/*
 * Check specified ring for outstanding IOCB on the SLI queue
 * Return true if iocb matches the specified nport
 */
static int
lpfc_check_sli_ndlp(struct lpfc_hba * phba,
		    struct lpfc_sli_ring * pring,
		    struct lpfc_iocbq * iocb, struct lpfc_nodelist * ndlp)
{
	struct lpfc_sli *psli;
	IOCB_t *icmd;

	psli = &phba->sli;
	icmd = &iocb->iocb;
	if (pring->ringno == LPFC_ELS_RING) {
		switch (icmd->ulpCommand) {
		case CMD_GEN_REQUEST64_CR:
			if (icmd->ulpContext == (volatile ushort)ndlp->nlp_rpi)
				return (1);
		case CMD_ELS_REQUEST64_CR:
		case CMD_XMIT_ELS_RSP64_CX:
			if (iocb->context1 == (uint8_t *) ndlp)
				return (1);
		}
	} else if (pring->ringno == psli->ip_ring) {

	} else if (pring->ringno == psli->fcp_ring) {
		/* Skip match check if waiting to relogin to FCP target */
	  	if ((ndlp->nlp_type & NLP_FCP_TARGET) &&
	  	  (ndlp->nlp_delayfunc.expires) &&
		  (ndlp->nlp_flag & NLP_DELAY_TMO)) {
			return (0);
		}
		if (icmd->ulpContext == (volatile ushort)ndlp->nlp_rpi) {
			return (1);
		}
	} else if (pring->ringno == psli->next_ring) {

	}
	return (0);
}

/*
 * Free resources / clean up outstanding I/Os
 * associated with nlp_rpi in the LPFC_NODELIST entry.
 */
int
lpfc_no_rpi(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp)
{
	struct lpfc_sli *psli;
	struct lpfc_sli_ring *pring;
	struct lpfc_iocbq *iocb, *next_iocb;
	IOCB_t *icmd;
	uint32_t rpi, i;

	psli = &phba->sli;
	rpi = ndlp->nlp_rpi;
	if (rpi) {
		/* Now process each ring */
		for (i = 0; i < psli->sliinit.num_rings; i++) {
			pring = &psli->ring[i];

			list_for_each_entry_safe(iocb, next_iocb, &pring->txq,
						list) {
				/*
				 * Check to see if iocb matches the nport we are
				 * looking for
				 */
				if ((lpfc_check_sli_ndlp
				     (phba, pring, iocb, ndlp))) {
					/* It matches, so deque and call compl
					   with an error */
					list_del(&iocb->list);
					pring->txq_cnt--;
					if (iocb->iocb_cmpl) {
						icmd = &iocb->iocb;
						icmd->ulpStatus =
						    IOSTAT_LOCAL_REJECT;
						icmd->un.ulpWord[4] =
						    IOERR_SLI_ABORTED;
						(iocb->iocb_cmpl) (phba,
								   iocb, iocb);
					} else {
						mempool_free(iocb,
						     phba->iocb_mem_pool);
					}
				}
			}
			/* Everything that matches on txcmplq will be returned
			 * by firmware with a no rpi error.
			 */
		}
	}
	return (0);
}

/*
 * Free resources / clean up outstanding I/Os
 * associated with a LPFC_NODELIST entry. This
 * routine effectively results in a "software abort".
 */
int
lpfc_driver_abort(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp)
{
	struct lpfc_sli *psli;
	struct lpfc_sli_ring *pring;
	struct lpfc_iocbq *iocb, *next_iocb;
	IOCB_t *icmd;
	uint32_t i, cmd;

	/* Abort outstanding I/O on NPort <nlp_DID> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_DISCOVERY,
			"%d:0201 Abort outstanding I/O on NPort x%x Data: x%x x%x x%x\n",
			phba->brd_no,
			ndlp->nlp_DID, ndlp->nlp_flag, ndlp->nlp_state,
			ndlp->nlp_rpi);

	psli = &phba->sli;
	/* Now process each ring */
	for (i = 0; i < psli->sliinit.num_rings; i++) {
		pring = &psli->ring[i];

		/* First check the txq */
		list_for_each_entry_safe(iocb, next_iocb, &pring->txq, list) {
			/* Check to see if iocb matches the nport we are looking
			   for */
			if ((lpfc_check_sli_ndlp(phba, pring, iocb, ndlp))) {
				/* It matches, so deque and call compl with an
				   error */
				list_del(&ndlp->nlp_listp);
				pring->txq_cnt--;
				if (iocb->iocb_cmpl) {
					icmd = &iocb->iocb;
					icmd->ulpStatus = IOSTAT_LOCAL_REJECT;
					icmd->un.ulpWord[4] = IOERR_SLI_ABORTED;
					(iocb->iocb_cmpl) (phba, iocb, iocb);
				} else {
					mempool_free(iocb, phba->iocb_mem_pool);
				}
			}
		}
		/* Everything on txcmplq will be returned by firmware with a no
		 * rpi / linkdown / abort error.  For ring 0, ELS discovery, we
		 * want to get rid of it right here.
		 */
		if (pring->ringno == LPFC_ELS_RING) {
			/* Next check the txcmplq */
			list_for_each_entry_safe(iocb, next_iocb,
						 &pring->txcmplq, list) {
				/* Check to see if iocb matches the nport we are
				   looking for */
				if ((lpfc_check_sli_ndlp
				     (phba, pring, iocb, ndlp))) {
					/* It matches, so deque and call compl
					   with an error */
					list_del(&iocb->list);
					pring->txcmplq_cnt--;

					icmd = &iocb->iocb;
					/* If the driver is completing an ELS
					 * command early, flush it out of the
					 * firmware.
					 */
					if ((icmd->ulpCommand ==
					     CMD_ELS_REQUEST64_CR)
					    && (icmd->un.elsreq64.bdl.
						ulpIoTag32)) {
						lpfc_sli_issue_abort_iotag32
						    (phba, pring, iocb);
					}
					if (iocb->iocb_cmpl) {
						icmd->ulpStatus =
						    IOSTAT_LOCAL_REJECT;
						icmd->un.ulpWord[4] =
						    IOERR_SLI_ABORTED;
						(iocb->iocb_cmpl) (phba,
								   iocb, iocb);
					} else {
						mempool_free(iocb,
						     phba->iocb_mem_pool);
					}
				}
			}
		}
	}

	/* If we are delaying issuing an ELS command, cancel it */
	if ((ndlp->nlp_delayfunc.expires) && (ndlp->nlp_flag & NLP_DELAY_TMO)) {
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		cmd = (uint32_t) ndlp->nlp_last_elscmd;
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;

		/* Allocate an IOCB and indicate an error completion */
		/* Allocate a buffer for the command iocb */
		if ((iocb = mempool_alloc(phba->iocb_mem_pool, GFP_ATOMIC))
		    == 0) {
			return (0);
		}
		memset(iocb, 0, sizeof (struct lpfc_iocbq));
		icmd = &iocb->iocb;
		icmd->ulpStatus = IOSTAT_LOCAL_REJECT;
		icmd->un.ulpWord[4] = IOERR_SLI_ABORTED;
		iocb->context1 = ndlp;

		switch (cmd) {
		case ELS_CMD_FLOGI:
			iocb->iocb_cmpl = lpfc_cmpl_els_flogi;
			break;
		case ELS_CMD_PLOGI:
			iocb->iocb_cmpl = lpfc_cmpl_els_plogi;
			break;
		case ELS_CMD_ADISC:
			iocb->iocb_cmpl = lpfc_cmpl_els_adisc;
			break;
		case ELS_CMD_PRLI:
			iocb->iocb_cmpl = lpfc_cmpl_els_prli;
			break;
		case ELS_CMD_LOGO:
			iocb->iocb_cmpl = lpfc_cmpl_els_logo;
			break;
		default:
			iocb->iocb_cmpl = lpfc_cmpl_els_cmd;
			break;
		}
		(iocb->iocb_cmpl) (phba, iocb, iocb);
	}
	return (0);
}

/*
 * Free resources associated with LPFC_NODELIST entry
 * so it can be freed.
 */
int
lpfc_freenode(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp)
{
	LPFC_MBOXQ_t *mbox;
	struct lpfc_sli *psli;

	/* The psli variable gets rid of the long pointer deference. */
	psli = &phba->sli;

	/* Cleanup node for NPort <nlp_DID> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_NODE,
			"%d:0900 Cleanup node for NPort x%x Data: x%x x%x x%x\n",
			phba->brd_no,
			ndlp->nlp_DID, ndlp->nlp_flag, ndlp->nlp_state,
			ndlp->nlp_rpi);

	/* NLP_FREED_NODE flag is to protect the node from being freed
	 * more then once. For driver_abort and other cases where the DSM
	 * calls itself recursively, its possible to free the node twice.
	 */
	if (ndlp->nlp_rflag & NLP_FREED_NODE) {
		return (0);
	}

	if (ndlp->nlp_flag & NLP_LIST_MASK) {
		lpfc_findnode_did(phba, (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
				  ndlp->nlp_DID);
	}

	if (ndlp->nlp_tmofunc.expires) {
		del_timer_sync(&ndlp->nlp_tmofunc);
		ndlp->nlp_tmofunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	}
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	}

	if (ndlp->nlp_rpi) {
		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			lpfc_unreg_login(phba, ndlp->nlp_rpi, mbox);
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mbox, phba->mbox_mem_pool);
			}
		}
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);
		lpfc_no_rpi(phba, ndlp);
		ndlp->nlp_rpi = 0;
		lpfc_set_failmask(phba, ndlp, LPFC_DEV_DISCONNECTED,
				  LPFC_SET_BITMASK);
	}
	return (0);
}

/*
 * Check to see if we can free the nlp back to the freelist.
 * If we are in the middle of using the nlp in the discovery state
 * machine, defer the free till we reach the end of the state machine.
 */
int
lpfc_nlp_remove(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp)
{
	if (ndlp->nlp_tmofunc.expires) {
		del_timer_sync(&ndlp->nlp_tmofunc);
		ndlp->nlp_tmofunc.expires = 0;
	}
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
	}
	if (ndlp->nlp_disc_refcnt) {
		ndlp->nlp_rflag |= NLP_DELAY_REMOVE;
	}
	else {
		lpfc_freenode(phba, ndlp);
		mempool_free( ndlp, phba->nlp_mem_pool);
	}
	return(0);
}

static int
lpfc_matchdid(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp, uint32_t did)
{
	D_ID mydid;
	D_ID ndlpdid;
	D_ID matchdid;
	int zero_did;

	if (did == Bcast_DID)
		return (0);

	zero_did = 0;
	if (ndlp->nlp_DID == 0) {
		return (0);
	}

	/* First check for Direct match */
	if (ndlp->nlp_DID == did)
		return (1);

	/* Next check for area/domain identically equals 0 match */
	mydid.un.word = phba->fc_myDID;
	if ((mydid.un.b.domain == 0) && (mydid.un.b.area == 0)) {
		goto out;
	}

	matchdid.un.word = did;
	ndlpdid.un.word = ndlp->nlp_DID;
	if (matchdid.un.b.id == ndlpdid.un.b.id) {
		if ((mydid.un.b.domain == matchdid.un.b.domain) &&
		    (mydid.un.b.area == matchdid.un.b.area)) {
			if ((ndlpdid.un.b.domain == 0) &&
			    (ndlpdid.un.b.area == 0)) {
				if (ndlpdid.un.b.id)
					return (1);
			}
			goto out;
		}

		matchdid.un.word = ndlp->nlp_DID;
		if ((mydid.un.b.domain == ndlpdid.un.b.domain) &&
		    (mydid.un.b.area == ndlpdid.un.b.area)) {
			if ((matchdid.un.b.domain == 0) &&
			    (matchdid.un.b.area == 0)) {
				if (matchdid.un.b.id)
					return (1);
			}
		}
	}
out:
	if (zero_did)
		ndlp->nlp_DID = 0;
	return (0);
}

/* Search for a nodelist entry on a specific list */
static struct lpfc_nodelist *
lpfc_findnode_scsiid(struct lpfc_hba * phba, uint32_t scsid)
{
	struct lpfc_nodelist *ndlp;
	struct lpfc_target *targetp;

	targetp = phba->device_queue_hash[scsid];
	/* First see if the SCSI ID has an allocated struct lpfc_target */
	if (targetp) {
		if (targetp->pnode) {
			return targetp->pnode;
		}
	}

	/* Now try the hard way */
	list_for_each_entry(ndlp, &phba->fc_nlpmap_list, nlp_listp) {
		if (scsid == ndlp->nlp_sid) {
			return (ndlp);
		}
	}

	/* no match found */
	return ((struct lpfc_nodelist *) 0);
}



/* Search for a nodelist entry on a specific list */
struct lpfc_nodelist *
lpfc_findnode_did(struct lpfc_hba * phba, uint32_t order, uint32_t did)
{
	struct lpfc_nodelist *ndlp, *next_ndlp;
	uint32_t data1;
	struct lpfc_bindlist *blp;

	blp = 0;
	if (order & NLP_SEARCH_UNMAPPED) {
		list_for_each_entry_safe(ndlp, next_ndlp,
					 &phba->fc_nlpunmap_list, nlp_listp) {
			if (lpfc_matchdid(phba, ndlp, did)) {
				data1 = (((uint32_t) ndlp->nlp_state << 24) |
					 ((uint32_t) ndlp->nlp_xri << 16) |
					 ((uint32_t) ndlp->nlp_type << 8) |
					 ((uint32_t) ndlp->nlp_rpi & 0xff));
				/* FIND node DID unmapped */
				lpfc_printf_log(phba,
						KERN_INFO,
						LOG_NODE,
						"%d:0929 FIND node DID unmapped Data: x%p x%x x%x x%x\n",
						phba->brd_no,
						ndlp, ndlp->nlp_DID,
						ndlp->nlp_flag, data1);
				if (order & NLP_SEARCH_DEQUE) {
					ndlp->nlp_flag &=
					    ~(NLP_UNMAPPED_LIST |
					      NLP_TGT_NO_SCSIID);
					phba->fc_unmap_cnt--;
					list_del(&ndlp->nlp_listp);
					phba->nport_event_cnt++;
				}

				return (ndlp);
			}
		}
	}

	if (order & NLP_SEARCH_MAPPED) {
		list_for_each_entry_safe(ndlp, next_ndlp, &phba->fc_nlpmap_list,
					nlp_listp) {
			if (lpfc_matchdid(phba, ndlp, did)) {

				data1 = (((uint32_t) ndlp->nlp_state << 24) |
					 ((uint32_t) ndlp->nlp_xri << 16) |
					 ((uint32_t) ndlp->nlp_type << 8) |
					 ((uint32_t) ndlp->nlp_rpi & 0xff));
				/* FIND node DID mapped */
				lpfc_printf_log(phba,
						KERN_INFO,
						LOG_NODE,
						"%d:0930 FIND node DID mapped Data: x%p x%x x%x x%x\n",
						phba->brd_no,
						ndlp, ndlp->nlp_DID,
						ndlp->nlp_flag, data1);
				if (order & NLP_SEARCH_DEQUE) {
					ndlp->nlp_flag &= ~NLP_MAPPED_LIST;
					phba->fc_map_cnt--;
					list_del(&ndlp->nlp_listp);
					phba->nport_event_cnt++;

					/* Must call before binding is
					   removed */
					lpfc_set_failmask(phba, ndlp,
							  LPFC_DEV_DISAPPEARED,
							  LPFC_SET_BITMASK);

					blp = ndlp->nlp_listp_bind;
					ndlp->nlp_listp_bind = 0;
					if (blp) {
						blp->nlp_Target =
						    ndlp->nlp_Target;
					}
					/* Keep Target and sid since
					 * LPFC_DEV_DISAPPEARED is a
					 * non-fatal error
					 */
					ndlp->nlp_flag &= ~NLP_SEED_MASK;
				}

				if (blp) {
					lpfc_nlp_bind(phba, blp);
				}
				return (ndlp);
			}
		}
	}

	if (order & NLP_SEARCH_PLOGI) {
		list_for_each_entry_safe(ndlp, next_ndlp, &phba->fc_plogi_list,
					nlp_listp) {
			if (lpfc_matchdid(phba, ndlp, did)) {

				data1 = (((uint32_t) ndlp->nlp_state << 24) |
					 ((uint32_t) ndlp->nlp_xri << 16) |
					 ((uint32_t) ndlp->nlp_type << 8) |
					 ((uint32_t) ndlp->nlp_rpi & 0xff));
				/* LOG change to PLOGI */
				/* FIND node DID bind */
				lpfc_printf_log(phba,
						KERN_INFO,
						LOG_NODE,
						"%d:0908 FIND node DID bind Data: x%p x%x x%x x%x\n",
						phba->brd_no,
						ndlp, ndlp->nlp_DID,
						ndlp->nlp_flag, data1);
				if (order & NLP_SEARCH_DEQUE) {
					ndlp->nlp_flag &= ~NLP_PLOGI_LIST;
					phba->fc_plogi_cnt--;
					list_del(&ndlp->nlp_listp);
				}

				return (ndlp);
			}
		}
	}

	if (order & NLP_SEARCH_ADISC) {
		list_for_each_entry_safe(ndlp, next_ndlp, &phba->fc_adisc_list,
					nlp_listp) {
			if (lpfc_matchdid(phba, ndlp, did)) {

				data1 = (((uint32_t) ndlp->nlp_state << 24) |
					 ((uint32_t) ndlp->nlp_xri << 16) |
					 ((uint32_t) ndlp->nlp_type << 8) |
					 ((uint32_t) ndlp->nlp_rpi & 0xff));
				/* LOG change to ADISC */
				/* FIND node DID bind */
				lpfc_printf_log(phba,
						KERN_INFO,
						LOG_NODE,
						"%d:0931 FIND node DID bind Data: x%p x%x x%x x%x\n",
						phba->brd_no,
						ndlp, ndlp->nlp_DID,
						ndlp->nlp_flag, data1);
				if (order & NLP_SEARCH_DEQUE) {
					ndlp->nlp_flag &= ~NLP_ADISC_LIST;
					phba->fc_adisc_cnt--;
					list_del(&ndlp->nlp_listp);
				}

				return (ndlp);
			}
		}
	}

	/* FIND node did <did> NOT FOUND */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_NODE,
			"%d:0932 FIND node did x%x NOT FOUND Data: x%x\n",
			phba->brd_no, did, order);

	/* no match found */
	return ((struct lpfc_nodelist *) 0);
}

/* Build a list of nodes to discover based on the loopmap */
void
lpfc_disc_list_loopmap(struct lpfc_hba * phba)
{
	struct lpfc_nodelist *ndlp;
	int j;
	uint32_t alpa, index;


	if (phba->hba_state <= LPFC_LINK_DOWN) {
		return;
	}
	if (phba->fc_topology != TOPOLOGY_LOOP) {
		return;
	}

	/* Check for loop map present or not */
	if (phba->alpa_map[0]) {
		for (j = 1; j <= phba->alpa_map[0]; j++) {
			alpa = phba->alpa_map[j];

			if (((phba->fc_myDID & 0xff) == alpa) || (alpa == 0)) {
				continue;
			}
			if ((ndlp = lpfc_findnode_did(phba,
						      (NLP_SEARCH_MAPPED |
						       NLP_SEARCH_UNMAPPED |
						       NLP_SEARCH_DEQUE),
						      alpa))) {
				/* Mark node for address authentication */
				lpfc_disc_state_machine(phba, ndlp, 0,
							NLP_EVT_DEVICE_ADD);
				continue;
			}
			/* Skip if the node is already in the plogi / adisc
			   list */
			if ((ndlp = lpfc_findnode_did(phba,
						      (NLP_SEARCH_PLOGI |
						       NLP_SEARCH_ADISC),
						      alpa))) {
				continue;
			}
			/* Cannot find existing Fabric ndlp, so allocate a new
			   one */
			if ((ndlp = mempool_alloc(phba->nlp_mem_pool,
						  GFP_ATOMIC)) == 0) {
				continue;
			}
			lpfc_nlp_init(phba, ndlp, alpa);
			ndlp->nlp_state = NLP_STE_UNUSED_NODE;

			/* Mark node for address discovery */
			lpfc_disc_state_machine(phba, ndlp, 0,
						NLP_EVT_DEVICE_ADD);
		}
	} else {
		/* No alpamap, so try all alpa's */
		for (j = 0; j < FC_MAXLOOP; j++) {
			if (phba->config[LPFC_CFG_SCAN_DOWN])
				index = FC_MAXLOOP - j - 1;
			else
				index = j;
			alpa = lpfcAlpaArray[index];
			if ((phba->fc_myDID & 0xff) == alpa) {
				continue;
			}

			if ((ndlp = lpfc_findnode_did(phba,
						      (NLP_SEARCH_MAPPED |
						       NLP_SEARCH_UNMAPPED |
						       NLP_SEARCH_DEQUE),
						      alpa))) {
				/* Mark node for address authentication */
				lpfc_disc_state_machine(phba, ndlp, 0,
							NLP_EVT_DEVICE_ADD);
				continue;
			}
			/* Skip if the node is already in the plogi / adisc
			   list */
			if ((ndlp = lpfc_findnode_did(phba,
						      (NLP_SEARCH_PLOGI |
						       NLP_SEARCH_ADISC),
						      alpa))) {
				continue;
			}
			/* Cannot find existing ndlp, so allocate a new one */
			if ((ndlp = mempool_alloc(phba->nlp_mem_pool,
						  GFP_ATOMIC)) == 0) {
				continue;
			}
			lpfc_nlp_init(phba, ndlp, alpa);
			ndlp->nlp_state = NLP_STE_UNUSED_NODE;

			/* Mark node for address discovery */
			lpfc_disc_state_machine(phba, ndlp, 0,
						NLP_EVT_DEVICE_ADD);
		}
	}
	return;
}

/* Start Link up / RSCN discovery on ADISC or PLOGI lists */
void
lpfc_disc_start(struct lpfc_hba * phba)
{
	struct lpfc_sli *psli;
	LPFC_MBOXQ_t *mbox;
	struct lpfc_nodelist *ndlp, *next_ndlp;
	uint32_t did_changed;
	uint32_t clear_la_pending;

	psli = &phba->sli;

	if (phba->hba_state <= LPFC_LINK_DOWN) {
		return;
	}
	if (phba->hba_state == LPFC_CLEAR_LA)
		clear_la_pending = 1;
	else
		clear_la_pending = 0;

	if (phba->hba_state < LPFC_HBA_READY) {
		phba->hba_state = LPFC_DISC_AUTH;
	}
	lpfc_set_disctmo(phba);

	if (phba->fc_prevDID == phba->fc_myDID) {
		did_changed = 0;
	} else {
		did_changed = 1;
	}
	phba->fc_prevDID = phba->fc_myDID;

	/* Start Discovery state <hba_state> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_DISCOVERY,
			"%d:0202 Start Discovery hba state x%x Data: x%x x%x x%x\n",
			phba->brd_no,
			phba->hba_state, phba->fc_flag, phba->fc_plogi_cnt,
			phba->fc_adisc_cnt);

	/* At this point, nothing should be on the mapped list, without
	 * NODEV_TMO timer running on it, link up discovery only.
	 */
	if (!(phba->fc_flag & FC_RSCN_MODE)) {
		list_for_each_entry_safe(ndlp, next_ndlp, &phba->fc_nlpmap_list,
					nlp_listp) {
			/* If nodev timer is not running, get rid of it */
			if (!(ndlp->nlp_flag & NLP_NODEV_TMO)) {
				lpfc_set_failmask(phba, ndlp,
						  LPFC_DEV_DISCONNECTED,
						  LPFC_SET_BITMASK);
				list_del(&ndlp->nlp_listp);
				lpfc_nlp_remove(phba, ndlp);
			}
		}
	}

	/* First do ADISC for authentication */
	if (phba->fc_adisc_cnt) {
		if (did_changed == 0) {
			phba->num_disc_nodes = 0;
			/* go thru ADISC list and issue ELS ADISCs */
			list_for_each_entry(ndlp, &phba->fc_adisc_list,
					    nlp_listp) {
				lpfc_issue_els_adisc(phba, ndlp, 0);
				ndlp->nlp_flag |= NLP_DISC_NODE;
				phba->num_disc_nodes++;
				if (phba->num_disc_nodes >=
				    phba->config[LPFC_CFG_DISC_THREADS]) {
					if (phba->fc_adisc_cnt >
					    phba->num_disc_nodes)
						phba->fc_flag |= FC_NLP_MORE;
					break;
				}
			}
			return;
		}
		/* Move these to PLOGI list instead */
		if (!list_empty(&phba->fc_adisc_list)) {
			list_for_each_entry_safe(ndlp, next_ndlp,
						 &phba->fc_adisc_list,
						 nlp_listp) {
				lpfc_nlp_plogi(phba, ndlp);
			}
		}
	}

	if ((phba->hba_state < LPFC_HBA_READY) && (!clear_la_pending)) {
		/* If we get here, there is nothing to ADISC */
		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			phba->hba_state = LPFC_CLEAR_LA;
			lpfc_clear_la(phba, mbox);
			mbox->mbox_cmpl = lpfc_mbx_cmpl_clear_la;
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mbox, phba->mbox_mem_pool);
				lpfc_disc_flush_list(phba);
				psli->ring[(psli->ip_ring)].flag &=
				    ~LPFC_STOP_IOCB_EVENT;
				psli->ring[(psli->fcp_ring)].flag &=
				    ~LPFC_STOP_IOCB_EVENT;
				psli->ring[(psli->next_ring)].flag &=
				    ~LPFC_STOP_IOCB_EVENT;
				phba->hba_state = LPFC_HBA_READY;
			}
		}
	} else {
		/* go thru PLOGI list and issue ELS PLOGIs */
		phba->num_disc_nodes = 0;
		if (phba->fc_plogi_cnt) {
			list_for_each_entry(ndlp, &phba->fc_plogi_list,
						nlp_listp) {
				if (ndlp->nlp_state == NLP_STE_UNUSED_NODE) {
					ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
					lpfc_issue_els_plogi(phba, ndlp, 0);
					ndlp->nlp_flag |= NLP_DISC_NODE;
					phba->num_disc_nodes++;
					if (phba->num_disc_nodes >=
					    phba->config[LPFC_CFG_DISC_THREADS]
					   ) {
						if (phba->fc_plogi_cnt >
						    phba->num_disc_nodes)
							phba->fc_flag |=
							    FC_NLP_MORE;
						break;
					}
				}
			}
		} else {
			if (phba->fc_flag & FC_RSCN_MODE) {
				/* Check to see if more RSCNs came in while we
				 * were processing this one.
				 */
				if ((phba->fc_rscn_id_cnt == 0) &&
				    (!(phba->fc_flag & FC_RSCN_DISCOVERY))) {
					lpfc_els_flush_rscn(phba);
				} else {
					lpfc_els_handle_rscn(phba);
				}
			}
		}
	}
	return;
}

void
lpfc_disc_flush_list(struct lpfc_hba * phba)
{
	struct lpfc_nodelist *ndlp, *next_ndlp;

	if (phba->fc_plogi_cnt) {
		list_for_each_entry_safe(ndlp, next_ndlp, &phba->fc_plogi_list,
					nlp_listp) {
			lpfc_set_failmask(phba, ndlp, LPFC_DEV_DISCONNECTED,
					  LPFC_SET_BITMASK);
			lpfc_free_tx(phba, ndlp);
			list_del(&ndlp->nlp_listp);
			lpfc_nlp_remove(phba, ndlp);
		}
	}
	if (phba->fc_adisc_cnt) {
		list_for_each_entry_safe(ndlp, next_ndlp, &phba->fc_adisc_list,
					nlp_listp) {
			lpfc_set_failmask(phba, ndlp, LPFC_DEV_DISCONNECTED,
					  LPFC_SET_BITMASK);
			lpfc_free_tx(phba, ndlp);
			list_del(&ndlp->nlp_listp);
			lpfc_nlp_remove(phba, ndlp);
		}
	}
	return;
}

/*****************************************************************************/
/*
 * NAME:     lpfc_disc_timeout
 *
 * FUNCTION: Fibre Channel driver discovery timeout routine.
 *
 * EXECUTION ENVIRONMENT: interrupt only
 *
 * CALLED FROM:
 *      Timer function
 *
 * RETURNS:
 *      none
 */
/*****************************************************************************/
void
lpfc_disc_timeout(unsigned long ptr)
{
	struct lpfc_hba *phba;
	struct lpfc_sli *psli;
	struct lpfc_nodelist *ndlp;
	LPFC_MBOXQ_t *mbox;
	unsigned long iflag;

	phba = (struct lpfc_hba *)ptr;
	if (!phba) {
		return;
	}
	spin_lock_irqsave(&(phba->drvrlock), iflag);
	phba->fc_disctmo.expires = 0;

	psli = &phba->sli;

	/* hba_state is identically LPFC_LOCAL_CFG_LINK while waiting for FAN */
	if (phba->hba_state == LPFC_LOCAL_CFG_LINK) {
		/* FAN timeout */
		lpfc_printf_log(phba,
				 KERN_WARNING,
				 LOG_DISCOVERY,
				 "%d:0221 FAN timeout\n",
				 phba->brd_no);

		/* Forget about FAN, Start discovery by sending a FLOGI
		 * hba_state is identically LPFC_FLOGI while waiting for FLOGI
		 * cmpl
		 */
		phba->hba_state = LPFC_FLOGI;
		lpfc_set_disctmo(phba);
		lpfc_initial_flogi(phba);
		goto out;
	}

	/* hba_state is identically LPFC_FLOGI while waiting for FLOGI cmpl */
	if (phba->hba_state == LPFC_FLOGI) {
		/* Initial FLOGI timeout */
		lpfc_printf_log(phba,
				 KERN_ERR,
				 LOG_DISCOVERY,
				 "%d:0222 Initial FLOGI timeout\n",
				 phba->brd_no);

		/* Assume no Fabric and go on with discovery.
		 * Check for outstanding ELS FLOGI to abort.
		 */

		/* FLOGI failed, so just use loop map to make discovery list */
		lpfc_disc_list_loopmap(phba);

		/* Start discovery */
		lpfc_disc_start(phba);
		goto out;
	}

	/* hba_state is identically LPFC_FABRIC_CFG_LINK while waiting for
	   NameServer login */
	if (phba->hba_state == LPFC_FABRIC_CFG_LINK) {
		/* Timeout while waiting for NameServer login */
		lpfc_printf_log(phba,
				 KERN_ERR,
				 LOG_DISCOVERY,
				 "%d:0223 Timeout while waiting for NameServer login\n",
				 phba->brd_no);

		/* Next look for NameServer ndlp */
		if ((ndlp = lpfc_findnode_did(phba,
				       (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
				       NameServer_DID))) {
			lpfc_nlp_remove(phba, ndlp);
		}
		/* Start discovery */
		lpfc_disc_start(phba);
		goto out;
	}

	/* Check for wait for NameServer Rsp timeout */
	if (phba->hba_state == LPFC_NS_QRY) {
		/* NameServer Query timeout */
		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_DISCOVERY,
				"%d:0224 NameServer Query timeout Data: x%x x%x\n",
				phba->brd_no,
				phba->fc_ns_retry, LPFC_MAX_NS_RETRY);

		if ((ndlp =
		     lpfc_findnode_did(phba, NLP_SEARCH_UNMAPPED,
				       NameServer_DID))) {
			if (phba->fc_ns_retry < LPFC_MAX_NS_RETRY) {
				/* Try it one more time */
				if (lpfc_ns_cmd(phba, ndlp, SLI_CTNS_GID_FT) ==
				    0) {
					goto out;
				}
			}
			phba->fc_ns_retry = 0;
		}

		/* Nothing to authenticate, so CLEAR_LA right now */
		if (phba->hba_state != LPFC_CLEAR_LA) {
			if ((mbox = mempool_alloc(phba->mbox_mem_pool,
						  GFP_ATOMIC))) {
				phba->hba_state = LPFC_CLEAR_LA;
				lpfc_clear_la(phba, mbox);
				mbox->mbox_cmpl = lpfc_mbx_cmpl_clear_la;
				if (lpfc_sli_issue_mbox
				    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
				    == MBX_NOT_FINISHED) {
					mempool_free(mbox, phba->mbox_mem_pool);
					goto clrlaerr;
				}
			} else {
				/* Device Discovery completion error */
				lpfc_printf_log(phba,
						KERN_ERR,
						LOG_DISCOVERY,
						"%d:0226 Device Discovery completion error\n",
						phba->brd_no);
				phba->hba_state = LPFC_HBA_ERROR;
			}
		}
		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			/* Setup and issue mailbox INITIALIZE LINK command */
			lpfc_linkdown(phba);
			lpfc_init_link(phba, mbox,
				       phba->config[LPFC_CFG_TOPOLOGY],
				       phba->config[LPFC_CFG_LINK_SPEED]);
			mbox->mb.un.varInitLnk.lipsr_AL_PA = 0;
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mbox, phba->mbox_mem_pool);
			}
		}
		goto out;
	}

	if (phba->hba_state == LPFC_DISC_AUTH) {
		/* Node Authentication timeout */
		lpfc_printf_log(phba,
				 KERN_ERR,
				 LOG_DISCOVERY,
				 "%d:0227 Node Authentication timeout\n",
				 phba->brd_no);
		lpfc_disc_flush_list(phba);
		if (phba->hba_state != LPFC_CLEAR_LA) {
			if ((mbox = mempool_alloc(phba->mbox_mem_pool,
						  GFP_ATOMIC))) {
				phba->hba_state = LPFC_CLEAR_LA;
				lpfc_clear_la(phba, mbox);
				mbox->mbox_cmpl = lpfc_mbx_cmpl_clear_la;
				if (lpfc_sli_issue_mbox
				    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
				    == MBX_NOT_FINISHED) {
					mempool_free(mbox, phba->mbox_mem_pool);
					goto clrlaerr;
				}
			}
		}
		goto out;
	}

	if (phba->hba_state == LPFC_CLEAR_LA) {
		/* CLEAR LA timeout */
		lpfc_printf_log(phba,
				 KERN_ERR,
				 LOG_DISCOVERY,
				 "%d:0228 CLEAR LA timeout\n",
				 phba->brd_no);
clrlaerr:
		lpfc_disc_flush_list(phba);
		psli->ring[(psli->ip_ring)].flag &= ~LPFC_STOP_IOCB_EVENT;
		psli->ring[(psli->fcp_ring)].flag &= ~LPFC_STOP_IOCB_EVENT;
		psli->ring[(psli->next_ring)].flag &= ~LPFC_STOP_IOCB_EVENT;
		phba->hba_state = LPFC_HBA_READY;
		goto out;
	}

	if ((phba->hba_state == LPFC_HBA_READY) &&
	    (phba->fc_flag & FC_RSCN_MODE)) {
		/* RSCN timeout */
		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_DISCOVERY,
				"%d:0231 RSCN timeout Data: x%x x%x\n",
				phba->brd_no,
				phba->fc_ns_retry, LPFC_MAX_NS_RETRY);

		/* Cleanup any outstanding ELS commands */
		lpfc_els_flush_cmd(phba);

		lpfc_els_flush_rscn(phba);
		lpfc_disc_flush_list(phba);
		goto out;
	}

out:
	spin_unlock_irqrestore(&(phba->drvrlock), iflag);
	return;
}

/*****************************************************************************/
/*
 * NAME:     lpfc_linkdown_timeout
 *
 * FUNCTION: Fibre Channel driver linkdown timeout routine.
 *
 * EXECUTION ENVIRONMENT: interrupt only
 *
 * CALLED FROM:
 *      Timer function
 *
 * RETURNS:
 *      none
 */
/*****************************************************************************/
void
lpfc_linkdown_timeout(unsigned long ptr)
{
	struct lpfc_hba *phba;
	struct lpfc_nodelist *ndlp;
	struct list_head *listp;
	struct list_head *node_list[4];
	struct lpfc_target *targetp;
	uint32_t tgt;
	unsigned long iflag;
	int i;

	phba = (struct lpfc_hba *)ptr;
	if (!phba) {
		return;
	}

	spin_lock_irqsave(&(phba->drvrlock), iflag);
	phba->fc_linkdown.expires = 0;


	/* Link Down timeout */
	lpfc_printf_log(phba,
			KERN_WARNING,
			LOG_LINK_EVENT | LOG_DISCOVERY,
			"%d:1306 Link Down timeout Data: x%x x%x x%x\n",
			phba->brd_no,
			phba->hba_state, phba->fc_flag, phba->fc_ns_retry);

	phba->fc_flag |= (FC_LD_TIMER | FC_LD_TIMEOUT);	/* indicate timeout */
	phba->hba_flag &= ~FC_LFR_ACTIVE;

	/* Issue a DEVICE REMOVE event to all nodes */
	node_list[0] = &phba->fc_plogi_list;
	node_list[1] = &phba->fc_adisc_list;
	node_list[2] = &phba->fc_nlpunmap_list;
	node_list[3] = &phba->fc_nlpmap_list;
	for (i = 0; i < 4; i++) {
		listp = node_list[i];
		if (list_empty(listp))
			continue;

		list_for_each_entry(ndlp, listp, nlp_listp) {
			/* Fabric nodes are not handled thru state machine for
			   link down tmo */
			if (!(ndlp->nlp_type & NLP_FABRIC)) {
				lpfc_disc_state_machine(phba, ndlp, 0,
							NLP_EVT_DEVICE_RM);
			}
		}
	}

	if (phba->config[LPFC_CFG_NODEV_TMO] == 0){
		for (tgt = 0; tgt < MAX_FCP_TARGET; tgt++) {
			targetp = phba->device_queue_hash[tgt];
			if ((targetp) && (targetp->targetFlags &
					  FC_NPR_ACTIVE)) {
				targetp->targetFlags &= ~FC_NPR_ACTIVE;

				/* flush the target */
				lpfc_sli_abort_iocb_tgt(phba,
							&phba->sli.ring[phba->
								sli.fcp_ring],
							targetp->scsi_id);

				if (targetp->tmofunc.expires) {
					del_timer_sync(&targetp->tmofunc);
					targetp->tmofunc.expires = 0;
				}
			}
		}

	}
	spin_unlock_irqrestore(&(phba->drvrlock), iflag);
	return;
}

static void
lpfc_nodev_timeout(unsigned long ptr)
{
	struct lpfc_hba *phba;
	struct lpfc_nodelist *ndlp;
	struct lpfc_target *targetp;
	unsigned long iflag;

	ndlp = (struct lpfc_nodelist *)ptr;
	phba = ndlp->nlp_phba;
	spin_lock_irqsave(&(phba->drvrlock), iflag);
	ndlp->nlp_tmofunc.expires = 0;

	/* Nodev timeout on NPort <nlp_DID> */
	lpfc_printf_log(phba,
			KERN_ERR,
			LOG_DISCOVERY,
			"%d:0203 Nodev timeout on NPort x%x Data: x%x x%x x%x\n",
			phba->brd_no,
			ndlp->nlp_DID, ndlp->nlp_flag, ndlp->nlp_state,
			ndlp->nlp_rpi);

	ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);

	ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	targetp = ndlp->nlp_Target;
	if (targetp) {
		targetp->targetFlags &= ~FC_NPR_ACTIVE;
		/* flush the target */
		lpfc_sli_abort_iocb_tgt(phba,
					&phba->sli.ring[phba->sli.fcp_ring],
					targetp->scsi_id);

		if (targetp->tmofunc.expires) {
			del_timer_sync(&targetp->tmofunc);
			targetp->tmofunc.expires = 0;
		}
	}
	ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);

	lpfc_disc_state_machine(phba, ndlp, 0, NLP_EVT_DEVICE_RM);
	spin_unlock_irqrestore(&(phba->drvrlock), iflag);
	return;
}

/*****************************************************************************/
/*
 * NAME:     lpfc_find_target
 *
 * FUNCTION: Fibre Channel target to struct lpfc_target lookup
 *
 * EXECUTION ENVIRONMENT:
 *
 * RETURNS:
 *      ptr to desired struct lpfc_target
 */
/*****************************************************************************/
struct lpfc_target *
lpfc_find_target(struct lpfc_hba * phba, uint32_t tgt, uint64_t lun,
	      int create_flag)
{
	struct lpfc_nodelist *nlp;
	struct lpfc_bindlist *blp;
	struct lpfc_target *targetp;

	targetp = phba->device_queue_hash[tgt];

	/* First see if the SCSI ID has an allocated struct lpfc_target */
	if (!targetp) {
		if (create_flag) {
			nlp = lpfc_findnode_scsiid(phba, tgt);
			if (nlp == 0) {
				return 0;
			}

			targetp = nlp->nlp_Target;
			if (targetp) {
				return targetp;
			}

			targetp = kmalloc(sizeof (struct lpfc_target),
					  GFP_ATOMIC);
			if (targetp == 0) {
				return (0);
			}

			memset(targetp, 0, sizeof (struct lpfc_target));
			init_timer(&targetp->tmofunc);
			targetp->tmofunc.function = lpfc_npr_timeout;

			targetp->scsi_id = tgt;
			targetp->max_lun = LPFC_MAX_LUN;
			targetp->pHba = phba;
			phba->device_queue_hash[tgt] = targetp;
			targetp->pnode = nlp;
			if (nlp) {

				/* Create SCSI Target <tgt> */
				lpfc_printf_log(phba,
						KERN_INFO,
						LOG_DISCOVERY | LOG_FCP,
						"%d:0204 Create SCSI Target %d\n",
						phba->brd_no,
						tgt);

				nlp->nlp_Target = targetp;
				if ((blp = nlp->nlp_listp_bind)) {
					blp->nlp_Target = targetp;
				}
			}
		}
	}
	return targetp;
}

/*
 *   lpfc_set_failmask
 *   Set, or clear, failMask bits in struct lpfc_nodelist
 */
void
lpfc_set_failmask(struct lpfc_hba * phba,
		  struct lpfc_nodelist * ndlp, uint32_t bitmask, uint32_t flag)
{
	struct lpfc_target *targetp;
	uint32_t oldmask;
	uint32_t changed;

	/* Failmask change on NPort <nlp_DID> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_DISCOVERY,
			"%d:0208 Failmask change on NPort x%x Data: x%x x%x x%x\n",
			phba->brd_no,
			ndlp->nlp_DID, ndlp->nlp_failMask, bitmask, flag);

	targetp = ndlp->nlp_Target;
	if (flag == LPFC_SET_BITMASK) {
		oldmask = ndlp->nlp_failMask;
		/* Set failMask event */
		ndlp->nlp_failMask |= bitmask;
		if (oldmask != ndlp->nlp_failMask) {
			changed = 1;
		} else {
			changed = 0;
		}

		if (oldmask == 0) {

		}
	} else {
		/* Clear failMask event */
		ndlp->nlp_failMask &= ~bitmask;
		changed = 1;
	}


	/* Since its fatal, now we can clear sid */
	if ((flag == LPFC_SET_BITMASK) && (bitmask & LPFC_DEV_FATAL_ERROR)) {
		ndlp->nlp_sid = 0;
	}
	return;
}

/*
 *  Ignore completion for all IOCBs on tx and txcmpl queue for ELS
 *  ring the match the sppecified nodelist.
 */
static void
lpfc_free_tx(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp)
{
	struct lpfc_sli *psli;
	IOCB_t     *icmd;
	struct lpfc_iocbq    *iocb, *next_iocb;
	struct lpfc_sli_ring *pring;
	struct lpfc_dmabuf   *mp;

	psli = &phba->sli;
	pring = &psli->ring[LPFC_ELS_RING];

	/* Error matching iocb on txq or txcmplq
	 * First check the txq.
	 */
	list_for_each_entry_safe(iocb, next_iocb, &pring->txq, list) {
		if (iocb->context1 != ndlp) {
			continue;
		}
		icmd = &iocb->iocb;
		if ((icmd->ulpCommand == CMD_ELS_REQUEST64_CR) ||
		    (icmd->ulpCommand == CMD_XMIT_ELS_RSP64_CX)) {

			list_del(&iocb->list);
			pring->txq_cnt--;
			lpfc_els_free_iocb(phba, iocb);
		}
	}

	/* Next check the txcmplq */
	list_for_each_entry_safe(iocb, next_iocb, &pring->txcmplq, list) {
		if (iocb->context1 != ndlp) {
			continue;
		}
		icmd = &iocb->iocb;
		if ((icmd->ulpCommand == CMD_ELS_REQUEST64_CR) ||
		    (icmd->ulpCommand == CMD_XMIT_ELS_RSP64_CX)) {

			iocb->iocb_cmpl = 0;
			/* context2 = cmd, context2->next = rsp, context3 =
			   bpl */
			if (iocb->context2) {
				/* Free the response IOCB before handling the
				   command. */

				mp = (struct lpfc_dmabuf *)
				     (((struct lpfc_dmabuf *) (iocb->context2))
				    ->list.next);
				if (mp) {
					/* Delay before releasing rsp buffer to
					 * give UNREG mbox a chance to take
					 * effect.
					 */
					list_add(&mp->list,
						&phba->freebufList);
				}
				lpfc_mbuf_free(phba,
					       ((struct lpfc_dmabuf *)
						iocb->context2)->virt,
					       ((struct lpfc_dmabuf *)
						iocb->context2)->phys);
				kfree(iocb->context2);
			}

			if (iocb->context3) {
				lpfc_mbuf_free(phba,
					       ((struct lpfc_dmabuf *)
						iocb->context3)->virt,
					       ((struct lpfc_dmabuf *)
						iocb->context3)->phys);
				kfree(iocb->context3);
			}
		}
	}

	return;
}

/*
 * This routine handles processing a NameServer REG_LOGIN mailbox
 * command upon completion. It is setup in the LPFC_MBOXQ
 * as the completion routine when the command is
 * handed off to the SLI layer.
 */
void
lpfc_mbx_cmpl_fdmi_reg_login(struct lpfc_hba * phba, LPFC_MBOXQ_t * pmb)
{
	struct lpfc_sli *psli;
	MAILBOX_t *mb;
	struct lpfc_dmabuf *mp;
	struct lpfc_nodelist *ndlp;

	psli = &phba->sli;
	mb = &pmb->mb;

	ndlp = (struct lpfc_nodelist *) pmb->context2;
	mp = (struct lpfc_dmabuf *) (pmb->context1);

	/* The mailbox was populated by the HBA.  Flush it to main store for the
	 * driver.  Note that all context buffers are from the driver's
	 * dma pool and have length LPFC_BPL_SIZE.
	 */
	pci_dma_sync_single_for_cpu(phba->pcidev, mp->phys, LPFC_BPL_SIZE,
			PCI_DMA_FROMDEVICE);

	pmb->context1 = 0;

	if (ndlp->nlp_rpi != 0)
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);
	ndlp->nlp_rpi = mb->un.varWords[0];
	lpfc_addnode_rpi(phba, ndlp, ndlp->nlp_rpi);
	ndlp->nlp_type |= NLP_FABRIC;
	lpfc_nlp_unmapped(phba, ndlp);
	ndlp->nlp_state = NLP_STE_PRLI_COMPL;

	/* Start issuing Fabric-Device Management Interface (FDMI)
	 * command to 0xfffffa (FDMI well known port)
	 */
	if (phba->config[LPFC_CFG_FDMI_ON] == 1) {
		lpfc_fdmi_cmd(phba, ndlp, SLI_MGMT_DHBA);
	} else {
		/*
		 * Delay issuing FDMI command if fdmi-on=2
		 * (supporting RPA/hostnmae)
		 */
		if (phba->fc_fdmitmo.expires) {
			del_timer_sync(&phba->fc_fdmitmo);
		}
		phba->fc_fdmitmo.expires = jiffies + HZ * 60;
		phba->fc_fdmitmo.data = (unsigned long) phba;
		add_timer(&phba->fc_fdmitmo);
	}

	lpfc_mbuf_free(phba, mp->virt, mp->phys);
	kfree(mp);
	mempool_free( pmb, phba->mbox_mem_pool);

	return;
}

/*
 * This routine looks up the ndlp hash
 * table for the given RPI. If rpi found
 * it return the node list pointer
 * else return 0.
 */
struct lpfc_nodelist *
lpfc_findnode_rpi(struct lpfc_hba * phba, uint16_t rpi)
{
	struct lpfc_nodelist *ret;

	ret = phba->fc_nlplookup[LPFC_RPI_HASH_FUNC(rpi)];
	while ((ret != 0) && (ret->nlp_rpi != rpi)) {
		ret = ret->nlp_rpi_hash_next;
	}
	return ret;
}

/*
 * This routine looks up the ndlp hash table for the
 * given RPI. If rpi found it return the node list
 * pointer else return 0 after deleting the entry
 * from hash table.
 */
struct lpfc_nodelist *
lpfc_findnode_remove_rpi(struct lpfc_hba * phba, uint16_t rpi)
{
	struct lpfc_nodelist *ret, *temp;;

	ret = phba->fc_nlplookup[LPFC_RPI_HASH_FUNC(rpi)];

	if (ret == 0) {
		return 0;
	}

	if (ret->nlp_rpi == rpi) {
		phba->fc_nlplookup[LPFC_RPI_HASH_FUNC(rpi)] =
		    ret->nlp_rpi_hash_next;
		ret->nlp_rpi_hash_next = 0;
		return ret;
	}

	while ((ret->nlp_rpi_hash_next != 0) &&
	       (ret->nlp_rpi_hash_next->nlp_rpi != rpi)) {
		ret = ret->nlp_rpi_hash_next;
	}

	if (ret->nlp_rpi_hash_next != 0) {
		temp = ret->nlp_rpi_hash_next;
		ret->nlp_rpi_hash_next = temp->nlp_rpi_hash_next;
		temp->nlp_rpi_hash_next = 0;
		return temp;
	} else {
		return 0;
	}
}

/*
 * This routine adds the node list entry to the
 * ndlp hash table.
 */
void
lpfc_addnode_rpi(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp,
		 uint16_t rpi)
{

	uint32_t index;

	index = LPFC_RPI_HASH_FUNC(rpi);
	ndlp->nlp_rpi_hash_next = phba->fc_nlplookup[index];
	phba->fc_nlplookup[index] = ndlp;
	return;
}

void
lpfc_nlp_init(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp,
		 uint32_t did)
{
	memset(ndlp, 0, sizeof (struct lpfc_nodelist));
	init_timer(&ndlp->nlp_tmofunc);
	ndlp->nlp_tmofunc.function = lpfc_nodev_timeout;
	init_timer(&ndlp->nlp_delayfunc);
	ndlp->nlp_delayfunc.function = lpfc_els_retry_delay;
	ndlp->nlp_DID = did;
	ndlp->nlp_phba = phba;
	return;
}

