/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_nportdisc.c 1.108 2004/07/22 20:34:01EDT sf_support Exp  $
 */

#include <linux/version.h>
#include <linux/blkdev.h>
#include <linux/dma-mapping.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <scsi/scsi_device.h>
#include "lpfc_sli.h"
#include "lpfc_disc.h"
#include "lpfc_scsi.h"
#include "lpfc_cfgparm.h"
#include "lpfc.h"
#include "lpfc_crtn.h"
#include "lpfc_hw.h"
#include "lpfc_logmsg.h"
#include "lpfc_mem.h"

extern uint8_t lpfcAlpaArray[];


static struct lpfc_bindlist *lpfc_assign_scsid(struct lpfc_hba *,
					struct lpfc_nodelist *, int);

static int
lpfc_check_adisc(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp,
		 struct lpfc_name * nn, struct lpfc_name * pn)
{
	if (memcmp(nn, &ndlp->nlp_nodename, sizeof (struct lpfc_name)) != 0)
		return 0;

	if (memcmp(pn, &ndlp->nlp_portname, sizeof (struct lpfc_name)) != 0)
		return 0;

	return 1;
}

static uint32_t
lpfc_rcv_plogi_unused_node(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	uint32_t *lp;
	IOCB_t *icmd;
	struct serv_parm *sp;
	LPFC_MBOXQ_t *mbox;
	struct ls_rjt stat;
	struct lpfc_node_farp_pend *pndlpfarp, *next_farp;
	FARP *pfarp;

	cmdiocb = (struct lpfc_iocbq *) arg;
	icmd = &cmdiocb->iocb;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	sp = (struct serv_parm *) ((uint8_t *) lp + sizeof (uint32_t));

	if ((phba->hba_state <= LPFC_FLOGI) ||
	    ((lpfc_check_sparm(phba, ndlp, sp, CLASS3) == 0))) {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		if (phba->hba_state <= LPFC_FLOGI) {
			stat.un.b.lsRjtRsnCodeExp = LSRJT_LOGICAL_BSY;
		} else {
			stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		}
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	} else {
		/* PLOGI chkparm OK */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_ELS,
				"%d:0114 PLOGI chkparm OK Data: x%x x%x x%x x%x\n",
				phba->brd_no,
				ndlp->nlp_DID, ndlp->nlp_state, ndlp->nlp_flag,
				ndlp->nlp_rpi);

		if ((phba->config[LPFC_CFG_FCP_CLASS] == 2) &&
		    (sp->cls2.classValid)) {
			ndlp->nlp_fcp_info |= CLASS2;
		} else {
			ndlp->nlp_fcp_info |= CLASS3;
		}

		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))
		    == 0) {
			goto out;
		}
		if ((phba->fc_flag & FC_PT2PT)
		    && !(phba->fc_flag & FC_PT2PT_PLOGI)) {
			/* The rcv'ed PLOGI determines what our NPortId will
			   be */
			phba->fc_myDID = icmd->un.rcvels.parmRo;
			lpfc_config_link(phba, mbox);
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mbox, phba->mbox_mem_pool);
				goto out;
			}
			if ((mbox = mempool_alloc(phba->mbox_mem_pool,
						  GFP_ATOMIC)) == 0) {
				goto out;
			}
		}
		if (lpfc_reg_login(phba, icmd->un.rcvels.remoteID,
				   (uint8_t *) sp, mbox, 0) == 0) {
			/* set_slim mailbox command needs to execute first,
			 * queue this command to be processed later.
			 */
			pfarp = (FARP *) pcmd;
			if ((pfarp->Rflags == FARP_REQUEST_PLOGI)
			    && (pfarp->Mflags == FARP_MATCH_PORT)) {
				struct list_head *head;
				head = &phba->fc_node_farp_list;

				list_for_each_entry_safe(pndlpfarp, next_farp,
						 &phba->fc_node_farp_list,
						 listentry) {
					list_del(&pndlpfarp->listentry);

					if ((memcmp(&pndlpfarp->rnode_addr,
						    &ndlp->nlp_portname,
						    sizeof (struct lpfc_name)) != 0) &&
					    (memcmp(&pndlpfarp->rnode_addr,
						    &ndlp->nlp_nodename,
						    sizeof (struct lpfc_name)) != 0)) {
						list_add(&pndlpfarp->listentry,
							 head);

					}
					else {
						break;

					}
				}
			}

			mbox->mbox_cmpl = lpfc_mbx_cmpl_reg_login;
			mbox->context2 = ndlp;
			ndlp->nlp_state = NLP_STE_REG_LOGIN_ISSUE;
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    != MBX_NOT_FINISHED) {
				lpfc_nlp_plogi(phba, ndlp);
				lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb,
						 ndlp, 0, 0);
				return (ndlp->nlp_state);	/* HAPPY PATH */
			}
			/* NOTE: we should have messages for unsuccessful
			   reglogin */
			mempool_free( mbox, phba->mbox_mem_pool);
		} else {
			mempool_free( mbox, phba->mbox_mem_pool);
		}
	}
out:
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_rcv_els_unused_node(struct lpfc_hba * phba,
			 struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	lpfc_issue_els_logo(phba, ndlp, 0);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_logo_unused_node(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;

	cmdiocb = (struct lpfc_iocbq *) arg;
	lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_cmpl_els_unused_node(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_cmpl_reglogin_unused_node(struct lpfc_hba * phba,
			       struct lpfc_nodelist * ndlp, void *arg,
			       uint32_t evt)
{
	/* dequeue, cancel timeout, unreg login */
	lpfc_nlp_remove(phba, ndlp);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_device_rm_unused_node(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	/* dequeue, cancel timeout, unreg login */
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_device_add_unused_node(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{
	if (ndlp->nlp_tmofunc.expires) {
		del_timer_sync(&ndlp->nlp_tmofunc);
		ndlp->nlp_tmofunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	}
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	}
	ndlp->nlp_state = NLP_STE_UNUSED_NODE;
	lpfc_nlp_plogi(phba, ndlp);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_device_unk_unused_node(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{
	/* dequeue, cancel timeout, unreg login */
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_rcv_plogi_plogi_issue(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, struct lpfc_iocbq *cmdiocb, uint32_t evt)
{
	struct lpfc_dmabuf *pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	uint32_t *lp = (uint32_t *) pcmd->virt;
	IOCB_t *icmd = &cmdiocb->iocb;
	struct serv_parm *sp = (struct serv_parm *) ((uint8_t *) lp + sizeof (uint32_t));
	LPFC_MBOXQ_t *mbox;
	struct ls_rjt stat;
	int port_cmp;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	if (ndlp->nlp_flag & NLP_LOGO_SND)
		return (ndlp->nlp_state);

	/* For a PLOGI, we only accept if our portname is less
	 * than the remote portname.
	 */
	phba->fc_stat.elsLogiCol++;
	port_cmp = memcmp(&phba->fc_portname, &sp->portName,
			  sizeof (struct lpfc_name));

	if (port_cmp >= 0) {
		/* Reject this request because the remote node will accept
		   ours */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		stat.un.b.lsRjtRsnCodeExp = LSEXP_CMD_IN_PROGRESS;
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	}
	else {
		if (lpfc_check_sparm(phba, ndlp, sp, CLASS3) == 0) {
			/* Reject this request because invalid parameters */
			stat.un.b.lsRjtRsvd0 = 0;
			stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
			stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
			stat.un.b.vendorUnique = 0;
			lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb,
					    ndlp);
		} else {
			/* PLOGI chkparm OK */
			lpfc_printf_log(phba,
					KERN_INFO,
					LOG_ELS,
					"%d:0120 PLOGI chkparm OK Data: x%x x%x x%x x%x\n",
					phba->brd_no,
					ndlp->nlp_DID, ndlp->nlp_state,
					ndlp->nlp_flag, ndlp->nlp_rpi);

			if ((phba->config[LPFC_CFG_FCP_CLASS] == 2) &&
			    (sp->cls2.classValid))
				ndlp->nlp_fcp_info |= CLASS2;
			else
				ndlp->nlp_fcp_info |= CLASS3;

			mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC);
			if (mbox) {
				if (!lpfc_reg_login (phba,
				      icmd->un.rcvels.remoteID,
				     (uint8_t *) sp, mbox, 0) == 0) {
					mempool_free(mbox, phba->mbox_mem_pool);
				} else {
					mbox->mbox_cmpl =
					    lpfc_mbx_cmpl_reg_login;
					mbox->context2 = ndlp;
					/* Issue Reg Login after successful
					   ACC */
					ndlp->nlp_flag |= NLP_ACC_REGLOGIN;

					/* Abort outstanding PLOGI */
					lpfc_driver_abort(phba, ndlp);

					lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI,
							 cmdiocb, ndlp, mbox,
							 0);
				}
			}
		}		/* if valid sparm */
	} /* if our portname was less */
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_prli_plogi_issue(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	/* software abort outstanding plogi, then send logout */
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	} else {
		lpfc_driver_abort(phba, ndlp);
	}
	lpfc_issue_els_logo(phba, ndlp, 0);

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_logo_plogi_issue(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq     *cmdiocb;
	struct lpfc_dmabuf         *pcmd;
	struct lpfc_target *targetp;
	unsigned long     tmo;
	int               issue_abort;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	/* software abort outstanding plogi before sending acc */
	issue_abort = 0;
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	} else {
		issue_abort = 1;
	}
	lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);

	/* resend plogi after 1 sec delay */
	targetp = ndlp->nlp_Target;
	if (targetp) {
		targetp->targetFlags |= FC_NPR_ACTIVE;
		if (targetp->tmofunc.expires) {
			del_timer_sync(&targetp->tmofunc);
			targetp->tmofunc.expires = 0;
		}
		if (phba->config[LPFC_CFG_NODEV_TMO] >
		   phba->config[LPFC_CFG_LINKDOWN_TMO]) {
			tmo = phba->config[LPFC_CFG_NODEV_TMO];
		} else {
			tmo = phba->config[LPFC_CFG_LINKDOWN_TMO];
		}
		targetp->tmofunc.expires = jiffies + HZ * tmo;
		targetp->tmofunc.data = (unsigned long)targetp;
		add_timer(&targetp->tmofunc);
	}
	ndlp->nlp_rflag |= NLP_NPR_ACTIVE;
	ndlp->nlp_flag |= NLP_DELAY_TMO;
	ndlp->nlp_retry = 0;

	ndlp->nlp_last_elscmd = (unsigned long)ELS_CMD_PLOGI;

	ndlp->nlp_delayfunc.expires = jiffies + HZ * 1;
	ndlp->nlp_delayfunc.data = (unsigned long)ndlp;
	add_timer(&ndlp->nlp_delayfunc);

	if (issue_abort) {
		lpfc_driver_abort(phba, ndlp);
	}
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_els_plogi_issue(struct lpfc_hba * phba,
			 struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	/* software abort outstanding plogi, then send logout */
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	} else {
		lpfc_driver_abort(phba, ndlp);
	}
	lpfc_issue_els_logo(phba, ndlp, 0);
	return (ndlp->nlp_state);
}

/*! lpfc_cmpl_plogi_plogi_issue
  *
  * \pre
  * \post
  * \param   phba
  * \param   ndlp
  * \param   arg
  * \param   evt
  * \return  uint32_t
  *
  * \b Description:
  *     This routine is envoked when we rcv a PLOGI completion from a node we
  *     tried to log into. We check the CSPs, and the ulpStatus. If successful
  *     change the state to REG_LOGIN_ISSUE and issue a REG_LOGIN. For failure,
  *     we free the nodelist entry.
  */

static uint32_t
lpfc_cmpl_plogi_plogi_issue(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb, *rspiocb;
	struct lpfc_dmabuf *pcmd, *prsp;
	uint32_t *lp;
	IOCB_t *irsp;
	struct serv_parm *sp;
	LPFC_MBOXQ_t *mbox;

	cmdiocb = (struct lpfc_iocbq *) arg;
	rspiocb = cmdiocb->context_un.rsp_iocb;

	if (ndlp->nlp_flag & NLP_ACC_REGLOGIN) {
		return (ndlp->nlp_state);
	}

	irsp = &rspiocb->iocb;

	if (irsp->ulpStatus == 0) {
		pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
		
		prsp = (struct lpfc_dmabuf *) pcmd->list.next;
		lp = (uint32_t *) prsp->virt;

		pci_dma_sync_single_for_cpu(phba->pcidev, prsp->phys,
			LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

		sp = (struct serv_parm *) ((uint8_t *) lp + sizeof (uint32_t));
		if ((lpfc_check_sparm(phba, ndlp, sp, CLASS3))) {
			/* PLOGI chkparm OK */
			lpfc_printf_log(phba,
					KERN_INFO,
					LOG_ELS,
					"%d:0121 PLOGI chkparm OK Data: x%x x%x x%x x%x\n",
					phba->brd_no,
					ndlp->nlp_DID, ndlp->nlp_state,
					ndlp->nlp_flag, ndlp->nlp_rpi);

			if ((phba->config[LPFC_CFG_FCP_CLASS] == 2) &&
			    (sp->cls2.classValid)) {
				ndlp->nlp_fcp_info |= CLASS2;
			} else {
				ndlp->nlp_fcp_info |= CLASS3;
			}

			if ((mbox = mempool_alloc(phba->mbox_mem_pool,
						  GFP_ATOMIC))) {
				if (lpfc_reg_login
				    (phba, irsp->un.elsreq64.remoteID,
				     (uint8_t *) sp, mbox, 0) == 0) {
					/* set_slim mailbox command needs to
					 * execute first, queue this command to
					 * be processed later.
					 */
					if (ndlp->nlp_DID == NameServer_DID) {
						mbox->mbox_cmpl =
						    lpfc_mbx_cmpl_ns_reg_login;
					} else if (ndlp->nlp_DID == FDMI_DID) {
						mbox->mbox_cmpl =
						   lpfc_mbx_cmpl_fdmi_reg_login;
					} else {
						mbox->mbox_cmpl =
						    lpfc_mbx_cmpl_reg_login;
					}
					mbox->context2 = ndlp;
					ndlp->nlp_state =
					    NLP_STE_REG_LOGIN_ISSUE;
					if (lpfc_sli_issue_mbox
					    (phba, mbox,
					     (MBX_NOWAIT | MBX_STOP_IOCB))
					    != MBX_NOT_FINISHED) {
						return (ndlp->nlp_state);
					}
					mempool_free(mbox, phba->mbox_mem_pool);
				} else {
					mempool_free(mbox, phba->mbox_mem_pool);
				}
			}
		}
	}

	/* If we are in the middle of discovery,
	 * take necessary actions to finish up.
	 */
	if (ndlp->nlp_DID == NameServer_DID) {
		/* Link up / RSCN discovery */
		lpfc_disc_start(phba);
	}

	/* If this is a driver initiated abort, to a FCP target,
	 * still honor the timer to relogin after a second.
	 */
	if ((irsp->ulpStatus == IOSTAT_LOCAL_REJECT) &&
	  (irsp->un.ulpWord[4] == IOERR_SLI_ABORTED) &&
	  (ndlp->nlp_delayfunc.expires) && (ndlp->nlp_flag & NLP_DELAY_TMO)) {
		return (ndlp->nlp_state);
	}
	ndlp->nlp_state = NLP_STE_UNUSED_NODE;

	/* Free this node since the driver cannot login or has the wrong
	   sparm */
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_cmpl_prli_plogi_issue(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{

	/* First ensure ndlp is on the plogi list */
	if (ndlp->nlp_flag & NLP_LIST_MASK) {
		lpfc_findnode_did(phba, (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
				  ndlp->nlp_DID);
	}
	lpfc_nlp_plogi(phba, ndlp);

	/* If a PLOGI is not already pending, issue one */
	if (!(ndlp->nlp_flag & NLP_PLOGI_SND)) {
		ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
		lpfc_issue_els_plogi(phba, ndlp, 0);
		ndlp->nlp_flag |= NLP_DISC_NODE;
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_logo_plogi_issue(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	if (!(ndlp->nlp_flag & NLP_PLOGI_SND)) {

		ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
		lpfc_issue_els_plogi(phba, ndlp, 0);
	}
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_adisc_plogi_issue(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{

	/* First ensure ndlp is on the plogi list */
	if (ndlp->nlp_flag & NLP_LIST_MASK) {
		lpfc_findnode_did(phba, (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
				  ndlp->nlp_DID);
	}
	lpfc_nlp_plogi(phba, ndlp);

	/* If a PLOGI is not already pending, issue one */
	if (!(ndlp->nlp_flag & NLP_PLOGI_SND)) {
		ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
		lpfc_issue_els_plogi(phba, ndlp, 0);
		ndlp->nlp_flag |= NLP_DISC_NODE;
	}
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_reglogin_plogi_issue(struct lpfc_hba * phba,
			       struct lpfc_nodelist * ndlp, void *arg,
			       uint32_t evt)
{
	LPFC_MBOXQ_t *pmb, *mbox;
	MAILBOX_t *mb;
	uint32_t ldata;
	uint16_t rpi;

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	pmb = (LPFC_MBOXQ_t *) arg;
	mb = &pmb->mb;
	ldata = mb->un.varWords[0];	/* rpi */
	rpi = (uint16_t) (le32_to_cpu(ldata) & 0xFFFF);

	/* first unreg node's rpi */
	if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
		/* now unreg rpi just got back from reg_login */
		lpfc_unreg_login(phba, rpi, mbox);
		if (lpfc_sli_issue_mbox(phba, mbox,
					(MBX_NOWAIT | MBX_STOP_IOCB))
		    == MBX_NOT_FINISHED) {
			mempool_free( mbox, phba->mbox_mem_pool);
		}
	}

	/* software abort outstanding plogi */
	lpfc_driver_abort(phba, ndlp);
	/* send a new plogi */
	ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
	lpfc_issue_els_plogi(phba, ndlp, 0);

	return (ndlp->nlp_state);
}

/*! lpfc_device_rm_plogi_issue
  *
  * \pre
  * \post
  * \param   phba
  * \param   ndlp
  * \param   arg
  * \param   evt
  * \return  uint32_t
  *
  * \b Description:
  *     This routine is envoked when we a request to remove a nport we are in
  *     the process of PLOGIing. We should issue a software abort on the
  *     outstanding PLOGI request, then issue a LOGO request. Change node state
  *     to UNUSED_NODE so it can be freed when LOGO completes.
  *
  */

static uint32_t
lpfc_device_rm_plogi_issue(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{

	/* software abort outstanding plogi, before sending LOGO */
	lpfc_driver_abort(phba, ndlp);

	/* If discovery processing causes us to remove a device, it is important
	 * that nothing gets sent to the device (soft zoning issues).
	 */
	ndlp->nlp_state = NLP_STE_UNUSED_NODE;
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_device_unk_plogi_issue(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{
	/* software abort outstanding plogi */
	lpfc_driver_abort(phba, ndlp);

	/* dequeue, cancel timeout, unreg login */
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_rcv_plogi_reglogin_issue(struct lpfc_hba * phba,
			      struct lpfc_nodelist * ndlp, void *arg,
			      uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	uint32_t *lp;
	struct serv_parm *sp;
	struct ls_rjt stat;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	sp = (struct serv_parm *) ((uint8_t *) lp + sizeof (uint32_t));

	if ((lpfc_check_sparm(phba, ndlp, sp, CLASS3) == 0)) {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	} else {
		/* PLOGI chkparm OK */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_ELS,
				"%d:0122 PLOGI chkparm OK Data: x%x x%x x%x x%x\n",
				phba->brd_no,
				ndlp->nlp_DID, ndlp->nlp_state, ndlp->nlp_flag,
				((struct lpfc_nodelist*) ndlp)->nlp_rpi);

		lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb, ndlp, 0, 0);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_prli_reglogin_issue(struct lpfc_hba * phba,
			     struct lpfc_nodelist * ndlp, void *arg,
			     uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	lpfc_els_rsp_prli_acc(phba, cmdiocb, ndlp);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_logo_reglogin_issue(struct lpfc_hba * phba,
			     struct lpfc_nodelist * ndlp, void *arg,
			     uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);

	/* resend plogi */
	ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
	lpfc_issue_els_plogi(phba, ndlp, 0);

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_padisc_reglogin_issue(struct lpfc_hba * phba,
			       struct lpfc_nodelist * ndlp, void *arg,
			       uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	uint32_t *lp;
	struct serv_parm *sp;
	struct ls_rjt stat;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	sp = (struct serv_parm *) ((uint8_t *) lp + sizeof (uint32_t));

	if ((lpfc_check_sparm(phba, ndlp, sp, CLASS3) == 0)) {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	} else {
		/* PLOGI chkparm OK */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_ELS,
				"%d:0123 PLOGI chkparm OK Data: x%x x%x x%x x%x\n",
				phba->brd_no,
				ndlp->nlp_DID, ndlp->nlp_state, ndlp->nlp_flag,
				((struct lpfc_nodelist*) ndlp)->nlp_rpi);

		lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb, ndlp, 0, 0);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_prlo_reglogin_issue(struct lpfc_hba * phba,
			     struct lpfc_nodelist * ndlp, void *arg,
			     uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;

	cmdiocb = (struct lpfc_iocbq *) arg;
	lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_logo_reglogin_issue(struct lpfc_hba * phba,
			      struct lpfc_nodelist * ndlp, void *arg,
			      uint32_t evt)
{
	/*
	 * don't really want to do anything since reglogin has not finished,
	 * and we won't let any els happen until the mb is finished.
	 */
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_adisc_reglogin_issue(struct lpfc_hba * phba,
			       struct lpfc_nodelist * ndlp, void *arg,
			       uint32_t evt)
{

	/* First ensure ndlp is on the plogi list */
	if (ndlp->nlp_flag & NLP_LIST_MASK) {
		lpfc_findnode_did(phba, (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
				  ndlp->nlp_DID);
	}
	lpfc_nlp_plogi(phba, ndlp);

	return (ndlp->nlp_state);
}

/*! lpfc_cmpl_reglogin_reglogin_issue
  *
  * \pre
  * \post
  * \param   phba
  * \param   ndlp
  * \param   arg
  * \param   evt
  * \return  uint32_t
  *
  * \b Description:
  *     This routine is envoked when the REG_LOGIN completes. If unsuccessful,
  *     we should send a LOGO ELS request and free the node entry. If
  *     successful, save the RPI assigned, then issue a PRLI request. The
  *     nodelist entry should be moved to the unmapped list.  If the NPortID
  *     indicates a Fabric entity, don't issue PRLI, just go straight into
  *     PRLI_COMPL.  PRLI_COMPL - for fabric entity
  */
static uint32_t
lpfc_cmpl_reglogin_reglogin_issue(struct lpfc_hba * phba,
				  struct lpfc_nodelist * ndlp,
				  void *arg, uint32_t evt)
{
	LPFC_MBOXQ_t *pmb;
	MAILBOX_t *mb;
	struct lpfc_nodelist *plogindlp;
	uint32_t did;

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	pmb = (LPFC_MBOXQ_t *) arg;
	mb = &pmb->mb;
	did = mb->un.varWords[1];
	if (mb->mbxStatus ||
	    ((plogindlp = lpfc_findnode_did(phba,
					     (NLP_SEARCH_PLOGI |
					      NLP_SEARCH_DEQUE), did)) == 0)
	    || (ndlp != plogindlp)) {
		/* RegLogin failed */
		lpfc_printf_log(phba,
				KERN_ERR,
				LOG_DISCOVERY,
				"%d:0246 RegLogin failed Data: x%x x%x x%x\n",
				phba->brd_no,
				did, mb->mbxStatus, phba->hba_state);

		if (ndlp->nlp_flag & NLP_LIST_MASK) {
			lpfc_findnode_did(phba,
					  (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
					  ndlp->nlp_DID);
		}
		ndlp->nlp_state = NLP_STE_UNUSED_NODE;
		lpfc_nlp_remove(phba, ndlp);
		return (NLP_STE_FREED_NODE);
	}

	if (ndlp->nlp_rpi != 0)
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);

	ndlp->nlp_rpi = mb->un.varWords[0];
	lpfc_addnode_rpi(phba, ndlp, ndlp->nlp_rpi);
	lpfc_nlp_unmapped(phba, ndlp);

	/* Only if we are not a fabric nport do we issue PRLI */
	if (!(ndlp->nlp_type & NLP_FABRIC)) {
		lpfc_issue_els_prli(phba, ndlp, 0);
		ndlp->nlp_state = NLP_STE_PRLI_ISSUE;
	} else {
		ndlp->nlp_state = NLP_STE_PRLI_COMPL;
	}

	return (ndlp->nlp_state);
}

/*! lpfc_device_rm_reglogin_issue
  *
  * \pre
  * \post
  * \param   phba
  * \param   ndlp
  * \param   arg
  * \param   evt
  * \return  uint32_t
  *
  * \b Description:
  *     This routine is envoked when we a request to remove a nport we are in
  *     the process of REG_LOGINing. We should issue a UNREG_LOGIN by did, then
  *     issue a LOGO request. Change node state to NODE_UNUSED, so it will be
  *     freed when LOGO completes.
  *
  */

static uint32_t
lpfc_device_rm_reglogin_issue(struct lpfc_hba * phba,
			      struct lpfc_nodelist * ndlp, void *arg,
			      uint32_t evt)
{
	LPFC_MBOXQ_t *mbox;

	if (ndlp->nlp_rpi) {
		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			lpfc_unreg_login(phba, ndlp->nlp_rpi, mbox);
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mbox, phba->mbox_mem_pool);
			}
		}
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);
		lpfc_no_rpi(phba, ndlp);
		ndlp->nlp_rpi = 0;
	}

	/* If discovery processing causes us to remove a device, it is important
	 * that nothing gets sent to the device (soft zoning issues).
	 */
	ndlp->nlp_state = NLP_STE_UNUSED_NODE;
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

/* DEVICE_ADD for REG_LOGIN_ISSUE is nodev */

static uint32_t
lpfc_device_unk_reglogin_issue(struct lpfc_hba * phba,
			       struct lpfc_nodelist * ndlp, void *arg,
			       uint32_t evt)
{
	/* dequeue, cancel timeout, unreg login */
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_rcv_plogi_prli_issue(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	uint32_t *lp;
	struct serv_parm *sp;
	struct ls_rjt stat;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	sp = (struct serv_parm *) ((uint8_t *) lp + sizeof (uint32_t));

	if ((lpfc_check_sparm(phba, ndlp, sp, CLASS3) == 0)) {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	} else {
		/* PLOGI chkparm OK */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_ELS,
				"%d:0124 PLOGI chkparm OK Data: x%x x%x x%x x%x\n",
				phba->brd_no,
				ndlp->nlp_DID, ndlp->nlp_state, ndlp->nlp_flag,
				((struct lpfc_nodelist*) ndlp)->nlp_rpi);

		lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb, ndlp, 0, 0);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_prli_prli_issue(struct lpfc_hba * phba,
			 struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	lpfc_els_rsp_prli_acc(phba, cmdiocb, ndlp);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_logo_prli_issue(struct lpfc_hba * phba,
			 struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	/* Software abort outstanding prli before sending acc */
	lpfc_driver_abort(phba, ndlp);

	/* Only call LOGO ACC for first LOGO, this avoids sending unnecessary
	 * PLOGIs during LOGO storms from a device.
	 */
	if (ndlp->nlp_flag & NLP_LOGO_ACC) {
		ndlp->nlp_flag &= ~NLP_LOGO_ACC;
		lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
		ndlp->nlp_flag |= NLP_LOGO_ACC;
	} else {
		ndlp->nlp_flag |= NLP_LOGO_ACC;
		lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
	}

	/* The driver has to wait until the ACC completes before it continues
	 * processing the LOGO.  The action will resume in
	 * lpfc_cmpl_els_logo_acc routine. Since part of processing includes an
	 * unreg_login, the driver waits so the ACC does not get aborted.
	 */
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_padisc_prli_issue(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	struct serv_parm *sp;		/* used for PDISC */
	ADISC *ap;		/* used for ADISC */
	uint32_t *lp;
	uint32_t cmd;
	struct lpfc_name *pnn, *ppn;
	struct ls_rjt stat;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;
	cmd = *lp++;

	if (cmd == ELS_CMD_ADISC) {
		ap = (ADISC *) lp;
		pnn = (struct lpfc_name *) & ap->nodeName;
		ppn = (struct lpfc_name *) & ap->portName;
	} else {
		sp = (struct serv_parm *) lp;
		pnn = (struct lpfc_name *) & sp->nodeName;
		ppn = (struct lpfc_name *) & sp->portName;
	}

	if (lpfc_check_adisc(phba, ndlp, pnn, ppn)) {
		if (cmd == ELS_CMD_ADISC) {
			lpfc_els_rsp_adisc_acc(phba, cmdiocb, ndlp);
		} else {
			lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb, ndlp,
					 0, 0);
		}
	} else {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	}

	return (ndlp->nlp_state);
}

/* This routine is envoked when we rcv a PRLO request from a nport
 * we are logged into.  We should send back a PRLO rsp setting the
 * appropriate bits.
 * NEXT STATE = PRLI_ISSUE
 */
static uint32_t
lpfc_rcv_prlo_prli_issue(struct lpfc_hba * phba,
			 struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;

	cmdiocb = (struct lpfc_iocbq *) arg;
	lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_prli_prli_issue(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb, *rspiocb;
	struct lpfc_dmabuf *pcmd, *prsp;
	uint32_t *lp;
	IOCB_t *irsp;
	PRLI *npr;
	struct lpfc_bindlist *blp;
	struct lpfc_target *targetp;

	cmdiocb = (struct lpfc_iocbq *) arg;
	rspiocb = cmdiocb->context_un.rsp_iocb;
	irsp = &rspiocb->iocb;
	if (irsp->ulpStatus) {
		ndlp->nlp_state = NLP_STE_PRLI_COMPL;
		return (ndlp->nlp_state);
	}
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	
	prsp = (struct lpfc_dmabuf *) pcmd->list.next;
	lp = (uint32_t *) prsp->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, prsp->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	npr = (PRLI *) ((uint8_t *) lp + sizeof (uint32_t));

	/* Check out PRLI rsp */
	if ((npr->acceptRspCode != PRLI_REQ_EXECUTED) ||
	    (npr->prliType != PRLI_FCP_TYPE) || (npr->targetFunc != 1)) {
		ndlp->nlp_state = NLP_STE_PRLI_COMPL;
		return (ndlp->nlp_state);
	}
	if (npr->Retry == 1) {
		ndlp->nlp_fcp_info |= NLP_FCP_2_DEVICE;
	}

	/* Can we assign a SCSI Id to this NPort */
	if ((blp = lpfc_assign_scsid(phba, ndlp, 0))) {
		lpfc_nlp_mapped(phba, ndlp, blp);
		ndlp->nlp_failMask = 0;
		targetp = ndlp->nlp_Target;
		if (targetp) {
			if (targetp->tmofunc.expires) {
				del_timer_sync(&targetp->tmofunc);
				targetp->tmofunc.expires = 0;
			}
		} else {
			/* new target to driver, allocate space to target <sid>
			   lun 0 */
			if (blp->nlp_Target == 0) {
				lpfc_find_target(phba, blp->nlp_sid, 0, 1);
				blp->nlp_Target =
				    phba->device_queue_hash[blp->nlp_sid];
			}
		}

		ndlp->nlp_state = NLP_STE_MAPPED_NODE;
	} else {
		ndlp->nlp_state = NLP_STE_PRLI_COMPL;
		ndlp->nlp_flag |= NLP_TGT_NO_SCSIID;
	}
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_logo_prli_issue(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	/* dequeue, cancel timeout, unreg login */
	lpfc_freenode(phba, ndlp);

	/* software abort outstanding prli, then send logout */
	lpfc_driver_abort(phba, ndlp);
	lpfc_issue_els_logo(phba, ndlp, 0);
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_cmpl_adisc_prli_issue(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_bindlist *blp;
	struct lpfc_target *targetp;

	/* First ensure ndlp is on the unmap list */
	if (ndlp->nlp_flag & NLP_LIST_MASK) {
		lpfc_findnode_did(phba, (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
				  ndlp->nlp_DID);
	}

	/* Can we reassign a SCSI Id to this NPort */
	if ((blp = lpfc_assign_scsid(phba, ndlp, 1))) {
		lpfc_nlp_mapped(phba, ndlp, blp);
		/* NOTE: can't we have this in lpfc_nlp_mapped */
		ndlp->nlp_state = NLP_STE_MAPPED_NODE;
		targetp = ndlp->nlp_Target;
		if (targetp) {
			targetp->targetFlags &= ~FC_NPR_ACTIVE;
			if (targetp->tmofunc.expires) {
				del_timer_sync(&targetp->tmofunc);
				targetp->tmofunc.expires = 0;
			}
		}
		else {
			/* new target to driver, allocate space to target <sid>
			   lun 0 */
			if (blp->nlp_Target == 0) {
				lpfc_find_target(phba, blp->nlp_sid, 0, 1);
				blp->nlp_Target =
					phba->device_queue_hash[blp->nlp_sid];
			}
		}
		lpfc_set_failmask(phba, ndlp, LPFC_DEV_ALL_BITS,
				  LPFC_CLR_BITMASK);
	}
	else {
		lpfc_nlp_unmapped(phba, ndlp);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_reglogin_prli_issue(struct lpfc_hba * phba,
			      struct lpfc_nodelist * ndlp, void *arg,
			      uint32_t evt)
{
	LPFC_MBOXQ_t *pmb, *mbox;
	MAILBOX_t *mb;
	uint32_t ldata;
	uint16_t rpi;

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	pmb = (LPFC_MBOXQ_t *) arg;
	mb = &pmb->mb;
	ldata = mb->un.varWords[0];	/* rpi */
	rpi = (uint16_t) (le32_to_cpu(ldata) & 0xFFFF);

	if (ndlp->nlp_rpi != rpi) {
		/* first unreg node's rpi */
		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			lpfc_unreg_login(phba, ndlp->nlp_rpi, mbox);
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mbox, phba->mbox_mem_pool);
			}
		}
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);
		lpfc_no_rpi(phba, ndlp);
		ndlp->nlp_rpi = 0;

		/* now unreg rpi just got back from reg_login */
		lpfc_unreg_login(phba, rpi, mbox);
		if (lpfc_sli_issue_mbox(phba, mbox,
					(MBX_NOWAIT | MBX_STOP_IOCB))
		    == MBX_NOT_FINISHED) {
			mempool_free( mbox, phba->mbox_mem_pool);
		}

		/* software abort outstanding prli */
		lpfc_driver_abort(phba, ndlp);

		/* send logout and put this node on plogi list */
		lpfc_issue_els_logo(phba, ndlp, 0);
		ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
		lpfc_nlp_plogi(phba, ndlp);
	}

	return (ndlp->nlp_state);
}

/*! lpfc_device_rm_prli_issue
  *
  * \pre
  * \post
  * \param   phba
  * \param   ndlp
  * \param   arg
  * \param   evt
  * \return  uint32_t
  *
  * \b Description:
  *    This routine is envoked when we a request to remove a nport we are in the
  *    process of PRLIing. We should software abort outstanding prli, unreg
  *    login, send a logout. We will change node state to UNUSED_NODE, put it
  *    on plogi list so it can be freed when LOGO completes.
  *
  */
static uint32_t
lpfc_device_rm_prli_issue(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{

	/* software abort outstanding prli */
	lpfc_driver_abort(phba, ndlp);

	ndlp->nlp_state = NLP_STE_UNUSED_NODE;
	/* dequeue, cancel timeout, unreg login */
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_device_add_prli_issue(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	}
	/* software abort outstanding prli */
	lpfc_driver_abort(phba, ndlp);

	lpfc_nlp_adisc(phba, ndlp);
	return (ndlp->nlp_state);
}

/*! lpfc_device_unk_prli_issue
  *
  * \pre
  * \post
  * \param   phba
  * \param   ndlp
  * \param   arg
  * \param   evt
  * \return  uint32_t
  *
  * \b Description:
  *    The routine is envoked when the state of a device is unknown, like
  *    during a link down. We should remove the nodelist entry from the
  *    unmapped list, issue a UNREG_LOGIN, do a software abort of the
  *    outstanding PRLI command, then free the node entry.
  */
static uint32_t
lpfc_device_unk_prli_issue(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	/* software abort outstanding prli */
	lpfc_driver_abort(phba, ndlp);

	/* dequeue, cancel timeout, unreg login */
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_rcv_plogi_prli_compl(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	uint32_t *lp;
	struct serv_parm *sp;
	struct ls_rjt stat;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	sp = (struct serv_parm *) ((uint8_t *) lp + sizeof (uint32_t));

	if ((phba->hba_state <= LPFC_FLOGI) ||
	    ((lpfc_check_sparm(phba, ndlp, sp, CLASS3) == 0))) {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		if (phba->hba_state <= LPFC_FLOGI) {
			stat.un.b.lsRjtRsnCodeExp = LSRJT_LOGICAL_BSY;
		} else {
			stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		}
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	} else {
		/* PLOGI chkparm OK */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_ELS,
				"%d:0125 PLOGI chkparm OK Data: x%x x%x x%x x%x\n",
				phba->brd_no,
				ndlp->nlp_DID, ndlp->nlp_state, ndlp->nlp_flag,
				((struct lpfc_nodelist*) ndlp)->nlp_rpi);

		lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb, ndlp, 0, 0);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_prli_prli_compl(struct lpfc_hba * phba,
			 struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	lpfc_els_rsp_prli_acc(phba, cmdiocb, ndlp);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_logo_prli_compl(struct lpfc_hba * phba,
			 struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	/* software abort outstanding adisc before sending acc */
	if (ndlp->nlp_flag & NLP_ADISC_SND) {
		lpfc_driver_abort(phba, ndlp);
	}
	/* Only call LOGO ACC for first LOGO, this avoids sending unnecessary
	 * PLOGIs during LOGO storms from a device.
	 */
	if (ndlp->nlp_flag & NLP_LOGO_ACC) {
		ndlp->nlp_flag &= ~NLP_LOGO_ACC;
		lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
		ndlp->nlp_flag |= NLP_LOGO_ACC;
	} else {
		ndlp->nlp_flag |= NLP_LOGO_ACC;
		lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
	}

	/* The driver has to wait until the ACC completes before we can continue
	 * processing the LOGO, the action will resume in
	 * lpfc_cmpl_els_logo_acc.  Since part of processing includes an
	 * unreg_login, the driver waits so the ACC does not get aborted.
	 */
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_padisc_prli_compl(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	struct serv_parm *sp;		/* used for PDISC */
	ADISC *ap;		/* used for ADISC */
	uint32_t *lp;
	uint32_t cmd;
	struct lpfc_name *pnn, *ppn;
	struct ls_rjt stat;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;
	cmd = *lp++;

	if (cmd == ELS_CMD_ADISC) {
		ap = (ADISC *) lp;
		pnn = (struct lpfc_name *) & ap->nodeName;
		ppn = (struct lpfc_name *) & ap->portName;
	} else {
		sp = (struct serv_parm *) lp;
		pnn = (struct lpfc_name *) & sp->nodeName;
		ppn = (struct lpfc_name *) & sp->portName;
	}

	if (lpfc_check_adisc(phba, ndlp, pnn, ppn)) {
		if (cmd == ELS_CMD_ADISC) {
			lpfc_els_rsp_adisc_acc(phba, cmdiocb, ndlp);
		} else {
			lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb, ndlp,
					 0, 0);
		}
	} else {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_prlo_prli_compl(struct lpfc_hba * phba,
			 struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;

	cmdiocb = (struct lpfc_iocbq *) arg;
	lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_logo_prli_compl(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	/* dequeue, cancel timeout, unreg login */
	lpfc_freenode(phba, ndlp);

	if (ndlp->nlp_flag & NLP_ADISC_SND) {
		/* software abort outstanding adisc */
		lpfc_driver_abort(phba, ndlp);
	}

	ndlp->nlp_state = NLP_STE_UNUSED_NODE;
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_cmpl_adisc_prli_compl(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb, *rspiocb;
	struct lpfc_dmabuf *pcmd, *prsp;
	uint32_t *lp;
	IOCB_t *irsp;
	struct lpfc_bindlist *blp;
	struct lpfc_target *targetp;
	ADISC *ap;

	cmdiocb = (struct lpfc_iocbq *) arg;
	rspiocb = cmdiocb->context_un.rsp_iocb;
	irsp = &rspiocb->iocb;

	/* First remove the ndlp from any list */
	if (ndlp->nlp_flag & NLP_LIST_MASK) {
		lpfc_findnode_did(phba, (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
				  ndlp->nlp_DID);
	}

	if (irsp->ulpStatus) {
		lpfc_freenode(phba, ndlp);
		ndlp->nlp_state = NLP_STE_UNUSED_NODE;
		lpfc_nlp_plogi(phba, ndlp);
		return (ndlp->nlp_state);
	}

	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	
	prsp = (struct lpfc_dmabuf *) pcmd->list.next;
	lp = (uint32_t *) prsp->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, prsp->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	ap = (ADISC *) ((uint8_t *) lp + sizeof (uint32_t));

	/* Check out ADISC rsp */
	if ((lpfc_check_adisc(phba, ndlp, &ap->nodeName, &ap->portName) == 0)) {
		lpfc_freenode(phba, ndlp);
		ndlp->nlp_state = NLP_STE_UNUSED_NODE;
		lpfc_nlp_plogi(phba, ndlp);
		return (ndlp->nlp_state);
	}
	if ((blp = lpfc_assign_scsid(phba, ndlp, 1))) {
		lpfc_nlp_mapped(phba, ndlp, blp);
		/* NOTE: can't we have this in lpfc_nlp_mapped */
		ndlp->nlp_state = NLP_STE_MAPPED_NODE;
		targetp = ndlp->nlp_Target;
		if (targetp) {
			targetp->targetFlags &= ~FC_NPR_ACTIVE;
			if (targetp->tmofunc.expires) {
				del_timer_sync(&targetp->tmofunc);
				targetp->tmofunc.expires = 0;
			}
		}
		else {
			/* new target to driver, allocate space to target <sid>
			   lun 0 */
			if (blp->nlp_Target == 0) {
				lpfc_find_target(phba, blp->nlp_sid, 0, 1);
				blp->nlp_Target =
					phba->device_queue_hash[blp->nlp_sid];
			}
		}
		lpfc_set_failmask(phba, ndlp, LPFC_DEV_ALL_BITS,
				  LPFC_CLR_BITMASK);
	}
	else {
		lpfc_nlp_unmapped(phba, ndlp);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_reglogin_prli_compl(struct lpfc_hba * phba,
			      struct lpfc_nodelist * ndlp, void *arg,
			      uint32_t evt)
{
	LPFC_MBOXQ_t *pmb, *mbox;
	MAILBOX_t *mb;
	uint32_t ldata;
	uint16_t rpi;

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	pmb = (LPFC_MBOXQ_t *) arg;
	mb = &pmb->mb;
	ldata = mb->un.varWords[0];	/* rpi */
	rpi = (uint16_t) (le32_to_cpu(ldata) & 0xFFFF);

	if (ndlp->nlp_rpi != rpi) {
		/* first unreg node's rpi */
		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			lpfc_unreg_login(phba, ndlp->nlp_rpi, mbox);
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mbox, phba->mbox_mem_pool);
			}
		}
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);
		lpfc_no_rpi(phba, ndlp);
		ndlp->nlp_rpi = 0;

		/* now unreg rpi just got back from reg_login */
		lpfc_unreg_login(phba, rpi, mbox);
		if (lpfc_sli_issue_mbox(phba, mbox,
					(MBX_NOWAIT | MBX_STOP_IOCB))
		    == MBX_NOT_FINISHED) {
			mempool_free( mbox, phba->mbox_mem_pool);
		}

		if (ndlp->nlp_flag & NLP_ADISC_SND) {
			/* software abort outstanding adisc */
			lpfc_driver_abort(phba, ndlp);
		}

		/* send logout and put this node on plogi list */
		lpfc_issue_els_logo(phba, ndlp, 0);
		ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
		lpfc_nlp_plogi(phba, ndlp);
	}

	return (ndlp->nlp_state);
}

/*! lpfc_device_rm_prli_compl
  *
  * \pre
  * \post
  * \param   phba
  * \param   ndlp
  * \param   arg
  * \param   evt
  * \return  uint32_t
  *
  * \b Description:
  *    This routine is envoked when we a request to remove a nport.
  *    It could be called when linkdown or nodev timer expires.
  *    If nodev timer is still running, we just want to exit.
  *    If this node timed out, we want to abort outstanding ADISC,
  *    unreg login, send logout, change state to UNUSED_NODE and
  *    place node on plogi list so it can be freed when LOGO completes.
  *
  */
static uint32_t
lpfc_device_rm_prli_compl(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{

	/* if nodev timer is running, then we just exit */
	if (!(ndlp->nlp_flag & NLP_NODEV_TMO)) {
		lpfc_set_failmask(phba, ndlp, LPFC_DEV_DISCONNECTED,
				  LPFC_SET_BITMASK);
		if (ndlp->nlp_flag & NLP_ADISC_SND) {
			/* software abort outstanding adisc */
			lpfc_driver_abort(phba, ndlp);
		}

		/* If discovery processing causes us to remove a device, it is
		 * important that nothing gets sent to the device (soft zoning
		 * issues).
		 */
		ndlp->nlp_state = NLP_STE_UNUSED_NODE;
		/* dequeue, cancel timeout, unreg login */
		lpfc_nlp_remove(phba, ndlp);
		return (NLP_STE_FREED_NODE);
	}
	lpfc_set_failmask(phba, ndlp, LPFC_DEV_DISAPPEARED, LPFC_SET_BITMASK);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_device_add_prli_compl(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	}
	lpfc_nlp_adisc(phba, ndlp);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_device_unk_prli_compl(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	unsigned long tmo;


	/* If we are in pt2pt mode, free node */
	if (phba->fc_flag & FC_PT2PT) {
		/* dequeue, cancel timeout, unreg login */
		lpfc_nlp_remove(phba, ndlp);
		return (NLP_STE_FREED_NODE);
	}

	/* linkdown timer should be running at this time.  Check to
	 * see if the driver needs to start a nodev timer
	 */
	if (phba->config[LPFC_CFG_NODEV_TMO]){
		/* if some timer's running, cancel it whether it's nodev timer
		 * or delay timer
		 */
		if (ndlp->nlp_tmofunc.expires) {
			del_timer_sync(&ndlp->nlp_tmofunc);
			ndlp->nlp_tmofunc.expires = 0;
			ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
			ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
		}

		/* now start a nodev timer */
		ndlp->nlp_flag |= NLP_NODEV_TMO;
		ndlp->nlp_rflag |= NLP_NPR_ACTIVE;

		if (phba->config[LPFC_CFG_NODEV_TMO] >
		   phba->config[LPFC_CFG_LINKDOWN_TMO]) {
			tmo = phba->config[LPFC_CFG_NODEV_TMO];
		} else {
			tmo = phba->config[LPFC_CFG_LINKDOWN_TMO];
		}

		ndlp->nlp_tmofunc.expires = jiffies + HZ * tmo;
		ndlp->nlp_tmofunc.data = (unsigned long)ndlp;
		add_timer(&ndlp->nlp_tmofunc);

		/* Start nodev timer */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_FCP,
				"%d:0706 Start nodev timer Data: x%x x%x x%x x%p\n",
				phba->brd_no,
				ndlp->nlp_DID, ndlp->nlp_flag, ndlp->nlp_state,
				ndlp);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_plogi_mapped_node(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	uint32_t *lp;
	struct serv_parm *sp;
	struct ls_rjt stat;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	sp = (struct serv_parm *) ((uint8_t *) lp + sizeof (uint32_t));

	if ((phba->hba_state <= LPFC_FLOGI) ||
	    ((lpfc_check_sparm(phba, ndlp, sp, CLASS3) == 0))) {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		if (phba->hba_state <= LPFC_FLOGI) {
			stat.un.b.lsRjtRsnCodeExp = LSRJT_LOGICAL_BSY;
		} else {
			stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		}
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	} else {
		/* PLOGI chkparm OK */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_ELS,
				"%d:0126 PLOGI chkparm OK Data: x%x x%x x%x x%x\n",
				phba->brd_no,
				ndlp->nlp_DID, ndlp->nlp_state, ndlp->nlp_flag,
				((struct lpfc_nodelist*) ndlp)->nlp_rpi);

		lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb, ndlp, 0, 0);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_prli_mapped_node(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	lpfc_els_rsp_prli_acc(phba, cmdiocb, ndlp);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_logo_mapped_node(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	/* software abort outstanding adisc before sending acc */
	if (ndlp->nlp_flag & NLP_ADISC_SND) {
		lpfc_driver_abort(phba, ndlp);
	}
	/* Only call LOGO ACC for first LOGO, this avoids sending unnecessary
	 * PLOGIs during LOGO storms from a device.
	 */
	if (ndlp->nlp_flag & NLP_LOGO_ACC) {
		ndlp->nlp_flag &= ~NLP_LOGO_ACC;
		lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
		ndlp->nlp_flag |= NLP_LOGO_ACC;
	} else {
		ndlp->nlp_flag |= NLP_LOGO_ACC;
		lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);
	}

	/* The driver has to wait until the ACC completes before we can continue
	 * processing the LOGO, the action will resume in
	 * lpfc_cmpl_els_logo_acc.  Since part of processing includes an
	 * unreg_login, the driver waits so the ACC does not get aborted.
	 */
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_padisc_mapped_node(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;
	struct serv_parm *sp;		/* used for PDISC */
	ADISC *ap;		/* used for ADISC */
	uint32_t *lp;
	uint32_t cmd;
	struct lpfc_name *pnn, *ppn;
	struct ls_rjt stat;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	lp = (uint32_t *) pcmd->virt;
	cmd = *lp++;

	if (cmd == ELS_CMD_ADISC) {
		ap = (ADISC *) lp;
		pnn = (struct lpfc_name *) & ap->nodeName;
		ppn = (struct lpfc_name *) & ap->portName;
	} else {
		sp = (struct serv_parm *) lp;
		pnn = (struct lpfc_name *) & sp->nodeName;
		ppn = (struct lpfc_name *) & sp->portName;
	}

	if (lpfc_check_adisc(phba, ndlp, pnn, ppn)) {
		if (cmd == ELS_CMD_ADISC) {
			lpfc_els_rsp_adisc_acc(phba, cmdiocb, ndlp);
		} else {
			lpfc_els_rsp_acc(phba, ELS_CMD_PLOGI, cmdiocb, ndlp,
					 0, 0);
		}
	} else {
		/* Reject this request because invalid parameters */
		stat.un.b.lsRjtRsvd0 = 0;
		stat.un.b.lsRjtRsnCode = LSRJT_UNABLE_TPC;
		stat.un.b.lsRjtRsnCodeExp = LSEXP_SPARM_OPTIONS;
		stat.un.b.vendorUnique = 0;
		lpfc_els_rsp_reject(phba, stat.un.lsRjtError, cmdiocb, ndlp);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_rcv_prlo_mapped_node(struct lpfc_hba * phba,
			  struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb;
	struct lpfc_dmabuf *pcmd;

	cmdiocb = (struct lpfc_iocbq *) arg;
	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	lpfc_els_rsp_acc(phba, ELS_CMD_ACC, cmdiocb, ndlp, 0, 0);

	/* save binding on binding list */
	if (ndlp->nlp_listp_bind) {
		lpfc_nlp_bind(phba, ndlp->nlp_listp_bind);

		/* flush the target */
		lpfc_sli_abort_iocb_tgt(phba,
				       &phba->sli.ring[phba->sli.fcp_ring],
				       ndlp->nlp_Target->scsi_id);

		ndlp->nlp_listp_bind = 0;
		ndlp->nlp_sid = 0;
		ndlp->nlp_Target = 0;
		ndlp->nlp_flag &= ~NLP_SEED_MASK;
	}

	ndlp->nlp_state = NLP_STE_PRLI_COMPL;
	lpfc_nlp_unmapped(phba, ndlp);

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_logo_mapped_node(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_target * targetp;

	/* save binding on binding list */
	if (ndlp->nlp_listp_bind) {
		lpfc_nlp_bind(phba, ndlp->nlp_listp_bind);
		if ((targetp = ndlp->nlp_Target)) {
		/* flush the target */
			lpfc_sli_abort_iocb_tgt(phba,
					&phba->sli.ring[phba->sli.fcp_ring],
					targetp->scsi_id);
		}

		ndlp->nlp_listp_bind = 0;
		ndlp->nlp_sid = 0;
		ndlp->nlp_Target = 0;
		ndlp->nlp_flag &= ~NLP_SEED_MASK;
	}

	/* dequeue, cancel timeout, unreg login */
	lpfc_freenode(phba, ndlp);

	/* software abort outstanding adisc */
	if (ndlp->nlp_flag & NLP_ADISC_SND) {
		lpfc_driver_abort(phba, ndlp);
	}
	lpfc_nlp_remove(phba, ndlp);
	return (NLP_STE_FREED_NODE);
}

static uint32_t
lpfc_cmpl_adisc_mapped_node(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{
	struct lpfc_iocbq *cmdiocb, *rspiocb;
	struct lpfc_dmabuf *pcmd, *prsp;
	uint32_t *lp;
	IOCB_t *irsp;
	struct lpfc_bindlist *blp;
	struct lpfc_target *targetp;
	ADISC *ap;

	cmdiocb = (struct lpfc_iocbq *) arg;
	rspiocb = cmdiocb->context_un.rsp_iocb;
	irsp = &rspiocb->iocb;

	/* First remove the ndlp from any list */
	if (ndlp->nlp_flag & NLP_LIST_MASK) {
		lpfc_findnode_did(phba, (NLP_SEARCH_ALL | NLP_SEARCH_DEQUE),
				  ndlp->nlp_DID);
	}
	if (irsp->ulpStatus) {
		/* If this is not a driver aborted ADISC, handle the recovery
		   here */
		if ((irsp->ulpStatus != IOSTAT_LOCAL_REJECT) ||
			!(irsp->un.ulpWord[4] & IOERR_DRVR_MASK)) {
			lpfc_freenode(phba, ndlp);
			ndlp->nlp_state = NLP_STE_UNUSED_NODE;
			lpfc_nlp_plogi(phba, ndlp);
		}
		return (ndlp->nlp_state);
	}

	pcmd = (struct lpfc_dmabuf *) cmdiocb->context2;
	
	prsp = (struct lpfc_dmabuf *) pcmd->list.next;
	lp = (uint32_t *) prsp->virt;

	pci_dma_sync_single_for_cpu(phba->pcidev, prsp->phys,
		LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	ap = (ADISC *) ((uint8_t *) lp + sizeof (uint32_t));

	/* Check out ADISC rsp */
	if ((lpfc_check_adisc(phba, ndlp, &ap->nodeName, &ap->portName) == 0)) {
		/* This is not a driver aborted ADISC, so handle the recovery
		   here */
		ndlp->nlp_state = NLP_STE_UNUSED_NODE;
		lpfc_freenode(phba, ndlp);
		lpfc_nlp_plogi(phba, ndlp);
		return (ndlp->nlp_state);
	}

	/* Can we reassign a SCSI Id to this NPort */
	if ((blp = lpfc_assign_scsid(phba, ndlp, 0))) {
		lpfc_nlp_mapped(phba, ndlp, blp);

		ndlp->nlp_state = NLP_STE_MAPPED_NODE;
		targetp = ndlp->nlp_Target;
		if (targetp) {
			targetp->targetFlags &= ~FC_NPR_ACTIVE;
			if (targetp->tmofunc.expires) {
				del_timer_sync(&targetp->tmofunc);
				targetp->tmofunc.expires = 0;
			}
		} else {
			/* new target to driver, allocate space to target <sid>
			   lun 0 */
			if (blp->nlp_Target == 0) {
				lpfc_find_target(phba, blp->nlp_sid, 0, 1);
				blp->nlp_Target =
				    phba->device_queue_hash[blp->nlp_sid];
			}
		}
		lpfc_set_failmask(phba, ndlp, LPFC_DEV_ALL_BITS,
				  LPFC_CLR_BITMASK);
	} else {
		lpfc_nlp_unmapped(phba, ndlp);
		ndlp->nlp_state = NLP_STE_PRLI_COMPL;
		ndlp->nlp_flag |= NLP_TGT_NO_SCSIID;
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_cmpl_reglogin_mapped_node(struct lpfc_hba * phba,
			       struct lpfc_nodelist * ndlp, void *arg,
			       uint32_t evt)
{
	LPFC_MBOXQ_t *pmb, *mbox;
	MAILBOX_t *mb;
	uint32_t ldata;
	uint16_t rpi;

	if (ndlp->nlp_flag & NLP_LOGO_SND) {
		return (ndlp->nlp_state);
	}

	pmb = (LPFC_MBOXQ_t *) arg;
	mb = &pmb->mb;
	ldata = mb->un.varWords[0];	/* rpi */
	rpi = (uint16_t) (le32_to_cpu(ldata) & 0xFFFF);

	if (ndlp->nlp_rpi != rpi) {
		/* first unreg node's rpi */
		if ((mbox = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC))) {
			lpfc_unreg_login(phba, ndlp->nlp_rpi, mbox);
			if (lpfc_sli_issue_mbox
			    (phba, mbox, (MBX_NOWAIT | MBX_STOP_IOCB))
			    == MBX_NOT_FINISHED) {
				mempool_free( mbox, phba->mbox_mem_pool);
			}
		}
		lpfc_findnode_remove_rpi(phba, ndlp->nlp_rpi);
		lpfc_no_rpi(phba, ndlp);
		ndlp->nlp_rpi = 0;

		/* now unreg rpi just got back from reg_login */
		lpfc_unreg_login(phba, rpi, mbox);
		if (lpfc_sli_issue_mbox(phba, mbox,
					(MBX_NOWAIT | MBX_STOP_IOCB))
		    == MBX_NOT_FINISHED) {
			mempool_free( mbox, phba->mbox_mem_pool);
		}

		/* save binding on binding list */
		if (ndlp->nlp_listp_bind) {
			lpfc_nlp_bind(phba, ndlp->nlp_listp_bind);

			/* flush the target */
			lpfc_sli_abort_iocb_tgt(phba,
					       &phba->sli.ring[
							phba->sli.fcp_ring],
					       ndlp->nlp_Target->scsi_id);

			ndlp->nlp_listp_bind = 0;
			ndlp->nlp_sid = 0;
			ndlp->nlp_Target = 0;
			ndlp->nlp_flag &= ~NLP_SEED_MASK;
		}


		/* software abort outstanding adisc */
		if (ndlp->nlp_flag & NLP_ADISC_SND) {
			lpfc_driver_abort(phba, ndlp);
		}

		/* send logout and put this node on plogi list */
		lpfc_issue_els_logo(phba, ndlp, 0);
		ndlp->nlp_state = NLP_STE_PLOGI_ISSUE;
		lpfc_nlp_plogi(phba, ndlp);
	}

	return (ndlp->nlp_state);
}

/*! lpfc_device_rm_mapped_node
  *
  * \pre
  * \post
  * \param   phba
  * \param   ndlp
  * \param   arg
  * \param   evt
  * \return  uint32_t
  *
  * \b Description:
  *    This routine is envoked when we a request to remove a nport.
  *    It could be called when linkdown or nodev timer expires.
  *    If nodev timer is still running, we just want to exit.
  *    If this node timed out, we want to abort outstanding ADISC,
  *    save its binding, unreg login, send logout, change state to
  *    UNUSED_NODE and place node on plogi list so it can be freed
  *    when LOGO completes.
  *
  */
static uint32_t
lpfc_device_rm_mapped_node(struct lpfc_hba * phba,
			   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	struct lpfc_target *targetp;
	unsigned long tmo;

	/* if nodev timer is running, then we just exit */
	if (!(ndlp->nlp_flag & NLP_NODEV_TMO)) {
		lpfc_set_failmask(phba, ndlp, LPFC_DEV_DISCONNECTED,
				  LPFC_SET_BITMASK);

		if (ndlp->nlp_flag & NLP_ADISC_SND) {
			/* software abort outstanding adisc */
			lpfc_driver_abort(phba, ndlp);
		}

		/* save binding info */
		if (ndlp->nlp_listp_bind) {
			targetp = ndlp->nlp_Target;
			lpfc_nlp_bind(phba, ndlp->nlp_listp_bind);
			if (targetp) {
				targetp->targetFlags &= ~FC_NPR_ACTIVE;
				if (targetp->tmofunc.expires) {
					del_timer_sync(&targetp->tmofunc);
					targetp->tmofunc.expires = 0;
				}
				/* flush the target */
				lpfc_sli_abort_iocb_tgt(phba,
						       &phba->sli.ring[phba->
								sli.fcp_ring],
						       targetp->scsi_id);

			}

			ndlp->nlp_listp_bind = 0;
			ndlp->nlp_sid = 0;
			ndlp->nlp_Target = 0;
			ndlp->nlp_flag &= ~NLP_SEED_MASK;
		}

		/* If discovery processing causes us to remove a device, it is
		 * important that nothing gets sent to the device (soft zoning
		 * issues).
		 */
		ndlp->nlp_state = NLP_STE_UNUSED_NODE;
		/* dequeue, cancel timeout, unreg login */
		lpfc_nlp_remove(phba, ndlp);
		return (NLP_STE_FREED_NODE);
	}
	targetp = ndlp->nlp_Target;
	if (targetp) {
		targetp->targetFlags |= FC_NPR_ACTIVE;
		if (targetp->tmofunc.expires) {
			del_timer_sync(&targetp->tmofunc);
			targetp->tmofunc.expires = 0;
		}
		if (phba->config[LPFC_CFG_NODEV_TMO] >
		   phba->config[LPFC_CFG_LINKDOWN_TMO]) {
			tmo = phba->config[LPFC_CFG_NODEV_TMO];
		} else {
			tmo = phba->config[LPFC_CFG_LINKDOWN_TMO];
		}

		targetp->tmofunc.expires = jiffies + HZ * tmo;
		targetp->tmofunc.data = (unsigned long)targetp;
		add_timer(&targetp->tmofunc);
	}
	lpfc_set_failmask(phba, ndlp, LPFC_DEV_DISAPPEARED, LPFC_SET_BITMASK);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_device_add_mapped_node(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{
	if (ndlp->nlp_tmofunc.expires) {
		del_timer_sync(&ndlp->nlp_tmofunc);
		ndlp->nlp_tmofunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	}
	if (ndlp->nlp_delayfunc.expires) {
		del_timer_sync(&ndlp->nlp_delayfunc);
		ndlp->nlp_delayfunc.expires = 0;
		ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		ndlp->nlp_rflag &= ~NLP_NPR_ACTIVE;
	}
	lpfc_nlp_adisc(phba, ndlp);
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_device_unk_mapped_node(struct lpfc_hba * phba,
			    struct lpfc_nodelist * ndlp, void *arg,
			    uint32_t evt)
{
	struct lpfc_target *targetp;
	unsigned long tmo;


	/* If we are in pt2pt mode, free node */
	if (phba->fc_flag & FC_PT2PT) {
		/* dequeue, cancel timeout, unreg login */
		lpfc_nlp_remove(phba, ndlp);
		return (NLP_STE_FREED_NODE);
	}

	targetp = ndlp->nlp_Target;
	if (targetp) {
		targetp->targetFlags |= FC_NPR_ACTIVE;
		if (targetp->tmofunc.expires) {
			del_timer_sync(&targetp->tmofunc);
			targetp->tmofunc.expires = 0;
		}

		if (phba->config[LPFC_CFG_NODEV_TMO] >
	 		phba->config[LPFC_CFG_LINKDOWN_TMO]) {
			tmo = phba->config[LPFC_CFG_NODEV_TMO];
		} else {
			tmo = phba->config[LPFC_CFG_LINKDOWN_TMO];
		}
		targetp->tmofunc.expires = jiffies + HZ * tmo;
		targetp->tmofunc.data = (unsigned long)targetp;
		add_timer(&targetp->tmofunc);
	}

	ndlp->nlp_rflag |= NLP_NPR_ACTIVE;
	/* linkdown timer should be running at this time.  Check to
	 * see if the driver has to start a nodev timer
	 */
	if (phba->config[LPFC_CFG_NODEV_TMO]){
		/* if some timer's running, cancel it whether it's nodev timer
		 * or delay timer
		 */
		if (ndlp->nlp_tmofunc.expires) {
			del_timer_sync(&ndlp->nlp_tmofunc);
			ndlp->nlp_tmofunc.expires = 0;
			ndlp->nlp_flag &= ~(NLP_NODEV_TMO | NLP_DELAY_TMO);
		}

		/* now start a nodev timer */
		ndlp->nlp_flag |= NLP_NODEV_TMO;


		if (phba->config[LPFC_CFG_NODEV_TMO] >
			phba->config[LPFC_CFG_LINKDOWN_TMO]) {
			tmo = phba->config[LPFC_CFG_NODEV_TMO];
		} else {
			tmo = phba->config[LPFC_CFG_LINKDOWN_TMO];
		}

		ndlp->nlp_tmofunc.expires = jiffies + HZ * tmo;
		ndlp->nlp_tmofunc.data = (unsigned long)ndlp;
		add_timer(&ndlp->nlp_tmofunc);

		/* Start nodev timer */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_FCP,
				"%d:0700 Start nodev timer Data: x%x x%x x%x x%p\n",
				phba->brd_no,
				ndlp->nlp_DID, ndlp->nlp_flag, ndlp->nlp_state,
				ndlp);
	}

	return (ndlp->nlp_state);
}

static uint32_t
lpfc_disc_nodev(struct lpfc_hba * phba,
		struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	/* This routine does nothing, just return the current state */
	return (ndlp->nlp_state);
}

static uint32_t
lpfc_disc_neverdev(struct lpfc_hba * phba,
		   struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	/* This routine does nothing, just return the current state */
	return (ndlp->nlp_state);
}

int
lpfc_check_sparm(struct lpfc_hba * phba,
		 struct lpfc_nodelist * ndlp, struct serv_parm * sp,
		 uint32_t class)
{
	volatile struct serv_parm *hsp = &phba->fc_sparam;
	/* First check for supported version */

	/* Next check for class validity */
	if (sp->cls1.classValid) {

		if (sp->cls1.rcvDataSizeMsb > hsp->cls1.rcvDataSizeMsb)
			sp->cls1.rcvDataSizeMsb = hsp->cls1.rcvDataSizeMsb;
		if (sp->cls1.rcvDataSizeLsb > hsp->cls1.rcvDataSizeLsb)
			sp->cls1.rcvDataSizeLsb = hsp->cls1.rcvDataSizeLsb;
	} else if (class == CLASS1) {
		return (0);
	}

	if (sp->cls2.classValid) {

		if (sp->cls2.rcvDataSizeMsb > hsp->cls2.rcvDataSizeMsb)
			sp->cls2.rcvDataSizeMsb = hsp->cls2.rcvDataSizeMsb;
		if (sp->cls2.rcvDataSizeLsb > hsp->cls2.rcvDataSizeLsb)
			sp->cls2.rcvDataSizeLsb = hsp->cls2.rcvDataSizeLsb;
	} else if (class == CLASS2) {
		return (0);
	}

	if (sp->cls3.classValid) {

		if (sp->cls3.rcvDataSizeMsb > hsp->cls3.rcvDataSizeMsb)
			sp->cls3.rcvDataSizeMsb = hsp->cls3.rcvDataSizeMsb;
		if (sp->cls3.rcvDataSizeLsb > hsp->cls3.rcvDataSizeLsb)
			sp->cls3.rcvDataSizeLsb = hsp->cls3.rcvDataSizeLsb;
	} else if (class == CLASS3) {
		return (0);
	}

	if (sp->cmn.bbRcvSizeMsb > hsp->cmn.bbRcvSizeMsb)
		sp->cmn.bbRcvSizeMsb = hsp->cmn.bbRcvSizeMsb;
	if (sp->cmn.bbRcvSizeLsb > hsp->cmn.bbRcvSizeLsb)
		sp->cmn.bbRcvSizeLsb = hsp->cmn.bbRcvSizeLsb;

	/* If check is good, copy wwpn wwnn into ndlp */
	memcpy(&ndlp->nlp_nodename, &sp->nodeName, sizeof (struct lpfc_name));
	memcpy(&ndlp->nlp_portname, &sp->portName, sizeof (struct lpfc_name));
	return (1);
}

static int
lpfc_binding_found(struct lpfc_bindlist * blp, struct lpfc_nodelist * ndlp)
{
	uint16_t bindtype = blp->nlp_bind_type;

	if ((bindtype & FCP_SEED_DID) && 
		   (ndlp->nlp_DID == be32_to_cpu(blp->nlp_DID))) { 
		return 1;
	} else if ((bindtype & FCP_SEED_WWPN) &&
		   (memcmp(&ndlp->nlp_portname, &blp->nlp_portname,
			   sizeof (struct lpfc_name)) == 0)) {
		return 1;
	} else if ((bindtype & FCP_SEED_WWNN) &&
		   (memcmp(&ndlp->nlp_nodename, &blp->nlp_nodename,
			    sizeof (struct lpfc_name)) == 0)) {
		return 1;
	}
	return 0;
}

static int
lpfc_binding_useid(struct lpfc_hba * phba, uint32_t sid)
{
	struct lpfc_bindlist *blp;

	list_for_each_entry(blp, &phba->fc_nlpbind_list, nlp_listp) {
		if (blp->nlp_sid == sid) {
			return (1);
		}
	}
	return (0);
}

static int
lpfc_mapping_useid(struct lpfc_hba * phba, uint32_t sid)
{
	struct lpfc_nodelist *mapnode;
	struct lpfc_bindlist *blp;

	list_for_each_entry(mapnode, &phba->fc_nlpmap_list, nlp_listp) {
		blp = mapnode->nlp_listp_bind;
		if (blp->nlp_sid == sid) {
			return (1);
		}
	}
	return (0);
}

static struct lpfc_bindlist *
lpfc_create_binding(struct lpfc_hba * phba,
		    struct lpfc_nodelist * ndlp, uint16_t index,
		    uint16_t bindtype)
{
	struct lpfc_bindlist *blp;

	if ((blp = mempool_alloc(phba->bind_mem_pool, GFP_ATOMIC))) {
		memset(blp, 0, sizeof (struct lpfc_bindlist));
		switch (bindtype) {
		case FCP_SEED_WWPN:
			blp->nlp_bind_type = FCP_SEED_WWPN;
			break;
		case FCP_SEED_WWNN:
			blp->nlp_bind_type = FCP_SEED_WWNN;
			break;
		case FCP_SEED_DID:
			blp->nlp_bind_type = FCP_SEED_DID;
			break;
		}
		blp->nlp_sid = index;
		blp->nlp_DID = ndlp->nlp_DID;
		blp->nlp_Target = phba->device_queue_hash[index];
		memcpy(&blp->nlp_nodename, &ndlp->nlp_nodename,
		       sizeof (struct lpfc_name));
		memcpy(&blp->nlp_portname, &ndlp->nlp_portname,
		       sizeof (struct lpfc_name));

		return (blp);
	}

	return (0);
}

static struct lpfc_bindlist *
lpfc_assign_scsid(struct lpfc_hba * phba, struct lpfc_nodelist * ndlp,
		  int prev_flag)
{
	struct lpfc_bindlist *blp, *next_blp;
	uint16_t index;


	/* check binding list */
	list_for_each_entry_safe(blp, next_blp, &phba->fc_nlpbind_list,
				nlp_listp) {
		if (lpfc_binding_found(blp, ndlp)) {
			ndlp->nlp_sid = blp->nlp_sid;
			ndlp->nlp_Target = blp->nlp_Target;
			ndlp->nlp_flag &= ~NLP_SEED_MASK;
			switch ((blp->nlp_bind_type & FCP_SEED_MASK)) {
			case FCP_SEED_WWPN:
				ndlp->nlp_flag |= NLP_SEED_WWPN;
				break;
			case FCP_SEED_WWNN:
				ndlp->nlp_flag |= NLP_SEED_WWNN;
				break;
			case FCP_SEED_DID:
				ndlp->nlp_flag |= NLP_SEED_DID;
				break;
			}
			if (blp->nlp_bind_type & FCP_SEED_AUTO) {
				ndlp->nlp_flag |= NLP_AUTOMAP;
			}

			/* take it off the binding list */
			phba->fc_bind_cnt--;
			list_del_init(&blp->nlp_listp);

			/* Reassign scsi id <sid> to NPort <nlp_DID> */
			lpfc_printf_log(phba,
					KERN_INFO,
					LOG_DISCOVERY | LOG_FCP,
					"%d:0213 Reassign scsi id x%x to NPort x%x Data: x%x x%x x%x x%x\n",
					phba->brd_no,
					blp->nlp_sid, ndlp->nlp_DID,
					blp->nlp_bind_type, ndlp->nlp_flag,
					ndlp->nlp_state, ndlp->nlp_rpi);

			return (blp);
		}
	}

	if (prev_flag)
		return(0);

	/* NOTE: if scan-down = 2 and we have private loop, then we use
	 * AlpaArray to determine sid.
	 */
	if ((phba->config[LPFC_CFG_BINDMETHOD] == 4) &&
	    ((phba->fc_flag & (FC_PUBLIC_LOOP | FC_FABRIC)) ||
	     (phba->fc_topology != TOPOLOGY_LOOP))) {
		/* Log message: ALPA based binding used on a non loop
		   topology */
		lpfc_printf_log(phba,
				KERN_WARNING,
				LOG_DISCOVERY,
				"%d:0245 ALPA based bind method used on an HBA which is in a nonloop topology Data: x%x\n",
				phba->brd_no,
				phba->fc_topology);
	}

	if ((phba->config[LPFC_CFG_BINDMETHOD] == 4) &&
	    !(phba->fc_flag & (FC_PUBLIC_LOOP | FC_FABRIC)) &&
	    (phba->fc_topology == TOPOLOGY_LOOP)) {
		for (index = 0; index < FC_MAXLOOP; index++) {
			if (ndlp->nlp_DID == (uint32_t) lpfcAlpaArray[index]) {
				if ((blp =
				     lpfc_create_binding(phba, ndlp, index,
							 FCP_SEED_DID))) {

					ndlp->nlp_sid = index;
					ndlp->nlp_Target =
					    phba->device_queue_hash[index];
					ndlp->nlp_flag &= ~NLP_SEED_MASK;
					ndlp->nlp_flag |= NLP_SEED_DID;
					ndlp->nlp_flag |= NLP_SEED_ALPA;

					/* Assign scandown scsi id <sid> to
					   NPort <nlp_DID> */
					lpfc_printf_log(phba,
						KERN_INFO,
						LOG_DISCOVERY | LOG_FCP,
						"%d:0216 Assign scandown scsi id x%x to NPort x%x Data: x%x x%x x%x x%x\n",
						phba->brd_no,
						blp->nlp_sid, ndlp->nlp_DID,
						blp->nlp_bind_type,
						ndlp->nlp_flag, ndlp->nlp_state,
						ndlp->nlp_rpi);

					return (blp);
				}
				goto errid;
			}
		}
	}

	if (phba->config[LPFC_CFG_AUTOMAP]) {
		while (1) {
			if ((lpfc_binding_useid(phba, phba->sid_cnt))
			     || (lpfc_mapping_useid (phba, phba->sid_cnt))) {

				phba->sid_cnt++;
			} else {
				if ((blp =
				     lpfc_create_binding(phba, ndlp,
							 phba->sid_cnt,
							 phba->fcp_mapping))) {
					ndlp->nlp_sid = blp->nlp_sid;
					ndlp->nlp_Target = blp->nlp_Target;
					ndlp->nlp_flag &= ~NLP_SEED_MASK;
					switch ((blp->
						 nlp_bind_type & FCP_SEED_MASK))
					{
					case FCP_SEED_WWPN:
						ndlp->nlp_flag |= NLP_SEED_WWPN;
						break;
					case FCP_SEED_WWNN:
						ndlp->nlp_flag |= NLP_SEED_WWNN;
						break;
					case FCP_SEED_DID:
						ndlp->nlp_flag |= NLP_SEED_DID;
						break;
					}
					blp->nlp_bind_type |= FCP_SEED_AUTO;
					ndlp->nlp_flag |= NLP_AUTOMAP;

					phba->sid_cnt++;

					/* Assign scsi id <sid> to NPort
					   <nlp_DID> */
					lpfc_printf_log(phba,
						KERN_INFO,
						LOG_DISCOVERY  | LOG_FCP,
						"%d:0229 Assign scsi ID x%x to NPort x%x Data: x%x x%x x%x x%x\n",
						phba->brd_no,
						blp->nlp_sid, ndlp->nlp_DID,
						blp->nlp_bind_type,
						ndlp->nlp_flag, ndlp->nlp_state,
						ndlp->nlp_rpi);

					return (blp);
				}
				goto errid;
			}
		}
	}
	/* if automap on */
errid:
	/* Cannot assign scsi id on NPort <nlp_DID> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_DISCOVERY | LOG_FCP,
			"%d:0230 Cannot assign scsi ID on NPort x%x Data: x%x x%x x%x\n",
			phba->brd_no,
			ndlp->nlp_DID, ndlp->nlp_flag, ndlp->nlp_state,
			ndlp->nlp_rpi);

	return (0);
}


/* This next section defines the NPort Discovery State Machine */

/* There are 4 different double linked lists nodelist entries can reside on.
 * The plogi list and adisc list are used when Link Up discovery or RSCN
 * processing is needed. Each list holds the nodes that we will send PLOGI
 * or ADISC on. These lists will keep track of what nodes will be effected
 * by an RSCN, or a Link Up (Typically, all nodes are effected on Link Up).
 * The unmapped_list will contain all nodes that we have successfully logged
 * into at the Fibre Channel level. The mapped_list will contain all nodes
 * that are mapped FCP targets.
 */
/*
 * The bind list is a list of undiscovered (potentially non-existent) nodes
 * that we have saved binding information on. This information is used when
 * nodes transition from the unmapped to the mapped list.
 */
/* For UNUSED_NODE state, the node has just been allocated .
 * For PLOGI_ISSUE and REG_LOGIN_ISSUE, the node is on
 * the PLOGI list. For REG_LOGIN_COMPL, the node is taken off the PLOGI list
 * and put on the unmapped list. For ADISC processing, the node is taken off
 * the ADISC list and placed on either the mapped or unmapped list (depending
 * on its previous state). Once on the unmapped list, a PRLI is issued and the
 * state changed to PRLI_ISSUE. When the PRLI completion occurs, the state is
 * changed to PRLI_COMPL. If the completion indicates a mapped
 * node, the node is taken off the unmapped list. The binding list is checked
 * for a valid binding, or a binding is automatically assigned. If binding
 * assignment is unsuccessful, the node is left on the unmapped list. If
 * binding assignment is successful, the associated binding list entry (if
 * any) is removed, and the node is placed on the mapped list.
 */
/*
 * For a Link Down, all nodes on the ADISC, PLOGI, unmapped or mapped
 * lists will receive a DEVICE_UNK event. If the linkdown or nodev timers
 * expire, all effected nodes will receive a DEVICE_RM event.
 */
/*
 * For a Link Up or RSCN, all nodes will move from the mapped / unmapped lists
 * to either the ADISC or PLOGI list.  After a Nameserver query or ALPA loopmap
 * check, additional nodes may be added (DEVICE_ADD) or removed (DEVICE_RM) to /
 * from the PLOGI or ADISC lists. Once the PLOGI and ADISC lists are populated,
 * we will first process the ADISC list.  32 entries are processed initially and
 * ADISC is initited for each one.  Completions / Events for each node are
 * funnelled thru the state machine.  As each node finishes ADISC processing, it
 * starts ADISC for any nodes waiting for ADISC processing. If no nodes are
 * waiting, and the ADISC list count is identically 0, then we are done. For
 * Link Up discovery, since all nodes on the PLOGI list are UNREG_LOGIN'ed, we
 * can issue a CLEAR_LA and reenable Link Events. Next we will process the PLOGI
 * list.  32 entries are processed initially and PLOGI is initited for each one.
 * Completions / Events for each node are funnelled thru the state machine.  As
 * each node finishes PLOGI processing, it starts PLOGI for any nodes waiting
 * for PLOGI processing. If no nodes are waiting, and the PLOGI list count is
 * indentically 0, then we are done. We have now completed discovery / RSCN
 * handling. Upon completion, ALL nodes should be on either the mapped or
 * unmapped lists.
 */

static void *lpfc_disc_action[NLP_STE_MAX_STATE * NLP_EVT_MAX_EVENT] = {
	/* Action routine                          Event       Current State  */
	(void *)lpfc_rcv_plogi_unused_node,	/* RCV_PLOGI   UNUSED_NODE    */
	(void *)lpfc_rcv_els_unused_node,	/* RCV_PRLI        */
	(void *)lpfc_rcv_logo_unused_node,	/* RCV_LOGO        */
	(void *)lpfc_rcv_els_unused_node,	/* RCV_ADISC       */
	(void *)lpfc_rcv_els_unused_node,	/* RCV_PDISC       */
	(void *)lpfc_rcv_els_unused_node,	/* RCV_PRLO        */
	(void *)lpfc_cmpl_els_unused_node,	/* CMPL_PLOGI      */
	(void *)lpfc_cmpl_els_unused_node,	/* CMPL_PRLI       */
	(void *)lpfc_cmpl_els_unused_node,	/* CMPL_LOGO       */
	(void *)lpfc_cmpl_els_unused_node,	/* CMPL_ADISC      */
	(void *)lpfc_cmpl_reglogin_unused_node,	/* CMPL_REG_LOGIN  */
	(void *)lpfc_device_rm_unused_node,	/* DEVICE_RM       */
	(void *)lpfc_device_add_unused_node,	/* DEVICE_ADD      */
	(void *)lpfc_device_unk_unused_node,	/* DEVICE_UNK      */
	(void *)lpfc_rcv_plogi_plogi_issue,	/* RCV_PLOGI   PLOGI_ISSUE    */
	(void *)lpfc_rcv_prli_plogi_issue,	/* RCV_PRLI        */
	(void *)lpfc_rcv_logo_plogi_issue,	/* RCV_LOGO        */
	(void *)lpfc_rcv_els_plogi_issue,	/* RCV_ADISC       */
	(void *)lpfc_rcv_els_plogi_issue,	/* RCV_PDISC       */
	(void *)lpfc_rcv_els_plogi_issue,	/* RCV_PRLO        */
	(void *)lpfc_cmpl_plogi_plogi_issue,	/* CMPL_PLOGI      */
	(void *)lpfc_cmpl_prli_plogi_issue,	/* CMPL_PRLI       */
	(void *)lpfc_cmpl_logo_plogi_issue,	/* CMPL_LOGO       */
	(void *)lpfc_cmpl_adisc_plogi_issue,	/* CMPL_ADISC      */
	(void *)lpfc_cmpl_reglogin_plogi_issue,	/* CMPL_REG_LOGIN  */
	(void *)lpfc_device_rm_plogi_issue,	/* DEVICE_RM       */
	(void *)lpfc_disc_nodev,	/* DEVICE_ADD      */
	(void *)lpfc_device_unk_plogi_issue,	/* DEVICE_UNK      */
	(void *)lpfc_rcv_plogi_reglogin_issue,	/* RCV_PLOGI  REG_LOGIN_ISSUE */
	(void *)lpfc_rcv_prli_reglogin_issue,	/* RCV_PLOGI       */
	(void *)lpfc_rcv_logo_reglogin_issue,	/* RCV_LOGO        */
	(void *)lpfc_rcv_padisc_reglogin_issue,	/* RCV_ADISC       */
	(void *)lpfc_rcv_padisc_reglogin_issue,	/* RCV_PDISC       */
	(void *)lpfc_rcv_prlo_reglogin_issue,	/* RCV_PRLO        */
	(void *)lpfc_disc_neverdev,	/* CMPL_PLOGI      */
	(void *)lpfc_disc_neverdev,	/* CMPL_PRLI       */
	(void *)lpfc_cmpl_logo_reglogin_issue,	/* CMPL_LOGO       */
	(void *)lpfc_cmpl_adisc_reglogin_issue,	/* CMPL_ADISC      */
	(void *)lpfc_cmpl_reglogin_reglogin_issue,	/* CMPL_REG_LOGIN  */
	(void *)lpfc_device_rm_reglogin_issue,	/* DEVICE_RM       */
	(void *)lpfc_disc_nodev,	/* DEVICE_ADD      */
	(void *)lpfc_device_unk_reglogin_issue,	/* DEVICE_UNK      */
	(void *)lpfc_rcv_plogi_prli_issue,	/* RCV_PLOGI   PRLI_ISSUE     */
	(void *)lpfc_rcv_prli_prli_issue,	/* RCV_PRLI        */
	(void *)lpfc_rcv_logo_prli_issue,	/* RCV_LOGO        */
	(void *)lpfc_rcv_padisc_prli_issue,	/* RCV_ADISC       */
	(void *)lpfc_rcv_padisc_prli_issue,	/* RCV_PDISC       */
	(void *)lpfc_rcv_prlo_prli_issue,	/* RCV_PRLO        */
	(void *)lpfc_disc_neverdev,	/* CMPL_PLOGI      */
	(void *)lpfc_cmpl_prli_prli_issue,	/* CMPL_PRLI       */
	(void *)lpfc_cmpl_logo_prli_issue,	/* CMPL_LOGO       */
	(void *)lpfc_cmpl_adisc_prli_issue,	/* CMPL_ADISC      */
	(void *)lpfc_cmpl_reglogin_prli_issue,	/* CMPL_REG_LOGIN  */
	(void *)lpfc_device_rm_prli_issue,	/* DEVICE_RM       */
	(void *)lpfc_device_add_prli_issue,	/* DEVICE_ADD      */
	(void *)lpfc_device_unk_prli_issue,	/* DEVICE_UNK      */
	(void *)lpfc_rcv_plogi_prli_compl,	/* RCV_PLOGI   PRLI_COMPL     */
	(void *)lpfc_rcv_prli_prli_compl,	/* RCV_PRLI        */
	(void *)lpfc_rcv_logo_prli_compl,	/* RCV_LOGO        */
	(void *)lpfc_rcv_padisc_prli_compl,	/* RCV_ADISC       */
	(void *)lpfc_rcv_padisc_prli_compl,	/* RCV_PDISC       */
	(void *)lpfc_rcv_prlo_prli_compl,	/* RCV_PRLO        */
	(void *)lpfc_disc_neverdev,	/* CMPL_PLOGI      */
	(void *)lpfc_disc_neverdev,	/* CMPL_PRLI       */
	(void *)lpfc_cmpl_logo_prli_compl,	/* CMPL_LOGO       */
	(void *)lpfc_cmpl_adisc_prli_compl,	/* CMPL_ADISC      */
	(void *)lpfc_cmpl_reglogin_prli_compl,	/* CMPL_REG_LOGIN  */
	(void *)lpfc_device_rm_prli_compl,	/* DEVICE_RM       */
	(void *)lpfc_device_add_prli_compl,	/* DEVICE_ADD      */
	(void *)lpfc_device_unk_prli_compl,	/* DEVICE_UNK      */
	(void *)lpfc_rcv_plogi_mapped_node,	/* RCV_PLOGI   MAPPED_NODE    */
	(void *)lpfc_rcv_prli_mapped_node,	/* RCV_PRLI        */
	(void *)lpfc_rcv_logo_mapped_node,	/* RCV_LOGO        */
	(void *)lpfc_rcv_padisc_mapped_node,	/* RCV_ADISC       */
	(void *)lpfc_rcv_padisc_mapped_node,	/* RCV_PDISC       */
	(void *)lpfc_rcv_prlo_mapped_node,	/* RCV_PRLO        */
	(void *)lpfc_disc_neverdev,	/* CMPL_PLOGI      */
	(void *)lpfc_disc_neverdev,	/* CMPL_PRLI       */
	(void *)lpfc_cmpl_logo_mapped_node,	/* CMPL_LOGO       */
	(void *)lpfc_cmpl_adisc_mapped_node,	/* CMPL_ADISC      */
	(void *)lpfc_cmpl_reglogin_mapped_node,	/* CMPL_REG_LOGIN  */
	(void *)lpfc_device_rm_mapped_node,	/* DEVICE_RM       */
	(void *)lpfc_device_add_mapped_node,	/* DEVICE_ADD      */
	(void *)lpfc_device_unk_mapped_node,	/* DEVICE_UNK      */
};

int
lpfc_disc_state_machine(struct lpfc_hba * phba,
			struct lpfc_nodelist * ndlp, void *arg, uint32_t evt)
{
	uint32_t cur_state, rc;
	uint32_t(*func) (struct lpfc_hba *, struct lpfc_nodelist *, void *,
			 uint32_t);

	ndlp->nlp_disc_refcnt++;
	cur_state = ndlp->nlp_state;

	/* DSM in event <evt> on NPort <nlp_DID> in state <cur_state> */
	lpfc_printf_log(phba,
		       KERN_INFO,
		       LOG_DISCOVERY,
		       "%d:0211 DSM in event x%x on NPort x%x in state %d Data: x%x\n",
		       phba->brd_no,
		       evt, ndlp->nlp_DID, cur_state, ndlp->nlp_flag);

	func = (uint32_t(*)(struct lpfc_hba *, struct lpfc_nodelist *, void *,
			    uint32_t))
	    lpfc_disc_action[(cur_state * NLP_EVT_MAX_EVENT) + evt];
	rc = (func) (phba, ndlp, arg, evt);

	/* DSM out state <rc> on NPort <nlp_DID> */
	lpfc_printf_log(phba,
		       KERN_INFO,
		       LOG_DISCOVERY,
		       "%d:0212 DSM out state %d on NPort x%x Data: x%x\n",
		       phba->brd_no,
		       rc, ndlp->nlp_DID, ndlp->nlp_flag);

	ndlp->nlp_disc_refcnt--;

	/* Check to see if ndlp removal is deferred */
	if ((ndlp->nlp_disc_refcnt == 0)
	    && (ndlp->nlp_rflag & NLP_DELAY_REMOVE)) {

		ndlp->nlp_rflag &= ~NLP_DELAY_REMOVE;
		lpfc_nlp_remove(phba, ndlp);
		return (NLP_STE_FREED_NODE);
	}
	if (rc == NLP_STE_FREED_NODE)
		return (NLP_STE_FREED_NODE);
	ndlp->nlp_state = rc;
	return (rc);
}
