 
#ifndef SONIXUSB_H
#define SONIXUSB_INIT_H

#define COMP 0x07
#define MCK_SIZE 0x33
#define MCK_INIT 0x63
#define SYS_CLK 0x04

static __u8
init640x480[]={
   SYS_CLK,0x03,0x00,0x00,0x00,0x00,0x00,0x20,0x11,0x00,0x00,0x00,0x00,0x00,
   0x00,0x01,0x00,0x68,0x0c,0x0a,0x28,0x1e,0x60,COMP,MCK_INIT,
   0x18,0x10,0x04,0x03,0x11,0x0c };
   
static __u8 i2c1[]= { 0x30,0x11,0x00,0xc0,0x08,0x00,0x00,0x10 };
static __u8 i2c2[]= { 0x30,0x11,0x02,0x20,0x70,0x00,0x00,0x10 };
static __u8 WriteControl[]= { 0x28,0x1e};//{ 0x50,0x3c}{ 0x28,0x1e}{ 0x16,0x12};
static __u8 SetActiveWindows[]= { 0x04,0x03,0x11,0x0c };

static int sonix_i2cwrite (struct usb_device *dev,__u8 *buffer,__u16 length)
{
	int retry = 60;
	__u8 ByteReceive=0x00;
	/* is i2c ready */
	if (length > 8 || !buffer) return -1;
	sonixRegWrite(dev,0x08,0x08,0x0000,buffer,length);
	while (retry--) {
	wait_ms (10);
	sonixRegRead(dev,0x00,0x08,0x0000,&ByteReceive,1);
	if(ByteReceive == 4) return 0;
	}
	return -1;
}

static __u16 sonix_getbrightness(struct usb_spca50x *spca50x)
{	/*FIXME hardcoded as we need to read register of the tasc */
	spca50x->brightness = 0x80 << 8;
	return (0x80 << 8);
}

static __u16 sonix_setbrightness(struct usb_spca50x *spca50x)
{
	__u8 value;
	__u8 i2c[]= { 0x30,0x11,0x02,0x20,0x70,0x00,0x00,0x10 };
	value = (0xFF - (spca50x->brightness >> 8));
	i2c[4] = value & 0xFF;
	PDEBUG(0,"brightness %d :%d ",value,i2c[4]);
	if(sonix_i2cwrite(spca50x->dev,i2c,8) < 0) PDEBUG(0,"i2c error brightness");
	return 0;
}

static int sonix_init(	struct usb_spca50x *spca50x )
{
	__u8 ByteReceive=0x00;
	spca50x->mode = 2;
	spca_set_interface(spca50x->dev, 0, 8);
	spca_set_interface(spca50x->dev, 0, 0);
	spca_clear_feature(spca50x->dev,3);
	spca_set_interface(spca50x->dev, 0, 8);
	sonixRegRead(spca50x->dev,0x00,0x00,0x0000,&ByteReceive,1);
	if(ByteReceive != 0x10) return -ENODEV;

	return 0;
}

static void sonix_start(struct usb_spca50x *spca50x )
{
	__u8 ByteSend = 0;
	__u8 compress = 0;
	 __u8 CompressCtrl[]= { 0,MCK_SIZE };
	compress = spca50x->mode << 4 | COMP;
	CompressCtrl[0] = compress;
	/* reg 0x01 bit 2 video transfert on */
	ByteSend=SYS_CLK;
	sonixRegWrite(spca50x->dev,0x08,0x01,0x0000,&ByteSend,1);
	/* reg 0x17 SensorClk enable inv Clk 0x60 */
	ByteSend=0x60;
	sonixRegWrite(spca50x->dev,0x08,0x17,0x0000,&ByteSend,1);

	sonixRegWrite(spca50x->dev,0x08,0x01,0x0000,init640x480,0x1f);

	if(sonix_i2cwrite(spca50x->dev,i2c1,8) < 0) PDEBUG(0,"i2c error i2c1");
	if(sonix_i2cwrite(spca50x->dev,i2c2,8) < 0) PDEBUG(0,"i2c error i2c2");

	/* H_size V_size  0x28,0x1e maybe 320x240 */	
	sonixRegWrite(spca50x->dev,0x08,0x15,0x0000,WriteControl,0x02);
	
	/* compression register */
	ByteSend = compress;
	sonixRegWrite(spca50x->dev,0x08,0x18,0x0000,&ByteSend,1);
	// H_start
	ByteSend=0x68;
	sonixRegWrite(spca50x->dev,0x08,0x12,0x0000,&ByteSend,1);
	// V_START
	ByteSend=0x0c;
	sonixRegWrite(spca50x->dev,0x08,0x13,0x0000,&ByteSend,1);

	ByteSend=0x60;
	sonixRegWrite(spca50x->dev,0x08,0x17,0x0000,&ByteSend,1);
	/*MCKSIZE ->3*/
	ByteSend=MCK_SIZE;
	sonixRegWrite(spca50x->dev,0x08,0x19,0x0000,&ByteSend,1);
	/* AE_STRX AE_STRY AE_ENDX AE_ENDY */
	sonixRegWrite(spca50x->dev,0x08,0x1c,0x0000,SetActiveWindows,4);

	/* Enable video transfert */
	ByteSend=SYS_CLK;
	sonixRegWrite(spca50x->dev,0x08,0x01,0x0000,&ByteSend,1);

	/* Compression */
	sonixRegWrite(spca50x->dev,0x08,0x18,0x0000,CompressCtrl,2);


}

static void sonix_stop(struct usb_spca50x *spca50x )
{  	__u8 ByteSend =0;
	
	ByteSend=0x00;
	sonixRegWrite(spca50x->dev,0x08,0x01,0x0000,&ByteSend,1);
}
#endif /* SONIXUSB_H */
