
#ifndef SPCAUSB_H
#define SPCAUSB_INIT_H

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 6)
#define wait_ms(a) msleep((a))
#endif

static void spca5xxRegRead(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) ;	
static void spca5xxRegWrite(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) ;	
static void sonixRegRead(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) ;	
static void sonixRegWrite(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) ;	
static int spca_set_interface(struct usb_device *dev,
				 int interface,
				 int alternate);

static int spca_clear_feature(struct usb_device *dev,
				 int endpoint)
{
	int inpipe;
	inpipe = usb_rcvintpipe(dev,endpoint);
	usb_clear_halt(dev, inpipe);
	return 0;
}
/***************************** Implementation ****************************/
static void spca5xxRegRead(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) 
{	
	int rc;
	__u8 *kbuffer = NULL;
	__u16 RegType;
	if(length >0) {
		kbuffer = (__u8 *) kmalloc (length,GFP_KERNEL);
		memcpy (kbuffer, buffer, length);
	}
	RegType = USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
	rc = usb_control_msg(dev,
		usb_rcvctrlpipe(dev,0),
		reg,
		RegType,
		(__u16)value, (__u16)index, kbuffer, length, HZ);
	if(length >0) {	
		memcpy (buffer, kbuffer, length);
		kfree (kbuffer);
	}
	if (buffer) { 
		PDEBUG(5, "reg read: 0x%02X, 0x%02X, 0x%02X, 0x%02X: 0x%04X",
		RegType ,reg ,value ,index ,(int) *buffer);
	} else { 
		PDEBUG(5, "reg read: 0x%02X, 0x%02X,0x%02X, 0x%02X ",RegType,reg,value,index );
	}
	return ;
}


static void spca5xxRegWrite(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length) 
{	int rc;
	__u16 RegType;
	__u8 *kbuffer = NULL;
	if(length >0) {
		kbuffer = (__u8 *) kmalloc (length,GFP_KERNEL);
		memcpy (kbuffer, buffer, length);
	}
	RegType = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
	rc = usb_control_msg(dev,
		usb_sndctrlpipe(dev,0),
		reg,
		RegType,
		(__u16)value, (__u16)index, kbuffer, length, HZ);
	if(length >0) {	
		memcpy (buffer, kbuffer, length);
		kfree (kbuffer);
	}
	if (buffer) { 
		PDEBUG(5, "reg write: 0x%02X, 0x%02X, 0x%02X, 0x%02X: 0x%04X",
		RegType ,reg ,value ,index ,(int) *buffer);
	} else { 
		PDEBUG(5, "reg write: 0x%02X, 0x%02X,0x%02X, 0x%02X ",RegType,reg,value,index );
	}
	return ;
}
static void sonixRegRead(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length)
{
	int rc;
	__u8 *kbuffer = NULL;
	__u16 RegType;
	if(length >0) {
		kbuffer = (__u8 *) kmalloc (length,GFP_KERNEL);
		memcpy (kbuffer, buffer, length);
	}
	RegType = USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_INTERFACE;
	rc = usb_control_msg(dev,
		usb_rcvctrlpipe(dev,0),
		reg,
		RegType,
		(__u16)value, (__u16)index, kbuffer, length, HZ);
	if(length >0) {
		memcpy (buffer, kbuffer, length);
		kfree (kbuffer);
	}
	if (buffer) {
		PDEBUG(5, "reg read: 0x%02X, 0x%02X, 0x%02X, 0x%02X: 0x%04X",
		RegType ,reg ,value ,index ,(int) *buffer);
	} else {
		PDEBUG(5, "reg read: 0x%02X, 0x%02X,0x%02X, 0x%02X ",RegType,reg,value,index );
	}
	return ;
}

static void sonixRegWrite(struct usb_device *dev,
				__u16 reg,
				__u16 value,
				__u16 index,
				__u8 *buffer,
				__u16 length)
{	int rc;
	__u16 RegType;
	__u8 *kbuffer = NULL;
	if(length >0) {
		kbuffer = (__u8 *) kmalloc (length,GFP_KERNEL);
		memcpy (kbuffer, buffer, length);
	}
	RegType = USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_INTERFACE;
	rc = usb_control_msg(dev,
		usb_sndctrlpipe(dev,0),
		reg,
		RegType,
		(__u16)value, (__u16)index, kbuffer, length, HZ);
	if(length >0) {
		memcpy (buffer, kbuffer, length);
		kfree (kbuffer);
	}
	if (buffer) {
		PDEBUG(5, "reg write: 0x%02X, 0x%02X, 0x%02X, 0x%02X: 0x%04X",
		RegType ,reg ,value ,index ,(int) *buffer);
	} else {
		PDEBUG(5, "reg write: 0x%02X, 0x%02X,0x%02X, 0x%02X ",RegType,reg,value,index );
	}
	return ;
}
static int spca_set_interface(struct usb_device *dev, int interface, int alternate)
{
	struct usb_interface *iface;
	int ret;

	iface = usb_ifnum_to_if(dev, interface);
	if (!iface) {
		warn("selecting invalid interface %d", interface);
		return -EINVAL;
	}

	/* 9.4.10 says devices don't need this, if the interface
	   only has one alternate setting */
	if (iface->num_altsetting == 1) {
		dbg("ignoring set_interface for dev %d, iface %d, alt %d",
			dev->devnum, interface, alternate);
		return 0;
	}

	if ((ret = usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
	    USB_REQ_SET_INTERFACE, USB_RECIP_INTERFACE, alternate,
	    interface, NULL, 0, HZ * 5)) < 0)
		return ret;

	
	return 0;

}	 
#endif /* SPCAUSB_H */
