

/***************************************************************************************************
*** MIDI interface
***************************************************************************************************/

#if 0
#ifdef EMIDI

static int snd_Echo_midi_input_open(snd_rawmidi_substream_t *substream)
{
	struct EchoChip *chip = substream->rmidi->private_data;
	unsigned long flags;

	spin_lock_irqsave(&chip->Midilock, flags);
	chip->MidiIn = substream;
	spin_unlock_irqrestore(&chip->Midilock, flags);
	printk("rawmidi_iopen\n");
	return (0);
}



static void snd_Echo_midi_input_trigger(snd_rawmidi_substream_t *substream, int up)
{
	struct EchoChip *chip = substream->rmidi->private_data;
	unsigned long flags;
	int st;

	spin_lock_irqsave(&chip->Midilock, flags);
	if (up)
		st = OpenMidiInput(chip->eg);
	else
		st = CloseMidiInput(chip->eg);
	spin_unlock_irqrestore(&chip->Midilock, flags);
	printk("rawmidi_itrigger(%d)=%d\n", up, st);
}



/* <-- snd_Echo_interrupt() */
static int snd_Echo_midi_input_read(struct EchoChip *chip)
{
	unsigned long flags;
	int n;

	spin_lock_irqsave(&chip->Midilock, flags);	/*@@not sure if this is needed */
	if (chip->MidiIn) {
		n = 0;
		while (ReadMidiByte(chip->eg, &chip->Midibuf[n]) == ECHOSTATUS_OK)
			n++;
		/*ECHOSTATUS_NO_DATA */
		if (n)
			snd_rawmidi_receive(chip->MidiIn, chip->Midibuf, n);
		printk("rawmidi_read=%d\n", n);
	}
	spin_unlock_irqrestore(&chip->Midilock, flags);
	return (0);
}



static int snd_Echo_midi_input_close(snd_rawmidi_substream_t *substream)
{
	struct EchoChip *chip = substream->rmidi->private_data;
	unsigned long flags;

	spin_lock_irqsave(&chip->Midilock, flags);
	chip->MidiIn = 0;
	spin_unlock_irqrestore(&chip->Midilock, flags);
	printk("rawmidi_iclose\n");
	return (0);
}



static int snd_Echo_midi_output_open(snd_rawmidi_substream_t *substream)
{
	struct EchoChip *chip = substream->rmidi->private_data;

	chip->MidiOut = substream;
	printk("rawmidi_oopen\n");
	return (0);
}



static void snd_Echo_midi_output_write(unsigned long data)
{
	struct EchoChip *chip = (struct EchoChip *) data;
	unsigned long flags;
	int nbytes, st;
	unsigned char buf[16];
	DWORD sent;

	/* No interrupts are involved: I have to check at regular intervals
	  if the card's buffer has room for new data. */
	sent = nbytes = 0;
	spin_lock_irqsave(&chip->Midilock, flags);
	if (chip->MidiOut) {
		if (!snd_rawmidi_transmit_empty(chip->MidiOut)) {
			nbytes = snd_rawmidi_transmit_peek(chip->MidiOut, buf, 16);
			if (nbytes > 0) {
				st = WriteMidi(chip->eg, nbytes, buf, &sent);
				snd_rawmidi_transmit_ack(chip->MidiOut, sent);
			}
		}
	}
	/* The kernel removes the timer when it expires. */
	if (chip->tinuse) {
		chip->timer.expires = ((HZ + 199) / 200) + jiffies;
		add_timer(&chip->timer);
	}

	spin_unlock_irqrestore(&chip->Midilock, flags);
	printk("rawmidi_owrite(%d)=%d\n", nbytes, (int) sent);
}



static void snd_Echo_midi_output_trigger(snd_rawmidi_substream_t *substream, int up)
{
	struct EchoChip *chip = substream->rmidi->private_data;
	unsigned long flags;

	spin_lock_irqsave(&chip->Midilock, flags);
	if (up) {
		if (!chip->tinuse) {
			init_timer(&chip->timer);
			chip->timer.function = snd_Echo_midi_output_write;
			chip->timer.data = (unsigned long) chip;
			chip->timer.expires = ((HZ + 199) / 200) + jiffies;	/* 1/200s == 5ms */
			add_timer(&chip->timer);
			chip->tinuse = 1;
		}
	} else {
		del_timer(&chip->timer);
		chip->tinuse = 0;
	}
	spin_unlock_irqrestore(&chip->Midilock, flags);

	if (up)
		snd_Echo_midi_output_write((unsigned long) chip);
	printk("rawmidi_otrigger\n");
}



static int snd_Echo_midi_output_close(snd_rawmidi_substream_t *substream)
{
	struct EchoChip *chip = substream->rmidi->private_data;

	chip->MidiOut = 0;
	printk("rawmidi_oclose\n");
	return (0);
}



snd_rawmidi_ops_t snd_Echo_midi_input = {
	.open = snd_Echo_midi_input_open,
	.close = snd_Echo_midi_input_close,
	.trigger = snd_Echo_midi_input_trigger,
};

snd_rawmidi_ops_t snd_Echo_midi_output = {
	.open = snd_Echo_midi_output_open,
	.close = snd_Echo_midi_output_close,
	.trigger = snd_Echo_midi_output_trigger,
};



static int __devinit snd_Echo_new_midi(snd_card_t *card, struct EchoChip *chip)
{
	char buf[32];

	spin_lock_init(&chip->Midilock);

	sprintf(buf, "%s MIDI 0", card->shortname);
	if (snd_rawmidi_new(card, buf, 0, 1, 1, &chip->rmidi) < 0) {
		return (-1);
	}

	sprintf(chip->rmidi->name, "%s MIDI 0", card->id);
	chip->rmidi->private_data = &chip;

	snd_rawmidi_set_ops(chip->rmidi, SNDRV_RAWMIDI_STREAM_INPUT, &snd_Echo_midi_input);
	snd_rawmidi_set_ops(chip->rmidi, SNDRV_RAWMIDI_STREAM_OUTPUT, &snd_Echo_midi_output);

	chip->rmidi->info_flags |= SNDRV_RAWMIDI_INFO_OUTPUT | SNDRV_RAWMIDI_INFO_INPUT | SNDRV_RAWMIDI_INFO_DUPLEX;
	DE_INIT(("MIDI ok\n"));
	    return (0);
}

#endif
#endif











////// _Dsp.c

#ifdef MIDI_SUPPORT

/****************************************************************************

	MIDI

 ****************************************************************************/

//===========================================================================
//
// Send a buffer full of MIDI data to the DSP
//
//===========================================================================
/*
ECHOSTATUS CDspCommObject::WriteMidi(struct EchoGals *eg, PBYTE pData,	// Ptr to data buffer
				     DWORD dwLength,	// How many bytes to write
				     PDWORD pdwActualCt	// Return how many actually written
  )
{
  ECHOSTATUS Status;
  DWORD dwLengthTemp, dwShift, dwPacked, i;

#ifdef ECHO_DEBUG
  DWORD dwOrgLgth = dwLength;
#endif

  //
  // Send all the MIDI data
  //
  *pdwActualCt = 0;
  Status = ECHOSTATUS_OK;

  while (dwLength > 0) {
    //
    // Wait for the last command to complete
    //
    if (WaitForHandshake()) {
      ECHO_DEBUGPRINTF(("CDspCommObject::WriteMidi - Timed out waiting for handshake\n"));

      Status = ECHOSTATUS_DSP_DEAD;
      break;
    }
    //
    // See if the DSP is ready to go
    //
    if (0 == (GetDspRegister(CHI32_STATUS_REG) & CHI32_STATUS_REG_HF4)) {
      //
      // MIDI output is full
      //
      break;
    }
    //
    // Pack the MIDI data three bytes at a time
    //
    if (dwLength > 3)
      dwLengthTemp = 3;
    else
      dwLengthTemp = dwLength;

    dwShift = 16;
    dwPacked = 0;
    for (i = 0; i < dwLengthTemp; i++) {
      dwPacked |= (DWORD) * pData++ << dwShift;
      dwShift -= 8;
    }

    dwPacked |= dwLengthTemp << 24;
    dwLength -= dwLengthTemp;

    //
    // Write the data to the DSP
    //
    DspCommPage->dwMIDIOutData = SWAP(dwPacked);
    ClearHandshake();
    SendVector(DSP_VC_MIDI_WRITE);

    ECHO_DEBUGPRINTF(("CDspCommObject::WriteMidi - wrote 0x%08lx\n", dwPacked));

    //
    // Update the count
    //
    *pdwActualCt += dwLengthTemp;

  }				// while( dwLength )

  if (0 != *pdwActualCt) {
    //
    // Save the current time - used to detect if MIDI out is currently busy
    //
    m_pOsSupport->OsGetSystemTime(&MidiOutTime);
    // Last time MIDI out occured
  }

  ECHO_DEBUGPRINTF(("CDspCommObject::WriteMidi: Actual %ld " "Expected %ld OK\n", *pdwActualCt, dwOrgLgth));

  return Status;

}				// ECHOSTATUS CDspCommObject::WriteMidi


//===========================================================================
//
// Called from the interrupt handler - get a MIDI input byte
//
//===========================================================================

ECHOSTATUS CDspCommObject::ReadMidi(struct EchoGals *eg, WORD wIndex,	// Buffer index
				    DWORD & dwData	// Return data
  )
{
  if (wIndex >= DSP_MIDI_BUFFER_SIZE - 1)
    return ECHOSTATUS_INVALID_INDEX;

  //
  // Get the data
  //      
  dwData = SWAP(DspCommPage->wMidiInData[wIndex]);

  //
  // Timestamp for the MIDI input activity indicator
  //
  m_pOsSupport->OsGetSystemTime(&MidiInTime);

  return ECHOSTATUS_OK;

}				// ECHOSTATUS CDspCommObject::ReadMidi


ECHOSTATUS CDspCommObject::SetMidiOn(struct EchoGals *eg, BOOL bOn)
{
  if (bOn) {
    if (0 == MidiOnCount) {
      if (WaitForHandshake())
	return ECHOSTATUS_DSP_DEAD;

      DspCommPage->Flags |= SWAP((DWORD) DSP_FLAG_MIDI_INPUT);

      ClearHandshake();
      SendVector(DSP_VC_UPDATE_FLAGS);
    }
    MidiOnCount++;
  } else {
    if (MidiOnCount == 0)
      return ECHOSTATUS_OK;

    if (0 == --MidiOnCount) {
      if (WaitForHandshake())
	return ECHOSTATUS_DSP_DEAD;

      DspCommPage->Flags &= SWAP((DWORD) ~ DSP_FLAG_MIDI_INPUT);

      ClearHandshake();
      SendVector(DSP_VC_UPDATE_FLAGS);
    }
  }

  return ECHOSTATUS_OK;

}				// ECHOSTATUS CDspCommObject::SetMidiOn
*/

#endif	// MIDI_SUPPORT
